/*

 Copyright (c) 2006 Pavel Novak.

 The use and distribution terms for this software are contained in the file named License.txt, 
 which can be found in the root of the NetSim distribution. By using this software 
 in any fashion, you are agreeing to be bound by the terms of this license.
 
 You must not remove this notice from this software.

*/

using System;
using System.Collections;

namespace NetSim.Core
{
	/// <summary>
	/// Collection used for storing bytes for packets. Currently does not support size changes.
	/// </summary>
	public class ByteArrayList : ICollection
	{
		/// <summary>
		/// Returns an array of bytes that are stored in this collection.
		/// </summary>
		/// <remarks>Currently returns refernce to an internal data buffer, do not change values
		/// in the returned array.</remarks>
		public byte[] Bytes { get { return data; } }

		/// <summary>
		/// Internal data storage.
		/// </summary>
		private byte[] data;

		/// <summary>
		/// Index of the first free byte. At this index the data will be appended next.
		/// </summary>
		private int position = 0;

		/// <summary>
		/// Creates new <see cref="ByteArrayList"/> with an initial capacity.
		/// </summary>
		/// <param name="capacity">Number of bytes that can be stored in this collection.</param>
		public ByteArrayList(int capacity)
		{
			data = new byte[capacity];
		}

		/// <summary>
		/// Clears all values in the collection. Adding data will start at the beginning.
		/// </summary>
		public void Clear()
		{
			position = 0;

			for (int i = 0; i < data.Length; i++)
				data[i] = 0;
		}

		/// <summary>
		/// Adds one byte just after the previously added data.
		/// </summary>
		/// <param name="b">Byte to append.</param>
		public void AddByte(byte b)
		{
			data[position] = b;
			position++;
		}

		/// <summary>
		/// Appends all values from the array just after previously added data.
		/// </summary>
		/// <param name="array">Values to append.</param>
		public void AddBytes(byte[] array)
		{
			AddBytes(array, 0, array.Length);
		}

		/// <summary>
		/// Appends <paramref name="len"/> values from <paramref name="array"/> to this collection,
		/// starting reading at <paramref name="offset"/>. Values will be written just after the
		/// previously appended data.
		/// </summary>
		/// <param name="array">An array to copy from.</param>
		/// <param name="offset">Offset in <paramref name="array"/> where to start.</param>
		/// <param name="len">How many bytes to copy.</param>
		public void AddBytes(byte[] array, int offset, int len)
		{
			Buffer.BlockCopy(array, offset, data, position, len);
			position += len;
		}

		/// <summary>
		/// Appends an <see cref="UInt16"/> value to this collection. Byte order is Big Endians.
		/// </summary>
		/// <param name="value">Value to append.</param>
		public void AddUInt16(UInt16 value)
		{
			data[position] = (byte) ( value / (byte.MaxValue+1) );
			data[position+1] = (byte) ( value % (byte.MaxValue+1) );
			position += 2;
		}

		/// <summary>
		/// Appends an <see cref="UInt32"/> value to this collection. Byte order is Big Endians.
		/// </summary>
		/// <param name="value">Value to append.</param>
		public void AddUInt32(UInt32 value)
		{
			data[position+3] = (byte) (value % (byte.MaxValue+1));
			value = value / (byte.MaxValue+1);
			data[position+2] = (byte) (value % (byte.MaxValue+1));
			value = value / (byte.MaxValue+1);
			data[position+1] = (byte) (value % (byte.MaxValue+1));
			value = value / (byte.MaxValue+1);
			data[position] = (byte)value;
			position += 4;
		}

		/// <summary>
		/// Computes a sum of 16-bit values starting at <paramref name="startAt"/>. Can be used
		/// further as an additional paramter to <see cref="OneComplementChecksum2"/> method.
		/// </summary>
		/// <param name="startAt">Where to start.</param>
		/// <param name="bytesCount">How many bytes to sum. If it is an odd number, last byte is
		/// treated as zero.</param>
		/// <returns>Sum of 16-bit values.</returns>
		public UInt32 SumForOneComplementComputation(int startAt, int bytesCount)
		{
			UInt32 sum = 0;
			while (bytesCount > 0)
			{
				// get first byte
				UInt32 val = data[startAt++];
				bytesCount--;
				val <<= 8;

				// get second byte if present
				val += (bytesCount > 0 ? data[startAt++] : (byte)0);
				bytesCount--;

				// add to the total sum
				sum += val;
			}

			return sum;
		}

		/// <summary>
		/// Computes one's complement checksum. Bytes starting at <paramref name="startAt"/> are treated
		/// as a sequence of <see cref="UInt16"/>, they are added in one's complement arithmetic and
		/// one complement of the sum is returned.
		/// </summary>
		/// <param name="startAt">Index to start computation.</param>
		/// <param name="bytesCount">How many bytes to count.</param>
		/// <returns>One's complement checksum.</returns>
		public UInt16 OneComplementChecksum2(int startAt, int bytesCount)
		{
			return OneComplementChecksum2(startAt, bytesCount, 0);
		}

		/// <summary>
		/// Computes one's complement checksum. Bytes starting at <paramref name="startAt"/> are treated
		/// as a sequence of <see cref="UInt16"/>, they are added in one's complement arithmetic and
		/// one complement of the sum is returned.
		/// </summary>
		/// <param name="startAt">Index to start computation.</param>
		/// <param name="bytesCount">How many bytes to count.</param>
		/// <param name="addToChecksum">Number that should be added to the sum before one's complement is made.</param>
		/// <returns>One's complement checksum.</returns>
		public UInt16 OneComplementChecksum2(int startAt, int bytesCount, UInt32 addToChecksum)
		{
			return OneComplementChecksum2(this.data, startAt, bytesCount, addToChecksum);
		}

		/// <summary>
		/// Computes one's complement checksum. Bytes starting at <paramref name="startAt"/> are treated
		/// as a sequence of <see cref="UInt16"/>, they are added in one's complement arithmetic and
		/// one complement of the sum is returned.
		/// </summary>
		/// <param name="data">An array of bytes from where the values are taken.</param>
		/// <param name="startAt">Index to start computation.</param>
		/// <param name="bytesCount">How many bytes to count.</param>
		/// <param name="addToChecksum">Number that should be added to the sum before one's complement is made.</param>
		/// <returns>One's complement checksum.</returns>
		public static UInt16 OneComplementChecksum2(byte[] data, int startAt, int bytesCount, UInt32 addToChecksum)
		{
			UInt32 sum = addToChecksum;
			while (bytesCount > 0)
			{
				UInt32 val = data[startAt++];
				bytesCount--;
				val <<= 8;

				val += (bytesCount > 0 ? data[startAt++] : (byte)0);
				sum += val;
				bytesCount--;
			}
			
			// there could be an overflow:
			sum = (sum & 0xFFFF) + (sum >> 16);
			sum = (sum & 0xFFFF) + (sum >> 16);

			return (UInt16)(~sum & 0xFFFF);
		}

		/// <summary>
		/// Set the value of <see cref="UInt16"/> at the specified position (two bytes are modified).
		/// Byte order is Big Endians.
		/// </summary>
		/// <param name="startAt">Where to put the value.</param>
		/// <param name="value">Value to set.</param>
		public void SetUInt16(int startAt, UInt16 value)
		{
			data[startAt] = (byte) ( value / (byte.MaxValue+1) );
			data[startAt+1] = (byte) ( value % (byte.MaxValue+1) );
		}

		#region Helper methods
		
		/// <summary>
		/// Gets two bytes from the <paramref name="data"/> and returns <see cref="UInt16"/> value.
		/// Byte order is Big Endians.
		/// </summary>
		/// <param name="data">An array from where the value is taken.</param>
		/// <param name="offset">Offset of the first byte in the array.</param>
		/// <returns>The 16-bit value.</returns>
		public static UInt16 ToUInt16(byte[] data, int offset)
		{
			return (UInt16) ( ((UInt16)byte.MaxValue+1)*data[offset] + data[offset+1] );
		}

		/// <summary>
		/// Gets four bytes from the <paramref name="data"/> and returns <see cref="UInt32"/> value.
		/// Byte order is Big Endians.
		/// </summary>
		/// <param name="data">An array from where the value is taken.</param>
		/// <param name="offset">Offset of the first byte in the array.</param>
		/// <returns>The 32-bit value.</returns>
		public static UInt32 ToUInt32(byte[] data, int offset)
		{
			UInt32 result = data[offset];
			result = result * ((UInt32)byte.MaxValue+1) + data[offset+1];
			result = result * ((UInt32)byte.MaxValue+1) + data[offset+2];
			result = result * ((UInt32)byte.MaxValue+1) + data[offset+3];

			return result;
		}

		/// <summary>
		/// Returns <b>true</b> if both arrays have the same length and identical values.
		/// </summary>
		/// <param name="arr1">First array to compare.</param>
		/// <param name="arr2">Second array to compare.</param>
		/// <returns><b>True</b> of both arrays are the same, <b>false</b> otherwise.</returns>
		public static bool IsEqual(byte[] arr1, byte[] arr2)
		{
			if (arr1.Length != arr2.Length)
				return false;

			for (int i = 0; i < arr1.Length; i++)
			{
				if (arr1[i] != arr2[i])
					return false;
			}

			return true;
		}
		#endregion

		/// <summary>
		/// Gets maximum number of bytes that can be stored in this collection.
		/// </summary>
		public int Capacity
		{
			get { return data.Length; }
		}

		#region ICollection Members

		/// <summary>
		/// Returns always <b>false</b>.
		/// </summary>
		public bool IsSynchronized
		{
			get { return false; }
		}

		/// <summary>
		/// Returns the number of bytes that have been appended to the collection. Ignores "set"
		/// operations where the value is directly inserted to specified position.
		/// </summary>
		public int Count
		{
			get { return position; }
		}

		/// <summary>
		/// Copies this collection to the <paramref name="array"/>.
		/// </summary>
		/// <param name="array">Destination array.</param>
		/// <param name="index">Start position in the destination array.</param>
		public void CopyTo(Array array, int index)
		{
			Buffer.BlockCopy(data, 0, array, index, position);
		}

		/// <summary>
		/// Returns always <b>null</b>. This collection cannot be synchronized.
		/// </summary>
		public object SyncRoot
		{
			get { return null; }
		}

		#endregion

		#region IEnumerable Members

		/// <summary>
		/// Returns an enumerator that can be used to enumerate bytes in the collection.
		/// </summary>
		/// <returns></returns>
		public IEnumerator GetEnumerator()
		{
			return data.GetEnumerator();
		}

		#endregion
	}
}
