/*

 Copyright (c) 2006 Pavel Novak.

 The use and distribution terms for this software are contained in the file named License.txt, 
 which can be found in the root of the NetSim distribution. By using this software 
 in any fashion, you are agreeing to be bound by the terms of this license.
 
 You must not remove this notice from this software.

*/

using System;
using System.Collections;
using System.Diagnostics;

namespace NetSim.Core
{
	/// <summary>
	/// A collection which provides mapping of keys to values and also values to keys.
	/// </summary>
	/// <remarks>
	/// Internally implemented with two <see cref="Hashtable"/>s.
	/// </remarks>
	public class DoubleHashtable
	{
		/// <summary>Hashtable hashing keys to values.</summary>
		private Hashtable keysToValues;
		/// <summary>Hashtable hashing values to keys.</summary>
		private Hashtable valuesToKeys;

		/// <summary>
		/// Object that can be used to synchronize access to this table.
		/// </summary>
		public virtual object SyncRoot { get { return this.syncRoot; } }
		private object syncRoot;

		/// <summary>
		/// Returns <b>true</b> if the access to the table is thread safe.
		/// </summary>
		public bool IsSynchronized { get { return isSynchronized; } }
		private bool isSynchronized;

		/// <summary>
		/// Returns thread-safe wrapper for the <see cref="DoubleHashtable"/>.
		/// </summary>
		/// <param name="table">The table to wrap.</param>
		/// <returns>Thread-safe wrapper of <paramref name="table"/>.</returns>
		public static DoubleHashtable Synchronized(DoubleHashtable table)
		{
			return new DoubleHashtable(table.keysToValues, table.valuesToKeys, table.syncRoot);
		}

		/// <summary>
		/// Used to synchronized wrapper creation - internal <see cref="Hashtable"/>s are shared.
		/// </summary>
		private DoubleHashtable(Hashtable keysToValues, Hashtable valuesToKeys, object syncRoot)
		{
			this.keysToValues = keysToValues;
			this.valuesToKeys = valuesToKeys;
			this.syncRoot = syncRoot;
			this.isSynchronized = true;
		}

		/// <summary>
		/// Creates new <see cref="DoubleHashtable"/> with default initial capacity.
		/// </summary>
		public DoubleHashtable()
		{
			this.keysToValues = new Hashtable();
			this.valuesToKeys = new Hashtable();
			this.syncRoot = new object();
			this.isSynchronized = false;
		}

		/// <summary>
		/// Creates new <see cref="DoubleHashtable"/> with the specified initial capacity.
		/// </summary>
		/// <param name="capacity">Initial capacity.</param>
		public DoubleHashtable(int capacity)
		{
			this.keysToValues = new Hashtable(capacity);
			this.valuesToKeys = new Hashtable(capacity);
			this.syncRoot = new object();
			this.isSynchronized = false;
		}

		/// <summary>
		/// Adds a new (key, value) pair to the table.
		/// </summary>
		/// <param name="key">Key.</param>
		/// <param name="value">Value.</param>
		public void Add(object key, object value)
		{
			if (isSynchronized)
			{
				lock (this.syncRoot)
				{
					AddInternal(key, value);
				}
			}
			else
			{
				AddInternal(key, value);
			}
		}

		/// <summary>
		/// Adds a new pair (key, value) to the table. Does not care about synchronization.
		/// </summary>
		/// <param name="key">Key.</param>
		/// <param name="value">Value.</param>
		private void AddInternal(object key, object value)
		{
			if (key == null)
				throw new ArgumentNullException("key");
			if (value == null)
				throw new ArgumentNullException("value");
			if (keysToValues.Contains(key))
				throw new ArgumentException("Key is already present", "key");
			if (valuesToKeys.Contains(value))
				throw new ArgumentException("Value is already present", "value");

			keysToValues.Add(key, value);
			valuesToKeys.Add(value, key);
		}

		/// <summary>
		/// Returns a value for specified key.
		/// </summary>
		/// <param name="key">The key.</param>
		/// <returns>Value for <paramref name="key"/>, or <b>null</b> if not found.</returns>
		public object GetValue(object key)
		{
			return keysToValues[key];
		}

		/// <summary>
		/// Returns a key for the specified value.
		/// </summary>
		/// <param name="value">The value.</param>
		/// <returns>A key for <paramref name="value"/>, or <b>null</b> if not found.</returns>
		public object GetKey(object value)
		{
			return valuesToKeys[value];
		}

		/// <summary>
		/// Removes a specified key from the table.
		/// </summary>
		/// <param name="key">Key to remove.</param>
		public void RemoveByKey(object key)
		{
			if (key == null)
				throw new ArgumentNullException("key");

			// remove the value
			object value = keysToValues[key];
			if (value != null) valuesToKeys.Remove(value);
			// remove the key
			keysToValues.Remove(key);
		}

		/// <summary>
		/// Removes a specified value from the table.
		/// </summary>
		/// <param name="value">Value to remove.</param>
		public void RemoveByValue(object value)
		{
			if (value == null)
				throw new ArgumentNullException("value");

			// remove the key
			object key = valuesToKeys[value];
			if (key != null) keysToValues.Remove(key);
			// remove the value
			valuesToKeys.Remove(value);
		}
	}
}
