//
// Source code downloaded from http://www.codeproject.com/csharp/PriorityQueue.asp
// Modified and "beautified" :)
//

using System;
using System.Collections;

namespace NetSim.Core
{
	/// <summary>
	/// Priority queue interface.
	/// </summary>
	public interface IPriorityQueue : ICollection, ICloneable, IList
	{
		int Push(IComparable key, object value);
		int Push(IPriorityQueueEntry entry);
		object Pop();
		object Peek();
	}

	/// <summary>
	/// One entry in the priority queue.
	/// </summary>
	public interface IPriorityQueueEntry : IComparable
	{
		IComparable Key { get; set; }
		object Value { get; set; }
	}

	/// <summary>
	/// One entry in the priority queue.
	/// </summary>
	public class PriorityQueueEntry : IPriorityQueueEntry
	{
		public IComparable Key
		{
			get { return _key; }
			set { _key = value; }
		}
		private IComparable _key;

		public object Value
		{
			get { return _value; }
			set { _value = value; }
		}
		private object _value;

		public PriorityQueueEntry(IComparable key, object value)
		{
			if (key is int)
				throw new Exception();

			this.Key = key;
			this.Value = value;
		}

		public override bool Equals(object ob)
		{
			if (ob is PriorityQueueEntry) 
			{
				PriorityQueueEntry entry = (PriorityQueueEntry) ob;
				return this.Key == entry.Key;
			}
			else 
			{
				return false;
			}
		}

		public override int GetHashCode()
		{
			return Key.GetHashCode();
		}

		#region IComparable Members

		public int CompareTo(object obj)
		{
				if (obj is PriorityQueueEntry) 
				{
					PriorityQueueEntry entry = (PriorityQueueEntry) obj;

					return this.Key.CompareTo(entry.Key);
				}
        
				throw new ArgumentException("Object to compare with is not a PriorityQueueEntry");    
		}

		#endregion
	}

	/// <summary>
	/// Priority queue implemented as binary heap.
	/// </summary>
	public class BinaryPriorityQueue : IPriorityQueue
	{
		protected ArrayList InnerList;

		#region contructors
		public BinaryPriorityQueue()
		{
			InnerList = new ArrayList();
		}

		public BinaryPriorityQueue(int capacity)
		{
			InnerList = new ArrayList(capacity);
		}

		protected BinaryPriorityQueue(ArrayList Core, bool Copy)
		{
			if (Copy)
				InnerList = Core.Clone() as ArrayList;
			else
				InnerList = Core;
		}

		#endregion

		protected void SwitchElements(int i, int j)
		{
			object h = InnerList[i];
			InnerList[i] = InnerList[j];
			InnerList[j] = h;
		}

		protected virtual int OnCompare(int i, int j)
		{
			return ((PriorityQueueEntry)InnerList[i]).CompareTo(InnerList[j]);
		}

		#region public methods
		/// <summary>
		/// Push an new entry into the PQ.
		/// </summary>
		/// <param name="entry">The new entry.</param>
		/// <returns>The index in the list where the object is _now_. This will change when objects are taken from or put onto the PQ.</returns>
		public int Push(IPriorityQueueEntry entry)
		{
			int p = InnerList.Count,p2;
			InnerList.Add(entry); // E[p] = O
			do
			{
				if(p==0)
					break;
				p2 = (p-1)/2;
				if(OnCompare(p,p2)<0)
				{
					SwitchElements(p,p2);
					p = p2;
				}
				else
					break;
			}while(true);
			return p;
		}

		public int Push(IComparable key, object value)
		{
			return Push(new PriorityQueueEntry(key, value));
		}

		/// <summary>
		/// Get the smallest object and remove it.
		/// </summary>
		/// <returns>The smallest object</returns>
		public object Pop()
		{
			object result = InnerList[0];
			int p = 0,p1,p2,pn;
			InnerList[0] = InnerList[InnerList.Count-1];
			InnerList.RemoveAt(InnerList.Count-1);
			do
			{
				pn = p;
				p1 = 2*p+1;
				p2 = 2*p+2;
				if(InnerList.Count>p1 && OnCompare(p,p1)>0) // links kleiner
					p = p1;
				if(InnerList.Count>p2 && OnCompare(p,p2)>0) // rechts noch kleiner
					p = p2;
				
				if(p==pn)
					break;
				SwitchElements(p,pn);
			}while(true);
			return result;
		}

		/// <summary>
		/// Notify the PQ that the object at position i has changed
		/// and the PQ needs to restore order.
		/// Since you dont have access to any indexes (except by using the
		/// explicit IList.this) you should not call this function without knowing exactly
		/// what you do.
		/// </summary>
		/// <param name="i">The index of the changed object.</param>
		protected void Update(int i)
		{
			int p = i,pn;
			int p1,p2;
			do	// aufsteigen
			{
				if(p==0)
					break;
				p2 = (p-1)/2;
				if(OnCompare(p,p2)<0)
				{
					SwitchElements(p,p2);
					p = p2;
				}
				else
					break;
			}while(true);
			if(p<i)
				return;
			do	   // absteigen
			{
				pn = p;
				p1 = 2*p+1;
				p2 = 2*p+2;
				if(InnerList.Count>p1 && OnCompare(p,p1)>0) // links kleiner
					p = p1;
				if(InnerList.Count>p2 && OnCompare(p,p2)>0) // rechts noch kleiner
					p = p2;
				
				if(p==pn)
					break;
				SwitchElements(p,pn);
			} while (true);
		}

		/// <summary>
		/// Get the smallest object without removing it.
		/// </summary>
		/// <returns>The smallest object</returns>
		public object Peek()
		{
			if (InnerList.Count > 0)
				return InnerList[0];

			return null;
		}

		public bool Contains(object value)
		{
			return InnerList.Contains(value);
		}

		public void Clear()
		{
			InnerList.Clear();
		}

		public int Count
		{
			get
			{
				return InnerList.Count;
			}
		}

		/// <summary>
		/// Returns an enumerator.
		/// </summary>
		/// <remarks>In this implementation values are not enumerated in the sorted order</remarks>
		/// <returns></returns>
		IEnumerator IEnumerable.GetEnumerator()
		{
			// Future version: Enumerator does not enumerate in the sorted order
			return InnerList.GetEnumerator();
		}

		/// <summary>
		/// Copies values into an array at specific index.
		/// </summary>
		/// <remarks>In this implementation the values are not copied are not sorted.</remarks>
		/// <param name="array"></param>
		/// <param name="index"></param>
		public void CopyTo(Array array, int index)
		{
			//Future version: The values copied are not sorted...
			InnerList.CopyTo(array,index);
		}

		public object Clone()
		{
			return new BinaryPriorityQueue(InnerList, true);	
		}

		public bool IsSynchronized
		{
			get
			{
				return InnerList.IsSynchronized;
			}
		}

		public object SyncRoot
		{
			get
			{
				return this;
			}
		}
		#endregion

		#region explicit implementation
		bool IList.IsReadOnly
		{
			get
			{
				return false;
			}
		}

		object IList.this[int index]
		{
			get
			{
				return InnerList[index];
			}
			set
			{
				InnerList[index] = value;
				Update(index);
			}
		}

		int IList.Add(object o)
		{
			PriorityQueueEntry entry = o as PriorityQueueEntry;
			if (entry == null)
				throw new ArgumentException("PriorityQueueEntry object expected");

			return Push(entry);
		}

		void IList.RemoveAt(int index)
		{
			throw new NotSupportedException();
		}

		void IList.Insert(int index, object value)
		{
			throw new NotSupportedException();
		}

		void IList.Remove(object value)
		{
			throw new NotSupportedException();
		}

		int IList.IndexOf(object value)
		{
			throw new NotSupportedException();
		}

		bool IList.IsFixedSize
		{
			get
			{
				return false;
			}
		}

		public static BinaryPriorityQueue Syncronized(BinaryPriorityQueue P)
		{
			return new BinaryPriorityQueue(ArrayList.Synchronized(P.InnerList), false);
		}
		public static BinaryPriorityQueue ReadOnly(BinaryPriorityQueue P)
		{
			return new BinaryPriorityQueue(ArrayList.ReadOnly(P.InnerList), false);
		}
		#endregion
	}
}
