/*

 Copyright (c) 2006 Pavel Novak.

 The use and distribution terms for this software are contained in the file named License.txt, 
 which can be found in the root of the NetSim distribution. By using this software 
 in any fashion, you are agreeing to be bound by the terms of this license.
 
 You must not remove this notice from this software.

*/

using System;
using System.Collections;

namespace NetSim.Core
{
	/// <summary>
	/// Sorted list allowing duplicate entries.
	/// </summary>
	public class SortedList2 : IEnumerable, ICollection
	{
		/// <summary>
		/// An array of keys for values in <see cref="innerList"/>.
		/// </summary>
		private ArrayList innerKeys;

		/// <summary>
		/// An array of values for keys in <see cref="innerKeys"/>.
		/// </summary>
		private ArrayList innerList;

		/// <summary>
		/// This memeber is set if current object is a thread-safe wrapper for the inner list.
		/// Otherwise, it is <b>null</b>.
		/// </summary>
		private SortedList2 innerNotSynchronized;

		/// <summary>
		/// Creates new <see cref="SortedList2"/>.
		/// </summary>
		public SortedList2()
		{
			innerList = new ArrayList();
			innerKeys = new ArrayList();
		}

		/// <summary>
		/// Creates new thread-safe wrapper around the <paramref name="innerNotSynchronized"/>.
		/// </summary>
		/// <param name="innerNotSynchronized"><see cref="SortedList2"/> to wrap.</param>
		private SortedList2(SortedList2 innerNotSynchronized)
		{
			this.innerNotSynchronized = innerNotSynchronized;
			this.innerKeys = ArrayList.Synchronized(innerNotSynchronized.innerKeys);
			this.innerList = ArrayList.Synchronized(innerNotSynchronized.innerList);
		}

		/// <summary>
		/// Adds new (key, value) pair to the list.
		/// </summary>
		/// <param name="key">The key.</param>
		/// <param name="value">The value.</param>
		public void Add(IComparable key, object value)
		{
			// synchronized wrapper
			if (innerNotSynchronized != null)
			{
				lock (SyncRoot)
				{
					innerNotSynchronized.Add(key, value);
					return;
				}
			}

			// perform binary search where to add the value
			int i = 0;
			int j = innerList.Count;
			int index = -1;

			while (i < j)
			{
				int k = (i + j) / 2;
				int compare_result = key.CompareTo(innerKeys[k]);

				if (compare_result < 0)
				{
					j = k;
				}
				else if (compare_result > 0)
				{
					i = k + 1;
				}
				else
				{
					index = k;
					break;
				}
			}

			if (index == -1)
				index = i;

			innerKeys.Insert(index, key);
			innerList.Insert(index, value);
		}

		/// <summary>
		/// Returns the smallest index of <paramref name="value"/>. If no match is found, returns -1.
		/// </summary>
		/// <param name="value">Value to search for.</param>
		/// <returns>The smallest index or -1.</returns>
		public int IndexOfValue(object value)
		{
			// synchronized wrapper
			if (innerNotSynchronized != null)
			{
				lock (SyncRoot)
				{
					return innerNotSynchronized.IndexOfValue(value);
				}
			}

			for (int i = 0; i < innerList.Count; i++)
			{
				if (value.Equals(innerList[i]))
					return i;
			}

			return -1;
		}

		/// <summary>
		/// Removes the (key, value) pair at specified index.
		/// </summary>
		/// <param name="index">The index where the values will be removed.</param>
		public void RemoveAt(int index)
		{
			// synchronized wrapper
			if (innerNotSynchronized != null)
			{
				lock (SyncRoot)
				{
					innerNotSynchronized.RemoveAt(index);
					return;
				}
			}

			innerList.RemoveAt(index);
			innerKeys.RemoveAt(index);
		}

		/// <summary>
		/// Returns a value at the specified index.
		/// </summary>
		/// <param name="index"></param>
		/// <returns></returns>
		public object GetByIndex(int index)
		{
			return innerList[index];
		}

		#region ICollection members

		/// <summary>
		/// Returns the number of (key, value) pairs in the collection.
		/// </summary>
		public int Count
		{
			get { return innerList.Count; }
		}

		/// <summary>
		/// Returns <b>true</b> if current collection is synchronized, <b>false</b> otherwise.
		/// </summary>
		public bool IsSynchronized
		{
			get { return innerNotSynchronized != null; }
		}

		/// <summary>
		/// Returns thread-safe wrapper for the <paramref name="list"/>.
		/// </summary>
		/// <param name="list">The list to wrap.</param>
		/// <returns>Thread-safe wrapper.</returns>
		public static SortedList2 Synchronized(SortedList2 list)
		{
			if (list == null)
				throw new ArgumentNullException("list");

			return new SortedList2(list);
		}

		/// <summary>
		/// Synchronization object for this collection.
		/// </summary>
		public object SyncRoot
		{
			get { return innerList.SyncRoot; }
		}

		/// <summary>
		/// Copies all values from the collection to the <paramref name="array"/>.
		/// </summary>
		/// <param name="array">The destination array.</param>
		/// <param name="index">Index where to start copy in the destination.</param>
		public void CopyTo(Array array, int index)
		{
			// synchronized wrapper
			if (innerNotSynchronized != null)
			{
				lock (SyncRoot)
				{
					innerNotSynchronized.CopyTo(array, index);
					return;
				}
			}

			if (array == null)
				throw new ArgumentNullException("array");
			if (index < 0)
				throw new ArgumentOutOfRangeException("index");
			if (array.Rank > 1 || index >= array.Length || this.Count > (array.Length - index))
				throw new ArgumentException("array", "Array is multi-dimensional or not large enough");

			foreach (object o in this)
			{
				array.SetValue(o, index);
				index++;
			}
		}
		#endregion

		#region IEnumerable members
		/// <summary>
		/// Returns an enumerator for this collection.
		/// </summary>
		/// <returns>An enumerator.</returns>
		IEnumerator IEnumerable.GetEnumerator()
		{
			return innerList.GetEnumerator();
		}
		#endregion

	}
}
