/*

 Copyright (c) 2006 Pavel Novak.

 The use and distribution terms for this software are contained in the file named License.txt, 
 which can be found in the root of the NetSim distribution. By using this software 
 in any fashion, you are agreeing to be bound by the terms of this license.
 
 You must not remove this notice from this software.

*/

using System;
using System.Collections;
using System.Diagnostics;
using NetSim.Core;
using NetSim.Library;

namespace NetSim.Ethernet
{
	/// <summary>
	/// Interface that should be implemented by every class that uses <see cref="ArpCache"/>.
	/// </summary>
	public interface IArpCacheOwner
	{
		/// <summary>
		/// Sends the packet to the link.
		/// </summary>
		/// <param name="pi">The <see cref="PacketInfo"/> containing the information about packet to send.</param>
		void SendPacket(PacketInfo pi);
		
		void ArpMappingAdded(NetworkAddress protocolAddr, NetworkAddress hwAddress);

		/// <summary>
		/// Returns <b>true</b> if the specified address is used in the settings for the specified protocol
		/// associated with this adapter.
		/// </summary>
		/// <param name="addr">Address to search.</param>
		/// <param name="packetGroup">Packet group (network layer) the address belnongs to.</param>
		/// <param name="protocol">The protocol at the specified layer.</param>
		/// <param name="locAddr">Found local address.</param>
		/// <returns><b>True</b> if the address is associated with this adapter.</returns>
		bool IsLocalAddress(byte[] addr, Packet.PacketGroup packetGroup, int protocol, out NetworkAddress locAddr);

		NetworkAddress GetLocalAddress(Packet.PacketGroup group, int type);

		/// <summary>
		/// Returns hardware address of the current adapter the cache is attached to.
		/// </summary>
		NetworkAddress HardwareAddress { get; }
	}

	/// <summary>
	/// Summary description for ArpCache.
	/// </summary>
	public class ArpCache
	{
		/// <summary>
		/// Arp cache entry.
		/// </summary>
		public class Entry
		{
			public NetworkAddress ProtocolAddress;
			public NetworkAddress HardwareAddress;

			/// <summary>
			/// Set to true if network and protocol address fields are not valid, but ARP request has been sent
			/// and we are waiting for the response
			/// </summary>
			public bool Waiting;
			public long Expires;

			public Entry(NetworkAddress protocol, NetworkAddress hw, long expires, bool waiting)
			{
				this.ProtocolAddress = protocol;
				this.HardwareAddress = hw;

				this.Waiting = waiting;
				this.Expires = expires;
			}

		}

		/// <summary>
		/// The object owning this ArpCache and providing methods to support the cache.
		/// </summary>
		private IArpCacheOwner owner;

		/// <summary>
		/// Keys: <see cref="NetworkAddress"/> as protocol address, Values: <see cref="ArpCache.Entry"/> objects.
		/// </summary>
		private Hashtable mapping;

		/// <summary>
		/// Type of underlying hardware according to ARP packet format definition. (Ethernet == 1)
		/// </summary>
		private UInt16 hardwareType;

		private long DefaultResponseWaitingTimeout = (long)2 * 1000 * 1000 * 10; // 2 seconds
		private long EntryExpirationTime = (long)10000000 * 60 * 5; // 5 minutes

		public ArpCache(IArpCacheOwner owner, UInt16 hardwareType)
		{
			this.owner = owner;
			this.mapping = Hashtable.Synchronized(new Hashtable());
			this.hardwareType = hardwareType;
		}

		/// <summary>
		/// 
		/// </summary>
		/// <param name="protocol"></param>
		/// <param name="hw"></param>
		protected void Add(NetworkAddress protocol, NetworkAddress hw)
		{
			mapping[protocol] = new Entry(protocol, hw, Simulator.Timer.GetTimeAfterDelay(EntryExpirationTime), false);
			owner.ArpMappingAdded(protocol, hw);
		}

		/// <summary>
		/// 
		/// </summary>
		/// <param name="protocol"></param>
		/// <param name="hwType"></param>
		/// <returns></returns>
		public NetworkAddress GetHwAddress(NetworkAddress protocol, UInt16 hwType)
		{
			Debug.Assert(protocol != null);

			if (hwType != this.hardwareType)
				return null; // not our hardware

			Entry entry = (Entry) mapping[protocol];
			if (entry == null) // not in the cache
				return null;

			if (entry.Expires < Simulator.Timer.Now) // expired?
			{
				// yes, delete the entry from the cache
				mapping[protocol] = null;
				return null;
			}
				
			if (entry.Waiting) // no response yet
				return null;

			// not expired, return the requested address
			return entry.HardwareAddress;
		}

		public void RequestMappingToHw(NetworkAddress searchedProtocolAddress)
		{
			if (searchedProtocolAddress.PacketGroup != Packet.PacketGroup.NetworkLayer)
				throw new ArgumentException("Only addresses for network layer can be added to the ArpCache");

			// is the entry in the cache?
			Entry entry = (Entry) mapping[searchedProtocolAddress];
			if (entry != null)
			{
				if (entry.Expires < Simulator.Timer.Now) // expired?
				{
					// yes, delete the entry from the cache
					mapping[searchedProtocolAddress] = null;
				}
				else
				{
					return; // not expired entry, do not send ARP request again
				}
			}

			NetworkAddress senderProtocolAddr = owner.GetLocalAddress(searchedProtocolAddress.PacketGroup, searchedProtocolAddress.PacketType);

			// if mapping is not in the cache, we should send ARP request
			ArpPacket p = ArpPacket.CreatePacket((ArpPacket.HardwareType)hardwareType, (Packet.NetworkLayerPacketType)searchedProtocolAddress.PacketType,
				owner.HardwareAddress.Length, searchedProtocolAddress.Length, ArpPacket.Operation.ArpRequest, owner.HardwareAddress.GetAddressBytes(),
				senderProtocolAddr.GetAddressBytes(), null, searchedProtocolAddress.GetAddressBytes());

			// add waiting entry to the cache
			mapping[searchedProtocolAddress] = new Entry(null, null, Simulator.Timer.GetTimeAfterDelay(this.DefaultResponseWaitingTimeout), true);

			PacketInfo pi = new PacketInfo(p);
			pi.NextHopDestination = null; // broadcast the packet
			owner.SendPacket(pi);
		}

		public void ProcessPacket(ArpPacket p)
		{
			switch (p.Op)
			{
				case ArpPacket.Operation.ArpRequest:
					ProcessArpRequest(p);
					break;

				case ArpPacket.Operation.ArpResponse:
					ProcessArpResponse(p);
					break;

				case ArpPacket.Operation.RArpRequest:
					//Future version
					break;

				case ArpPacket.Operation.RArpResponse:
					//Future version
					break;
			}
		}

		public void ProcessArpRequest(ArpPacket p)
		{
			NetworkAddress loc_address;
			if (owner.IsLocalAddress(p.TargetProtocolAddress, Packet.PacketGroup.NetworkLayer, (int)p.ProtocolTypeField, out loc_address)) // request for us
			{
				// add the sender to our own cache
				NetworkAddress sender_protocol_addr = loc_address.CreateNewAddress(p.SenderProtocolAddress);
				this.Add(sender_protocol_addr, owner.HardwareAddress.CreateNewAddress(p.SenderHardwareAddress));

				// reply to the request
				ArpPacket reply = ArpPacket.SwitchToArpResponse(p, owner.HardwareAddress);
				PacketInfo pi = new PacketInfo(reply);
				pi.NextHopDestination = sender_protocol_addr;
				owner.SendPacket(pi);
			}
			else
			{
				//Do not remember sender mappings from all arp requests...
			}
		}

		public void ProcessArpResponse(ArpPacket p)
		{
			NetworkAddress loc_address;
			if (owner.IsLocalAddress(p.TargetProtocolAddress, Packet.PacketGroup.NetworkLayer, (int)p.ProtocolTypeField, out loc_address)) // response for us
			{
				// add the sender to our own cache
				NetworkAddress sender_protocol_addr = loc_address.CreateNewAddress(p.SenderProtocolAddress);
				this.Add(sender_protocol_addr, owner.HardwareAddress.CreateNewAddress(p.SenderHardwareAddress));
			}
			else
			{
				//Do not remember anything from other responses
			}
		}
	}
}
