/*

 Copyright (c) 2006 Pavel Novak.

 The use and distribution terms for this software are contained in the file named License.txt, 
 which can be found in the root of the NetSim distribution. By using this software 
 in any fashion, you are agreeing to be bound by the terms of this license.
 
 You must not remove this notice from this software.

*/

using System;
using NetSim.Core;

namespace NetSim.Ethernet
{

	/// <summary>
	/// Ethernet frame.
	/// </summary>
	public class EthFrame : Packet
	{
		/// <summary>
		/// MAC address of the sender.
		/// </summary>
		public EthMacAddress Source { get { return source; } }
		private EthMacAddress source;

		/// <summary>
		/// MAC address of the receiver.
		/// </summary>
		public EthMacAddress Destination { get { return destination; } }
		private EthMacAddress destination;

		/// <summary>
		/// Type of the Ethernet frame - corresponds to "Type" field in the frame.
		/// </summary>
		private ushort frameType;

		public static EthFrame CreatePacket(byte[] data, int offset)
		{
			return new EthFrame(data, offset);
		}

		public static EthFrame CreatePacket(EthMacAddress source, EthMacAddress destination, ushort frameType, Packet higherLayerPacket)
		{
			return new EthFrame(source, destination, frameType, higherLayerPacket);
		}

		protected EthFrame(byte[] data, int offset)
		{
			// destination address
			destination = EthMacAddress.CreateAddress(data, offset);
			offset += EthMacAddress.MacAddressLen;
			// source address
			source = EthMacAddress.CreateAddress(data, offset);
			offset += EthMacAddress.MacAddressLen;
			// frame type
			frameType = ByteArrayList.ToUInt16(data, offset);
			offset += 2;

			if (frameType < 1500) // 802.3 frame, otherwise Ethernet II
			{
				// LLC header
				byte dest_sap = data[offset++];
				byte source_sap = data[offset++];
				byte control_byte = data[offset++];

				if (dest_sap == 0xAA && source_sap == 0xAA) // SNAP frame
				{
					// skip OUI
					offset += 3;
					// frame type is here..
					frameType = ByteArrayList.ToUInt16(data, offset);
					offset += 2;
				}
				else // others from 802.3 are not supported
				{
					// IEEE charges $1000 fee for retrieving the complete list :)
					// see http://www.iana.org/assignments/ieee-802-numbers
					return;
				}
			}

			HigherLayerPacket = Simulator.GetInstance().CreatePacket(Packet.PacketGroup.NetworkLayer, frameType, data, offset);
		}

		public EthFrame(EthMacAddress source, EthMacAddress destination, ushort frameType, Packet higherLayerPacket)
		{
			this.source = source;
			this.destination = destination;
			this.frameType = frameType;
			this.HigherLayerPacket = higherLayerPacket;
		}

		protected override void AppendBytesHeader(ByteArrayList buf)
		{
			buf.AddBytes(destination.GetAddressBytes());
			buf.AddBytes(source.GetAddressBytes());
			buf.AddUInt16(frameType);
		}

		protected override void AppendBytesFooter(ByteArrayList buf)
		{
			// nothing
		}

		protected override int BytesCount()
		{
			return
				2 * EthMacAddress.MacAddressLen // source + destination MAC
				+ 2                             // frame type
				;
		}

		public override Packet.PacketGroup Group
		{
			get { return Packet.PacketGroup.LinkLayer; }
		}

		public override int Type
		{
			get { return (int)Packet.LinkLayerPacketType.DLT_EN10MB; }
		}

		public override string ToString()
		{
			return String.Format("EthFrame (src:{0}, dst:{1})", Source, Destination);
		}

	}


	/// <summary>
	/// Summary description for PacketsFactory.
	/// </summary>
	public class EthernetPacketFactory : PacketFactory
	{
		public override Packet CreatePacket(Packet.PacketGroup group, int identification, byte[] data, int offset, int len)
		{
			// we can create only ethernet packets
			if (group != Packet.PacketGroup.LinkLayer || identification != (int)Packet.LinkLayerPacketType.DLT_EN10MB)
				throw new PacketCreationException(String.Format("Packet group number '{0}' and type '{1}' is not supported by EthernetPacketFactory", group, identification));

			return EthFrame.CreatePacket(data, offset);
		}

		public override PacketFactory.SupportedPacket[] Supports
		{
			get
			{
				return new PacketFactory.SupportedPacket[] {
					new PacketFactory.SupportedPacket(Packet.PacketGroup.LinkLayer, (int)Packet.LinkLayerPacketType.DLT_EN10MB)
																									 };
			}
		}

	}

}
