/*

 Copyright (c) 2006 Pavel Novak.

 The use and distribution terms for this software are contained in the file named License.txt, 
 which can be found in the root of the NetSim distribution. By using this software 
 in any fashion, you are agreeing to be bound by the terms of this license.
 
 You must not remove this notice from this software.

*/

using System;
using NetSim.Core;

namespace NetSim.Ethernet
{
	public class EthMacAddress : NetworkAddress
	{
		/// <summary>
		/// Length of MAC address in bytes.
		/// </summary>
		public const byte MacAddressLen = 6;

		/// <summary>
		/// Array
		/// </summary>
		public override byte[] GetAddressBytes()
		{
			return address;
		}
		private byte[] address;
		
		public static EthMacAddress CreateAddress(byte[] data)
		{
			return CreateAddress(data, 0);
		}

		public static EthMacAddress CreateAddress(byte[] data, int offset)
		{
			if (data.Length < offset + MacAddressLen)
				return null;

			return new EthMacAddress(data, offset);
		}

		public override NetworkAddress CreateNewAddress(byte[] bytes)
		{
			if (this.Length != bytes.Length)
				throw new ArgumentException(String.Format("Byte array of {0} bytes expected", this.Length));

			return EthMacAddress.CreateAddress(bytes);
		}

		protected EthMacAddress(byte[] data, int offset)
		{
			this.address = new byte[MacAddressLen];
			Buffer.BlockCopy(data, offset, address, 0, MacAddressLen);
		}

		public override byte Length
		{
			get { return MacAddressLen; }
		}

		public static EthMacAddress Broadcast
		{
			get
			{
				//Future version: faster implementaion
				return Parse("FF:FF:FF:FF:FF:FF");
			}
		}


		public static EthMacAddress Parse(string str)
		{
			try
			{
				// array of hexadecimal strings for bytes..
				string[] bytes_strings = str.Split(':', '-');
				if (bytes_strings.Length > 1 && bytes_strings.Length != MacAddressLen)
					throw new FormatException(String.Format("Wrong MAC address format: expecting {0} hexadecimal numbers separated by ':' or '-'", MacAddressLen));

				byte[] bytes = new byte[MacAddressLen];
				for (int i = 0; i < MacAddressLen; i++)
				{
					bytes[i] = System.Byte.Parse(bytes_strings[i], System.Globalization.NumberStyles.HexNumber);
				}

				return new EthMacAddress(bytes, 0);
			}
			catch (OverflowException e)
			{
				throw new FormatException(String.Format("Wrong MAC address format: expecting {0} hexadecimal numbers separated by ':' or '-'", MacAddressLen), e);
			}
			catch (FormatException e)
			{
				throw new FormatException(String.Format("Wrong MAC address format: expecting {0} hexadecimal numbers separated by ':' or '-'", MacAddressLen), e);
			}
		}

		public override Packet.PacketGroup PacketGroup
		{
			get {	return Packet.PacketGroup.LinkLayer; }
		}

		public override int PacketType
		{
			get { return (int)Packet.LinkLayerPacketType.DLT_EN10MB; }
		}

		public override bool Equals(object obj)
		{
			EthMacAddress second = obj as EthMacAddress;
			if (second == null)
				return false;

			return ByteArrayList.IsEqual(this.address, second.address);
		}

		public override int GetHashCode()
		{
			// treat lower 4 bytes as an integer
			return unchecked((int)ByteArrayList.ToUInt32(this.address, 2));
		}

		public override string ToString()
		{
			bool first = true;
			System.Text.StringBuilder str = new System.Text.StringBuilder();
			foreach (byte b in this.GetAddressBytes())
			{
				if (!first)
					str.Append(":");
				str.Append(b.ToString("x"));
					
				first = false;
			}

			return str.ToString();
		}

	}

//	public class EthMacAddressFactory : AddressFactory
//	{
//	}
}
