/*

 Copyright (c) 2006 Pavel Novak.

 The use and distribution terms for this software are contained in the file named License.txt, 
 which can be found in the root of the NetSim distribution. By using this software 
 in any fashion, you are agreeing to be bound by the terms of this license.
 
 You must not remove this notice from this software.

*/

using System;
using NetSim.Core;
using NetSim.Library;
using System.Xml;
using System.Collections;

namespace NetSim.Ethernet
{
	/// <summary>
	/// Summary description for Adapter.
	/// </summary>
	public class EthernetAdapter : Adapter, IArpCacheOwner
	{
		/// <summary>
		/// Default maximum transfer unit.
		/// </summary>
		public const int DefaultMTU = 1500;

		/// <summary>
		/// Whether the adapter is in promiscuous mode or not.
		/// </summary>
		private bool promiscuousMode = false;

		/// <summary>
		/// The link this adapter is connected to.
		/// </summary>
		public override Link Link
		{
			get { return base.Link; }
			set
			{
				if (!(value is EthernetLink))
				{
					ethLink = null;
				}
				else
				{
					ethLink = (EthernetLink)value;
				}

				base.Link = value;
			}
		}

		/// <summary>
		/// Set if the link this adapter is connected to is <see cref="EthernetLink"/>. Otherwise is <b>null</b>.
		/// </summary>
		/// <remarks>
		/// The adapter can connect also to the non-ethernet link, for example <see cref="SimpleLink"/> in the
		/// case that it connects to real network. If non-ethernet link is used, ethernet specific behavior
		/// is not used, of course. Frames are sent immediately, discarded if the send operation fails.
		/// </remarks>
		protected EthernetLink ethLink;

		/// <summary>
		/// MAC address of this adapter.
		/// </summary>
		EthMacAddress macAddress;
		public override NetworkAddress HardwareAddress
		{
			get { return macAddress; }
		}

		/// <summary>
		/// Gets maximum transfer unit used by this adapter.
		/// </summary>
		public override int MTU
		{
			get { return this.maxTransferUnit; }
		}
		private int maxTransferUnit = DefaultMTU;

		/// <summary>
		/// Packets that cannot be sent immediately, because hardware address of the destination is not known.
		/// </summary>
		// generics: <PacketInfo>
		private ArrayList pendingPackets;

		protected ArpCache ArpCache
		{
			get { return arpCache; }
		}
		private ArpCache arpCache;

		/// <summary>
		/// The packets are sent sequentionally. If there is a new request to send the packet while previous packet
		/// is being sent, the packet is enqueued and send immediately after the send operation finishes (either
		/// the packet comes back as an acknowledgement or <see cref="PacketSendingFailed"/> is called).
		/// </summary>
		private Queue waitingPackets;

		public EthernetAdapter()
		{
			pendingPackets = ArrayList.Synchronized(new ArrayList());
			waitingPackets = Queue.Synchronized(new Queue());
			arpCache = new ArpCache(this, (UInt16)ArpPacket.HardwareType.Ethernet);
		}

		public override void XmlConfigure(System.Xml.XmlDocument document)
		{
			// configure protocols
			base.XmlConfigure(document);

			// find and set MAC address
			XmlNode n = document.DocumentElement.FirstChild;
			while (n != null)
			{
				if (n.Name == "MacAddress")
				{
					if (macAddress != null)
						throw new InitializationException("MacAdress specified twice");

					macAddress = EthMacAddress.Parse(n.InnerText);
				}
				else if (n.Name == "Promiscuous")
				{
					try { promiscuousMode = Boolean.Parse(n.InnerText); }
					catch (FormatException e)
					{
						throw new InitializationException("Invalid value for 'Promiscuous'", e);
					}
				}

				n = n.NextSibling;
			}

			if (macAddress == null)
				throw new InitializationException("MacAddress not specified");
		}

		public override void SendPacket(PacketInfo pi)
		{
			if (!Enabled)
				return;

			// if raw packet is send, it shouldn't be wrapped by the ethernet packet
			Packet to_send = pi.RawPacket;

			// create ethernet packet if necessary
			if (to_send == null)
			{
				EthMacAddress destination;
				if (pi.NextHopDestination == null) // no destination specified, use broadcast
				{
					destination = EthMacAddress.Broadcast;
				}
				else
				{
					destination = this.ArpCache.GetHwAddress(pi.NextHopDestination, (UInt16)ArpPacket.HardwareType.Ethernet) as EthMacAddress;
				}

				if (destination == null) // no mapping known
				{
					// remember the packet
					this.pendingPackets.Add(pi);
					// tell the cache that it should fill the table with this mapping
					this.ArpCache.RequestMappingToHw(pi.NextHopDestination);
					return;
				}
				
				// get the type of higher layer packet
				UInt16 frame_type;
				try { frame_type = (UInt16)pi.Packet.Type; }
				catch (OverflowException e)
				{
					Simulator.GetInstance().Events.Notice(String.Format("The packet with frame type '{0}' cannot be inserted into the Ethernet frame", pi.Packet.Type), e);
					return;
				}

				// create new ethernet packet
				to_send = EthFrame.CreatePacket(this.macAddress, destination, frame_type, pi.Packet);
			}

			// the packet will come back as confirmation and shouldn't be received...
			// MUST be called before sending the packet since the sending task could be done by the different thread
			// before the packet is written to the send packets collection
			RememberSentPacket(to_send);

			waitingPackets.Enqueue(to_send);
			
			// test after enqueuing the packet, because if the packet would be enqueued after this test, it can be
			// missed and wait in the queue possibly forever

			if (sentPackets.Count > 1) // do not send the packet right now, another packet is being sent
				return;

			// send the next packet from the queue
			SendWaitingPacket();
		}
		
		/// <summary>
		/// Sends new packet from the <see cref="waitingPackets"/> queue if one is present.
		/// </summary>
		private void SendWaitingPacket()
		{
			Packet p = null;
			lock (waitingPackets.SyncRoot)
			{
				if (waitingPackets.Count > 0)
					p = (Packet)waitingPackets.Dequeue();
			}

			if (p == null) // nothing to send
				return;

			// wait for the link to become available and send
			EthernetLink ethLink = Link as EthernetLink;
			if (ethLink == null) // not ethernet link
			{
				// send immediately, we do not know the link...
				Link.SendPacket(this, p);
			}
			else
			{
				long available = ((EthernetLink)Link).LinkAvailableAt;
				if (available <= Simulator.Timer.Now) // link is available immediately, send
					Link.SendPacket(this, p);
				else // we have to wait
					ManagedThreadPool.QueueUserWorkItemAt(new System.Threading.WaitCallback(this.SendPacketCallback),
						p, available);
			}
		}

		private void SendPacketCallback(object packetObj)
		{
			Packet p = (Packet)packetObj;
			Link.SendPacket(this, p);
		}

		/// <summary>
		/// Called by the link to notify the adapter that the packet cannot be send.
		/// </summary>
		/// <param name="p">Packet that cannot be sent.</param>
		/// <param name="retry"><b>True</b> if retrying to send makes sense.</param>
		public override void PacketSendingFailed(Packet p, bool retry)
		{
			// Ethernet behavior: next send at r^th collision slot where 0 < r < 2^k, k = min(n, 10),
			// n == number of tries, stop after 16 tries
			
			// get the number of tries
			object tries_obj = sentPackets[p];
			System.Diagnostics.Debug.Assert(tries_obj != null);
			int tries = (int)tries_obj;
			System.Diagnostics.Debug.Assert(tries > 0);

			if (!retry || ethLink == null || tries >= 16) // stop trying to send
			{
				// remove from sent packets
				CheckPacketComingBack(p);
				// send next packet
				SendWaitingPacket();
				return;
			}

			int exponent = Math.Min(10, tries);
			int slot_count = Simulator.Random.Next(0, (1 << exponent) + 1);

			ManagedThreadPool.QueueUserWorkItem(new System.Threading.WaitCallback(this.SendPacketCallback),
				p, slot_count * ethLink.CollisionSlot);
		}

		/// <summary>
		/// Called by the <see cref="Link"/> while the packet should be received by this node through
		/// this adapter.
		/// </summary>
		/// <param name="p">Packet to receive.</param>
		public override void ReceivePacket(Packet p)
		{
			// if this is not an ethernet frame, ignore
			EthFrame p_eth = p as EthFrame;
			if (p_eth == null)
				return;

			if (CheckPacketComingBack(p)) // our packet coming back..
			{
				SendWaitingPacket();
				return;
			}
			
			if (!Enabled)
				return;

			if (!promiscuousMode &&
				!(p_eth.Destination.Equals(EthMacAddress.Broadcast) || this.HardwareAddress.Equals(p_eth.Destination))
				)
				return;

			// if this is the ARP packet, notify the ARP cache
			if (p.HigherLayerPacket is ArpPacket)
			{
				this.ArpCache.ProcessPacket(p.HigherLayerPacket as ArpPacket);
				//if (!promiscuousMode) // allow reception of the ARP packet by the upper layers while promiscuous mode..
				//	return;
			}

			// pass the packet to the node
			PacketInfo pi = new PacketInfo(p.HigherLayerPacket, this);
			pi.RawPacket = p;
			node.ReceivePacket(pi);
		}

		/// <summary>
		/// Returns <b>true</b> if the specified address is used in the settings for the specified protocol
		/// associated with this adapter.
		/// </summary>
		/// <param name="addr">Address to search.</param>
		/// <param name="packetGroup">Packet group where the packet containing the address belongs.</param>
		/// <param name="packetType">Packet type within the group.</param>
		/// <param name="locAddr">Out parameter containing the <see cref="NetworkAddress"/> instance of the
		/// address if the function returns <b>true</b>.</param>
		/// <returns><b>True</b> if the address is associated with this adapter.</returns>
		public bool IsLocalAddress(byte[] addr, Packet.PacketGroup packetGroup, int packetType, out NetworkAddress locAddr)
		{
			locAddr = null;
			foreach (ProtocolSettings settings in protocolSettings)
			{
				if (
					(settings.Address.PacketGroup == packetGroup && packetType == settings.Address.PacketType)
					// address is the same
					&& ByteArrayList.IsEqual(settings.Address.GetAddressBytes(), addr)
					)
				{
					locAddr = settings.Address;
					return true;
				}
			}

			return false;
		}

		public NetworkAddress GetLocalAddress(Packet.PacketGroup group, int type)
		{
			foreach (ProtocolSettings settings in protocolSettings)
			{
				if (settings.Address.PacketGroup == group && type == settings.Address.PacketType)
				{
					return settings.Address;
				}
			}

			return null;
		}

		public void ArpMappingAdded(NetworkAddress protocolAddr, NetworkAddress hwAddress)
		{
			lock (pendingPackets.SyncRoot)
			{
				int i = 0;
				while (i < pendingPackets.Count)
				{
					PacketInfo pi = (PacketInfo) pendingPackets[i];
					if (pi.NextHopDestination.Equals(protocolAddr))
					{
						// the added address applies to this packet - remove it from pending packets and send
						pendingPackets.RemoveAt(i);
						this.SendPacket(pi);
					}
					else
					{
						i++;
					}
				}
			}
		}

	}
}
