/*

 Copyright (c) 2006 Pavel Novak.

 The use and distribution terms for this software are contained in the file named License.txt, 
 which can be found in the root of the NetSim distribution. By using this software 
 in any fashion, you are agreeing to be bound by the terms of this license.
 
 You must not remove this notice from this software.

*/

using System;
using System.Threading;
using NetSim.Core;

namespace NetSim.Ethernet
{
	/// <summary>
	/// Base Ethernet link implementation. Separate Ethernet technologies derives from this link.
	/// </summary>
	public abstract class EthernetLink : Link
	{
		/// <summary>
		/// Link speed in bytes per second (not bits!). Overriding in the descendant class is required.
		/// </summary>
		public abstract long Speed
		{
			get;
		}

		/// <summary>
		/// Collision slot time lenght in ticks (100ns intervals). Overriding in the descendant class is required.
		/// </summary>
		public abstract int CollisionSlot
		{
			get;
		}

		/// <summary>
		/// Returns the time by <see cref="NetSim.Core.SimulationTimer"/> when is the link available for
		/// data transfer. The returned time can be also in the past, in that case the link is available
		/// immediately
		/// </summary>
		public virtual long LinkAvailableAt
		{
			get { return nextAvailableAt; }
		}
		protected long nextAvailableAt;

		/// <summary>
		/// Function for collision detection. Is called when the packet is about to send; we are waiting
		/// the collision slot time if another packet will come. In such case there is a collision - none
		/// of the packets would succeed and the adapters should be notified that the link hasn't sent
		/// the packets. After the collision slot time is out <see cref="FrameToSendDec"/> should be called.
		/// </summary>
		protected void FrameToSendInc()
		{
			if (Interlocked.Increment(ref framesToSend) > 1)
				collision = true; // more than one frame is sent within a collision slot
		}

		/// <summary>
		/// Called after the collision slot time is out. Dual function to <see cref="FrameToSendInc"/>.
		/// </summary>
		/// <returns><b>True</b> if a collision took place, <b>false</b> otherwise.</returns>
		protected bool FrameToSendDec()
		{
			System.Diagnostics.Debug.Assert(framesToSend > 0);

			// remember the collision value
			bool result = collision;

			// decrement the counter and reset the collision member if appropriate
			if (Interlocked.Decrement(ref framesToSend) == 0)
				collision = false; // 0 frames are waiting to be sent, cannot be a collision

			return result;
		}

		private bool collision = false;
		private int framesToSend = 0;

		/// <summary>
		/// Default constructor allowing the link creation via reflection according to the configuration file.
		/// </summary>
		public EthernetLink()
		{
		}

		public override void SendPacket(IAdapter adapter, Packet p)
		{
			long now = Simulator.Timer.Now;

			// 1) If the link is disabled, we cannot send packet
			// 2) If the link is busy sending packet, we cannot send.. This shouldn't occur very often thanks
			//    to testing if the link is busy in the ethernet adapter. If it is the case, the simulation
			//    behaviour is not accurate, because also the packet being sent in this time interval should
			//    have a collision.. We don't care here about such situation.
			if (!Enabled || this.LinkAvailableAt > now)
			{
				adapter.PacketSendingFailed(p, Enabled);
				return;
			}

			// wait a collision slot time before the packet will be sent.
			FrameToSendInc();
			ManagedThreadPool.QueueUserWorkItem(new WaitCallback(this.SendPacketWaitCollisionTime),
				new SendPacketState(adapter, p), this.CollisionSlot);
		}

		private void SendPacketWaitCollisionTime(object stateObj)
		{
			// remember the time right here at the beginning, we won't to add delay by executing code
			// or even worse by switching to another thread...
			long now = Simulator.Timer.Now;

			bool collision = FrameToSendDec();
			SendPacketState state = (SendPacketState)stateObj;
			
			if (collision)
			{
				state.Adapter.PacketSendingFailed(state.Packet, true);
			}
			else // no collision
			{
				// set the link busy time according to the packet length
				int bytes = state.Packet.GetBytesCount();
				long ticks_delay = ((long)bytes * 10000000) / this.Speed; // won't overflow, long is large enough
				this.nextAvailableAt = Simulator.Timer.GetTimeAfterDelay(ticks_delay, now);

				// send to all adapters
				try
				{
					// send the packet to all adapters
					foreach (IAdapter a in adapters)
					{
						// for each adapter process by a separate thread - receiving the packet can take some time...
						ManagedThreadPool.QueueUserWorkItem(new WaitCallback(this.SendPacketCallback),
							new SendPacketState(a, state.Packet));
					}
				}
				catch (InvalidOperationException)
				{
					// the collection was modified - ignore.
					// This allows changing network structure while running - some packets won't be
					// delivered just after the change - it is not crucial.
				}
			}
		}
	}

	/// <summary>
	/// 10Base2 Ethernet implementation.
	/// </summary>
	public class Ethernet10Base2 : EthernetLink
	{
		/// <summary>
		/// Link speed in bytes per second (not bits!);
		/// </summary>
		public override long Speed
		{
			get { return 10L * 1024 * 1024 / 8; }
		}

		public override int CollisionSlot
		{
			get { return 512; }
		}

	}

	/// <summary>
	/// 100TX Ethernet implementation.
	/// </summary>
	public class Ethernet100TX : EthernetLink
	{
		/// <summary>
		/// Link speed in bytes per second (not bits!);
		/// </summary>
		public override long Speed
		{
			get { return 100L * 1024 * 1024 / 8; }
		}

		public override int CollisionSlot
		{
			get { return 512; }
		}

	}

}
