/*

 Copyright (c) 2006 Pavel Novak.

 The use and distribution terms for this software are contained in the file named License.txt, 
 which can be found in the root of the NetSim distribution. By using this software 
 in any fashion, you are agreeing to be bound by the terms of this license.
 
 You must not remove this notice from this software.

*/

using System;
using NetSim.Core;
using System.Collections;

namespace NetSim.Ethernet
{
	/// <summary>
	/// If this module is present at the node, it resends ethernet packets to one or all adapters
	/// depending on that if it knows where the destination MAC address resides. (Simple learning switch
	/// behavior.)
	/// </summary>
	public class LearningSwitch : Module
	{
		/// <summary>
		/// Mapping between <see cref="EthMacAddress"/> objects and <see cref="IAdapter"/>s.
		/// </summary>
		// generics: <EthMacAddress, IAdapter>
		private Hashtable addrToAdapters;

		/// <summary>
		/// Default constructor required by the simulator.
		/// </summary>
		public LearningSwitch()
		{
			// synchronized collection is required because more than one thread can call ReceivePacket
			// concurrently
			addrToAdapters = Hashtable.Synchronized(new Hashtable());
		}

		public override void ReceivePacket(ref PacketInfo pInfo)
		{
			EthFrame eth_frame = pInfo.RawPacket as EthFrame;
			if (eth_frame == null)
				return; // do nothing, eth switch cannot send non-ethernet data..

			// remember the sender address
			addrToAdapters[eth_frame.Source] = pInfo.UseAdapter;

			// try to find target adapter
			IAdapter target_adapter = (IAdapter)addrToAdapters[eth_frame.Destination];

			if (target_adapter != pInfo.UseAdapter) // send only if the destination is behind different adapter
			{                                       // (this condition works also if target_adapter == null)

				pInfo.ExcludeAdapter = pInfo.UseAdapter; // do not send back
				pInfo.UseAdapter = target_adapter; // send only here (can be null)

				Node.NewPacketToSend(this.ID, pInfo);
			}

			// do not send to upper layers
			pInfo = null;
		}

	}
}
