/*

 Copyright (c) 2006 Pavel Novak.

 The use and distribution terms for this software are contained in the file named License.txt, 
 which can be found in the root of the NetSim distribution. By using this software 
 in any fashion, you are agreeing to be bound by the terms of this license.
 
 You must not remove this notice from this software.

*/

using System;
using System.Collections;

namespace NetSim.Core
{
	/// <summary>
	/// Represents one event stored in <see cref="EventLog"/>.
	/// </summary>
	[Serializable]
	public class Event
	{
		/// <summary>
		/// Type of event.
		/// </summary>
		[Serializable]
		public enum EventType { Warning, Error, Notice }

		/// <summary>
		/// Type of event (warning, error, notice).
		/// </summary>
		public EventType Type { get { return type; } }
		private EventType type;

		/// <summary>
		/// Time when the event was raised.
		/// </summary>
		public DateTime Time { get { return time; } }
		private DateTime time;

		/// <summary>
		/// Event description message.
		/// </summary>
		public string Message { get { return message; } }
		private string message;

		/// <summary>
		/// An exception that was thrown in the application that raised the event. Can be <b>null</b>.
		/// </summary>
		public Exception Exception { get { return exception; } }
		private Exception exception;

		/// <summary>
		/// Creates new event.
		/// </summary>
		/// <param name="type">Event type.</param>
		/// <param name="message">Message describing the event.</param>
		public Event(EventType type, string message) : this(type, message, null)
		{
		}

		/// <summary>
		/// Creates new event.
		/// </summary>
		/// <param name="type">Event type.</param>
		/// <param name="message">Message describing the event.</param>
		/// <param name="exception"><see cref="Exception"/> that caused the event.</param>
		public Event(EventType type, string message, Exception exception)
		{
			this.type = type;
			this.time = DateTime.Now;
			this.message = message;
			this.exception = exception;
		}
	}

	/// <summary>
	/// EventLog collects information about errors or other events that occured during the simulation.
	/// </summary>
	public class EventLog
	{
		public delegate void NewEntryDelegate(Event.EventType eventType);
		public event NewEntryDelegate NewEntryEvent;

		/// <summary>
		/// Collection of <see cref="Event"/>s.
		/// </summary>
		// generics: <Event>
		ArrayList events = ArrayList.Synchronized(new ArrayList());

		/// <summary>
		/// Creates new <see cref="EventLog"/> object.
		/// </summary>
		public EventLog()
		{
		}

		/// <summary>
		/// Inserts new event of type <see cref="Event.EventType.Warning"/> to the event log.
		/// </summary>
		/// <param name="message">Message describing the event.</param>
		public void Warning(string message)
		{
			Warning(message, null);
		}

		/// <summary>
		/// Inserts new event of type <see cref="Event.EventType.Warning"/> to the event log. If there
		/// is an exception that caused the event, it can be stored too. Otherwise set
		/// <paramref name="exception"/> to <b>null</b>.
		/// </summary>
		/// <param name="message">Message describing the event.</param>
		/// <param name="exception"><see cref="Exception"/> that caused the event.</param>
		public void Warning(string message, Exception exception)
		{
			events.Add(new Event(Event.EventType.Warning, message, exception));
			
			if (NewEntryEvent != null)
				NewEntryEvent(Event.EventType.Warning);
		}

		/// <summary>
		/// Inserts new event of type <see cref="Event.EventType.Error"/> to the event log.
		/// </summary>
		/// <param name="message">Message describing the event.</param>
		public void Error(string message)
		{
			Error(message, null);
		}

		/// <summary>
		/// Inserts new event of type <see cref="Event.EventType.Error"/> to the event log. If there
		/// is an exception that caused the event, it can be stored too. Otherwise set
		/// <paramref name="exception"/> to <b>null</b>.
		/// </summary>
		/// <param name="message">Message describing the event.</param>
		/// <param name="exception"><see cref="Exception"/> that caused the event.</param>
		public void Error(string message, Exception exception)
		{
			events.Add(new Event(Event.EventType.Error, message, exception));
			
			if (NewEntryEvent != null)
				NewEntryEvent(Event.EventType.Error);
		}
	
		/// <summary>
		/// Inserts new event of type <see cref="Event.EventType.Notice"/> to the event log.
		/// </summary>
		/// <param name="message">Message describing the event.</param>
		public void Notice(string message)
		{
			Notice(message, null);
		}

		/// <summary>
		/// Inserts new event of type <see cref="Event.EventType.Notice"/> to the event log. If there
		/// is an exception that caused the event, it can be stored too. Otherwise set
		/// <paramref name="exception"/> to <b>null</b>.
		/// </summary>
		/// <param name="message">Message describing the event.</param>
		/// <param name="exception"><see cref="Exception"/> that caused the event.</param>
		public void Notice(string message, Exception exception)
		{
			events.Add(new Event(Event.EventType.Notice, message, exception));
			
			if (NewEntryEvent != null)
				NewEntryEvent(Event.EventType.Notice);
		}

		/// <summary>
		/// Returns the first log entry as an out parameter.
		/// </summary>
		/// <param name="e">Out parameter that will be set to the <see cref="Event"/> object.</param>
		/// <param name="remove"><b>True</b> if the first entry should be removed from the log.</param>
		/// <returns><b>True</b> if an entry has been found, <b>false</b> otherwise.</returns>
		public bool GetLogEntry(out NetSim.Core.Event e, bool remove)
		{
			e = null;
			if (this.events.Count == 0)
				return false;
			
			lock (events.SyncRoot)
			{
				e = (Event)events[0];
				if (remove)
					this.events.RemoveAt(0);
			}

			return true;
		}
	}

}
