/*

 Copyright (c) 2006 Pavel Novak.

 The use and distribution terms for this software are contained in the file named License.txt, 
 which can be found in the root of the NetSim distribution. By using this software 
 in any fashion, you are agreeing to be bound by the terms of this license.
 
 You must not remove this notice from this software.

*/

using System;

namespace NetSim.Core
{
	/// <summary>
	/// Common ancestor for classes that provides creation of <see cref="Packet"/> objects.
	/// </summary>
	public abstract class PacketFactory
	{
		/// <summary>
		/// Default constructor allowing creation of the packet factory with no parameters while reading
		/// simulator configuration. Inherited classes should have default constructor too.
		/// </summary>
		public PacketFactory()
		{
		}

		/// <summary>
		/// Entries returned by <see cref="PacketFactory.Supports"/> property.
		/// </summary>
		public struct SupportedPacket
		{
			/// <summary>
			/// Packet type group (which network layer).
			/// </summary>
			public Packet.PacketGroup Group;

			/// <summary>
			/// Identification number within the group.
			/// </summary>
			public int Identification;

			/// <summary>
			/// Creates new <see cref="SupportedPacket"/> object.
			/// </summary>
			/// <param name="g">Group of a packet.</param>
			/// <param name="id">Packet identification.</param>
			public SupportedPacket(Packet.PacketGroup g, int id)
			{
				this.Group = g;
				this.Identification = id;
			}

			/// <summary>
			/// Returns a string representation of this object. Useful for debugging purposes only.
			/// </summary>
			/// <returns>String containing information about this object.</returns>
			public override string ToString()
			{
				return String.Format("PacketFactory.SupportedPacket: Group = {0}; Identification = {1}", Group, Identification);
			}
		}

		/// <summary>
		/// Returns an array of <see cref="SupportedPacket"/>s that can be created by this factory.
		/// </summary>
		public abstract SupportedPacket[] Supports { get; }

		/// <summary>
		/// Creates a <see cref="Packet"/> instance.
		/// </summary>
		/// <param name="group">Group (type) of the packet to create.</param>
		/// <param name="identification">Identification number within the group.</param>
		/// <param name="data">Raw contents of the packet.</param>
		/// <param name="offset">Offset where to start in the <paramref name="data"/> array.</param>
		/// <param name="len">Length of data to add to the packet.</param>
		/// <returns>Newly created Packet.</returns>
		public abstract Packet CreatePacket(Packet.PacketGroup group, int identification, byte[] data, int offset, int len);

		/// <summary>
		/// Creates a <see cref="Packet"/> instance.
		/// </summary>
		/// <param name="group">Group (type) of the packet to create.</param>
		/// <param name="identification">Identification number within the group.</param>
		/// <param name="data">Raw contents of the packet.</param>
		/// <param name="offset">Offset where to start in the <paramref name="data"/> array.</param>
		/// <returns>Newly created Packet.</returns>
		public virtual Packet CreatePacket(Packet.PacketGroup group, int identification, byte[] data, int offset)
		{
			return this.CreatePacket(group, identification, data, offset, data.Length - offset);
		}

		/// <summary>
		/// Creates a <see cref="Packet"/> instance.
		/// </summary>
		/// <param name="group">Group (type) of the packet to create.</param>
		/// <param name="identification">Identification number within the group.</param>
		/// <param name="data">Raw contents of the packet.</param>
		/// <returns>Newly created Packet.</returns>
		public virtual Packet CreatePacket(Packet.PacketGroup group, int identification, byte[] data)
		{
			return this.CreatePacket(group, identification, data, 0, data.Length);
		}
	}

}
