//
// Downloaded from http://www.gotdotnet.com/Community/UserSamples/Details.aspx?SampleGuid=BF59C98E-D708-4F8E-9795-8BAE1825C3B6
// and modified.
//

//
// Author of the original code:
// Stephen Toub, stoub@microsoft.com
// Very simple threadpool in C#.
// 4/27/04
//

#region Namespaces
using System;
using System.Threading;
using System.Collections;
#endregion

namespace NetSim.Core
{
	/// <summary>Implementation of Dijkstra's PV Semaphore based on the Monitor class.</summary>
	public class Semaphore
	{
		#region Member Variables
		/// <summary>The number of units alloted by this semaphore.</summary>
		private int _count;
		/// <summary>Lock for the semaphore.</summary>
		private object _semLock = new object();
		#endregion

		#region Construction
		/// <summary> Initialize the semaphore as a binary semaphore.</summary>
		public Semaphore() : this(1) 
		{
		}

		/// <summary> Initialize the semaphore as a counting semaphore.</summary>
		/// <param name="count">Initial number of threads that can take out units from this semaphore.</param>
		/// <exception cref="ArgumentException">Throws if the count argument is less than 0.</exception>
		public Semaphore(int count) 
		{
			if (count < 0) throw new ArgumentException("Semaphore must have a count of at least 0.", "count");
			_count = count;
		}
		#endregion

		#region Synchronization Operations
		/// <summary>V the semaphore (add 1 unit to it).</summary>
		public void AddOne() { V(); }

		/// <summary>P the semaphore (take out 1 unit from it).</summary>
		public void WaitOne() { P(); }

		/// <summary>P the semaphore (take out 1 unit from it).</summary>
		public void P() 
		{
			// Lock so we can work in peace.  This works because lock is actually
			// built around Monitor.
			lock(_semLock) 
			{
				// Wait until a unit becomes available.  We need to wait
				// in a loop in case someone else wakes up before us.  This could
				// happen if the Monitor.Pulse statements were changed to Monitor.PulseAll
				// statements in order to introduce some randomness into the order
				// in which threads are woken.
				while(_count <= 0) Monitor.Wait(_semLock, Timeout.Infinite);
				_count--;
			}
		}

		/// <summary>V the semaphore (add 1 unit to it).</summary>
		public void V() 
		{
			// Lock so we can work in peace.  This works because lock is actually
			// built around Monitor.
			lock(_semLock) 
			{
				// Release our hold on the unit of control.  Then tell everyone
				// waiting on this object that there is a unit available.
				_count++;
				Monitor.Pulse(_semLock);
			}
		}

		/// <summary>Resets the semaphore to the specified count.  Should be used cautiously.</summary>
		public void Reset(int count)
		{
			lock(_semLock) { _count = count; }
		}
		#endregion
	}

	/// <summary>Managed thread pool.</summary>
	public class ManagedThreadPool
	{
		#region Constants
		/// <summary>Maximum number of threads the thread pool has at its disposal.</summary>
		private const int _maxWorkerThreads = 25;
		#endregion

		#region Member Variables
		/// <summary>Queue of all the callbacks waiting to be executed.</summary>
		private static IPriorityQueue _waitingCallbacks;
		/// <summary>
		/// Used to signal that a worker thread is needed for processing.  Note that multiple
		/// threads may be needed simultaneously and as such we use a semaphore instead of
		/// an auto reset event.
		/// </summary>
		private static Semaphore _workerThreadNeeded;
		/// <summary>List of all worker threads at the disposal of the thread pool.</summary>
		private static ArrayList _workerThreads;
		/// <summary>Number of threads currently active.</summary>
		private static int _inUseThreads;
		/// <summary>Lockable object for the pool.</summary>
		private static object _poolLock = new object();

		private static AutoResetEvent waitingForTimerEvent = new AutoResetEvent(true);
		private static int waitingForTimer = 0;
		#endregion

		#region Construction and Finalization
		/// <summary>Initialize the thread pool.</summary>
		static ManagedThreadPool() { Initialize(); }

		/// <summary>Initializes the thread pool.</summary>
		private static void Initialize()
		{
			// Create our thread stores; we handle synchronization ourself
			// as we may run into situtations where multiple operations need to be atomic.
			// We keep track of the threads we've created just for good measure; not actually
			// needed for any core functionality.
			_waitingCallbacks = new BinaryPriorityQueue();
			_workerThreads = new ArrayList();
			_inUseThreads = 0;

			// Create our "thread needed" event
			_workerThreadNeeded = new Semaphore(0);
			
			// Create all of the worker threads
			for(int i=0; i<_maxWorkerThreads; i++)
			{
				// Create a new thread and add it to the list of threads.
				Thread newThread = new Thread(new ThreadStart(ProcessQueuedItems));
				_workerThreads.Add(newThread);

				// Configure the new thread and start it
				newThread.Name = "ManagedPoolThread #" + i.ToString();
				newThread.IsBackground = true;
				newThread.Priority = ThreadPriority.AboveNormal;
				newThread.Start();
			}

			// set higher priority to the process running these thread-pool threads
			System.Diagnostics.Process.GetCurrentProcess().PriorityClass = System.Diagnostics.ProcessPriorityClass.AboveNormal;
		}
		#endregion

 		#region Public Methods
		/// <summary>Queues a user work item to the thread pool.</summary>
		/// <param name="callback">
		/// A WaitCallback representing the delegate to invoke when the thread in the 
		/// thread pool picks up the work item.
		/// </param>
		public static void QueueUserWorkItem(WaitCallback callback)
		{
			// Queue the delegate with no state and process immediately
			QueueUserWorkItem(callback, null, 0);
		}

		/// <summary>Queues a user work item to the thread pool.</summary>
		/// <param name="callback">
		/// A WaitCallback representing the delegate to invoke when the thread in the 
		/// thread pool picks up the work item.
		/// </param>
		/// <param name="state">
		/// The object that is passed to the delegate when serviced from the thread pool.
		/// </param>
		public static void QueueUserWorkItem(WaitCallback callback, object state)
		{
			// Queue the delegate with state and process immediately
			QueueUserWorkItem(callback, state, 0);
		}

		/// <summary>Queues a user work item to the thread pool.</summary>
		/// <param name="callback">
		/// A WaitCallback representing the delegate to invoke when the thread in the 
		/// thread pool picks up the work item.
		/// </param>
		/// <param name="state">
		/// The object that is passed to the delegate when serviced from the thread pool.
		/// </param>
		/// <param name="waitTime">
		/// Time to wait before the callback delegate is executed. Default units are ticks - 100ns intervals,
		/// however, they can be changed by setting the <see cref="SimulationTimer.Speed"/> property of the member
		/// <see cref="Timer"/>.
		/// </param>
		public static void QueueUserWorkItem(WaitCallback callback, object state, long waitTime)
		{
			if (waitTime == 0) // special attention to this case, no delay
			{
				// Create a waiting callback that contains the delegate and its state.
				// At it to the processing queue, and signal that data is waiting.
				WaitingCallback waiting = new WaitingCallback(callback, state);

				lock(_poolLock) { _waitingCallbacks.Push(0L, waiting); }
				_workerThreadNeeded.AddOne();

				// the new item is before all the others, so it should be processed
				waitingForTimerEvent.Set();
			}
			else // some delay required
			{
				// count time when the work should be executed
				long time_after = Simulator.Timer.GetTimeAfterDelay(waitTime);

				QueueUserWorkItemAt(callback, state, time_after);
			}
		}

		/// <summary>Queues a user work item to the thread pool.</summary>
		/// <param name="callback">
		/// A WaitCallback representing the delegate to invoke when the thread in the 
		/// thread pool picks up the work item.
		/// </param>
		/// <param name="state">
		/// The object that is passed to the delegate when serviced from the thread pool.
		/// </param>
		/// <param name="timeToExecute">
		/// Time when the callback delegate should be executed. The time units are not specified,
		/// it should be obtained by calling <see cref="SimulationTimer"/> methods.
		/// </param>
		public static void QueueUserWorkItemAt(WaitCallback callback, object state, long timeToExecute)
		{
			// Create a waiting callback that contains the delegate and its state.
			// At it to the processing queue, and signal that data is waiting.
			WaitingCallback waiting = new WaitingCallback(callback, state);

			bool first = false;
			lock(_poolLock) { first = _waitingCallbacks.Push(timeToExecute, waiting) == 0; }
			_workerThreadNeeded.AddOne();

			// the new item is possibly before all the others, so it should be processed
			if (first)
				waitingForTimerEvent.Set();
		}

		/// <summary>Empties the work queue of any queued work items.  Resets all threads in the pool.</summary>
		public static void Reset()
		{
			lock(_poolLock) 
			{ 
				// Cleanup any waiting callbacks
				try 
				{
					// Try to dispose of all remaining state
					foreach(object obj in _waitingCallbacks)
					{
						WaitingCallback callback = (WaitingCallback) ((IPriorityQueueEntry)obj).Value;
						if (callback.State is IDisposable) ((IDisposable)callback.State).Dispose();
					}
				} 
				catch{}

				// Shutdown all existing threads
				try 
				{
					foreach(Thread thread in _workerThreads) 
					{
						if (thread != null) thread.Abort("reset");
					}
				}
				catch{}

				// Reinitialize the pool (create new threads, etc.)
				Initialize();
			}
		}

		public static void Pause()
		{
			lock (_poolLock)
			{
				try
				{
					foreach (Thread thread in _workerThreads)
					{
						if (thread != null)
							try
							{
								thread.Suspend();
							}
							catch (ThreadStateException)
							{
							}
					}
				}
				catch {}
			}
		}

		public static void Resume()
		{
			lock (_poolLock)
			{
				try
				{
					foreach (Thread thread in _workerThreads)
					{
						if (thread != null)
							try
							{
								thread.Resume();
							}
							catch (ThreadStateException)
							{
							}
					}
				}
				catch {}
			}
		}
		#endregion

		#region Properties
		/// <summary>Gets the number of threads at the disposal of the thread pool.</summary>
		public static int MaxThreads { get { return _maxWorkerThreads; } }
		/// <summary>Gets the number of currently active threads in the thread pool.</summary>
		public static int ActiveThreads { get { return _inUseThreads; } }
		/// <summary>Gets the number of callback delegates currently waiting in the thread pool.</summary>
		public static int WaitingCallbacks { get { lock(_poolLock) { return _waitingCallbacks.Count; } } }
		#endregion

		#region Thread Processing
		/// <summary>Event raised when there is an exception on a threadpool thread.</summary>
		public static event UnhandledExceptionEventHandler UnhandledException;

		/// <summary>A thread worker function that processes items from the work queue.</summary>
		private static void ProcessQueuedItems()
		{
			TimeSpan timeout;
			TimeSpan infinite = new TimeSpan(1, 0, 0, 0, 0);

			// Process indefinitely
			while (true)
			{
				_workerThreadNeeded.WaitOne();

				// Get the next item in the queue.  If there is nothing there, go to sleep
				// for a while until we're woken up when a callback is waiting.
				WaitingCallback callback = null;

				// Test the time when the next item should be executed and eventually remove the entry
				IPriorityQueueEntry queueEntry;
				do
				{
					queueEntry = null;
					lock(_poolLock)
					{
						try { queueEntry = (IPriorityQueueEntry)_waitingCallbacks.Peek(); }
						catch {}
						if (queueEntry != null && (long)queueEntry.Key <= Simulator.Timer.Now)
						{
							// the callback should be executed, remove the entry
							callback = (WaitingCallback)((IPriorityQueueEntry)_waitingCallbacks.Pop()).Value;

							// let other thread know that there is a next entry which should be possibly processed
							waitingForTimerEvent.Set();
						}
					}

					if (queueEntry != null && callback == null) // we should wait
					{
						timeout = infinite;
						if (Interlocked.Exchange(ref waitingForTimer, 1) == 0) // let one thread go and check the pool
						{
							long delay = (long)queueEntry.Key - Simulator.Timer.Now;
							
							if (delay < 0)
								timeout = TimeSpan.Zero;
							else
								timeout = new TimeSpan(delay);
						}

						waitingForTimerEvent.WaitOne(timeout, false);

						if (!timeout.Equals(infinite)) // if we have set waitingForTimer, reset it
							Interlocked.Exchange(ref waitingForTimer, 0);
					}
						
				} while (queueEntry != null && callback == null);

				if (callback == null)
					continue;

				// We now have a callback.  Execute it.  Make sure to accurately
				// record how many callbacks are currently executing.
				try 
				{
					Interlocked.Increment(ref _inUseThreads);
					callback.Callback(callback.State);
				} 
				catch(Exception exc)
				{
					try
					{
						UnhandledExceptionEventHandler handler = UnhandledException;
						if (handler != null) handler(typeof(ManagedThreadPool), new UnhandledExceptionEventArgs(exc, false));
					}
					catch{}
				}
				finally
				{
					Interlocked.Decrement(ref _inUseThreads);
				}
			}
		}
		#endregion

		/// <summary>Used to hold a callback delegate and the state for that delegate.</summary>
		private class WaitingCallback
		{
			#region Member Variables
			/// <summary>Callback delegate for the callback.</summary>
			private WaitCallback _callback;
			/// <summary>State with which to call the callback delegate.</summary>
			private object _state;
			#endregion

			#region Construction
			/// <summary>Initialize the callback holding object.</summary>
			/// <param name="callback">Callback delegate for the callback.</param>
			/// <param name="state">State with which to call the callback delegate.</param>
			public WaitingCallback(WaitCallback callback, object state)
			{
				_callback = callback;
				_state = state;
			}
			#endregion

			#region Properties
			/// <summary>Gets the callback delegate for the callback.</summary>
			public WaitCallback Callback { get { return _callback; } }
			/// <summary>Gets the state with which to call the callback delegate.</summary>
			public object State { get { return _state; } }
			#endregion
		}
	}
}
