/*

 Copyright (c) 2006 Pavel Novak.

 The use and distribution terms for this software are contained in the file named License.txt, 
 which can be found in the root of the NetSim distribution. By using this software 
 in any fashion, you are agreeing to be bound by the terms of this license.
 
 You must not remove this notice from this software.

*/

using System;
using System.Collections;
using System.Net.Sockets;
using System.Net;
using ManualResetEvent = System.Threading.ManualResetEvent;
using Timeout = System.Threading.Timeout;
using System.Diagnostics;
using System.Threading;

namespace NetSim.Core
{
	/// <summary>
	/// The interface that must be implemented by the class that manages <see cref="NetSimSocket"/>
	/// instances.
	/// </summary>
	public interface ISocketModule
	{
		/// <summary>
		/// Checks if the <paramref name="socket"/> binding is correct.
		/// </summary>
		/// <remarks>
		/// Requires socket local endpoint to be set. Checks if it contains local address and unused port
		/// number. Note that an endpoint can contain <see cref="System.Net.IPAddress.Any"/> or port 0.
		/// In case of any it is left to be assigned later, in case of port equal to zero unused port number
		/// is assigned.
		/// </remarks>
		/// <param name="socket">The socket to be bound to the endpoint.</param>
		/// <returns><b>True</b> if the binding is correct.</returns>
		bool CheckBind(NetSimSocket socket);

		/// <summary>
		/// Creates new TCP connection to the <paramref name="remoteEP"/>.
		/// </summary>
		/// <param name="socket">Socket to connect.</param>
		/// <param name="remoteEP">Remote end point to connect to.</param>
		void Connect(NetSimSocket socket, IPEndPoint remoteEP);

		/// <summary>
		/// Sends data in <paramref name="buffer"/> to the <paramref name="socket"/>.
		/// </summary>
		/// <param name="socket">Socket to use.</param>
		/// <param name="buffer">An array containing data to send.</param>
		/// <param name="offset">Index to <paramref name="buffer"/> where the data start.</param>
		/// <param name="size">Length of data to send.</param>
		/// <param name="socketFlags">Flags that should be used to modify send operation behavior.</param>
		/// <param name="remoteEP">Remote destination where the data should be sent.</param>
		/// <returns>Number of bytes sent.</returns>
		int SendTo(NetSimSocket socket, byte[] buffer, int offset, int size, SocketFlags socketFlags, IPEndPoint remoteEP);

		/// <summary>
		/// Closes the socket and releases the port number, which can be used by another socket.
		/// </summary>
		/// <param name="socket">Socket to close.</param>
		void Close(NetSimSocket socket);

		/// <summary>
		/// Registers new socket with the socket manager.
		/// </summary>
		/// <param name="socket">New socket to be registered.</param>
		void RegisterNewSocket(NetSimSocket socket);

		/// <summary>
		/// Sends notification to the remote endpoint that we have accepted the connection (SYN + ACK).
		/// </summary>
		/// <param name="socket">The newly created socket for this connection.</param>
		void Accept(NetSimSocket socket);

		/// <summary>
		/// Notifies the remote endpoint that we have sent all data on the specified socket.
		/// </summary>
		/// <param name="socket">Socket to use.</param>
		void ShutdownSend(NetSimSocket socket);

		/// <summary>
		/// We have received information that a remote endpoint finished sending; this mehtod changes
		/// a state of the socket appropriately.
		/// </summary>
		/// <param name="socket">Socket to use.</param>
		void ShutdownReceive(NetSimSocket socket);
	}

	/// <summary>
	/// Contains info about the connection state, sequence numbers, window sizes etc. While modifying, the instance
	/// should be locked.
	/// </summary>
	public class ConnectionInfo
	{
		/// <summary>
		/// Creates new <see cref="ConnectionInfo"/>; instances can be created only within this assembly.
		/// </summary>
		/// <param name="socket">The socket about which the connection info has information.</param>
		internal ConnectionInfo(NetSimSocket socket)
		{
			this.socket = socket;
		}

		/// <summary>
		/// Possible states of the socket. See Comer, 5-th edition, page 223, for details and transitions between states.
		/// </summary>
		public enum State
		{
			Closed,
			Listen,
			SynRecvd,
			SynSent,
			Established,
			CloseWait,
			LastAck,
			FinWait1,
			FinWait2,
			Closing,
			TimedWait
		}

		/// <summary>
		/// Gets or sets current socket state. The socket is initially closed (<see cref="State.Closed"/>).
		/// </summary>
		public State CurrentState
		{
			get { return currentState; }
			set
			{
				if (currentState == value)
					return;

				currentState = value;

				// do some required actions
				switch (currentState)
				{
					case State.Established:
						// notify the thread waiting for connection
						this.socket.ConnectCallback();
						break;
				}
			}
		}
		private State currentState = State.Closed;

		/// <summary>
		/// The start position of the window for outgoing stream. It is equal to the first unacknowledged byte.
		/// </summary>
		/// <remarks>
		/// This starts always at 0 as opposed to the specification. This behaviour might change in future releases.
		/// </remarks>
		//Future version: Should be initialized by random number, according to the specification? Check RFCs.
		// (if so, assign the same number to OutNewDataSeqNum.
		public UInt32 OutWindowStart
		{
			get { return outWindowStart; }
			set
			{
				if (outWindowStart != value) // new acknowledgement received, remove acknowledged packets
					this.SentPackets.RemoveToSeqNum(value);

				outWindowStart = value;
			}
		}
		private UInt32 outWindowStart = 0;

		/// <summary>
		/// The length of the window for outgoing stream. This length is changed according to the value
		/// specified by the remote side in the "window" field of incoming packets.
		/// </summary>
		public UInt16 OutWindowLen;

		/// <summary>
		/// The sequence number in outgoing stream that should be assigned to new data that will be sent.
		/// </summary>
		private UInt32 outNewDataSeqNum = 0;
		public UInt32 OutNewDataSeqNum(UInt32 shift)
		{
			lock (this)
			{
				UInt32 result = outNewDataSeqNum;
				unchecked { outNewDataSeqNum += shift; } // allow overflow
				return result;
			}
		}

		/// <summary>
		/// The start position of the window for incoming stream. This is equal to the "ack number" that
		/// is being sent to the remote side - it means the byte that is expected to be received next.
		/// </summary>
		public UInt32 InWindowStart;
		public void InWindowMove(UInt32 how)
		{
			lock (this)
			{
				InWindowStart = PacketReorder.SeqNumIncrement(InWindowStart, how);
			}
		}

		/// <summary>
		/// The length of the window for incoming stream. It is equal to the value assigned to "window"
		/// field advertised to the remote side - the length of our buffer.
		/// </summary>
		public UInt16 InWindowLen = UInt16.MaxValue; // default value - maximum possible

		/// <summary>
		/// The collection of packets received out of order, ordered by the sequence number.
		/// </summary>
		public PacketReorder InOutOfOrderPackets = new PacketReorder();

		/// <summary>
		/// Wheather the outgoing stream has been closed.
		/// </summary>
		public bool SendEnd = false;

		/// <summary>
		/// Wheather the incoming stream has been closed.
		/// </summary>
		public bool ReceiveEnd
		{
			get
			{
				return receiveEnd;
			}
			set
			{
				receiveEnd = value;
				if (receiveEnd)
					socket.waitForData.Set(); // notify threads waiting for data that no data are available
			}
		}
		private bool receiveEnd = false;

		/// <summary>
		/// Collection of packets that were sent and are stored for possible retransmission.
		/// </summary>
		public PacketsCache SentPackets
		{
			get { return sentPackets; }
		}
		private PacketsCache sentPackets = new PacketsCache();

		/// <summary>
		/// The <see cref="NetSimSocket"/> that references this <see cref="ConnectionInfo"/>.
		/// </summary>
		private NetSimSocket socket;
	}

	/// <summary>
	/// Storage for <see cref="PacketInfo"/> objects.
	/// </summary>
	public class PacketsCache
	{
		/// <summary>
		/// Entry for one packet in the <see cref="PacketsCache"/>.
		/// </summary>
		private class Entry
		{
			/// <summary>
			/// The time of <see cref="SimulationTimer"/> when the packet was sent.
			/// </summary>
			public long Time;

			/// <summary>
			/// <see cref="PacketInfo"/> that was sent.
			/// </summary>
			public PacketInfo PacketInfo;

			/// <summary>
			/// Sequence number acknowledgement that is needed for acking this packet
			/// </summary>
			public UInt32 SeqNumAckNeeded;

			/// <summary>
			/// How many times the packet should be resent.
			/// </summary>
			public int ResendCount;

			public Entry(long time, PacketInfo pInfo, UInt32 seqNumAckNeeded, int resendCount)
			{
				this.Time = time;
				this.PacketInfo = pInfo;
				this.SeqNumAckNeeded = seqNumAckNeeded;
				this.ResendCount = resendCount;
			}
		}

		/// <summary>
		/// List of <see cref="PacketsCache.Entry"/> objects, ordered by the insertion order,
		/// which is assumed to be the same as sequence numbers order.
		/// </summary>
		// generics: <TcpPacketsCache.Entry>
		private ArrayList packets = ArrayList.Synchronized(new ArrayList());

		public void Add(PacketInfo pInfo, UInt32 seqNumAckNeeded, int resendCount)
		{
			if (resendCount > 0) // it makes sense to remember the packet only if it should be resent
				packets.Add(new Entry(Simulator.Timer.Now, pInfo, seqNumAckNeeded, resendCount));
		}

		/// <summary>
		/// Removes all packets that are older than the sequnece number specified.
		/// </summary>
		/// <param name="seqNum">Last acknowledged sequence number.</param>
		public void RemoveToSeqNum(UInt32 seqNum)
		{
			lock (packets.SyncRoot)
			{
				while (true)
				{
					if (packets.Count == 0) // empty collection
						break;

					Entry e = (Entry)packets[0];
					if (PacketReorder.SeqNumCompare(e.SeqNumAckNeeded,	seqNum) > 0)
						break; // the packet is newer

					// remove the packet and continue
					packets.RemoveAt(0);
				}
			}
		}

		/// <summary>
		/// If there is a packet older than <paramref name="ticksDelay"/>, its PacketInfo is returned.
		/// Otherwise returns <b>null</b>.
		/// </summary>
		/// <param name="ticksDelay">The min age of the packet.</param>
		/// <param name="lastTry">Set to <b>true</b> if the returned <code>PacketInfo</code> should be
		/// sent last time and if no ack comes, no further retransmission should be performed.</param>
		/// <returns>First old packet found.</returns>
		public PacketInfo GetOldPacket(long ticksDelay, out bool lastTry)
		{
			lastTry = false;
			long now = Simulator.Timer.Now;

			lock (packets)
			{
				foreach (Entry e in packets)
				{
					if (Simulator.Timer.GetTimeAfterDelay(ticksDelay, e.Time) < now)
					{
						// old packet found, update time and return it
						e.Time = now;
						e.ResendCount--;
						if (e.ResendCount == 0)
						{
							lastTry = true;
							packets.Remove(e);
						}

						return e.PacketInfo;
					}
				}
			}

			// no old packet found
			return null;
		}

	}

	/// <summary>
	/// Class reordering the incoming packets that came out-of-order. Incoming packets are added to the
	/// instance of this class, it stores them until all previous packets come and than they can be removed
	/// in the right order.
	/// </summary>
	public class PacketReorder
	{
		/// <summary>
		/// Data delivered out of order, sorted by the sequence number.
		/// </summary>
		// generics: <UInt32, byte[]>
		private SortedList queuedPackets = new SortedList();
		
		/// <summary>
		/// The sequence number just after the last byte in the stream. Valid only if <see cref="endOfData"/>
		/// is set to <b>true</b>.
		/// </summary>
		private UInt32 lastSeqNum;

		/// <summary>
		/// <b>True</b> if the end-of-data notification has been received. In that case <see cref="lastSeqNum"/>
		/// member is valid.
		/// </summary>
		private bool endOfData = false;

		/// <summary>
		/// Adds new incoming packet to the packet reorder.
		/// </summary>
		/// <param name="packetSeqNum">Sequence number of the incoming packet.</param>
		/// <param name="data">Incoming packet data.</param>
		/// <param name="expectedSeqNum">Sequence number that was expected (the next that should be passed to the application).</param>
		/// <param name="doNotReturnPacket"><b>True</b> if no packet should be returned in <paramref name="outData"/>,
		/// even if there are some available in the right order.</param>
		/// <param name="lastPacketIn"><b>True</b> if this packet is the last packet.</param>
		/// <param name="outData">Data that are available in the right order. They will NOT be returned again!</param>
		/// <param name="lastDataOut">Returns <b>true</b> if there will be no data to receive.</param>
		/// <returns>Number of bytes returned in <paramref name="outData"/>.</returns>
		[System.Runtime.CompilerServices.MethodImpl(System.Runtime.CompilerServices.MethodImplOptions.Synchronized)]
		public int Add(UInt32 packetSeqNum, byte[] data, UInt32 expectedSeqNum, bool doNotReturnPacket,
			bool lastPacketIn, out byte[] outData, out bool lastDataOut)
		{
			outData = null;
			lastDataOut = false;

			if (lastPacketIn)
			{
				endOfData = true;
				lastSeqNum = SeqNumIncrement(packetSeqNum, (UInt32)data.Length);
			}

			if (SeqNumCompare(SeqNumIncrement(packetSeqNum, (UInt32)data.Length), expectedSeqNum) < 0)
				// incoming data are old
				return 0;

			// insert to the queue
			byte[] present = (byte[])queuedPackets[packetSeqNum];
			if (present != null)
			{
				// sequence number already in the queue, if the incoming data are longer, replace it, otherwise
				// do nothing, this is duplicate
				if (data.Length > present.Length)
					queuedPackets[packetSeqNum] = data;
			}
			else if (data.Length > 0)// not present right now, just add (it may overlap with another, that's OK)
			{
				queuedPackets[packetSeqNum] = data;
			}
			
			int len = 0;

			if (!doNotReturnPacket) // try to extract next data
				len = ExtractData(expectedSeqNum, out outData);
			
			if (endOfData)
			{
				UInt32 next_seq_num = SeqNumIncrement(expectedSeqNum, (UInt32)len);
				if (SeqNumCompare(lastSeqNum, next_seq_num) <= 0)
					lastDataOut = true;
			}

			return len;
		}

		/// <summary>
		/// Compares two sequence numbers. Takes overflow in consideration.
		/// </summary>
		/// <remarks>
		/// Important for comparation is the distance between those two numbers, zero immediately following
		/// <see cref="UInt32.MaxValue"/>. Where the distance is shorter, the first number (in the sequence
		/// 0 - 1 - 2 - ... 500 ... UInt32.MaxValue - 0 - 1 ...) is considered smaller than the second.
		/// Hence. <see cref="UInt32.MaxValue"/> is "smaller" than zero, even UInt32.MaxValue/2+1 is smaller
		/// than zero.
		/// </remarks>
		/// <param name="seq1">First sequence number to compare.</param>
		/// <param name="seq2">Second sequence number.</param>
		/// <returns>Zero if <paramref name="seq1"/> is equal to <paramref name="seq2"/>, number less than zero
		/// if <paramref name="seq1"/> is less than <paramref name="seq2"/>, otherwise the number greater than
		/// zero.</returns>
		public static int SeqNumCompare(UInt32 seq1, UInt32 seq2)
		{
			if (seq1 == seq2)
				return 0;

			UInt32 dist = seq1 < seq2 ? seq2 - seq1 : seq1 - seq2;
			if (dist <= UInt32.MaxValue/2) // we can compare as usual
				return seq1 < seq2 ? -1 : 1;
			else // distance over UInt32.MaxValue is smaller - compare the opposite way
				return seq1 < seq2 ? 1 : -1;
		}

		/// <summary>
		/// Increments the sequence number <paramref name="seqNum"/> by the value <paramref name="how"/>
		/// and returns the value.
		/// </summary>
		/// <param name="seqNum">Sequence number to increment.</param>
		/// <param name="how">The value how much the number should be incremented.</param>
		/// <returns>New value.</returns>
		public static UInt32 SeqNumIncrement(UInt32 seqNum, UInt32 how)
		{
			unchecked // allow overflow
			{
				return seqNum + how;
			}
		}

		/// <summary>
		/// Returns the distance between two sequence numbers. Takes overflow into account.
		/// </summary>
		/// <remarks>
		/// Returns the value Z that <c>last == SeqNumIncrement(first, Z);</c>. (This is valid
		/// for positive numbers since <see cref="SeqNumIncrement"/> does not allow for negative number.
		/// </remarks>
		/// <param name="first">First sequence number.</param>
		/// <param name="last">Second sequence number.</param>
		/// <returns>The distance between numbers.</returns>
		public static int SeqNumDifference(UInt32 first, UInt32 last)
		{
			if (first <= last)
				return (int)(last - first);

			// sequence numbers have overflown here
			return (int) ((UInt32.MaxValue - (first - last)) + 1);
		}

		/// <summary>
		/// Returns the number of bytes available to be read.
		/// </summary>
		/// <param name="startSeqNum">Sequence number where the data should start.</param>
		/// <returns>Nuber of bytes that are available.</returns>
		[System.Runtime.CompilerServices.MethodImpl(System.Runtime.CompilerServices.MethodImplOptions.Synchronized)]
		public int Available(UInt32 startSeqNum)
		{
			UInt32 last_seq_num = SeqNumIncrement(startSeqNum, 1); // just following sequence number
			for (int i = 0; i < queuedPackets.Count; i++)
			{
				UInt32 i_key =  (UInt32) queuedPackets.GetKey(i);
				byte[] i_value = (byte[]) queuedPackets.GetByIndex(i);

				UInt32 right_border = SeqNumIncrement(i_key, (UInt32)i_value.Length);
				if (SeqNumCompare(i_key, last_seq_num) < 0
					&& SeqNumCompare(last_seq_num, right_border) < 0)
				{
					last_seq_num = right_border;
				}
			}

			return SeqNumDifference(startSeqNum, last_seq_num) - 1;
		}

		/// <summary>
		/// Tries to find as much as possible data available starting at <paramref name="startSeqNum"/>.
		/// </summary>
		/// <param name="startSeqNum">Sequence number of the first byte that will be returned.</param>
		/// <param name="data">Returns an array of available bytes.</param>
		/// <returns>Number of bytes in <paramref name="data"/> array.</returns>
		private int ExtractData(UInt32 startSeqNum, out byte[] data)
		{
			int found_index = -1;
			for (int i = 0; i < queuedPackets.Count; i++)
			{
				UInt32 i_key =  (UInt32) queuedPackets.GetKey(i);
				byte[] i_value = (byte[]) queuedPackets.GetByIndex(i);
				if (SeqNumCompare(i_key, startSeqNum) <= 0
					&& SeqNumCompare(startSeqNum, SeqNumIncrement(i_key, (UInt32)i_value.Length)) < 0)
				{
					found_index = i;
					break;
				}
			}

			if (found_index == -1)
			{
				data = null;
				return 0;
			}

			// there is entry at index "found_index" that overlaps with requested seq number
			int len = GetContinuousBytesCount(found_index, startSeqNum);

			if (len == 0)
			{
				data = null;
				return 0;
			}

			data = new byte[len];
			int len2 = GetDataFrom(found_index, startSeqNum, data);
			Debug.Assert(len == len2); // buffer exactly filled out

			return len2;
		}

		private int GetDataFrom(int index, UInt32 seqNumStart, byte[] data_out)
		{
			int data_out_pos = 0;
			UInt32 cur_seq_num = seqNumStart;

			while (true)
			{
				UInt32 seq_num = (UInt32) queuedPackets.GetKey(index);
				byte[] data = (byte[]) queuedPackets.GetByIndex(index);

				if (SeqNumCompare(seq_num, cur_seq_num) <= 0) // this packet seq number is previous - data are for us
				{
					UInt32 next_seq_num = SeqNumIncrement(seq_num, (UInt32)data.Length);
					if (SeqNumCompare(cur_seq_num, next_seq_num) < 0) // this packet overlap
					{
						int buf_start_at = (int) unchecked(cur_seq_num - seq_num);
						int len = data.Length - buf_start_at;

						Buffer.BlockCopy(data, buf_start_at, data_out, data_out_pos, len);
						data_out_pos += len;
						cur_seq_num = SeqNumIncrement(cur_seq_num, (UInt32)len);
					}
					else // this whole packet is duplicate of some data in the past, do nothing, it will be removed..
					{
					}

					queuedPackets.RemoveAt(index);

					if (queuedPackets.Count == 0)
						break; // no more data
					else
						index = index % queuedPackets.Count; // one packet removed, no shift, but this is needed for end of the queue handling
				}
				else
				{
					break; // no more data available
				}
			}
			
			return data_out_pos;
		}

		/// <summary>
		/// Returns the number of bytes that are available starting at index <paramref name="index"/> in
		/// the <see cref="queuedPackets"/>.
		/// </summary>
		/// <param name="index">Queued packet index to start searching.</param>
		/// <param name="seqNumStart">Sequence number of the first byte of continuous sequence of available data.</param>
		/// <returns></returns>
		private int GetContinuousBytesCount(int index, UInt32 seqNumStart)
		{
			int start_index = index;
			int sum = 0;
			UInt32 cur_seq_num = seqNumStart;

			while (true)
			{
				UInt32 seq_num = (UInt32) queuedPackets.GetKey(index);
				byte[] data = (byte[]) queuedPackets.GetByIndex(index);

				if (SeqNumCompare(seq_num, cur_seq_num) <= 0) // this packet seq number is previous - data are for us
				{
					UInt32 next_seq_num = SeqNumIncrement(seq_num, (UInt32)data.Length);
					if (SeqNumCompare(cur_seq_num, next_seq_num) < 0) // this packet overlap
					{
						int add = data.Length - (int)unchecked(cur_seq_num - seq_num);
						sum += add;
						cur_seq_num = SeqNumIncrement(cur_seq_num, (UInt32)add);

					}
					else // this whole packet is duplicate of some data in the past, do nothing, it will be removed..
					{
					}

					index = (index+1) % queuedPackets.Count; // go to the next packet
					if (start_index == index)
						break; // all packets processed
				}
				else
				{
					break; // no more data available
				}
			}
			
			return sum;

		}
	}

	/// <summary>
	/// Socket at the virtual node. Implements the same interface as <see cref="Socket"/>. Used directly
	/// by virtual applications written direcly for the simulator running in the same process as the
	/// simulator, or indirecly by <see cref="RemoteSocketsManager"/>.
	/// </summary>
	public class NetSimSocket
	{
		#region AsyncIOState, IOAsyncResult

		internal class AsyncIOState
		{
			public byte[] Buffer;
			public int Offset;
			public int Size;
			public SocketFlags SocketFlags;
			public AsyncCallback Callback;
			public IOAsyncResult AsyncResult;

			public AsyncIOState(byte[] buffer, int offset, int size, SocketFlags socketFlags,
				AsyncCallback callback, IOAsyncResult asyncResult)
			{
				this.Buffer = buffer;
				this.Offset = offset;
				this.Size = size;
				this.SocketFlags = socketFlags;
				this.Callback = callback;
				this.AsyncResult = asyncResult;
			}
		}

		internal class IOAsyncResult : IAsyncResult
		{
			private object state;
			private bool completedSynchronously = false;
			private ManualResetEvent waitHandle;
			private bool isCompleted = false;
			public int BytesTransferred;
			public EndPoint EndPoint;

			/// <summary>
			/// If an error occured, the appropriate exception is set here. Otherwise this field remains <b>null</b>.
			/// </summary>
			public Exception Exception;

			public IOAsyncResult(object state)
			{
				this.state = state;
			}

			#region IAsyncResult Members
			
			public object AsyncState
			{
				get { return this.state; }
			}

			public bool CompletedSynchronously
			{
				get { return this.completedSynchronously; }
			}

			public WaitHandle AsyncWaitHandle
			{
				get
				{
					lock (this)
					{
            this.waitHandle = new ManualResetEvent(isCompleted);
						return waitHandle;
					}
				}
			}

			public bool IsCompleted
			{
				get { return this.isCompleted; }
			}

			#endregion

			internal void SetCompleted()
			{
				lock (this)
				{
					this.isCompleted = true;
					if (waitHandle != null)
						waitHandle.Set();
				}
			}
		}
		#endregion

		/// <summary>
		/// Basic information about the connection. Is <b>null</b> for connectionless sockets.
		/// </summary>
		public ConnectionInfo CInfo { get { return this.connectionInfo; } }
		private ConnectionInfo connectionInfo;

		#region Constants
		/// <summary>
		/// The name of this module specified in the configuration file should be always this name.
		/// It is used for retrieving reference to the module while creating a socket (for example from
		/// the remote computer).
		/// </summary>
		private const string StdSocketModuleName = "SocketModule";
		
		/// <summary>
		/// How long wait for connecting an connection-oriented socket (in miliseconds).
		/// </summary>
		private const int ConnectTimeout = 30000;
		#endregion

		#region Windows Sockets Error Codes

		//
		// Windows socket error codes used to provide information when SocketException is thrown.
		//
		internal const int WSAEINTR = 10004;
		internal const int WSAEACCES = 10013;
		internal const int WSAEFAULT = 10014;
		internal const int WSAEINVAL = 10022;
		internal const int WSAEMFILE = 10024;
		internal const int WSAEWOULDBLOCK = 10035;
		internal const int WSAEINPROGRESS = 10036;
		internal const int WSAEALREADY = 10037;
		internal const int WSAENOTSOCK = 10038;
		internal const int WSAEDESTADDRREQ = 10039;
		internal const int WSAEMSGSIZE = 10040;
		internal const int WSAEPROTOTYPE = 10041;
		internal const int WSAENOPROTOOPT = 10042;
		internal const int WSAEPROTONOSUPPORT = 10043;
		internal const int WSAESOCKTNOSUPPORT = 10044;
		internal const int WSAEOPNOTSUPP = 10045;
		internal const int WSAEPFNOSUPPORT = 10046;
		internal const int WSAEAFNOSUPPORT = 10047;
		internal const int WSAEADDRINUSE = 10048;
		internal const int WSAEADDRNOTAVAIL = 10049;
		internal const int WSAENETDOWN = 10050;
		internal const int WSAENETUNREACH = 10051;
		internal const int WSAENETRESET = 10052;
		internal const int WSAECONNABORTED = 10053;
		internal const int WSAECONNRESET = 10054;
		internal const int WSAENOBUFS = 10055;
		internal const int WSAEISCONN = 10056;
		internal const int WSAENOTCONN = 10057;
		internal const int WSAESHUTDOWN = 10058;
		internal const int WSAETIMEDOUT = 10060;
		internal const int WSAECONNREFUSED = 10061;
		internal const int WSAEHOSTDOWN = 10064;
		internal const int WSAEHOSTUNREACH = 10065;
		internal const int WSAEPROCLIM = 10067;
		internal const int WSASYSNOTREADY = 10091;
		internal const int WSAVERNOTSUPPORTED = 10092;
		internal const int WSANOTINITIALISED = 10093;
		internal const int WSAEDISCON = 10101;
		internal const int WSATYPE_NOT_FOUND = 10109;
		internal const int WSAHOST_NOT_FOUND = 11001;
		internal const int WSATRY_AGAIN = 11002;
		internal const int WSANO_RECOVERY = 11003;
		internal const int WSANO_DATA = 11004;

		/// <summary>Value of INVALID_SOCKET from WinSock library.</summary>
		public const int INVALID_SOCKET = -1;

		#endregion

		#region Socket options

		/// <summary>
		/// Send buffer length. Default value is 64kB.
		/// </summary>
		private int OptSendBuffer = 64 * 1024;

		#endregion

		/// <summary>
		/// Counter that is incremented by one and specified the unique ID of NetSim socket.
		/// </summary>
		private static int idCounter = 1;
		private static object _idCounterLock = new object();
		private static int GetNextID()
		{
			lock (_idCounterLock) { return idCounter++; }
		}

		/// <summary>
		/// Unique number of this socket - used for identification passed to remote computers etc.
		/// </summary>
		public int ID { get { return this.id; } }
		private int id;

		//
		// State of the socket
		//

		/// <summary>Indicates whether the socket has been closed.</summary>
		private bool closed; // = false
		/// <summary>Whether the socket is connected.</summary>
		public bool Connected
		{
			get { return connected; }
		}
		private bool connected; // = false

		//
		// Members initialized by the constructor
		//

		public AddressFamily AddressFamily { get { return this.addressFamily; } }
		private AddressFamily addressFamily;

		public SocketType SocketType { get { return this.socketType; } }
		private SocketType socketType;

		public ProtocolType ProtocolType { get { return this.protocolType; } }
		private ProtocolType protocolType;

		public bool SupportsIPv4 { get { return true; } }
		public bool SupportsIPv6 { get { return false; } } // IPv6 not supported yet

		//
		// synchronization objects for blocking user threads while waiting for data etc.
		//

		/// <summary>
		/// Used to block the thread calling <see cref="Connect"/> at the connection-oriented socket.
		/// </summary>
		internal ManualResetEvent waitForConnect = new ManualResetEvent(true);

		/// <summary>
		/// Used to block the thread calling <see cref="Receive"/> if no data is available and the socket
		/// is in blocking mode.
		/// </summary>
		internal ManualResetEvent waitForData = new ManualResetEvent(true);

		//
		// Queues for data
		//

		/// <summary>
		/// Entry used in the queues for data waiting to be sent or received.
		/// </summary>
		protected class DataQueueEntry
		{
			public byte[] Data;
			public int Offset;
			public int Length;
			public IPEndPoint RemoteEP;

			public DataQueueEntry(byte[] data, int offset, int length, IPEndPoint remoteEP)
			{
				this.Data = data;
				this.Offset = offset;
				this.Length = length;
				this.RemoteEP = remoteEP;
			}
		}

		// generics: <DataQueueEntry>
		protected Queue incomingData;
		// generics: <DataQueueEntry>
		protected Queue incomingOutOfBandData;

		protected class ConnectQueueEntry
		{
			public NetSimSocket Socket;

			public ConnectQueueEntry(NetSimSocket sock)
			{
				this.Socket = sock;
			}
		}

		// generics: <ConnectQueueEntry>
		protected Queue incomingConnectionRequests;

		/// <summary>
		/// Gets the amount of data that has been received from the network and is available to be read.
		/// </summary>
		public int Available
		{
			get
			{
				return CInfo.InOutOfOrderPackets.Available(CInfo.InWindowStart);
			}
		}

		public bool Blocking
		{
			get { return blocking; }
			set { blocking = value; }
		}
		private bool blocking = true; // default is to block

		/// <summary>
		/// Original documentation says: Gets the operating system handle for the Socket.
		/// Of course, this property return the simulator internal identification, not the OS handle.
		/// </summary>
		public IntPtr Handle
		{
			get { return new IntPtr(this.ID); }
		}

		/// <summary>
		/// Returns <b>true</b> if the socket was bound to address <see cref="System.Net.IPAddress.Any"/>.
		/// In that case, we should receive packets from any interface, even if the local EP would have
		/// been set to some of them in the future. While selecting the socket for packet reception,
		/// this property is taken in consideration.
		/// </summary>
		public bool BoundToAny
		{
			get { return this.boundToAny; }
		}
		private bool boundToAny;

		public EndPoint LocalEndPoint
		{
			get { return localEndPoint; }
		}
		private EndPoint localEndPoint;

		/// <summary>
		/// See the note for <see cref="Bind"/> with two arguments.
		/// </summary>
		private bool keepAny = false;

		public EndPoint RemoteEndPoint
		{
			get { return remoteEndPoint; }
		}
		private EndPoint remoteEndPoint;

		/// <summary>
		/// The <see cref="ISocketModule"/> instance this socket is attached to - data sent and received
		/// by this socket are sent through this module.
		/// </summary>
		private ISocketModule owningModule;

		public NetSimSocket(AddressFamily addressFamily, SocketType socketType,
			ProtocolType protocolType, Node node)
		{
			incomingData = Queue.Synchronized(new Queue());
			incomingOutOfBandData = Queue.Synchronized(new Queue());

			// get SocketModule
			IModule module = (IModule)node.GetItem(NodeItemKind.Module, StdSocketModuleName);
			if (module == null)
				throw new SocketException(NetSimSocket.WSAENETDOWN);
			ISocketModule sm = module as ISocketModule;
			if (sm == null)
				throw new SocketException(NetSimSocket.WSAENETDOWN);

			this.owningModule = sm;
			this.id = GetNextID();
			SetSocketParams(addressFamily, socketType, protocolType);

			if (socketType == SocketType.Stream)
			{
				connectionInfo = new ConnectionInfo(this);
				incomingConnectionRequests = Queue.Synchronized(new Queue());
			}

			this.owningModule.RegisterNewSocket(this);
		}

		private NetSimSocket(AddressFamily addressFamily, SocketType socketType,
			ProtocolType protocolType, ISocketModule module)
		{
			incomingData = Queue.Synchronized(new Queue());
			incomingOutOfBandData = Queue.Synchronized(new Queue());

			this.owningModule = module;

			this.id = GetNextID();
			SetSocketParams(addressFamily, socketType, protocolType);

			if (protocolType == ProtocolType.Tcp)
			{
				connectionInfo = new ConnectionInfo(this);
				incomingConnectionRequests = Queue.Synchronized(new Queue());
			}

			this.owningModule.RegisterNewSocket(this);
		}

		/// <summary>
		/// Called by the constructor to initialize socket type.
		/// </summary>
		/// <param name="addressFamily">Address family of the new socket.</param>
		/// <param name="socketType">Socket type of the new socket.</param>
		/// <param name="protocolType">Protocol type of the new socket.</param>
		private void SetSocketParams(AddressFamily addressFamily, SocketType socketType,
			ProtocolType protocolType)
		{
			bool error = false;
			switch (addressFamily)
			{
					//---------------------
					// IPv4
				case AddressFamily.InterNetwork:
				switch (protocolType)
				{
					case ProtocolType.Udp:
						if (socketType != SocketType.Dgram) // UDP requires datagram socket type
							error = true;
						break;
					case ProtocolType.Tcp:
						if (socketType != SocketType.Stream) // TCP requires stream socket type
							error = true;
						break;
					case ProtocolType.IP:
						if (socketType == SocketType.Stream)
							protocolType = ProtocolType.Tcp;
						else if (socketType == SocketType.Dgram)
							protocolType = ProtocolType.Udp;
                        else if (socketType == SocketType.Raw)
						{
							// no direct communication with underlying protocol supported yet
							error = true;
						}
						else
							error = true;

						break;
				}
					break;
				
					//--------------------
					// IPv6
				case AddressFamily.InterNetworkV6:
					goto default; // IPv6 not supported yet

					//--------------------
					// Others
				default:
					error = true; // others not supported
					break;
			}

			if (error)
				throw new SocketException(NetSimSocket.WSAENETDOWN);

			// remember values
			this.addressFamily = addressFamily;
			this.socketType = socketType;
			this.protocolType = protocolType;
		}

		public override int GetHashCode()
		{
			return this.ID; // this unique ID is very well hash code
		}

		/// <summary>
		/// Creates a new Socket for a newly created connection.
		/// </summary>
		/// <returns></returns>
		public NetSimSocket Accept()
		{
			ConnectQueueEntry queue_entry = null;

			if (this.closed)
				throw new ObjectDisposedException(this.GetType().FullName);
			if (this.ProtocolType != ProtocolType.Tcp) // only TCP supports accepting connections
				throw new SocketException(NetSimSocket.WSAENETDOWN);
			if (CInfo.CurrentState != ConnectionInfo.State.Listen) // only listening socket can accept connections
				throw new SocketException(NetSimSocket.WSAENETDOWN);
			if (this.LocalEndPoint == null) // socket not bound
				throw new InvalidOperationException("Socket is not bound, call Bind() first");

			while (queue_entry == null)
			{
				// check first while not locked
				if (incomingConnectionRequests.Count == 0)
				{
					if (this.blocking)
					{
						waitForConnect.Reset();
						waitForConnect.WaitOne();
					}
					else
					{
						throw new SocketException(NetSimSocket.WSAENETDOWN);
					}
				}

				lock (incomingConnectionRequests.SyncRoot)
				{
					if (incomingConnectionRequests.Count > 0) // double check
					{
						// queue is locked, there is at least one item
						queue_entry = (ConnectQueueEntry) this.incomingConnectionRequests.Dequeue();
					}
				}
			}

			
			return queue_entry.Socket;
		}

		/// <summary>
		/// Begins an asynchronous operation to accept an incoming connection attempt.
		/// </summary>
		/// <param name="callback"></param>
		/// <param name="state"></param>
		/// <returns></returns>
		public IAsyncResult BeginAccept(AsyncCallback callback, object state)
		{
			throw new NotImplementedException();
		}

		/// <summary>
		/// Begins an asynchronous request for a remote host connection.
		/// </summary>
		/// <param name="remoteEP"></param>
		/// <param name="callback"></param>
		/// <param name="state"></param>
		/// <returns></returns>
		public IAsyncResult BeginConnect(EndPoint remoteEP,	AsyncCallback callback,	object state)
		{
			throw new NotImplementedException();
		}

		/// <summary>
		/// Begins to asynchronously receive data from a connected Socket.
		/// </summary>
		/// <param name="buffer"></param>
		/// <param name="offset"></param>
		/// <param name="size"></param>
		/// <param name="socketFlags"></param>
		/// <param name="callback"></param>
		/// <param name="state"></param>
		/// <returns></returns>
		public IAsyncResult BeginReceive(byte[] buffer, int offset, int size,
			SocketFlags socketFlags, AsyncCallback callback, object state)
		{
			// check parameters
			if (buffer == null)
				throw new ArgumentNullException("buffer", "Buffer is null.");
			if (this.closed)
				throw new ObjectDisposedException(this.GetType().FullName);
			if (offset < 0 || offset >= buffer.Length)
				throw new ArgumentOutOfRangeException("offset");
			if (size < 0 || (size > buffer.Length - offset))
				throw new ArgumentOutOfRangeException("size");

			// create end point
			IPEndPoint remote_ep = (IPEndPoint)RemoteEndPoint;
			EndPoint ep = new IPEndPoint(remote_ep.Address, remote_ep.Port);

			return BeginReceiveFrom(buffer, offset, size, socketFlags, ref ep, callback, state);
		}

		/// <summary>
		/// Begins to asynchronously receive data from a specified network device.
		/// </summary>
		/// <param name="buffer"></param>
		/// <param name="offset"></param>
		/// <param name="size"></param>
		/// <param name="socketFlags"></param>
		/// <param name="remoteEP"></param>
		/// <param name="callback"></param>
		/// <param name="state"></param>
		/// <returns></returns>
		public IAsyncResult BeginReceiveFrom(byte[] buffer, int offset, int size,
			SocketFlags socketFlags, ref EndPoint remoteEP, AsyncCallback callback, object state)
		{
			// check parameters
			if (buffer == null)
				throw new ArgumentNullException("buffer", "Buffer is null.");
			if (remoteEP == null)
				throw new ArgumentNullException("remoteEP", "Remote endpoint is null");
			if (this.closed)
				throw new ObjectDisposedException(this.GetType().FullName);
			if (offset < 0 || offset >= buffer.Length)
				throw new ArgumentOutOfRangeException("offset");
			if (size < 0 || (size > buffer.Length - offset))
				throw new ArgumentOutOfRangeException("size");

			IOAsyncResult async_res = new IOAsyncResult(state);
			async_res.EndPoint = remoteEP;
			AsyncIOState astate = new AsyncIOState(buffer, offset, size, socketFlags, callback, async_res);
			ThreadPool.QueueUserWorkItem(new WaitCallback(this.ProcessBeginReceiveFrom), astate);
			
			return async_res;
		}

		private void ProcessBeginReceiveFrom(object s)
		{
			// get data
			AsyncIOState state = (AsyncIOState)s;

			// perform receive
			try
			{
				state.AsyncResult.BytesTransferred = ReceiveFrom(state.Buffer, state.Offset, state.Size, state.SocketFlags, ref state.AsyncResult.EndPoint);
			}
			catch (SocketException e)
			{
				state.AsyncResult.BytesTransferred = 0;
				state.AsyncResult.Exception = e;
			}
						
			// notify the calling application
			state.AsyncResult.SetCompleted();
			state.Callback(state.AsyncResult);
		}

		/// <summary>
		/// Sends data asynchronously to a connected Socket.
		/// </summary>
		/// <param name="buffer"></param>
		/// <param name="offset"></param>
		/// <param name="size"></param>
		/// <param name="socketFlags"></param>
		/// <param name="callback"></param>
		/// <param name="state"></param>
		/// <returns></returns>
		public IAsyncResult BeginSend(byte[] buffer, int offset, int size,
			SocketFlags socketFlags, AsyncCallback callback, object state)
		{
			return BeginSendTo(buffer, offset, size, socketFlags, this.RemoteEndPoint, callback, state);
		}

		/// <summary>
		/// Sends data asynchronously to a specific remote host.
		/// </summary>
		/// <param name="buffer"></param>
		/// <param name="offset"></param>
		/// <param name="size"></param>
		/// <param name="socketFlags"></param>
		/// <param name="remoteEP"></param>
		/// <param name="callback"></param>
		/// <param name="state"></param>
		/// <returns></returns>
		public IAsyncResult BeginSendTo(byte[] buffer, int offset, int size,
			SocketFlags socketFlags, EndPoint remoteEP, AsyncCallback callback, object state)
		{
			IOAsyncResult async_res = new IOAsyncResult(state);
			async_res.EndPoint = remoteEP;
			AsyncIOState astate = new AsyncIOState(buffer, offset, size, socketFlags, callback, async_res);
			ThreadPool.QueueUserWorkItem(new WaitCallback(this.ProcessBeginSendTo), astate);

			return async_res;
		}

		private void ProcessBeginSendTo(object s)
		{
			// get data
			AsyncIOState state = (AsyncIOState)s;

			// perform send
			try
			{
				state.AsyncResult.BytesTransferred = SendTo(state.Buffer, state.Offset, state.Size,
					state.SocketFlags, state.AsyncResult.EndPoint);
			}
			catch (SocketException e)
			{
				state.AsyncResult.BytesTransferred = 0;
				state.AsyncResult.Exception = e;
			}
						
			// notify the calling application
			state.AsyncResult.SetCompleted();
			state.Callback(state.AsyncResult);
		}

		/// <summary>
		/// Associates a Socket with a local endpoint.
		/// </summary>
		/// <param name="localEP"></param>
		public void Bind(EndPoint localEP)
		{
			if (localEP == null)
				throw new ArgumentNullException("localEP");
			
			IPEndPoint ip_ep = localEP as IPEndPoint;
			if (ip_ep == null)
				throw new ArgumentException("IPEndPoint instance expected", "localEP");

			if (this.closed)
				throw new ObjectDisposedException(this.GetType().FullName);

			if (localEndPoint != null) // socket is already bound
				throw new SocketException(NetSimSocket.WSAENETDOWN);

			// if at any time the local address is set to Any, it should receive packets from any interface
			if (ip_ep.Address.Equals(System.Net.IPAddress.Any))
				boundToAny = true;

			localEndPoint = ip_ep;

			if (!owningModule.CheckBind(this))
			{
				localEndPoint = null;
				boundToAny = false;
				throw new SocketException(NetSimSocket.WSAENETDOWN);
			}
		}

		/// <summary>
		/// Associates a Socket with a local endpoint.
		/// </summary>
		/// <remarks>
		/// This is a NetSim extension to the regular <see cref="System.Net.Sockets.Socket.Bind"/>
		/// method. Requires <paramref name="localEP"/> to be set to <see cref="System.Net.IPAddress.Any"/>
		/// if <paramref name="keepAny"/> is set to <b>true</b>. This never assigns the local endpoint
		/// to any valid address - the result is that any outgoing packet will have source address
		/// set to the address of an interface where it is sent. Useful for some network communication
		/// such as routing etc. Can be used only with connectionless sockets.
		/// </remarks>
		/// <param name="localEP">Local endpoint to bind to.</param>
		/// <param name="keepAny">Set to <b>true</b> if local address won't be assigned.</param>
		public void Bind(EndPoint localEP, bool keepAny)
		{
			if (!keepAny)
			{
				Bind(localEP);
				return;
			}

			IPEndPoint ip_ep = localEP as IPEndPoint;
			if (ip_ep == null)
				throw new ArgumentException("Requires an instance of IPEndPoint", "localEP");
			if (!ip_ep.Address.Equals(System.Net.IPAddress.Any))
				throw new ArgumentException("Requires IPAddress.Any", "localEP");
			if (this.SocketType == SocketType.Stream)
				throw new InvalidOperationException("Cannot be used with connection-oriented socket");

			Bind(localEP);
			this.keepAny = true;
		}

		/// <summary>
		/// Disconnects the socket immediately, remote side sent "reset".
		/// </summary>
		public void Reset()
		{
			// async select handling
			if ((currentAsyncSelectEvents & AsyncSelectEvent.Close) != 0)
			{
				asyncSelectCallback(this, AsyncSelectEvent.Close);
				currentAsyncSelectEvents &= ~AsyncSelectEvent.Close; // send close only once
			}

			this.Close();
		}

		/// <summary>
		/// Closes the Socket connection and releases all associated resources.
		/// </summary>
		public void Close()
		{
			this.connected = false;
			this.closed = true;

			// release the port number
			this.owningModule.Close(this);

			// discard all pending data
			this.incomingData.Clear();
			this.incomingOutOfBandData.Clear();
		}

		/// <summary>
		/// Establishes a connection to a remote host.
		/// </summary>
		/// <param name="remoteEP"></param>
		public void Connect(EndPoint remoteEP)
		{
			if (remoteEP == null)
				throw new ArgumentNullException("remoteEP");

			if (closed)
				throw new ObjectDisposedException(typeof(NetSimSocket).FullName);

			IPEndPoint ip_ep = remoteEP as IPEndPoint;
			if (ip_ep == null) // only IP end point is supported
				throw new SocketException(NetSimSocket.WSAENETDOWN);

			if (this.ProtocolType == ProtocolType.Tcp) // connection to the remote point needed
			{
				if (!this.Blocking)
					throw new SocketException(NetSimSocket.WSAENETDOWN); // need time to connect

				this.owningModule.Connect(this, ip_ep);
				waitForConnect.Reset();
				if (!waitForConnect.WaitOne(ConnectTimeout, false)) // timeout elapsed
					throw new SocketException(NetSimSocket.WSAENETDOWN);
			}

			this.remoteEndPoint = remoteEP;
			this.connected = true;
		}

		/// <summary>
		/// Called by the <see cref="ISocketModule"/> to notify the socket that it has been connected.
		/// </summary>
		internal void ConnectCallback()
		{
			waitForConnect.Set();
		}

		/// <summary>
		/// Asynchronously accepts an incoming connection attempt and creates a new Socket
		/// to handle remote host communication.
		/// </summary>
		/// <param name="asyncResult"></param>
		/// <returns></returns>
		public Socket EndAccept(
			IAsyncResult asyncResult
			)
		{
			throw new NotImplementedException();
		}

		/// <summary>
		/// Ends a pending asynchronous connection request.
		/// </summary>
		/// <param name="asyncResult"></param>
		public void EndConnect(
			IAsyncResult asyncResult
			)
		{
			throw new NotImplementedException();
		}

		/// <summary>
		/// Ends a pending asynchronous read.
		/// </summary>
		/// <param name="asyncResult"></param>
		/// <returns></returns>
		public int EndReceive(IAsyncResult asyncResult)
		{
			// create end point
			IPEndPoint remote_ep = (IPEndPoint)RemoteEndPoint;
			EndPoint ep = new IPEndPoint(remote_ep.Address, remote_ep.Port);

			return EndReceiveFrom(asyncResult, ref ep);
		}

		/// <summary>
		/// Ends a pending asynchronous read from a specific endpoint.
		/// </summary>
		/// <param name="asyncResult"></param>
		/// <param name="endPoint"></param>
		/// <returns></returns>
		public int EndReceiveFrom(IAsyncResult asyncResult, ref EndPoint endPoint)
		{
			IOAsyncResult async_res = asyncResult as IOAsyncResult;
			if (async_res == null)
				throw new ArgumentException("Invalid value.", "asyncResult");

			if (!async_res.IsCompleted) // block until it is completed
			{
				async_res.AsyncWaitHandle.WaitOne();
			}

			if (async_res.Exception != null) // error?
				throw async_res.Exception;

			endPoint = async_res.EndPoint;
			return async_res.BytesTransferred;
		}

		/// <summary>
		/// Ends a pending asynchronous send.
		/// </summary>
		/// <param name="asyncResult"></param>
		/// <returns></returns>
		public int EndSend(IAsyncResult asyncResult)
		{
			return EndSendTo(asyncResult);
		}

		/// <summary>
		/// Ends a pending asynchronous send to a specific location.
		/// </summary>
		/// <param name="asyncResult"></param>
		/// <returns></returns>
		public int EndSendTo(IAsyncResult asyncResult)
		{
			IOAsyncResult async_res = asyncResult as IOAsyncResult;
			if (async_res == null)
				throw new ArgumentException("Invalid value.", "asyncResult");

			if (!async_res.IsCompleted) // block until it is completed
			{
				async_res.AsyncWaitHandle.WaitOne();
			}

			if (async_res.Exception != null) // error?
				throw async_res.Exception;

			return async_res.BytesTransferred;
		}

		/// <summary>
		/// Returns the value of a specified Socket option, represented as an object.
		/// </summary>
		/// <param name="optionLevel"></param>
		/// <param name="optionName"></param>
		/// <returns></returns>
		public object GetSocketOption(SocketOptionLevel optionLevel, SocketOptionName optionName)
		{
			switch (optionLevel)
			{
					//
					// SOCKET option level
					//

				case SocketOptionLevel.Socket:
				switch (optionName)
				{
					case SocketOptionName.SendBuffer:
						return OptSendBuffer;
					
					case SocketOptionName.Type:
					switch (this.SocketType)
					{
						case SocketType.Stream: return 1;
						case SocketType.Dgram: return 2;
						case SocketType.Raw: return 3;
						case SocketType.Rdm: return 4;
						case SocketType.Seqpacket: return 5;
					}
						break;
				}
					break;

					//
					// IP option level
					//

				case SocketOptionLevel.IP:
				switch (optionName)
				{
					default:
						break;
				}
					break;

					//
					// IPv6 option level
					//

				case SocketOptionLevel.IPv6:
				switch (optionName)
				{
					default:
						break;
				}
					break;

					//
					// TCP option level
					//

				case SocketOptionLevel.Tcp:
				switch (optionName)
				{
					default:
						break;
				}
					break;

					//
					// UDP option level
					//

				case SocketOptionLevel.Udp:
				switch (optionName)
				{
					default:
						break;
				}
					break;
			}

			// no such option
			return null;
		}

		/// <summary>
		/// Returns the specified Socket option setting, represented as a byte array.
		/// </summary>
		/// <param name="optionLevel"></param>
		/// <param name="optionName"></param>
		/// <param name="optionValue"></param>
		public void GetSocketOption(SocketOptionLevel optionLevel, SocketOptionName optionName,	byte[] optionValue)
		{
			object val = GetSocketOption(optionLevel, optionName);

			if (val is Int32)
			{
				if (optionValue.Length >= 4)
				{
					Int32 v = (Int32)val;
					v = IPAddress.HostToNetworkOrder(v);
					optionValue[3] = (byte)(v & 0xFF); v >>= 1;
					optionValue[2] = (byte)(v & 0xFF); v >>= 1;
					optionValue[1] = (byte)(v & 0xFF); v >>= 1;
					optionValue[0] = (byte)(v & 0xFF);
				}
			}
			else
			{
					throw new SocketException(NetSimSocket.WSAENETDOWN);
			}
		}

		/// <summary>
		/// Returns the value of the specified Socket option in an array.
		/// </summary>
		/// <param name="optionLevel"></param>
		/// <param name="optionName"></param>
		/// <param name="optionLength"></param>
		/// <returns></returns>
		public byte[] GetSocketOption(SocketOptionLevel optionLevel, SocketOptionName optionName, int optionLength)
		{
			byte[] val = new byte[optionLength];
			GetSocketOption(optionLevel, optionName, val);
			return val;
		}

		/// <summary>
		/// Sets low-level operating modes for the Socket.
		/// </summary>
		/// <param name="ioControlCode"></param>
		/// <param name="optionInValue"></param>
		/// <param name="optionOutValue"></param>
		/// <returns></returns>
		public int IOControl(
			int ioControlCode,
			byte[] optionInValue,
			byte[] optionOutValue
			)
		{
			throw new NotImplementedException();
		}

		/// <summary>
		/// Places a Socket in a listening state.
		/// </summary>
		/// <param name="backlog"></param>
		public void Listen(int backlog)
		{
			if (this.closed)
				throw new ObjectDisposedException(this.GetType().FullName);
			
			if (this.ProtocolType != ProtocolType.Tcp) // only TCP supports listening
				throw new SocketException(NetSimSocket.WSAENETDOWN);

			if (this.LocalEndPoint == null)
				throw new SocketException(NetSimSocket.WSAENETDOWN); // socket not bound

			if (this.CInfo.CurrentState != ConnectionInfo.State.Closed)
				throw new SocketException(NetSimSocket.WSAENETDOWN); // only closed socket can change its state to listening mode

			this.CInfo.CurrentState = ConnectionInfo.State.Listen;
		}

		/// <summary>
		/// Determines the status of the Socket.
		/// </summary>
		/// <param name="microSeconds"></param>
		/// <param name="mode"></param>
		/// <returns></returns>
		public bool Poll(
			int microSeconds,
			SelectMode mode
			)
		{
			throw new NotImplementedException();
		}

		/// <summary>
		/// Receives data from a bound Socket into a receive buffer.
		/// </summary>
		/// <param name="buffer"></param>
		/// <returns></returns>
		public int Receive(byte[] buffer)
		{
			return Receive(buffer, 0, buffer.Length, SocketFlags.None);
		}

		/// <summary>
		/// Receives data from a bound Socket into a receive buffer, using the specified SocketFlags.
		/// </summary>
		/// <param name="buffer"></param>
		/// <param name="socketFlags"></param>
		/// <returns></returns>
		public int Receive(byte[] buffer, SocketFlags socketFlags)
		{
			return Receive(buffer, 0, buffer.Length, socketFlags);
		}

		/// <summary>
		/// Receives the specified number of bytes of data from a bound Socket into a receive buffer,
		/// using the specified SocketFlags.
		/// </summary>
		/// <param name="buffer"></param>
		/// <param name="size"></param>
		/// <param name="socketFlags"></param>
		/// <returns></returns>
		public int Receive(byte[] buffer, int size, SocketFlags socketFlags)
		{
			return Receive(buffer, 0, size, socketFlags);
		}

		/// <summary>
		/// Receives the specified number of bytes from a bound Socket into the specified
		/// offset position of the receive buffer, using the specified SocketFlags.
		/// </summary>
		/// <param name="buffer"></param>
		/// <param name="offset"></param>
		/// <param name="size"></param>
		/// <param name="socketFlags"></param>
		/// <returns></returns>
		public int Receive(byte[] buffer, int offset, int size, SocketFlags socketFlags)
		{
			// Is RemoteEndPoint set? ie. has Connect or Accept been called?
			if (this.RemoteEndPoint == null)
				throw new SocketException(NetSimSocket.WSAENETDOWN);
			
			IPEndPoint ip_ep = new IPEndPoint(System.Net.IPAddress.Any, 0);
			EndPoint ep = (EndPoint)ip_ep;
			return ReceiveFrom(buffer, offset, size, socketFlags, ref ep);
		}

		/// <summary>
		/// Receives a datagram into the data buffer and stores the endpoint.
		/// </summary>
		/// <param name="buffer"></param>
		/// <param name="remoteEP"></param>
		/// <returns></returns>
		public int ReceiveFrom(byte[] buffer, ref EndPoint remoteEP)
		{
			return ReceiveFrom(buffer, 0, buffer.Length, SocketFlags.None, ref remoteEP);
		}

		/// <summary>
		/// Receives a datagram into the data buffer, using the specified SocketFlags, and stores the endpoint.
		/// </summary>
		/// <param name="buffer"></param>
		/// <param name="socketFlags"></param>
		/// <param name="remoteEP"></param>
		/// <returns></returns>
		public int ReceiveFrom(byte[] buffer, SocketFlags socketFlags, ref EndPoint remoteEP)
		{
			return ReceiveFrom(buffer, 0, buffer.Length, socketFlags, ref remoteEP);
		}

		/// <summary>
		/// Receives the specified number of bytes into the data buffer, using the specified SocketFlags, and stores the endpoint.
		/// </summary>
		/// <param name="buffer"></param>
		/// <param name="size"></param>
		/// <param name="socketFlags"></param>
		/// <param name="remoteEP"></param>
		/// <returns></returns>
		public int ReceiveFrom(byte[] buffer, int size, SocketFlags socketFlags, ref EndPoint remoteEP)
		{
			return ReceiveFrom(buffer, 0, size, socketFlags, ref remoteEP);
		}

		/// <summary>
		/// Receives the specified number of bytes of data into the specified location of the data buffer,
		/// using the specified SocketFlags, and stores the endpoint.
		/// </summary>
		/// <param name="buffer"></param>
		/// <param name="offset"></param>
		/// <param name="size"></param>
		/// <param name="socketFlags"></param>
		/// <param name="remoteEP"></param>
		/// <returns></returns>
		public int ReceiveFrom(byte[] buffer, int offset, int size, SocketFlags socketFlags, ref EndPoint remoteEP)
		{
			IPEndPoint ip_remote = remoteEP as IPEndPoint;

			// if connection-oriented, remoteEP is ignored
			if (this.ProtocolType == ProtocolType.Tcp)
			{
				ip_remote = (IPEndPoint)this.RemoteEndPoint;
				if (ip_remote == null || !connected)
					throw new SocketException(NetSimSocket.WSAENETDOWN); // socket not connected
			}

			//
			// check arguments
			//
			if (buffer == null)
				throw new ArgumentNullException("buffer");
			if (remoteEP == null)
				throw new ArgumentNullException("remoteEP");
			if (offset < 0 || offset >= buffer.Length)
				throw new ArgumentOutOfRangeException("offset");
			if (size < 0 || (size > buffer.Length - offset))
				throw new ArgumentOutOfRangeException("size");
			if (ip_remote == null)
				throw new ArgumentException("IPEndPoint expected", "remoteEP");

			if (this.closed)
				throw new ObjectDisposedException(this.GetType().FullName);
			if (this.LocalEndPoint == null)
				throw new SocketException(NetSimSocket.WSAENETDOWN);

			// Out-of-band data?
			Queue q = this.incomingData;
			if ((socketFlags & SocketFlags.OutOfBand) != 0)
				q = this.incomingOutOfBandData;

			DataQueueEntry entry;
			int copy_offset, copy_count;

			while (true)
			{
				if (q.Count == 0) // no data available
				{
					if (CInfo != null && CInfo.ReceiveEnd) // end of data
					{
						if ((asyncSelectEvents & AsyncSelectEvent.Close) != 0)
							asyncSelectCallback(this, AsyncSelectEvent.Close);

						return 0;
					}

					if (blocking)
					{
						bool should_wait = false;
						lock (q.SyncRoot)
						{
							if (q.Count == 0) // double check
							{
								should_wait = true;
								waitForData.Reset(); // set non-signaled state
							}
						}
						//TODO: this is not right solution - imagine that right here is something added to the queue
						if (should_wait)
						{
							this.waitForData.WaitOne(); // and wait
						}
					}
					else
						throw new SocketException(NetSimSocket.WSAENETDOWN);
				}
				else
				{
					lock (q.SyncRoot)
					{
						if (q.Count > 0) // double check
						{
							// data available
							entry = (DataQueueEntry)q.Peek();
							copy_offset = entry.Offset;
							copy_count = Math.Min(entry.Length, size);
							entry.Offset += copy_count;
							entry.Length -= copy_count;
							if ((entry.Length == 0 || this.SocketType == SocketType.Dgram) // all entry copied, or rest of dgram should be discarded
								&& (socketFlags & SocketFlags.Peek) == 0) // and no Peek flag is specified
								q.Dequeue();

							break; // exit the infinite loop
						}
					}
				}
			}

			// async select handling
			if (q == incomingData && (asyncSelectEvents & AsyncSelectEvent.Read) != 0) // regular data
			{
				if (q.Count > 0) // more data - send notification
					this.asyncSelectCallback(this, AsyncSelectEvent.Read);
				else // no more data - enable notification
					this.currentAsyncSelectEvents |= (asyncSelectEvents & AsyncSelectEvent.Read);
			}
			if ((asyncSelectEvents & AsyncSelectEvent.OutOfBand) != 0) // oob data
			{
				if (incomingOutOfBandData.Count > 0) // more oob data - send notification
					this.asyncSelectCallback(this, AsyncSelectEvent.OutOfBand);
				else // no more data - enable notification
					this.currentAsyncSelectEvents |= (asyncSelectEvents & AsyncSelectEvent.OutOfBand);
			}

			// copy data
			Buffer.BlockCopy(entry.Data, copy_offset, buffer, offset, copy_count);
			remoteEP = entry.RemoteEP;

			// if this is datagram transfer and not all data have been copied, we should notify the caller
			// that some data were discarded
			if (entry.Length > 0 && (this.SocketType == SocketType.Dgram || this.SocketType == SocketType.Rdm))
				throw new SocketException(NetSimSocket.WSAENETDOWN);

			return copy_count;
		}

		/// <summary>
		/// Determines the status of one or more sockets.
		/// </summary>
		/// <param name="checkRead"></param>
		/// <param name="checkWrite"></param>
		/// <param name="checkError"></param>
		/// <param name="microSeconds"></param>
		public static void Select(IList checkRead, IList checkWrite, IList checkError, int microSeconds)
		{
			if (
				(checkRead == null || checkRead.Count == 0)
				&& (checkWrite == null || checkWrite.Count == 0)
				&& (checkError == null || checkError.Count == 0)
				)
				throw new ArgumentNullException();

			// lists of sockets that will be selected
			ArrayList r = new ArrayList();
			ArrayList w = new ArrayList();
			ArrayList e = new ArrayList();

			// if we would select any socket, and no timeout elapsed, we will set the counter that is being
			// decremented by one each loop to "downcount_init" and keep trying checking sockets until the
			// counter reaches zero
			int downcount = int.MaxValue;
			int downcount_init = 0;

			long timeout = Simulator.Timer.GetTimeAfterDelay(10*(long)microSeconds);
			while (true)
			{
				// check read sockets
				if (checkRead != null)
				{
					int i = 0;
					while (i < checkRead.Count)
					{
						NetSimSocket sock = (NetSimSocket)checkRead[i];
						if (sock.CheckSelectRead())
						{
							r.Add(sock);
							checkRead.RemoveAt(i);
							if (downcount > downcount_init)
								downcount = downcount_init;
						}
						else 
						{
							i++; // increment only if no socket removed (removal shifts indexes to the left by one)
						}
					}
				}

				// check write sockets
				if (checkWrite != null)
				{
					int i = 0;
					while (i < checkWrite.Count)
					{
						NetSimSocket sock = (NetSimSocket)checkWrite[i];
						if (sock.CheckSelectWrite())
						{
							w.Add(sock);
							checkWrite.RemoveAt(i);
							if (downcount > downcount_init)
								downcount = downcount_init;
						}
						else
						{
							i++; // increment only if no socket removed (removal shifts indexes to the left by one)
						}
					}
				}

				// check error sockets
				if (checkError != null)
				{
					int i = 0;
					while (i < checkError.Count)
					{
						NetSimSocket sock = (NetSimSocket)checkError[i];
						if (sock.CheckSelectError())
						{
							e.Add(sock);
							checkError.RemoveAt(i);
							if (downcount > downcount_init)
								downcount = downcount_init;
						}
						else
						{
							i++; // increment only if no socket removed (removal shifts indexes to the left by one)
						}
					}
				}

				// do we have some socket selected and should we stop trying find any next one?
				if (downcount == 0)
					break;

				// does our time elapsed?
				// WORKING: downcount > downcount_init (= always try find more sockets)
				if (downcount > downcount_init && timeout <= Simulator.Timer.Now)
					break;

				// otherwise wait some time and check again
				Thread.Sleep(1);

				downcount--;
			}

			// clear original lists
			checkRead.Clear();
			checkWrite.Clear();
			checkError.Clear();

			// and fill them with selected sockets
			foreach (object o in r)
				checkRead.Add(o);
			foreach (object o in w)
				checkWrite.Add(o);
			foreach (object o in e)
				checkError.Add(o);
		}

		/// <summary>
		/// Returns <b>true</b> if the socket should remain in "read" list while calling the <see cref="Select"/>
		/// method.
		/// </summary>
		/// <returns><b>True</b> if the socket is readable.</returns>
		private bool CheckSelectRead()
		{
			if (CInfo != null) // connection oriented
			{
				if (this.CInfo.CurrentState == ConnectionInfo.State.Listen)
				{
					if (this.incomingConnectionRequests.Count > 0) // call to Accept won't block
						return true;
				}
				else if (this.CInfo.ReceiveEnd)
				{
					return true; // incoming stream closed => receive will return just 0
				}
			}

			if (this.incomingData.Count > 0)
			{
				return true; // some data available
			}
			//WORKING: what about CInfo.InOutOfOrderPackets.Available(CInfo.InWindowStart) > 0

			return false;
		}

		/// <summary>
		/// Returns <b>true</b> if the socket should remain in "write" list while calling the <see cref="Select"/>
		/// method.
		/// </summary>
		/// <returns><b>True</b> if the socket is writeable.</returns>
		private bool CheckSelectWrite()
		{
			// Future version: implement finite buffer
			return this.Connected && (CInfo == null || !this.CInfo.SendEnd);
		}

		/// <summary>
		/// Returns <b>true</b> if the socket should remain in "error" list while calling the <see cref="Select"/>
		/// method.
		/// </summary>
		/// <returns><b>True</b> if the socket has an error condition.</returns>
		private bool CheckSelectError()
		{
			return this.CInfo.CurrentState == ConnectionInfo.State.Closed;
		}

		/// <summary>
		/// Sends data to a connected Socket.
		/// </summary>
		/// <param name="buffer"></param>
		/// <returns></returns>
		public int Send(byte[] buffer)
		{
			return Send(buffer, 0, buffer.Length, SocketFlags.None);
		}

		/// <summary>
		/// Sends data to a connected Socket using the specified SocketFlags.
		/// </summary>
		/// <param name="buffer"></param>
		/// <param name="socketFlags"></param>
		/// <returns></returns>
		public int Send(byte[] buffer, SocketFlags socketFlags)
		{
			return Send(buffer, 0, buffer.Length, socketFlags);
		}

		/// <summary>
		/// Sends the specified number of bytes of data to a connected Socket, using the specified SocketFlags.
		/// </summary>
		/// <param name="buffer"></param>
		/// <param name="size"></param>
		/// <param name="socketFlags"></param>
		/// <returns></returns>
		public int Send(byte[] buffer, int size, SocketFlags socketFlags)
		{
			return Send(buffer, 0, size, socketFlags);
		}

		/// <summary>
		/// Sends the specified number of bytes of data to a connected Socket, starting at the specified offset,
		/// and using the specified SocketFlags.
		/// </summary>
		/// <param name="buffer"></param>
		/// <param name="offset"></param>
		/// <param name="size"></param>
		/// <param name="socketFlags"></param>
		/// <returns></returns>
		public int Send(byte[] buffer, int offset, int size, SocketFlags socketFlags)
		{
			// Is RemoteEndPoint set? ie. has Connect or Accept been called?
			if (this.RemoteEndPoint == null)
				throw new SocketException(NetSimSocket.WSAENETDOWN);

			return SendTo(buffer, offset, size, socketFlags, this.RemoteEndPoint);
		}

		/// <summary>
		/// Sends data to the specified endpoint.
		/// </summary>
		/// <param name="buffer"></param>
		/// <param name="remoteEP"></param>
		/// <returns></returns>
		public int SendTo(byte[] buffer, EndPoint remoteEP)
		{
			return SendTo(buffer, 0, buffer.Length, SocketFlags.None, remoteEP);
		}

		/// <summary>
		/// Sends data to a specific endpoint using the specified SocketFlags.
		/// </summary>
		/// <param name="buffer"></param>
		/// <param name="socketFlags"></param>
		/// <param name="remoteEP"></param>
		/// <returns></returns>
		public int SendTo(byte[] buffer, SocketFlags socketFlags, EndPoint remoteEP)
		{
			return SendTo(buffer, 0, buffer.Length, socketFlags, remoteEP);
		}

		/// <summary>
		/// Sends the specified number of bytes of data to the specified endpoint using the specified SocketFlags.
		/// </summary>
		/// <param name="buffer"></param>
		/// <param name="size"></param>
		/// <param name="socketFlags"></param>
		/// <param name="remoteEP"></param>
		/// <returns></returns>
		public int SendTo(byte[] buffer, int size, SocketFlags socketFlags, EndPoint remoteEP)
		{
			return SendTo(buffer, 0, size, socketFlags, remoteEP);
		}

		/// <summary>
		/// Sends the specified number of bytes of data to the specified endpoint, starting at the specified
		/// location in the buffer, and using the specified SocketFlags.
		/// </summary>
		/// <param name="buffer"></param>
		/// <param name="offset"></param>
		/// <param name="size"></param>
		/// <param name="socketFlags"></param>
		/// <param name="remoteEP"></param>
		/// <returns></returns>
		public int SendTo(byte[] buffer, int offset, int size, SocketFlags socketFlags, EndPoint remoteEP)
		{
			IPEndPoint ip_remote = remoteEP as IPEndPoint;

			// if connection-oriented, remoteEP is ignored
			if (this.ProtocolType == ProtocolType.Tcp)
			{
				ip_remote = (IPEndPoint)this.RemoteEndPoint;
				if (ip_remote == null || !connected)
					throw new SocketException(NetSimSocket.WSAENETDOWN); // socket not connected

				if (CInfo != null && CInfo.SendEnd) // outgoing stream closed
					throw new SocketException(NetSimSocket.WSAENETDOWN);
			}

			// check parameters
			if (buffer == null)
				throw new ArgumentNullException("buffer");
			if (remoteEP == null)
				throw new ArgumentNullException("remoteEP");
			if (offset < 0 || offset >= buffer.Length)
				throw new ArgumentOutOfRangeException("offset");
			if (size < 0 || (size > buffer.Length - offset))
				throw new ArgumentOutOfRangeException("size");
			if (ip_remote == null)
				throw new ArgumentException("IPEndPoint expected", "remoteEP");

			// check if the socket is still open
			if (this.closed)
				throw new ObjectDisposedException(this.GetType().FullName);

			return this.owningModule.SendTo(this, buffer, offset, size, socketFlags, ip_remote);
		}

		internal void SetSocketOption(SocketOptionLevel optionLevel, SocketOptionName optionName,
			IntPtr optVal, int optLen)
		{
			optLen = optLen;
		}

		/// <summary>
		/// Sets the specified Socket option to the specified value, represented as a byte array.
		/// </summary>
		/// <param name="optionLevel"></param>
		/// <param name="optionName"></param>
		/// <param name="optionValue"></param>
		public void SetSocketOption(
			SocketOptionLevel optionLevel,
			SocketOptionName optionName,
			byte[] optionValue
			)
		{
			throw new NotImplementedException();
		}

		/// <summary>
		/// Sets the specified Socket option to the specified integer value.
		/// </summary>
		/// <param name="optionLevel"></param>
		/// <param name="optionName"></param>
		/// <param name="optionValue"></param>
		public void SetSocketOption(
			SocketOptionLevel optionLevel,
			SocketOptionName optionName,
			int optionValue
			)
		{
			throw new NotImplementedException();
		}

		/// <summary>
		/// Sets the specified Socket option to the specified value, represented as an object.
		/// </summary>
		/// <param name="optionLevel"></param>
		/// <param name="optionName"></param>
		/// <param name="optionValue"></param>
		public void SetSocketOption(
			SocketOptionLevel optionLevel,
			SocketOptionName optionName,
			object optionValue
			)
		{
			throw new NotImplementedException();
		}

		/// <summary>
		/// Disables sends and receives on a Socket.
		/// </summary>
		/// <param name="how"></param>
		public void Shutdown(SocketShutdown how)
		{
			if (how == SocketShutdown.Send || how == SocketShutdown.Both)
			{
				owningModule.ShutdownSend(this);
			}

			if (how == SocketShutdown.Receive || how == SocketShutdown.Both)
			{
				owningModule.ShutdownReceive(this);
			}
		}

		public delegate void AsyncSelectCallbackDelegate(NetSimSocket s, AsyncSelectEvent e);
		private AsyncSelectCallbackDelegate asyncSelectCallback = null;
		
		/// <summary>
		/// Events monitored, exactly as the user specified.
		/// </summary>
		/// <remarks>The code modifying these enumerations is not thread safe, however, at
		/// worst sometime the message will be sent if not necessary, the code won't allow
		/// not sending the message if needed.</remarks>
		private AsyncSelectEvent asyncSelectEvents = AsyncSelectEvent.None;

		/// <summary>
		/// Sometime, even if the event should be monitored, if the event occurs, the message
		/// shouldn't be generated. In most cases, this is because the message would duplicate
		/// previous message.
		/// If the event is temporary disabled, it is disabled in this enumeration. This enumeration
		/// should be always a subset of <see cref="asyncSelectEvents"/>.
		/// </summary>
		private AsyncSelectEvent currentAsyncSelectEvents = AsyncSelectEvent.None;

		/// <summary>
		/// Called to notify the socket that it should send asynchronous select notifications.
		/// Which events occured is returned in callback state parameter, ie. the state param
		/// can be casted to <see cref="AsyncSelectEvent"/> enumeration.
		/// </summary>
		/// <param name="events"></param>
		/// <param name="callback"></param>
		public void AsyncSelect(AsyncSelectEvent events, AsyncSelectCallbackDelegate callback)
		{
			// sets the socket to non-blocking mode
			this.blocking = false;

			if (events != AsyncSelectEvent.None)
			{
				if (callback == null)
					throw new ArgumentNullException("callback", "callback shouldn't be null if some event is specified");

				this.asyncSelectCallback = callback; // remember callback
			}
			else
			{
				// reset callback if no event should be monitored (ignore parameter callback)
				this.asyncSelectCallback = null;
			}

			// remember events
			asyncSelectEvents = events;
			currentAsyncSelectEvents = events;
		}

		#region Communication with SocketModule

		/// <summary>
		/// If the local endpoint is not set, assigns specified parameters.
		/// Otherwise does nothing.
		/// </summary>
		/// <param name="address"></param>
		/// <param name="port"></param>
		public void SetLocalEndPoint(System.Net.IPAddress address, UInt16 port)
		{
			if (address == null || keepAny)
				address = System.Net.IPAddress.Any;

			if (this.localEndPoint == null)
			{
				this.localEndPoint = new IPEndPoint(address, port);
			}
			else
			{
				IPEndPoint ip_ep = (IPEndPoint)localEndPoint;
				IPAddress new_addr = ip_ep.Address;
				int new_port = ip_ep.Port;
				bool change = false;

				if (ip_ep.Address.Equals(System.Net.IPAddress.Any))
				{
					new_addr = address;
					change = true;
				}
					
				if (ip_ep.Port == 0)
				{
					new_port = port;
					change = true;
				}

				if (change)
					this.localEndPoint = new IPEndPoint(new_addr, new_port);
			}
		}

		/// <summary>
		/// If the local endpoint is not set, assigns specified parameters.
		/// Otherwise does nothing.
		/// </summary>
		/// <param name="address"></param>
		/// <param name="port"></param>
		public void SetRemoteEndPoint(System.Net.IPAddress address, UInt16 port)
		{
			if (this.remoteEndPoint == null)
				this.remoteEndPoint = new IPEndPoint(address, port);
			else
			{
				IPEndPoint ip_ep = (IPEndPoint)remoteEndPoint;
				IPAddress new_addr = ip_ep.Address;
				int new_port = ip_ep.Port;
				bool change = false;

				if (ip_ep.Address.Equals(System.Net.IPAddress.Any))
				{
					new_addr = address;
					change = true;
				}
					
				if (ip_ep.Port == 0)
				{
					new_port = port;
					change = true;
				}

				if (change)
					this.remoteEndPoint = new IPEndPoint(new_addr, new_port);
			}
		}

		/// <summary>
		/// Notifies the socket that new data are available.
		/// </summary>
		/// <param name="data">Incoming data.</param>
		/// <param name="flags">Flags modifying the connection or data processing.</param>
		/// <param name="remoteEP">The origin of the data.</param>
		public void NewData(byte[] data, SocketFlags flags, IPEndPoint remoteEP)
		{
			if (data.Length == 0)
				return;

			if (!ShouldAcceptFrom(remoteEP))
				return;
			
			if (CInfo != null && CInfo.ReceiveEnd)
				this.Close(); // no data should arrive, incoming stream was closed

			// determine the queue the data should be inserted in
			Queue q = ((flags & SocketFlags.OutOfBand) != 0) ? this.incomingOutOfBandData : this.incomingData;

			// enqueue data
			DataQueueEntry entry = new DataQueueEntry(data, 0, data.Length, remoteEP);
			q.Enqueue(entry);

			// wakeup any sleeping thread waiting for data
			waitForData.Set();

			// notify the application if async select was called
			if ((flags & SocketFlags.OutOfBand) == 0 && (currentAsyncSelectEvents & AsyncSelectEvent.Read) != 0) // not oob
			{
				this.asyncSelectCallback(this, AsyncSelectEvent.Read);
				currentAsyncSelectEvents &= ~AsyncSelectEvent.Read;
			}
			else if ((flags & SocketFlags.OutOfBand) != 0 && (currentAsyncSelectEvents & AsyncSelectEvent.OutOfBand) != 0) // not oob
			{
				this.asyncSelectCallback(this, AsyncSelectEvent.OutOfBand);
				currentAsyncSelectEvents &= ~AsyncSelectEvent.OutOfBand;
			}
		}

		/// <summary>
		/// Enqueues new incoming connection and wakes up any sleeping thread waiting for it.
		/// </summary>
		/// <remarks>
		/// This function is called by the socket module only if the socket is listening at given endpoint.
		/// We need to notify the caller that the connection will be accepted, therefore we create the
		/// socket here to store the information about the connection (not after <see cref="Accept"/> call).
		/// </remarks>
		/// <param name="remoteEP">Remote endpoint requesting the connection.</param>
		/// <param name="localEP">Local endpoint.</param>
		/// <param name="seqNumber">Sequence number of the request.</param>
		/// <param name="window">TCP window size announced.</param>
		public void NewConnectionRequest(IPEndPoint remoteEP, IPEndPoint localEP, UInt32 seqNumber, UInt16 window)
		{
			//
			// create new socket
			//
			NetSimSocket result = new NetSimSocket(this.AddressFamily, this.SocketType, this.ProtocolType, this.owningModule);

			// set endpoints
			result.SetLocalEndPoint(localEP.Address, (UInt16)localEP.Port);
			result.SetRemoteEndPoint(remoteEP.Address, (UInt16)remoteEP.Port);

			// original socket remains in listening state, new one moves to the next one
			result.CInfo.CurrentState = ConnectionInfo.State.SynRecvd;
			result.CInfo.InWindowStart = unchecked(seqNumber + 1);
			result.CInfo.InWindowLen = window;

			result.connected = true;

			// register new socket with socket module
			owningModule.RegisterNewSocket(result);

			//
			// let the remote endpoint know that we have accepted the connection
			//
			owningModule.Accept(result);

			this.incomingConnectionRequests.Enqueue(new ConnectQueueEntry(result));
			this.waitForConnect.Set();
		}

		/// <summary>
		/// Determines whether the data coming from <paramref name="remoteEP"/> should be accepted by
		/// this socket.
		/// </summary>
		/// <param name="remoteEP"></param>
		/// <returns><b>True</b> if the socket should accept the data, <b>false</b> otherwise.</returns>
		private bool ShouldAcceptFrom(IPEndPoint remoteEP)
		{
			if (this.RemoteEndPoint == null) // socket not connected to an address
			{
				if (this.SocketType == SocketType.Stream)
					return false;
				else
					return true; // dgram delivery doesn't need to be connected
			}
			else
			{
				IPEndPoint ep = (IPEndPoint)this.RemoteEndPoint;
				if (
					(ep.Address.Equals(System.Net.IPAddress.Any) || ep.Address.Equals(remoteEP.Address))
					&& (ep.Port == remoteEP.Port)
					)
					return true; // our remote EP match the parameter or accepts any IP address
				else
					return false;
			}
		}

		#endregion
	}


}
