/*

 Copyright (c) 2006 Pavel Novak.

 The use and distribution terms for this software are contained in the file named License.txt, 
 which can be found in the root of the NetSim distribution. By using this software 
 in any fashion, you are agreeing to be bound by the terms of this license.
 
 You must not remove this notice from this software.

*/

using System;

namespace NetSim.Core
{
	/// <summary>
	/// The instance of this object might be remotely accessed to control the simulator.
	/// </summary>
	public class RemoteControl : MarshalByRefObject
	{
		#region Nested types
		public interface ICallback
		{
			void NewEvent(SimulatorEvent e);
			void EnabledDisabled(BasicObjectKind objKind, string name, bool enabled);
		}

		public enum SimulatorEvent
		{
			NewLogEntry
		}

		/// <summary>
		/// Information about object status and which way it can be changed.
		/// </summary>
		[Serializable]
			public struct SuspendInfo
		{
			public bool CanDisable;
			public bool CanReconfigure;
			public bool Enabled;

			public SuspendInfo(bool canDisable, bool canReconfigure, bool enabled)
			{
				this.CanDisable = canDisable;
				this.CanReconfigure = canReconfigure;
				this.Enabled = enabled;
			}
		}
		#endregion

		private static ICallback callback;

		/// <summary>
		/// Creates new instance supporting the control from the other process.
		/// </summary>
		public RemoteControl()
		{
			Simulator.GetInstance().Events.NewEntryEvent += new EventLog.NewEntryDelegate(newLogEntry);
		}

		/// <summary>
		/// Registers callback object which methods are called by this control.
		/// </summary>
		/// <param name="c"></param>
		public void RegisterCallback(ICallback c)
		{
			callback = c;
		}

		/// <summary>
		/// Called by the simulator log when new entry is added.
		/// </summary>
		/// <param name="eventType"></param>
		private void newLogEntry(Event.EventType eventType)
		{
			// let the application or whatever know that new entry is in the log
			if (callback == null)
				return;

			try
			{
				callback.NewEvent(SimulatorEvent.NewLogEntry);
			}
			catch (Exception) // make sure not to fail here, remoting may be down or whatever
			{}
		}

		/// <summary>
		/// Pauses the simulator. It can be resumed by calling <see cref="Resume"/>.
		/// </summary>
		public void Pause()
		{
			Simulator.GetInstance().Pause();
		}

		/// <summary>
		/// Resumes the simulator previously paused by <see cref="Pause"/>.
		/// </summary>
		public void Resume()
		{
			Simulator.GetInstance().Resume();
		}

		/// <summary>
		/// Stops the simulator execution and frees all resources.
		/// </summary>
		public void Stop()
		{
			Simulator.GetInstance().Stop();
		}

		/// <summary>
		/// Returns <b>null</b>, therefore the lifetime lease never expires.
		/// </summary>
		/// <returns><b>Null</b>.</returns>
		public override object InitializeLifetimeService()
		{
			return null; // never expires
		}

		/// <summary>
		/// Returns one log entry from the log as out parameter.
		/// </summary>
		/// <param name="e">After the method returns <b>true</b>, contains a log entry.</param>
		/// <param name="remove"><b>True</b> if the log entry should be removed from the log file, <b>false</b> otherwise.</param>
		/// <returns><b>True</b> if the log entry has been succesfuly returned, <b>false</b> if there is no
		/// log entry available.</returns>
		public bool GetLogEntry(out NetSim.Core.Event e, bool remove)
		{
			return Simulator.GetInstance().Events.GetLogEntry(out e, remove);
		}

		public void GetSuspendInfo(out SuspendInfo info, BasicObjectKind basicKind, string basicName)
		{
			ISuspendable item = Simulator.GetInstance().GetItem(basicKind, basicName);
			if (item == null)
				throw new RemoteControlException("No such item found.");

			info = new SuspendInfo(item.CanDisable, item.CanReconfigure, item.Enabled);
		}

		public void GetSuspendInfo(out SuspendInfo info, NodeItemKind nodeItemKind, string itemName, string nodeName)
		{
			Node node = Simulator.GetInstance().GetNode(nodeName);
			if (node == null)
				throw new RemoteControlException("No such node found.");

			ISuspendable item = node.GetItem(nodeItemKind, itemName);
			if (item == null)
				throw new RemoteControlException(String.Format("No such item at the node {0} found.", nodeName));

			info = new SuspendInfo(item.CanDisable, item.CanReconfigure, item.Enabled);
		}

		public void SetEnabled(BasicObjectKind basicKind, string basicName, bool enabled)
		{
			ISuspendable item = Simulator.GetInstance().GetItem(basicKind, basicName);
			if (item == null)
				throw new RemoteControlException("No such item found.");

			try
			{
				item.Enabled = enabled;
			}
			catch (InvalidOperationException e)
			{
				throw new RemoteControlException("Cannot change enabled/disabled: " + e.Message, e);
			}
		}

		public void SetEnabled(NodeItemKind nodeItemKind, string itemName, string nodeName, bool enabled)
		{
			Node node = Simulator.GetInstance().GetNode(nodeName);
			if (node == null)
				throw new RemoteControlException("No such node found.");

			ISuspendable item = node.GetItem(nodeItemKind, itemName);
			if (item == null)
				throw new RemoteControlException(String.Format("No such item at the node {0} found.", nodeName));

			try
			{
				item.Enabled = enabled;
			}
			catch (InvalidOperationException e)
			{
				throw new RemoteControlException("Cannot change enabled/disabled: " + e.Message, e);
			}
		}

		public void Reconfigure(BasicObjectKind basicKind, string basicName)
		{
			ISuspendable item = Simulator.GetInstance().GetItem(basicKind, basicName);
			if (item == null)
				throw new RemoteControlException("No such item found.");

			try
			{
				item.Reconfigure();
			}
			catch (InvalidOperationException e)
			{
				throw new RemoteControlException("Cannot reconfigure: " + e.Message, e);
			}
		}

		public void Reconfigure(NodeItemKind nodeItemKind, string itemName, string nodeName)
		{
			Node node = Simulator.GetInstance().GetNode(nodeName);
			if (node == null)
				throw new RemoteControlException("No such node found.");

			ISuspendable item = node.GetItem(nodeItemKind, itemName);
			if (item == null)
				throw new RemoteControlException(String.Format("No such item at the node {0} found.", nodeName));

			try
			{
				item.Reconfigure();
			}
			catch (InvalidOperationException e)
			{
				throw new RemoteControlException("Cannot reconfigure: " + e.Message, e);
			}
		}

		public void EnabledDisabled(BasicObjectKind objKind, string name, bool enabled)
		{
			if (callback == null)
				return;

			try
			{
				callback.EnabledDisabled(objKind, name, enabled);
			}
			catch (Exception) // make sure not to fail here, remoting may be down or whatever
			{}
		}
	}

	public class RemoteControlException : ApplicationException
	{
		public RemoteControlException() : base() {}
		public RemoteControlException(string msg) : base(msg) {}
		public RemoteControlException(string msg, Exception inner) : base(msg, inner) {}
	}
}
