/*

 Copyright (c) 2006 Pavel Novak.

 The use and distribution terms for this software are contained in the file named License.txt, 
 which can be found in the root of the NetSim distribution. By using this software 
 in any fashion, you are agreeing to be bound by the terms of this license.
 
 You must not remove this notice from this software.

*/

using System;
using System.Collections;
using System.Net;
using System.Net.Sockets;
using System.Diagnostics;
using System.Runtime.Remoting;
using Marshal = System.Runtime.InteropServices.Marshal;

namespace NetSim.Core
{
	#region RemoteSocketsManager

	public class SocketEntry
	{
		public SocketEntry(NetSimSocket socket, SocketCallbackObject callback)
		{
			this.Socket = socket;
			this.Callback = callback;
		}

		/// <summary>
		/// The socket
		/// </summary>
		public NetSimSocket Socket;

		/// <summary>
		/// Remote object in LSPWraper assembly that is called if data should be send or
		/// received etc.
		/// </summary>
		public SocketCallbackObject Callback;

		// AsyncSelect options
		public AsyncSelectEvent AsyncSelectEvents; // which events are monitored
		public IntPtr AsyncSelecetHWnd; // handle of window that should receive the message
		public uint AsyncSelectMessage; // message being sent to the window
	}

	/// <summary>
	/// Controls sockets created remotely on all the virtual network.
	/// </summary>
	[Serializable]
	public class RemoteSocketsManager : MarshalByRefObject
	{
		/// <summary>
		/// Hashtable containing <see cref="SocketEntry"/> objects for sockets that are currently open.
		/// </summary>
		private Hashtable openSockets;
		/// <summary>
		/// Thread-safe wrapper to the <see cref="openSockets"/>.
		/// </summary>
		private Hashtable openSocketsThreadSafe;

		public RemoteSocketsManager()
		{
			openSockets = new Hashtable();
			openSocketsThreadSafe = Hashtable.Synchronized(openSockets);
		}

		~RemoteSocketsManager()
		{
			// close all sockets
			foreach (DictionaryEntry de in openSockets)
			{
				try { ((SocketEntry)de.Value).Socket.Close(); }
				catch (Exception) { /* nothing to do if failed... */ }
			}
		}

		public int Socket(AddressFamily addressFamily, SocketType socketType,
			ProtocolType protocolType,	ObjRef callbackObjRef, string virtualNodeName)
		{
			// get the node this socket should be created
			Node node = Simulator.GetInstance().GetNode(virtualNodeName);
			if (node == null)
			{
				Simulator.GetInstance().Events.Error(String.Format("Virtual node '{0}' doesn't exist. Cannot create socket.",
					virtualNodeName));
				return NetSimSocket.INVALID_SOCKET;
			}

			// create proxy for the callback
			SocketCallbackObject callback = (SocketCallbackObject)RemotingServices.Unmarshal(callbackObjRef);
			// open new socket
			NetSimSocket s = new NetSimSocket(addressFamily, socketType, protocolType, node);

			// remember new entry for the socket
			openSocketsThreadSafe.Add(s.ID, new SocketEntry(s, callback));

			return s.ID;
		}

		public void Test(ObjRef callbackObjRef)
		{
			SocketCallbackObject callback = (SocketCallbackObject)RemotingServices.Unmarshal(callbackObjRef);
			callback.Test();
		}

		public void Bind(int socket, EndPoint ep)
		{
			NetSimSocket s = ((SocketEntry)openSockets[socket]).Socket;
			s.Bind(ep);
		}

		public void Listen(int socket, int backlog)
		{
			NetSimSocket s = ((SocketEntry)openSockets[socket]).Socket;
			s.Listen(backlog);
		}

		public int Accept(int socket, out EndPoint localEndPoint, ObjRef callbackObjRef)
		{
			NetSimSocket s = ((SocketEntry)openSockets[socket]).Socket;
			
			NetSimSocket sock_accepted = s.Accept();
			int result = sock_accepted.ID;

			// put new socket to the open sockets list
			SocketCallbackObject callback = (SocketCallbackObject)RemotingServices.Unmarshal(callbackObjRef);
			openSocketsThreadSafe.Add(result, new SocketEntry(sock_accepted, callback));

			localEndPoint = sock_accepted.LocalEndPoint;
			return result;
		}

		public void GetSockOpt(int socket, SocketOptionLevel optionLevel, SocketOptionName optionName,
			ref byte[] optionValue, ref int optionLen)
		{
			NetSimSocket s = ((SocketEntry)openSockets[socket]).Socket;

			optionValue = s.GetSocketOption(optionLevel, optionName, optionLen);
			optionLen = optionValue.Length;
		}

		public void SetSockOpt(int socket, SocketOptionLevel optionLevel, SocketOptionName optionName,
			IntPtr optionValue, int optionLen)
		{
			NetSimSocket s = ((SocketEntry)openSockets[socket]).Socket;
			s.SetSocketOption(optionLevel, optionName, optionValue, optionLen);
		}

		public EndPoint GetLocalEndPoint(int socket)
		{
			NetSimSocket s = ((SocketEntry)openSockets[socket]).Socket;
			return s.LocalEndPoint;
		}

		public EndPoint GetRemoteEndPoint(int socket)
		{
			NetSimSocket s = ((SocketEntry)openSockets[socket]).Socket;
			return s.RemoteEndPoint;
		}

		public void Connect(int socket, EndPoint ep)
		{
			NetSimSocket s = ((SocketEntry)openSockets[socket]).Socket;
			s.Connect(ep);
		}

		public void CloseSocket(int socket)
		{
			NetSimSocket s = ((SocketEntry)openSockets[socket]).Socket;
			s.Close();
			openSocketsThreadSafe.Remove(socket);
		}

		internal class ReceiveCallState
		{
			public ReceiveCallState(int socketHash, byte[] buffer, int context)
			{
				this.SocketHash = socketHash;
				this.Buffer = buffer;
				this.Context = context;
			}

			public int SocketHash;
			public byte[] Buffer;
			public int Context;
		}

		internal class SendCallState
		{
			public SendCallState(int socketHash, int count, int context)
			{
				this.SocketHash = socketHash;
				this.Count = count;
				this.Context = context;
			}

			public int SocketHash;
			public int Count;
			public int Context;
		}

		/// <summary>
		/// Starts receiving asynchronously from the socket.
		/// </summary>
		/// <param name="socket">The socket.</param>
		/// <param name="bytes">Max bytes count to receive.</param>
		/// <param name="context">Context.</param>
		public void Receive(int socket, int bytes, int context)
		{
			NetSimSocket s = ((SocketEntry)openSockets[socket]).Socket;

			byte[] buffer = new byte[bytes];
			ReceiveCallState state = new ReceiveCallState(socket, buffer, context);
			AsyncCallback callback = new AsyncCallback(this.ReceiveCallback);
			s.BeginReceive(buffer, 0, bytes, SocketFlags.None, callback, state);
		}

		/// <summary>
		/// Callback method called by the socket when asynchronous receive operation
		/// is about to complete.
		/// </summary>
		/// <param name="result"></param>
		public void ReceiveCallback(IAsyncResult result)
		{
			ReceiveCallState state = (ReceiveCallState)result.AsyncState;
			SocketEntry se = (SocketEntry)openSockets[state.SocketHash];

			// finish receive call
			int bytes_transfered = 0;
			try
			{
				bytes_transfered = se.Socket.EndReceive(result);

				// notify the LSPWrapper that data are available
				se.Callback.RecvCallback(state.Context, state.Buffer, bytes_transfered);
			}
			catch (Exception e)
			{
				Simulator.GetInstance().Events.Warning(String.Format("Receive operation on remote socket '{0}' with local address '{2}' failed: {1}",
					se.Socket.ID, e.Message, se.Socket.LocalEndPoint));
			}
		}

		/// <summary>
		/// Starts asynchronous send operation.
		/// </summary>
		/// <param name="socket">Socket identification.</param>
		/// <param name="buffer">Data to send.</param>
		/// <param name="count">Bytes count in <paramref name="buffer"/> to send.</param>
		/// <param name="context">Context passed back in the callback.</param>
		public void Send(int socket, byte[] buffer, int count, int context)
		{
			NetSimSocket s = ((SocketEntry)openSockets[socket]).Socket;

			SendCallState state = new SendCallState(socket, count, context);
			AsyncCallback callback = new AsyncCallback(this.SendCallback);
			s.BeginSend(buffer, 0, count, SocketFlags.None, callback, state);
		}

		/// <summary>
		/// Called by the socket when asynchronous send operation finishes.
		/// </summary>
		/// <param name="result">Result containing <see cref="SendCallState"/> in its AsyncState member.</param>
		public void SendCallback(IAsyncResult result)
		{
			SendCallState state = (SendCallState)result.AsyncState;
			SocketEntry se = (SocketEntry)openSockets[state.SocketHash];

			// finish receive call
			int bytes_transfered = 0;
			try
			{
				bytes_transfered = se.Socket.EndSend(result);

				// notify the LSPWrapper that data are available
				se.Callback.SendCallback(state.Context, bytes_transfered);
			}
			catch (Exception e)
			{
				Simulator.GetInstance().Events.Warning(String.Format("Send operation on remote socket '{0}' with local address '{2}' failed: {1}",
					se.Socket.ID, e.Message, se.Socket.LocalEndPoint));
			}
		}

		/// <summary>
		/// Performs select operation on the sockets.
		/// </summary>
		/// <param name="readfds">Sockets to be checked for readability.</param>
		/// <param name="writefds">Sockets to be checked for writeability.</param>
		/// <param name="exceptfds">Sockets to be checked for errors.</param>
		/// <param name="microSeconds">Timeout in microseconds; -1 indicates infinite timeout.</param>
		/// <returns>Number of sockets that satisfy the condition.</returns>
		public int Select(ref ArrayList readfds, ref ArrayList writefds, ref ArrayList exceptfds, int microSeconds)
		{
			Debug.Assert(readfds != null && writefds != null && exceptfds != null);

			//
			// convert hash codes to sockets
			//

			// arrays for Socket instances that will be passed to Select method
			ArrayList read_sockets = new ArrayList(readfds.Count);
			ArrayList write_sockets = new ArrayList(writefds.Count);
			ArrayList except_sockets = new ArrayList(exceptfds.Count);
			// arrays of socket hashcodes that will not be passed to Select
			ArrayList e_sock_no_select = new ArrayList(exceptfds.Count);
			
			for (int i = 0; i < readfds.Count; i++)
			{
				SocketEntry se = (SocketEntry)openSockets[(int)readfds[i]];
				if (se != null)
					read_sockets.Add(se.Socket);
			}
			for (int i = 0; i < writefds.Count; i++)
			{
				SocketEntry se = (SocketEntry)openSockets[(int)writefds[i]];
				if (se != null)
					write_sockets.Add(se.Socket);
			}
			for (int i = 0; i < exceptfds.Count; i++)
			{
				SocketEntry se = (SocketEntry)openSockets[(int)exceptfds[i]];
				if (se != null)
					except_sockets.Add(se.Socket);
				else
					e_sock_no_select.Add(exceptfds[i]);
			}

			NetSimSocket.Select(read_sockets, write_sockets, except_sockets, microSeconds);

			// convert back hascodes
			readfds.Clear();
			writefds.Clear();
			exceptfds.Clear();

			// add sockets from select
			for (int i = 0; i < read_sockets.Count; i++)
				readfds.Add(((NetSimSocket)read_sockets[i]).ID);
			for (int i = 0; i < write_sockets.Count; i++)
				writefds.Add(((NetSimSocket)write_sockets[i]).ID);
			for (int i = 0; i < except_sockets.Count; i++)
				exceptfds.Add(((NetSimSocket)except_sockets[i]).ID);
			// add sockets that weren't passed to select
			for (int i = 0; i < e_sock_no_select.Count; i++)
				exceptfds.Add(e_sock_no_select[i]);

			return readfds.Count + writefds.Count + exceptfds.Count;
		}

		public void AsyncSelect(int socket, IntPtr hWnd, uint wMsg, AsyncSelectEvent events)
		{
			SocketEntry se = (SocketEntry)openSockets[socket];
			if (se == null)
				return;

			se.AsyncSelecetHWnd = hWnd;
			se.AsyncSelectMessage = wMsg;
			se.AsyncSelectEvents = events;

			se.Socket.AsyncSelect(events, new NetSimSocket.AsyncSelectCallbackDelegate(this.AsyncSelectCallback));
		}

		public void AsyncSelectCallback(NetSimSocket socket, AsyncSelectEvent events)
		{
			SocketEntry se = (SocketEntry)openSockets[socket.ID];
			if (se == null)
				return;

			// restrict events only to those being monitored
			events = events & se.AsyncSelectEvents;
			if (events != AsyncSelectEvent.None && se.Callback != null)
			{
				se.Callback.AsyncSelectCallback(se.AsyncSelecetHWnd, se.AsyncSelectMessage, events);
			}
		}

		public void Shutdown(int socket, SocketShutdown how)
		{
			NetSimSocket s = ((SocketEntry)openSockets[socket]).Socket;
			s.Shutdown(how);
		}

		public override object InitializeLifetimeService()
		{
			return null; // lease never expires
		}
	}

	
	#endregion
	
	/// <summary>
	/// This delegate is createdy by the LSP Wrapper and passed to the RemoteSockets.
	/// It is called after receive function completion.
	/// </summary>
	public delegate void WrapperReceiveDelegate(byte[] data, int bytesTransferred, object state);

	/// <summary>
	/// This delegate is createdy by the LSP Wrapper and passed to the RemoteSockets.
	/// It is called after send function completion.
	/// </summary>
	public delegate void WrapperSendDelegate(int bytesTransferred, object state);

	/// <summary>
	/// This delegate is createdy by the LSP Wrapper and passed to the RemoteSockets.
	/// It is called while notification requested by AsyncSelect call is needed.
	/// </summary>
	public delegate void WrapperAsyncSelectDelegate(int socket, IntPtr hWnd, uint wMsg, AsyncSelectEvent events);

	public class SocketCallbackObject : MarshalByRefObject
	{
		public int i = 0;
		public void Test()
		{
			i = 1;
		}

		private int hash;
		private Hashtable pendingRecvOperations;
		private Hashtable pendingSendOperations;

		private static WrapperReceiveDelegate wrapperReceiveDlg;
		public static void SetWrapperReceiveDelegate(WrapperReceiveDelegate dlg)
		{
			wrapperReceiveDlg = dlg;
		}

		private static WrapperSendDelegate wrapperSendDlg;
		public static void SetWrapperSendDelegate(WrapperSendDelegate dlg)
		{
			wrapperSendDlg = dlg;
		}

		private static WrapperAsyncSelectDelegate wrapperAsyncSelectDlg;
		public static void SetWrapperAsyncSelectDelegate(WrapperAsyncSelectDelegate dlg)
		{
			wrapperAsyncSelectDlg = dlg;
		}

		public SocketCallbackObject()
		{
			pendingRecvOperations = new Hashtable();
			pendingSendOperations = new Hashtable();
		}
		
		public void SetHashCode(int hash)
		{
			this.hash = hash;
		}

		public void SetPendingReceive(int key, object state)
		{
			pendingRecvOperations[key] = state;
		}

		public void SetPendingSend(int key, object state)
		{
			pendingSendOperations[key] = state;
		}

		public void RecvCallback(int key, byte[] data, int bytesTransfered)
		{
			object state = pendingRecvOperations[key];
			pendingRecvOperations.Remove(key);

			wrapperReceiveDlg(data, bytesTransfered, state);
		}

		public void SendCallback(int key, int bytesTransfered)
		{
			object state = pendingSendOperations[key];
			pendingSendOperations.Remove(key);

			wrapperSendDlg(bytesTransfered, state);
		}

		public void AsyncSelectCallback(IntPtr hWnd, uint wMsg, AsyncSelectEvent events)
		{
			wrapperAsyncSelectDlg(this.hash, hWnd, wMsg, events);
		}

		public override object InitializeLifetimeService()
		{
			return null; // lease never expires
		}
	}

	[Flags]
	[Serializable]
	public enum AsyncSelectEvent : int
	{
		None          = 0,
		Read          = 1 << 0,
		Write         = 1 << 1,
		OutOfBand    = 1 << 2,
		Accept        = 1 << 3,
		Connect       = 1 << 4,
		Close         = 1 << 5,
		Qos           = 1 << 6,
		GroupQos      = 1 << 7,
		RoutingIfaceGange = 1 << 8, 
		AddressListChange = 1 << 9
	}

	public class NestingCounter
	{
		static object _lock = new object();

		static int count = 0;

		public static int Count(int add)
		{
			lock(_lock)
			{
				count += add;
				return count;
			}
		}

		public static UInt32 GetThreadStartAddress(int pid, int tid)
		{
			string query = String.Format("SELECT * FROM Win32_Thread WHERE ProcessHandle = {0} AND Handle = {1}", pid, tid);
			System.Management.ManagementObjectSearcher searcher = new System.Management.ManagementObjectSearcher(query); 
			System.Management.ManagementObjectCollection m_objects = searcher.Get();

			if (m_objects.Count != 1) // we expect just one thread
				return 0;

			// we can get an object only through an enumerator
			foreach (System.Management.ManagementObject mo in m_objects)
			{
				return (UInt32)mo.Properties["StartAddress"].Value;
			}

			return 0;
		}
	}


}
