/*

 Copyright (c) 2006 Pavel Novak.

 The use and distribution terms for this software are contained in the file named License.txt, 
 which can be found in the root of the NetSim distribution. By using this software 
 in any fashion, you are agreeing to be bound by the terms of this license.
 
 You must not remove this notice from this software.

*/

using System;
using System.Runtime.InteropServices;

namespace NetSim.Core
{
	/// <summary>
	/// Summary description for SimulationTimer.
	/// </summary>
	public class SimulationTimer
	{
		/// <summary>
		/// The frequency of high-resolution timer. If the high-resolution timer is not supported, contains
		/// frequency for standard ticks - ie. how many ticks are in one second.
		/// </summary>
		private long highResTimerFrequency = 10 * 1000 * 1000;
		private bool highResTimerSupported = false;

#if MONO
		[StructLayout(LayoutKind.Sequential)]
		private struct TIMEVAL
		{
			public int sec;
			public int usec;
		};

		[StructLayout(LayoutKind.Sequential)]
		private struct TIMEZONE
		{
			public int sec;
			public int usec;
		};

		[DllImport("libc.so.6")]
		private extern static int gettimeofday(ref TIMEVAL tv, ref TIMEZONE tz);

		// for performance counter on mono is standard libc gettimeofday used
		private static int QueryPerformanceFrequency(ref long /* LARGE_INTEGER* */ frequency)
		{
			frequency = 1000000; // microseconds
			return 1;
		}
		private static int QueryPerformanceCounter(ref long /* LARGE_INTEGER* */ count)
		{
			TIMEVAL tv = new TIMEVAL();
			TIMEZONE tz = new TIMEZONE();
			gettimeofday(ref tv, ref tz);

			count = ((long)tv.sec) * 1000000 + (long)tv.usec;
			return 1;
		}
#else
		[DllImport("Kernel32.dll")]
		private extern static int QueryPerformanceFrequency(ref long /* LARGE_INTEGER* */ frequency);

		[DllImport("Kernel32.dll")]
		private extern static int QueryPerformanceCounter(ref long /* LARGE_INTEGER* */ count);
#endif

		public SimulationTimer() : this(1.0f)
		{}

		public SimulationTimer(float speed)
		{
			this.speed = speed;

			highResTimerSupported = QueryPerformanceFrequency(ref highResTimerFrequency) != 0;

			if (!highResTimerSupported)
				Simulator.GetInstance().Events.Warning("Your hardware doesn't support high resolution timer. The times measured by the simulator and hence times of packet delays will be very inaccurate");
		}

		/// <summary>Gets or sets the multiplier of the time the time goes on.</summary>
		public float Speed
		{
			get { return speed; }
			set { speed = value; }
		}
		/// <summary>
		/// Changes the time in which the events occur. The time is multiplied by this value:
		/// therefore the value of one means no change, the value of 2 means two-times slower execution
		/// etc.
		/// </summary>
		private float speed;

		/// <summary>
		/// Returns the time after <paramref name="ticksDelay"/> from now.
		/// </summary>
		/// <param name="ticksDelay">The delay in ticks.</param>
		/// <returns>The time after the delay.</returns>
		public long GetTimeAfterDelay(long ticksDelay)
		{
			return GetTimeAfterDelay(ticksDelay, this.Now);
		}

		/// <summary>
		/// Returns the time after <paramref name="ticksDelay"/> from the <paramref name="time"/>.
		/// </summary>
		/// <param name="ticksDelay">The delay in ticks.</param>
		/// <param name="time">The time from when is the delay counted.</param>
		/// <returns>The time after the delay.</returns>
		public long GetTimeAfterDelay(long ticksDelay, long time)
		{
			if (highResTimerSupported) // convert ticks paramter to ticks of perf. counter
				ticksDelay = (ticksDelay * highResTimerFrequency) / 10000000;

			return time + (long)(speed * (double)ticksDelay);
		}

		/// <summary>
		/// Returns the time right now.
		/// </summary>
		public long Now
		{
			get
			{
				long result = 0;
				if (highResTimerSupported)
					QueryPerformanceCounter(ref result);
				else
					result = DateTime.Now.Ticks;

				return result;
			}
		}
	}
}
