/*

 Copyright (c) 2006 Pavel Novak.

 The use and distribution terms for this software are contained in the file named License.txt, 
 which can be found in the root of the NetSim distribution. By using this software 
 in any fashion, you are agreeing to be bound by the terms of this license.
 
 You must not remove this notice from this software.

*/

using System;
using System.Runtime.Remoting;
using System.Runtime.Remoting.Channels;
using System.Runtime.Remoting.Channels.Tcp;
using System.Xml;
using System.Collections;
using System.Reflection;
using System.Diagnostics;

namespace NetSim.Core
{
	/// <summary>
	/// The main class for simulation. Provides basic functionality to manage the simulator.
	/// </summary>
	public class Simulator
	{
		#region Nested types
		public delegate void StartEventDelegate();
		public delegate void StopEventDelegate();
		#endregion

		#region Constants
		public const string RemoteSocketsManagerObjectName = "SimulatorRemoteSocketsManager";
		public static readonly string RemoteControlObjectName = "SimulatorRemoteControl";

		public const string ExtensionXmlNodeName = "Extension";
		public const string NetworkStructureXmlNodeName = "Network";
		public const string NodesXmlNodeName = "Nodes";
		public const string LinksXmlNodeName = "Links";
		public const string NodeXmlNodeName = "Node";
		public const string LinkXmlNodeName = "Link";
		public const string AdapterXmlNodeName = "Adapter";
		public const string ModuleXmlNodeName = "Module";
		public const string ApplicationXmlNodeName = "Application";
		#endregion

		#region Properties and member variables
		/// <summary>
		/// The only one instance of simulator.
		/// </summary>
		private static Simulator simulator;

		/// <summary>
		/// <b>True</b> if verbose messages should be written.
		/// </summary>
		public static bool Verbose = true;

		/// <summary>
		/// Absolute path to main configuration file for the simulator.
		/// </summary>
		public string MainConfigFile
		{
			get { return mainConfigFile; }
		}
		private string mainConfigFile;

		/// <summary>
		/// Returns <b>true</b> if the simulator is running, <b>false</b> otherwise.
		/// </summary>
		public bool Running
		{
			get { return running; }
		}
		private bool running;

		/// <summary>
		/// Gets or sets whether remote sockets via TCP are enabled. The value can be set only once.
		/// </summary>
		public bool UseRemoteSocketsTcp
		{
			get
			{
				return useRemoteSocketsTcp == 1;
			}
			set
			{
				if (useRemoteSocketsTcp != -1) // value has been set
					throw new InitializationException("UseRemoteSocketsTcp can be set only once.");
				useRemoteSocketsTcp = value ? 1 : 0;
			}
		}
		// handled internally by integer to recognise if the value has been set
		// -1 : not set
		// 0 : do not use
		// 1 : use
		private int useRemoteSocketsTcp = -1;

		/// <summary>
		/// Gets or sets whether remote sockets via TCP are enabled. The value can be set only once.
		/// </summary>
		public bool UseRemoteSocketsShm
		{
			get
			{
				return useRemoteSocketsShm == 1;
			}
			set
			{
				if (useRemoteSocketsShm != -1) // value has been set
					throw new InitializationException("UseRemoteSocketsShm can be set only once.");
				useRemoteSocketsShm = value ? 1 : 0;
			}
		}
		// handled internally by integer to recognise if the value has been set
		// -1 : not set
		// 0 : do not use
		// 1 : use
		private int useRemoteSocketsShm = -1;

		/// <summary>
		/// <see cref="RemoteSocketsManager"/> receives calls to sockets that were opened by the remote side
		/// on this computer. Only one manager exists in the simulator.
		/// </summary>
		private RemoteSocketsManager remoteSocketsManager;

		/// <summary>
		/// Channel for remote communication with <see cref="remoteSocketsManager"/> via TCP.
		/// </summary>
		private IChannel remoteSocketsTcpChannel;

		/// <summary>
		/// Channel for remote communication with <see cref="remoteSocketsManager"/> via shared memory.
		/// </summary>
		private IChannel remoteSocketsShmChannel;

		/// <summary>
		/// Object for remote controlling of the simulator. Is accessed via <see cref="remoteSocketsShmChannel"/>.
		/// </summary>
		private RemoteControl remoteControl;
		
		/// <summary>
		/// Gets or sets whether the remote control via shared memory should be used. The value can be set only once.
		/// </summary>
		public bool UseRemoteControl
		{
			get
			{
				return useRemoteControl == 1;
			}
			set
			{
				if (useRemoteControl != -1) // value has been set
					throw new InitializationException("UseRemoteControl can be set only once.");
				useRemoteControl = value ? 1 : 0;
			}
		}
		// handled internally by integer to recognise if the value has been set
		// -1 : not set
		// 0 : do not use
		// 1 : use
		private int useRemoteControl = -1;

		/// <summary>
		/// List of <see cref="Assembly"/> classes with extensions.
		/// </summary>
		private ArrayList extensions;

		/// <summary>
		/// Returns node of the name <paramref name="name"/> or <b>null</b> if such node doesn't exist.
		/// </summary>
		/// <param name="name">Node name.</param>
		/// <returns><see cref="Node"/> object or <b>null</b>.</returns>
		public Node GetNode(string name)
		{
			return (Node)nodes[name];
		}

		public void EnableDisableNode(string name, bool enable)
		{
			Node n = GetNode(name);
			if (n != null && n.Enabled != enable)
			{
				// change status
				n.Enabled = enable;

				// notify the remote side
				if (remoteControl != null)
					remoteControl.EnabledDisabled(BasicObjectKind.Node, name, enable);
			}
		}

		public void EnableDisableLink(string name, bool enable)
		{
			Link l = GetLink(name);
			if (l != null && l.Enabled != enable)
			{
				// change status
				l.Enabled = enable;

				// notify the remote side
				if (remoteControl != null)
					remoteControl.EnabledDisabled(BasicObjectKind.Link, name, enable);
			}
		}

		/// <summary>
		/// Returns the list of <see cref="Node"/> objects that are present in the simulated network.
		/// </summary>
		public IList Nodes
		{
			get
			{
				// return a copy of internal list => prevent its modification
				Node[] result = new Node[nodes.Count];
				int i = 0;
				foreach (DictionaryEntry de in nodes)
				{
					result[i] = (Node) de.Value;
					i++;
				}
				return result;
			}
		}

		/// <summary>
		/// Keys: node names, values: <see cref="Node"/> objects.
		/// </summary>
		//gererics: <Node>
		private Hashtable nodes;

		public Link GetLink(string name)
		{
			return (Link)links[name];
		}

		/// <summary>
		/// Returns the list of <see cref="Link"/> objects that are present in the simulated network.
		/// </summary>
		public IList Links
		{
			get
			{
				// return a copy of internal list => prevent its modification
				Link[] result = new Link[links.Count];
				int i = 0;
				foreach (DictionaryEntry de in links)
				{
					result[i] = (Link) de.Value;
					i++;
				}
				return result;
			}
		}

		/// <summary>
		/// Keys: link names, values: <see cref="Link"/> objects.
		/// </summary>
		//generics: <Link>
		private Hashtable links;

		public IStructureObject GetItem(BasicObjectKind kind, string name)
		{
			switch (kind)
			{
				case BasicObjectKind.Node:
					return GetNode(name);
				case BasicObjectKind.Link:
					return GetLink(name);
			}
			return null;
		}

		private Hashtable[] packetFactories;

		/// <summary>
		/// Gets the collection of events that occured during the simulation.
		/// </summary>
		public EventLog Events { get { return eventLog; } }
		private EventLog eventLog;

		/// <summary>
		/// Can be used to wait for simulator exit.
		/// </summary>
		public readonly System.Threading.ManualResetEvent WaitForStop = new System.Threading.ManualResetEvent(false);

		/// <summary>
		/// Timer that can be used for calculating delays in the simulator.
		/// </summary>
		public static SimulationTimer Timer { get { return timer; } }
		private static SimulationTimer timer;

		/// <summary>
		/// Random number generator. Call it whenever you need a random number.
		/// </summary>
		public static readonly Random Random = new Random();

		#endregion

		#region Construction and initialization
		
		static Simulator()
		{
			simulator = new Simulator();
			simulator.Initialize();
		}

		/// <summary>
		/// Private constructor allowing only one instance of simulator.
		/// </summary>
		private Simulator()
		{
		}

		private void Initialize()
		{
			// init event log first to be able to log events while initializing
			eventLog = new EventLog();

			extensions = new ArrayList();
			// add core assembly to the list - this assembly is searched first in all cases
			extensions.Add(Assembly.GetExecutingAssembly());

			// Future version: add support for adding and removing while the simulator is running, therefore
			// synchronized wrappers needed here
			nodes = new Hashtable();
			links = new Hashtable();

			timer = new SimulationTimer();

			packetFactories = new Hashtable[(int)Packet.PacketGroup._max];
			for (int i = 0; i < (int)Packet.PacketGroup._max - 1; i++)
				packetFactories[i] = new Hashtable();
		}

		/// <summary>
		/// Returns reference to simulator singleton.
		/// </summary>
		/// <returns></returns>
		public static Simulator GetInstance()
		{
			// simulator instance shouldn't be null since it is initialized in the static constructor
			Debug.Assert(simulator != null, "Simulator is null - why? It is initialized in the type constructor...");

			return simulator;
		}
		#endregion

		#region Configuration
		public void Configure(string configPath)
		{
			this.mainConfigFile = System.IO.Path.GetFullPath(configPath);

			XmlValidatingReader reader;

			// create validating XML reader
			try
			{
				XmlTextReader text_reader = new XmlTextReader(configPath);
				reader = new XmlValidatingReader(text_reader);
			}
			catch (System.IO.FileNotFoundException e)
			{
				throw new InitializationException(String.Format("Main configuration file {0} was not found", configPath), e);
			}

			try
			{
				// parse document
				XmlDocument document = new XmlDocument();
				document.Load(reader);
				XmlElement root_element = document.DocumentElement;

				// load extensions
				LoadExtensions(root_element.GetElementsByTagName(ExtensionXmlNodeName));

				// load network structure
				XmlNodeList network_list = root_element.GetElementsByTagName(NetworkStructureXmlNodeName);
				if (network_list.Count != 1)
					throw new InitializationException(String.Format("Only one Network element expected in {0}", configPath));
				CreateNetworkStructure(network_list[0]);
			}
			catch (System.Xml.Schema.XmlSchemaException e)
			{
				throw new InitializationException(String.Format("Main configuration file {0} does not conform to the XML schema", configPath), e);
			}
		}

		private void LoadExtensions(XmlNodeList list)
		{
			foreach (XmlNode node in list)
			{
				try
				{
					// load extension
					XmlAttribute pathAttr = node.Attributes["path"];
					if (pathAttr == null)
						throw new InitializationException("'path' attribute is required for node 'Extension'");

					Assembly a = Assembly.LoadFrom(pathAttr.InnerText);
					extensions.Add(a);

					// check for packet factory
					if (node.HasChildNodes)
					{
						XmlNode n_child = node.FirstChild;
						while (n_child != null)
						{
							if (n_child.Name == "PacketFactory") // the only one node we want
							{
								XmlAttribute classAttr = n_child.Attributes["class"];
								if (classAttr == null)
									throw new InitializationException("'PacketFactory' node should have 'class' attribute");

								// create the factory instance
								PacketFactory pf = (PacketFactory) GetInstanceByNameFromAssembly(classAttr.InnerText,
									typeof(PacketFactory), a);
								if (pf == null)
									throw new InitializationException(String.Format("Cannot create '{0}' instance from {1} assembly",
										classAttr.InnerText, a.GetName()));

								// insert the packet factory for all supported data link types
								foreach (PacketFactory.SupportedPacket supported in pf.Supports)
								{
									try { this.packetFactories[(int)supported.Group].Add(supported.Identification, pf); }
									catch (ArgumentException e)
									{
										Simulator.GetInstance().Events.Notice(String.Format(
											"Packet factory for link of type '{0}' is already present, ignoring the second occurence",
											supported), e);
									}
								}
							}
//							else if (n_child.Name == "AddressFactory")
//							{
//								XmlAttribute classAttr = n_child.Attributes["class"];
//								if (classAttr == null)
//									throw new InitializationException("'AddressFactory' node should have 'class' attribute");
//
//								// create the factory instance
//								AddressFactory af = (AddressFactory) GetInstanceByNameFromAssembly(classAttr.InnerText,
//									typeof(AddressFactory), a);
//								if (af == null)
//									throw new InitializationException(String.Format("Cannot create '{0}' instance from {1} assembly",
//										classAttr.InnerText, a.GetName()));
//
//								// insert the address factory for all supported data link types
//								foreach (int protocol_type in af.ProtocolTypes)
//								{
//									try { this.addressFactories.Add(protocol_type, af); }
//									catch (ArgumentException e)
//									{
//										Simulator.GetInstance().Events.Warning(String.Format(
//											"Address factory for addresses of type '{0}' is already present, ignoring the second occurence",
//											protocol_type), e);
//									}
//								}
//							}
							n_child = n_child.NextSibling;
						}
					}
				}
				catch (System.IO.FileNotFoundException e)
				{
					throw new InitializationException(String.Format("Extension file {0} cannot be found", node.InnerText), e);
				}
			}
		}

		private void CreateNetworkStructure(XmlNode networkNode)
		{
			// find nodes and links
			XmlNode nodes = null;
			XmlNode links = null;

			XmlNode child = networkNode.FirstChild;
			while (child != null)
			{
				if (child.Name == NodesXmlNodeName)
					nodes = child;
				else if (child.Name == LinksXmlNodeName)
					links = child;

				child = child.NextSibling;
			}

			if (nodes == null || links == null)
				throw new InitializationException(String.Format("No {0} or {1} node under {2} in main configuration file",
					NodesXmlNodeName, LinksXmlNodeName, NetworkStructureXmlNodeName));

			// create links
			CreateNetworkLinks(links);
			// create nodes after links - they are being connected to the links at this stage
			CreateNetworkNodes(nodes);
		}

		private void CreateNetworkNodes(XmlNode nodesXml)
		{
			XmlNode node = nodesXml.FirstChild;
			while (node != null)
			{
				// process only our Nodes
				if (node.Name == NodeXmlNodeName)
				{
					CreateNetworkNode(node);
				}

				node = node.NextSibling;
			}
		}

		private void CreateNetworkNode(XmlNode node)
		{
			XmlAttribute configAttribute = node.Attributes["config"];
			XmlAttribute nameAttribute = node.Attributes["name"];
			XmlAttribute classAttribute = node.Attributes["class"];

			if (nameAttribute == null)
				throw new InitializationException(String.Format("Attribute 'name' not present in node '{0}'", node.Name));
			if (classAttribute == null)
				throw new InitializationException(String.Format("Attribute 'class' not present in node '{0}'", node.Name));

			Node node_object = (Node) GetInstanceByName(classAttribute.InnerText,
				typeof(Node));

			// set node name
			node_object.Name = nameAttribute.InnerText;

			// configure node
			if (configAttribute != null && configAttribute.InnerText != String.Empty)
				node_object.Configure(configAttribute.InnerText);

			// add to the list
			try
			{
				nodes.Add(nameAttribute.InnerText, node_object);
			}
			catch (ArgumentException e)
			{
				throw new InitializationException(String.Format("Duplicate node name '{0}'", nameAttribute.InnerText), e);
			}

			// load adapters, modules and connect to the links
			XmlNode child = node.FirstChild;
			while (child != null)
			{
				if (child.Name == AdapterXmlNodeName)
					LoadNodeAdapter(node_object, child);
				else if (child.Name == ModuleXmlNodeName)
					LoadNodeModule(node_object, child);
				else if (child.Name == ApplicationXmlNodeName)
					LoadNodeApplication(node_object, child);

				child = child.NextSibling;
			}
		}

		private void LoadNodeAdapter(Node nodeInstance, XmlNode adapter)
		{
			XmlAttribute configAttribute = adapter.Attributes["config"];
			XmlAttribute nameAttribute = adapter.Attributes["name"];
			XmlAttribute classAttribute = adapter.Attributes["class"];
			XmlAttribute linkAttribute = adapter.Attributes["link"];

			if (nameAttribute == null)
				throw new InitializationException(String.Format("Attribute 'name' not present in adapter at node '{0}'", nodeInstance.Name));
			if (classAttribute == null)
				throw new InitializationException(String.Format("Attribute 'class' not present in adapter '{0}' at node '{1}'", nameAttribute.InnerText, nodeInstance.Name));

			IAdapter adapter_object = (Adapter) GetInstanceByName(classAttribute.InnerText,
				typeof(Adapter));

			// set adapter name
			adapter_object.Name = nameAttribute.InnerText;

			// configure adapter
			if (configAttribute != null && configAttribute.InnerText != String.Empty)
				adapter_object.Configure(configAttribute.InnerText);

			// connect to the link
			Link link = null;
			if (linkAttribute != null && linkAttribute.InnerText != String.Empty)
			{
				link = (Link) links[linkAttribute.InnerText];
				if (link == null)
					throw new InitializationException(String.Format("Link '{0}' cannot be found", linkAttribute.InnerText));
			}

			if (link != null)
			{
				// assign to the Node
				nodeInstance.AddAdapter(adapter_object);
				// assign link to adapter
				adapter_object.Link = link;
				// assign adapter to the link
				link.RegisterAdapter(adapter_object);
			}
		}

		private void LoadNodeModule(Node nodeInstance, XmlNode module)
		{
			XmlAttribute classAttr = module.Attributes["class"];
			if (classAttr == null)
				throw new InitializationException(String.Format("Required attribute 'class' in '{0}' node is missing.", ModuleXmlNodeName));

			IModule module_object = (IModule) GetInstanceByName(classAttr.InnerText,
				typeof(IModule));

			// name of the module
			XmlAttribute nameAttr = module.Attributes["name"];
			if (nameAttr != null)
				module_object.Name = nameAttr.InnerText;

			// node should be set here, it is required for some modules configurations
			module_object.Node = nodeInstance;

			// configure module
			XmlAttribute configAttr = module.Attributes["config"];
			if (configAttr != null)
				module_object.Configure(configAttr.InnerText);

			// assign to the Node
			nodeInstance.AddModule(module_object);
		}

		private void LoadNodeApplication(Node nodeInstance, XmlNode app)
		{
			XmlAttribute classAttr = app.Attributes["class"];
			if (classAttr == null)
				throw new InitializationException(String.Format("Required attribute 'class' in '{0}' node is missing.", ApplicationXmlNodeName));

			INodeApplication app_object = (INodeApplication) GetInstanceByName(classAttr.InnerText,
				typeof(INodeApplication));

			app_object.SetNode(nodeInstance);

			// configure module
			XmlAttribute configAttr = app.Attributes["config"];
			if (configAttr != null)
				app_object.Configure(configAttr.InnerText);

			// assign to the Node
			nodeInstance.AddNodeApplication(app_object);
		}

		/// <summary>
		/// Searches core assembly and extensions for the specified class and checkes if the class is 
		/// </summary>
		/// <param name="className"></param>
		/// <param name="objectType"></param>
		/// <returns></returns>
		internal Object GetInstanceByName(string className, Type objectType)
		{
			foreach (Assembly a in extensions)
			{
				object result = GetInstanceByNameFromAssembly(className, objectType, a);
				if (result != null)
					return result;
			}

			// not found
			throw new InitializationException(String.Format("Type {0} cannot be found in any extension assembly.", className));
		}

		private Object GetInstanceByNameFromAssembly(string className, Type objectType, Assembly assembly)
		{
			Object result = null;
			Type t = assembly.GetType(className, false);
			if (t != null)
			{
				ConstructorInfo ci = t.GetConstructor(Type.EmptyTypes);
				if (ci == null)
					throw new InitializationException(String.Format("Type {0} in assembly {1} does not have default constructor",
						className, assembly.GetName()));

				if (t != objectType // different types
					&& !t.IsSubclassOf(objectType) // not subclass
					&& (objectType.IsInterface && t.FindInterfaces(new TypeFilter(InterfaceFilter), objectType).Length == 0) // or not implementing the interface
					)
					throw new InitializationException(String.Format("Type {0} in assembly {1} should inherit from {2}",
						className, assembly.GetName(), objectType.FullName));

				result = ci.Invoke(null);
			}
			
			return result;
		}

		// filter for GetInstanceByNameFromAssembly() method
		private static bool InterfaceFilter(Type typeObj, object criteriaObj)
		{
			if (typeObj.Equals( (Type)criteriaObj ))
				return true;
			else
				return false;
		}

		private void CreateNetworkLinks(XmlNode linksXml)
		{
			XmlNode link = linksXml.FirstChild;
			while (link != null)
			{
				// process only our Nodes
				if (link.Name == LinkXmlNodeName)
				{
					CreateNetworkLink(link);
				}

				link = link.NextSibling;
			}
		}

		private void CreateNetworkLink(XmlNode link)
		{
			XmlAttribute classAttr = link.Attributes["class"];
			if (classAttr == null)
				throw new InitializationException("'class' attribute expected");

			XmlAttribute configAttr = link.Attributes["config"];

			XmlAttribute nameAttr = link.Attributes["name"];
			if (nameAttr == null)
				throw new InitializationException("'name' attribute expected");

			Link link_object = (Link) GetInstanceByName(classAttr.InnerText,
				typeof(Link));

			// set link name
			link_object.Name = nameAttr.InnerText;

			// configure the link
			if (configAttr != null && configAttr.InnerText != String.Empty)
				link_object.Configure(configAttr.InnerText);

			// add to the list
			links.Add(nameAttr.InnerText, link_object);
		}
		#endregion

		#region Public methods
		public Packet CreatePacket(Packet.PacketGroup group, int packetType, byte[] data)
		{
			return CreatePacket(group, packetType, data, 0, data.Length);
		}

		public Packet CreatePacket(Packet.PacketGroup group, int packetType, byte[] data, int offset)
		{
			return CreatePacket(group, packetType, data, offset, data.Length - offset);
		}

		public Packet CreatePacket(Packet.PacketGroup group, int packetType, byte[] data, int offset, int len)
		{
			PacketFactory pf = (PacketFactory) packetFactories[(int)group][packetType];
			try
			{
				if (pf != null)
					return pf.CreatePacket(group, packetType, data, offset, len);
			}
			catch (Exception e) // make sure not to fail here - some error while creating the packet - this shouldn't break the simulation
			{
				Simulator.GetInstance().Events.Error(String.Format("Packet factory {0} threw an exception while creating the packet", pf), e);
			}

			// no factory for this type
			return null;
		}

//		public NetworkAddress CreateAddress(ushort addressType, byte[] bytes)
//		{
//			AddressFactory af = (AddressFactory) addressFactories[addressType];
//			try
//			{
//				if (af != null)
//					return af.CreateAddress(addressType, bytes);
//			}
//			catch (Exception e) // make sure not to fail here - some error while creating the address - this shouldn't break the simulation
//			{
//				Simulator.GetInstance().Events.Error(String.Format("Address factory {0} threw an exception while creating the address", af), e);
//			}
//
//			// no factory for this type
//			return null;
//
//		}

		#endregion

		#region Start, Stop

		public event StartEventDelegate StartEvent;
		public event StopEventDelegate StopEvent;

		public event StartEventDelegate SuspendEvent;
		public event StopEventDelegate ResumeEvent;

		/// <summary>
		/// Starts the simulator.
		/// </summary>
		public void Start()
		{
			InitializeChannels();

			if (UseRemoteSocketsTcp || UseRemoteSocketsShm)
				StartRemoteSocket();
			if (UseRemoteControl)
				StartRemoteControl();

			StartEvent();
			this.running = true;
		}

		/// <summary>
		/// Stops the simulator.
		/// </summary>
		public void Stop()
		{
			this.running = false;

			// notify all components that the simulator is being stopped
			StopEvent();

			if (UseRemoteControl)
				StopRemoteControl();
			if (UseRemoteSocketsTcp || UseRemoteSocketsShm)
				StopRemoteSocket();

			// stop all threads in our thread-pool and reintialize them - they will be ready for the next start
			ManagedThreadPool.Reset();

			ShutdownChannels();
		}

		public void Pause()
		{
			if (!running) // stopped already
				return;

			this.running = false;
			ManagedThreadPool.Pause();
			SuspendEvent();
		}

		public void Resume()
		{
			if (running) // started already
				return;

			ResumeEvent();
			ManagedThreadPool.Resume();
			this.running = true;
		}

		private void InitializeChannels()
		{
			if (UseRemoteSocketsTcp) // tcp channel only if needed
			{
				BinaryServerFormatterSinkProvider serverProv = new BinaryServerFormatterSinkProvider();
				serverProv.TypeFilterLevel = System.Runtime.Serialization.Formatters.TypeFilterLevel.Full;
				System.Collections.Specialized.ListDictionary channelProperties = new System.Collections.Specialized.ListDictionary();
				channelProperties.Add("port", 9001);
				remoteSocketsTcpChannel = new TcpChannel(channelProperties, new BinaryClientFormatterSinkProvider(), serverProv);
				ChannelServices.RegisterChannel(remoteSocketsTcpChannel);
			}

			// ShmChannel doesn't work on Mono - it is just windows based.
#if !MONO
			// shm channel always, it is needed by both remote sockets and remote control
			{
				BinaryServerFormatterSinkProvider serverProv = new BinaryServerFormatterSinkProvider();
				serverProv.TypeFilterLevel = System.Runtime.Serialization.Formatters.TypeFilterLevel.Full;
				System.Collections.Specialized.ListDictionary channelProperties = new System.Collections.Specialized.ListDictionary();
				channelProperties.Add("name", "SimulatorShmChannel");
				channelProperties.Add("section", "NetSimServer");
				remoteSocketsShmChannel = new PHP.Core.ShmChannel(channelProperties, new BinaryClientFormatterSinkProvider(), serverProv);
				ChannelServices.RegisterChannel(remoteSocketsShmChannel);
			}
#endif
		}

		private void ShutdownChannels()
		{
			// unregister the channel
			if (remoteSocketsTcpChannel != null)
				ChannelServices.UnregisterChannel(remoteSocketsTcpChannel);
			if (remoteSocketsShmChannel != null)
				ChannelServices.UnregisterChannel(remoteSocketsShmChannel);

			remoteSocketsTcpChannel = null;
			remoteSocketsShmChannel = null;
		}

		private void StartRemoteSocket()
		{
			try
			{
				// This is the only one instance of WrapperControl we create
				RemotingConfiguration.RegisterWellKnownServiceType(typeof(RemoteSocketsManager), RemoteSocketsManagerObjectName,  
					WellKnownObjectMode.Singleton);
				remoteSocketsManager = new RemoteSocketsManager();
			}
			catch (System.Net.Sockets.SocketException exc)
			{
				Simulator.GetInstance().Events.Error(String.Format("Unable to start remote sockets: {0}", exc.Message));
			}
		}

		private void StopRemoteSocket()
		{
			if (remoteSocketsManager != null) // remote sockets have been started
			{
				// unregister the socket object from publication
				RemotingServices.Disconnect(remoteSocketsManager);
			}
		}

		private void StartRemoteControl()
		{
			Debug.WriteLineIf(Simulator.Verbose, "Starting remote control.");

			// This is the only one instance of WrapperControl we create
			RemotingConfiguration.RegisterWellKnownServiceType(typeof(RemoteControl), RemoteControlObjectName, WellKnownObjectMode.Singleton);
			remoteControl = new RemoteControl();
		}

		private void StopRemoteControl()
		{
			Debug.WriteLineIf(Simulator.Verbose, "Stopping remote control.");

			if (remoteControl != null) // remote control has been started
			{
				// unregister the control object from publication
				RemotingServices.Disconnect(remoteControl);
			}
		}

		#endregion

	}
}
