/*

 Copyright (c) 2006 Pavel Novak.

 The use and distribution terms for this software are contained in the file named License.txt, 
 which can be found in the root of the NetSim distribution. By using this software 
 in any fashion, you are agreeing to be bound by the terms of this license.
 
 You must not remove this notice from this software.

*/

using System;
using System.Collections;
using System.Xml;

namespace NetSim.Core
{
	/// <summary>
	/// An interface that should be implemented by all adapters.
	/// </summary>
	/// <remarks>
	/// You do not have to care about this interface if you derive new class from <see cref="Adapter"/>
	/// abstract class.
	/// </remarks>
	public interface IAdapter : IStructureObject
	{
		/// <summary>
		/// Gets or sets the adapter name.
		/// </summary>
		new string Name { get; set; } // we add setter

		/// <summary>
		/// Gets or sets a <see cref="Link"/> the adapter connects to.
		/// </summary>
		Link Link { get; set; }

		/// <summary>
		/// Gets the maximum transfer unit for the adapter.
		/// </summary>
		int MTU	{	get; }

		/// <summary>
		/// Configures the adapter according to <paramref name="configPath"/> configuration file.
		/// </summary>
		/// <param name="configPath">Path to the configuration file.</param>
		void Configure(string configPath);

		/// <summary>
		/// Forces the adapter to send a packet.
		/// </summary>
		/// <param name="pi">Packet to send.</param>
		void SendPacket(PacketInfo pi);

		/// <summary>
		/// Called by the link when an adapter should receive a packet.
		/// </summary>
		/// <param name="p">Packet to receive.</param>
		void ReceivePacket(Packet p);

		/// <summary>
		/// Called by the link when send operation previously requested failed.
		/// </summary>
		/// <param name="p">The packet that was sent.</param>
		/// <param name="retry"><b>True</b> if retrying the send may help, <b>false</b> otherwise.</param>
		void PacketSendingFailed(Packet p, bool retry);

		/// <summary>
		/// Bounds the adapter to the specified node.
		/// </summary>
		/// <param name="node">Node that will be the owner of the adapter.</param>
		void SetNode(Node node);

		/// <summary>
		/// Returns <b>true</b> if the adapter knows how to send packets to the <paramref name="address"/>,
		/// <b>false</b> otherwise. If <b>true</b> is returned, local address (of the same type as
		/// <paramref name="address"/>) is returned in out <paramref name="thisMachineAddress"/> parameter.
		/// </summary>
		/// <param name="address">Destination address.</param>
		/// <param name="thisMachineAddress">Local address.</param>
		/// <returns><b>True</b> if the destination address can be reached by the adapter.</returns>
		bool CanReach(NetworkAddress address, out NetworkAddress thisMachineAddress);

		/// <summary>
		/// Return <b>true</b> if <paramref name="address"/> is bound to this adapter.
		/// </summary>
		/// <param name="address">Address to check if it is a local address.</param>
		/// <returns><b>True</b> if the address is local address.</returns>
		bool IsLocalAddress(NetworkAddress address);

		/// <summary>
		/// Returns <b>true</b> if this adapter should receive a packet that has <paramref name="address"/>
		/// set as destination.
		/// </summary>
		/// <remarks>Note that <b>true</b> is returned not only if <paramref name="address"/> matches
		/// local address, but also if it is for example broadcast address or an adapter is set to
		/// promiscuous mode.</remarks>
		/// <param name="address">Destination address.</param>
		/// <param name="localAddress">Local address (may be different than <paramref name="address"/>).</param>
		/// <returns><b>True</b> if the packet should be received.</returns>
		bool ShouldReceive(NetworkAddress address, out NetworkAddress localAddress);
	}

	/// <summary>
	/// Common ancestor of all adapters.
	/// </summary>
	public abstract class Adapter : StructureObject, IAdapter
	{
		/// <summary>
		/// Gets maximum transfer unit for the adapter.
		/// </summary>
		public abstract int MTU	{	get; }

		/// <summary>
		/// The link this adapter is attached to.
		/// </summary>
		public virtual Link Link
		{
			get { return link; }
			set { link = value; }
		}
		protected Link link;

		/// <summary>
		/// Whether the adapter is enabled or disabled.
		/// </summary>
		public override bool Enabled
		{
			get { return enabled; }
			set { enabled = value; }
		}
		private bool enabled = true;

		/// <summary>
		/// Returns always <b>true</b> since <see cref="Enabled"/> property can be set. For more information
		/// see <see cref="ISuspendable"/>.
		/// </summary>
		public override bool CanDisable
		{
			get { return true; }
		}

		/// <summary>
		/// The node this adapter belogns to.
		/// </summary>
		protected Node node;

		/// <summary>
		/// Packets that have been sent, they should return and it will confirm that the
		/// packet has been sent. Keys are packets hash codes.
		/// </summary>
		// generics: <Packet, int>
		protected Hashtable sentPackets;

		/// <summary>
		/// Collection of <see cref="ProtocolSettings"/> that are configured for this adapter.
		/// </summary>
		// generics: <ProtocolSettings>
		protected ArrayList protocolSettings;

		/// <summary>
		/// Sets the owner node of this adapter.
		/// </summary>
		/// <param name="node">The Node.</param>
		public void SetNode(Node node)
		{
			this.node = node;
		}

		/// <summary>
		/// Gets adapter's hardware address.
		/// </summary>
		public abstract NetworkAddress HardwareAddress { get; }

		/// <summary>
		/// Default constructor allowing creation of the adapter according to the configuration file.
		/// </summary>
		public Adapter()
		{
			sentPackets = Hashtable.Synchronized(new Hashtable());
			protocolSettings = new ArrayList();
		}

		/// <summary>
		/// This default implementation checkes the configuration file for <b>Protocols</b> node and
		/// adds settings for individual protocols associated with this adapter.
		/// </summary>
		/// <param name="document">XML document containing the configuration.</param>
		public override void XmlConfigure(System.Xml.XmlDocument document)
		{
			// try to find Protocols node
			XmlNode node = document.DocumentElement.FirstChild;
			while (node != null)
			{
				if (node.Name == "Protocols")
				{
					// configure the adapter - insert settings for the protocols specified
					XmlConfigureProtocols(node);
				}
				node = node.NextSibling;
			}
		}

		/// <summary>
		/// Configures the adapter for specified protocol.
		/// </summary>
		/// <param name="protocolsNode">XML node of the protocol configuration.</param>
		protected virtual void XmlConfigureProtocols(XmlNode protocolsNode)
		{
			XmlNode settings_node = protocolsNode.FirstChild;
			while (settings_node != null)
			{
				if (settings_node.Name == "Settings")
				{
					XmlAttribute class_attr = settings_node.Attributes["class"];
					if (class_attr == null)
					{
						Simulator.GetInstance().Events.Warning("'class' attribute in protocol settings not found, settings ignored");
						continue;
					}
					ProtocolSettings settings = (ProtocolSettings) Simulator.GetInstance().GetInstanceByName(class_attr.InnerText, typeof(ProtocolSettings));
					settings.Configure(settings_node);
					this.AddProtocolSettings(settings);
				}
				settings_node = settings_node.NextSibling;
			}

		}

		/// <summary>
		/// Adds settings for the specified protocol to the adapter.
		/// </summary>
		/// <param name="settings">Protocol settings to associate with the adapter.</param>
		public void AddProtocolSettings(ProtocolSettings settings)
		{
			this.protocolSettings.Add(settings);
		}

		/// <summary>
		/// <p>Sends the packet to the link.</p>
		/// <p>This method implements default behaviour - all packets will be sent to the link.
		/// Descendants should override this method and send only those packets that has addresses that
		/// should be sent over this adapter.</p>
		/// </summary>
		/// <param name="pi">The <see cref="PacketInfo"/> containing the information about packet to send.</param>
		public virtual void SendPacket(PacketInfo pi)
		{
			if (!Enabled)
				return;

			// default behaviour - send all packets - no header
			Packet to_send = pi.RawPacket != null ? pi.RawPacket : pi.Packet;
			RememberSentPacket(to_send);
			link.SendPacket(this, to_send);
		}

		/// <summary>
		/// Called just before the packet will be sent to the link. Remebers the packet.
		/// </summary>
		/// <param name="p">Packet to remember.</param>
		protected void RememberSentPacket(Packet p)
		{
			// store packet and remember that it is the first try it is being sent
			sentPackets.Add(p, 1);
		}

		/// <summary>
		/// Called by the link to notify the adapter that the packet cannot be send.
		/// </summary>
		/// <remarks>
		/// This method implements the default behavior: the packet is just removed from the list
		/// of sent packets.
		/// </remarks>
		/// <param name="p">Packet that cannot be sent.</param>
		/// <param name="retry"><b>True</b> if retrying to send makes sense.</param>
		public virtual void PacketSendingFailed(Packet p, bool retry)
		{
			// default behaviour: do nothing, just remove from our list of sent packets
			CheckPacketComingBack(p);
		}

		/// <summary>
		/// <p>Receives packet from the link and passes it to the node.</p>
		/// <p>Although this function won't block, it should be called by a separate thread, because
		/// the receiving can take some time and packets should be delivered to all adapters in the link
		/// at about the same time.</p>
		/// </summary>
		/// <param name="p">The received packet.</param>
		public virtual void ReceivePacket(Packet p)
		{
			if (p == null)
				return; 
			if (CheckPacketComingBack(p)) // our packet coming back..
				return;

			if (!Enabled)
				return;

			// pass the packet to the node
			PacketInfo pi = new PacketInfo(p, this);
			pi.RawPacket = p;
			node.ReceivePacket(pi);

			//Future version: notifying node about undelivered packet?
		}

		/// <summary>
		/// If the packet <paramref name="p"/> is our packet coming back as an acknowledgement,
		/// removes it from the list of sent packets and return <b>true</b>.
		/// </summary>
		/// <param name="p">Incoming packet.</param>
		/// <returns><b>True</b> if the packet is our packet coming back, <b>false</b> otherwise.</returns>
		protected bool CheckPacketComingBack(Packet p)
		{
			object num_tries = sentPackets[p];
			if (num_tries != null) // this packet is returned as confirmation of delivery
			{
				// remove from our list of sent packets and return
				sentPackets.Remove(p);
				return true;
			}

			return false;
		}

		/// <summary>
		/// Connects the adapter to the link. Remebers the reference to the link and register the
		/// adapter by the link.
		/// </summary>
		/// <param name="link"></param>
		public void ConnectToLink(Link link)
		{
			// store that we are connected to this link
			this.link = link;
			// let the adapter know that we are connected
			link.RegisterAdapter(this);
		}

		/// <summary>
		/// Returns <b>true</b> if the <paramref name="address"/> is associated (configured) with
		/// this adapter.
		/// </summary>
		/// <param name="address">The address.</param>
		/// <returns><b>True</b> if the address is local, otherwise <b>false</b>.</returns>
		public bool IsLocalAddress(NetworkAddress address)
		{
			foreach (ProtocolSettings ps in protocolSettings)
			{
				if (ps.Address.Equals(address))
					return true;
			}

			// not found - the address is not associated with any adapter
			return false;
		}

		/// <summary>
		/// Returns <b>true</b> if the packet with <paramref name="address"/> should be received by this node.
		/// Note that intermediate nodes (routers) need to receive not only packets that will return true here.
		/// </summary>
		/// <param name="address">Destination address of the packet that is being received.</param>
		/// <param name="localAddress">Local address of the interface that will receive this packet.</param>
		/// <returns><b>True</b> if this adapter (node) is destination of this packet.</returns>
		public bool ShouldReceive(NetworkAddress address, out NetworkAddress localAddress)
		{
			localAddress = null;
			foreach (ProtocolSettings ps in protocolSettings)
			{
				if (ps.ShouldReceive(address, out localAddress))
					return true;
			}

			// not found - the address is not associated with any adapter
			return false;
		}

		/// <summary>
		/// Returns <b>true</b> if the adapter knows how to send packets to the <paramref name="address"/>,
		/// <b>false</b> otherwise. If <b>true</b> is returned, local address (of the same type as
		/// <paramref name="address"/>) is returned in out <paramref name="thisMachineAddress"/> parameter.
		/// </summary>
		/// <param name="address">Destination address.</param>
		/// <param name="thisMachineAddress">Local address.</param>
		/// <returns><b>True</b> if the destination address can be reached by the adapter.</returns>
		public bool CanReach(NetworkAddress address, out NetworkAddress thisMachineAddress)
		{
			foreach (ProtocolSettings ps in protocolSettings)
			{
				if (ps.CanSendPacketTo(address))
				{
					thisMachineAddress = ps.Address;
					return true;
				}
			}

			thisMachineAddress = null;
			return false;
		}
	}


	#region SimpleAdapter
	/// <summary>
	/// Simple adapter that just sends what is coming, does not modify the packet, does not add own
	/// link layer protocol. Does not resend packets.
	/// </summary>
	public class SimpleAdapter : Adapter
	{
		/// <summary>
		/// Returns <see cref="Int32.MaxValue"/>, allowing any size of the packet.
		/// </summary>
		public override int MTU
		{
			// allow all packet sizes
			get { return Int32.MaxValue; }
		}

		/// <summary>
		/// Returns <see cref="SimpleLinkLayerAddress"/> object.
		/// </summary>
		public override NetworkAddress HardwareAddress
		{
			get { return address; }
		}
		private SimpleLinkLayerAddress address = new SimpleLinkLayerAddress();
	}
	#endregion
}
