/*

 Copyright (c) 2006 Pavel Novak.

 The use and distribution terms for this software are contained in the file named License.txt, 
 which can be found in the root of the NetSim distribution. By using this software 
 in any fashion, you are agreeing to be bound by the terms of this license.
 
 You must not remove this notice from this software.

*/

using System;
using NetSim.Core;
using Thread = System.Threading.Thread;
using ArrayList = System.Collections.ArrayList;

namespace NetSim.Core
{
	/// <summary>
	/// Interface for those who do not want to inherit from <see cref="NodeApplication"/>. Instead, you
	/// can implement this interface.
	/// </summary>
	/// <remarks>
	/// Be sure that you implement the methods the way that they cannot fail. It is recommended to add
	/// Run method and in the <see cref="Start"/> just call the Run wrapped into try-catch block:
	/// <code>
	/// try
	/// {
	///		Run();
	/// }
	/// catch(Exception e)
	/// {
	///		Simulator.GetInstance().Events.Error(String.Format("Application threw an exception: {0}", e.Message), e);
	/// }
	/// </code>
	/// </remarks>
	public interface INodeApplication : IStructureObject
	{
		/// <summary>
		/// Configures the application according to <paramref name="configPath"/> configuration file.
		/// </summary>
		/// <param name="configPath">Configuration file.</param>
		void Configure(string configPath);

		/// <summary>
		/// Sets the owner of this application object.
		/// </summary>
		/// <param name="n">Node that will be the owner of this application.</param>
		void SetNode(Node n);

		/// <summary>
		/// Starts the application.
		/// </summary>
		void Start();

		/// <summary>
		/// Gets or sets the main thread of the application.
		/// </summary>
		/// <remarks>Each application runs in a separate thread. The thread is created by the simulator
		/// and set before the application is started.</remarks>
		Thread MainThread { get; set; }

		/// <summary>
		/// List of threads other than the <see cref="MainThread"/> that the application is using. They
		/// should be accessible to the simulator to be able to suspend the application execution.
		/// </summary>
		Thread[] Threads { get; }
	}
	
	/// <summary>
	/// The class allowing the user to add his/her own application running on the virtual node.
	/// </summary>
	public abstract class NodeApplication : StructureObject, INodeApplication
	{
		/// <summary>
		/// Gets the node this application runs on.
		/// </summary>
		protected Node Node { get { return this.node; } }
		private Node node;

		void INodeApplication.SetNode(Node n)
		{
			if (n == null)
				throw new ArgumentNullException("n");

			this.node = n;
		}

		/// <summary>
		/// The main thread that is created and started/stopped by the simulator.
		/// </summary>
		Thread INodeApplication.MainThread
		{
			get { return mainThread; }
			set { mainThread = value; }
		}
		private Thread mainThread;

		/// <summary>
		/// Gets the list of threads other than <see cref="INodeApplication.MainThread"/>.
		/// </summary>
		Thread[] INodeApplication.Threads
		{
			get { return (Thread[])ThreadList.ToArray(typeof(Thread)); }
		}
		protected ArrayList ThreadList = new ArrayList();

		/// <summary>
		/// Default constructor allowing the simulator create an instance without knowing anything about
		/// the class.
		/// </summary>
		public NodeApplication()
		{
		}

		/// <summary>
		/// If you need to do some initialization prior to call the <see cref="Run"/> method, you can override
		/// this method. Do not forget to call Initialize() on the base class!
		/// </summary>
		protected virtual void Initialize()
		{
		}

		/// <summary>
		/// This method is called by the simulator to start the application. It calls the <see cref="Run"/>
		/// method, override <see cref="Run"/> to add your own implementation.
		/// </summary>
		public void Start()
		{
			// it is required to set the Node property by calling SetNode() first
			if (this.Node == null)
			{
				Simulator.GetInstance().Events.Error(String.Format("Cannot start an application {0}: Node property has not been set", this.GetType().FullName));
				return;
			}

			int return_code;
			try
			{
				Initialize();
				return_code = this.Run();
			}
			catch (Exception e)
			{
				Simulator.GetInstance().Events.Error(String.Format("Application ({0}) threw an exception", this.GetType().FullName), e);
				return;
			}

			// code 0 means succesful completion
			if (return_code != 0)
				Simulator.GetInstance().Events.Warning(String.Format("Application ({0}) exited with code {1}", this.GetType().FullName, return_code));
			else
				Simulator.GetInstance().Events.Notice(String.Format("Application ({0}) exited with code {1}", this.GetType().FullName, return_code));
		}

		/// <summary>
		/// Override this method to add your own implementation. The simulator creates new thread to run your
		/// code here.
		/// </summary>
		/// <remarks>
		/// If you need more threads, you can use <see cref="System.Threading.ThreadPool"/> class. If you comlete your
		/// work, return zero, otherwise an warning will be written to the event log.
		/// </remarks>
		/// <returns></returns>
		protected abstract int Run();

	}
}
