/*

 Copyright (c) 2006 Pavel Novak.

 The use and distribution terms for this software are contained in the file named License.txt, 
 which can be found in the root of the NetSim distribution. By using this software 
 in any fashion, you are agreeing to be bound by the terms of this license.
 
 You must not remove this notice from this software.

*/

using System;
using System.Collections;
using System.Threading;

namespace NetSim.Core
{
	#region Link
	/// <summary>
	/// Common ancestor of all link classes.
	/// </summary>
	public abstract class Link : StructureObject
	{
		#region SendPacketState
		/// <summary>
		/// Instances of this class are passed to the thread-pool and than passed back
		/// as state parameter to the delegate function.
		/// </summary>
		public class SendPacketState
		{
			/// <summary>
			/// Packet to send.
			/// </summary>
			public Packet Packet;

			/// <summary>
			/// Adapter where to send.
			/// </summary>
			public IAdapter Adapter;

			/// <summary>
			/// Creates new <see cref="SendPacketState"/> object.
			/// </summary>
			/// <param name="a">Adapter where the packet should be sent.</param>
			/// <param name="p">Packet to send.</param>
			public SendPacketState(IAdapter a, Packet p)
			{
				this.Adapter = a;
				this.Packet = p;
			}
		}
		#endregion

		/// <summary>
		/// Whether the link is enabled or disabled.
		/// </summary>
		public override bool Enabled
		{
			get { return enabled; }
			set { enabled = value; }
		}
		private bool enabled = true;

		/// <summary>
		/// Returns always <b>true</b> since <see cref="Enabled"/> property can be set. For more information
		/// see <see cref="ISuspendable"/>.
		/// </summary>
		public override bool CanDisable
		{
			get { return true; }
		}

		/// <summary>
		/// Gets or sets the maximum number of adapters allowed by the link. Default value is
		/// <see cref="Int32.MaxValue"/>.
		/// </summary>
 		public int MaxAdapters
		{
			get { return maxAdapters; }
			set { maxAdapters = value; }
		}
		private int maxAdapters = Int32.MaxValue;

		/// <summary>
		/// Adapters that are connected to this link.
		/// </summary>
		protected ArrayList adapters;

		/// <summary>
		/// Default constructor allows creation according to configuration file.
		/// </summary>
		public Link()
		{
			adapters = new ArrayList();

			Name = String.Empty;
			MaxAdapters = Int32.MaxValue;
		}

		/// <summary>
		/// Registers the adapter with the link. From this moment the link will send all packets
		/// also to this adapter.
		/// </summary>
		/// <remarks>This method is thread-safe.</remarks>
		/// <param name="a">An adapter to register.</param>
		public void RegisterAdapter(IAdapter a)
		{
			lock (adapters.SyncRoot)
			{
				if (adapters.Count >= MaxAdapters)
					throw new InvalidOperationException("Maximum number of adapters exceeded.");

				adapters.Add(a);
			}
		}

		/// <summary>
		/// Removes adapter from the list of registered adapters.
		/// This method is thread-safe.
		/// </summary>
		/// <param name="a"></param>
		public void DeregisterAdapter(IAdapter a)
		{
			lock (adapters.SyncRoot)
			{
				adapters.Remove(a);
			}
		}

		/// <summary>
		/// Sends the packet to all adapters connected to the link.
		/// </summary>
		/// <param name="adapter">Adapter that is sending the packet.</param>
		/// <param name="p">Packet to send</param>
		public virtual void SendPacket(IAdapter adapter, Packet p)
		{
			// do nothing if the link is disabled
			if (!Enabled)
				return;

			try
			{
				// default behaviour - send packet immediately to all adapters.
				foreach (IAdapter a in adapters)
				{
					// for each adapter process by a separate thread - receiving the packet can take some time...
					ManagedThreadPool.QueueUserWorkItem(new WaitCallback(this.SendPacketCallback),
						new SendPacketState(a, p));
				}
			}
			catch (InvalidOperationException)
			{
				// the collection was modified - ignore.
				// This allows changing network structure while running - some packets won't be
				// delivered just after the change - it is not crucial.
			}
		}

		/// <summary>
		/// Callback method for processing packet sending.
		/// </summary>
		/// <param name="state"></param>
		protected virtual void SendPacketCallback(object state)
		{
			SendPacketState send_state = (SendPacketState) state;

			// send the packet to the adapter
			send_state.Adapter.ReceivePacket(send_state.Packet);
		}
	}
	#endregion

	#region SimpleLink
	/// <summary>
	/// Simple link implementation that delivers packets with no delay, no loss...
	/// </summary>
	public class SimpleLink : Link
	{
		// thanks to default implementation in the abstract base class,
		// nothing is needed here...
	}
	#endregion
}
