/*

 Copyright (c) 2006 Pavel Novak.

 The use and distribution terms for this software are contained in the file named License.txt, 
 which can be found in the root of the NetSim distribution. By using this software 
 in any fashion, you are agreeing to be bound by the terms of this license.
 
 You must not remove this notice from this software.

*/

using System;

namespace NetSim.Core
{
	/// <summary>
	/// An interface that should be implemented by all modules that attach to a <see cref="Node"/>.
	/// </summary>
	/// <remarks>You do not have to care about this interface if you implement new module that
	/// inherits from <see cref="Module"/>.</remarks>
	public interface IModule : IStructureObject
	{
		/// <summary>
		/// The name of the module. Can be <b>null</b>. If it is not <b>null</b>, it must be unique for
		/// the modules associtated with one node.
		/// </summary>
		new string Name { get; set; } // we add setter

		/// <summary>
		/// ID that the module should pass every time it interacts with other code.
		/// </summary>
		object ID { get; set; }

		/// <summary>
		/// The node this module belongs to. You do not need to care about this property if you implement
		/// a module independent of node or adapters connected to that node.
		/// </summary>
		Node Node { get; set; }

		/// <summary>
		/// Configures the module according to the configuration file specified.
		/// </summary>
		/// <param name="configPath">Configuration file absolute or relative path.</param>
		void Configure(string configPath);

		/// <summary>
		/// Method called when the module should process the packet that is being received.
		/// </summary>
		/// <param name="pInfo">The <see cref="PacketInfo"/> instance. Module might change the packet, the change
		/// will be passed to other modules, or set it to <b>null</b>, in that case the packet won't be
		/// passed to the next modules.</param>
		void ReceivePacket(ref PacketInfo pInfo);

		/// <summary>
		/// <p>Method called when the module should process the packet that is being sent.</p>
		/// <p>The module might change the value to some other <see cref="Packet"/>. This will be done
		/// typically by the modules representing one layer in the protocol stack - the incomming packet
		/// will be encapsulated by the new type of packet of the lower layer.</p>
		/// </summary>
		/// <param name="pInfo">The <see cref="PacketInfo"/> instance. It can be changed.</param>
		void SendPacket(ref PacketInfo pInfo);

		/// <summary>
		/// Event raised by the module if the module generates the new packet that should be sent.
		/// </summary>
		event Module.NewPacketDelegate SendNewPacketEvent;

		/// <summary>
		/// Event raised by the module if the module generates the new packet that should be received.
		/// </summary>
		event Module.NewPacketDelegate ReceiveNewPacketEvent;
	}

	/// <summary>
	/// An abstract class for modules.
	/// </summary>
	/// <remarks>New modules can be implemented either by inheritting from this class or by
	/// implementing <see cref="IModule"/> interface.</remarks>
	public abstract class Module : StructureObject, IModule
	{
		/// <summary>
		/// Delegate for <see cref="SendNewPacketEvent"/> and <see cref="ReceiveNewPacketEvent"/>.
		/// </summary>
		public delegate void NewPacketDelegate(object moduleID, PacketInfo pInfo);

		/// <summary>
		/// The Node this module belongs to.
		/// </summary>
		public Node Node
		{
			get { return node; }
			set { node = value; }
		}
		private Node node;

		/// <summary>
		/// Module identifier set by the node at the time the module is created and associated wit the node.
		/// </summary>
		public object ID
		{
			get { return id; }
			set { id = value; }
		}
		private object id;

		/// <summary>
		/// Default constructor allowing creation of the module according to the configuration file.
		/// </summary>
		public Module()
		{
		}

		/// <summary>
		/// <p>Process the packet that is being received. This implementation makes no change to the packet.</p>
		/// <p>For more information <see cref="IModule.ReceivePacket"/>.</p>
		/// </summary>
		/// <param name="pInfo">The <see cref="PacketInfo"/> instance.</param>
		public virtual void ReceivePacket(ref PacketInfo pInfo)
		{
			// default behaviour - do nothing with the packet
		}

		/// <summary>
		/// <p>Process the packet that is being sent. This implementation makes no change to the packet.</p>
		/// <p>For more information <see cref="IModule.SendPacket"/>.</p>
		/// </summary>
		/// <param name="pInfo">The <see cref="PacketInfo"/> instance.</param>
		public virtual void SendPacket(ref PacketInfo pInfo)
		{
			// default behaviour - do nothing with the packet
		}

		/// <summary>
		/// Event that the module owner registers to and that is fired if this module wants to send
		/// completely new packet.
		/// </summary>
		public event NewPacketDelegate SendNewPacketEvent;

		/// <summary>
		/// Fires an <see cref="SendNewPacketEvent"/> event.
		/// </summary>
		/// <param name="pInfo"></param>
		protected void SendNewPacket(PacketInfo pInfo)
		{
			ManagedThreadPool.QueueUserWorkItem(new System.Threading.WaitCallback(this.SendNewPacketCallback), pInfo);
		}

		/// <summary>
		/// Callback function called by a thread-pool thread when new packet should be sent. Fires 
		/// </summary>
		/// <param name="state"></param>
		private void SendNewPacketCallback(object state)
		{
			this.SendNewPacketEvent(this.ID, (PacketInfo)state);
		}

		/// <summary>
		/// An Event that the module owner registers to and that is fired when this module wants the node
		/// to receive completely new packet.
		/// </summary>
		public event NewPacketDelegate ReceiveNewPacketEvent;

		/// <summary>
		/// Fires an <see cref="ReceiveNewPacketEvent"/> event.
		/// </summary>
		/// <param name="moduleID"></param>
		/// <param name="pInfo"></param>
		protected void ReceiveNewPacket(object moduleID, PacketInfo pInfo)
		{
			this.ReceiveNewPacketEvent(moduleID, pInfo);
		}
	}

}
