/*

 Copyright (c) 2006 Pavel Novak.

 The use and distribution terms for this software are contained in the file named License.txt, 
 which can be found in the root of the NetSim distribution. By using this software 
 in any fashion, you are agreeing to be bound by the terms of this license.
 
 You must not remove this notice from this software.

*/

using System;
using System.Diagnostics;

namespace NetSim.Core
{
	/// <summary>
	/// Common ancestor for classes used for addresses in the simulator.
	/// </summary>
	public abstract class NetworkAddress
	{
		/// <summary>
		/// Equals and GetHashCode should be overriden to compare bytes and not references..
		/// A descendant of this class has to override this method.
		/// </summary>
		/// <param name="obj">Object to compare to.</param>
		/// <returns><b>True</b> if this instance is equal to <paramref name="obj"/>.</returns>
		public override bool Equals(object obj)
		{
			Debug.Fail("NetworkAddress descendant does not override Equals()");
			return false;
		}

		/// <summary>
		/// Equals and GetHashCode should be overriden to compare bytes and not references..
		/// A descendant of this class has to override this method.
		/// </summary>
		/// <returns>Hash code.</returns>
		public override int GetHashCode()
		{
			Debug.Fail("NetworkAddress descendant does not override GetHashCode()");
			return 0;
		}

		/// <summary>
		/// Gets a length of the address in bytes.
		/// </summary>
		public abstract byte Length { get; }

		/// <summary>
		/// Returns an array of bytes of the address.
		/// </summary>
		/// <returns>Bytes of the address.</returns>
		public abstract byte[] GetAddressBytes();

		/// <summary>
		/// Returns new address object of the same type as current instance, but it will contain
		/// <paramref name="bytes"/> bytes as the address.
		/// </summary>
		/// <param name="bytes">Bytes of the address.</param>
		/// <returns>New object of the same type.</returns>
		public abstract NetworkAddress CreateNewAddress(byte[] bytes);

		/// <summary>
		/// Gets a group of packets for which the address is suitable.
		/// </summary>
		public abstract Packet.PacketGroup PacketGroup { get; }

		/// <summary>
		/// Gets a packet type within a group for which the address is suitable.
		/// </summary>
		public abstract int PacketType { get; }

		/// <summary>
		/// Returns a string representation of the address. Default is an IPv4-like representation;
		/// bytes as decimal numbers separated by dots.
		/// </summary>
		/// <returns>String representation of the address.</returns>
		public override string ToString()
		{
			byte[] bytes = GetAddressBytes();
			System.Text.StringBuilder sb = new System.Text.StringBuilder();
			for (int i=0; i < bytes.Length; i++)
			{
				if (i > 0) sb.Append('.');
				sb.Append(bytes[i]);
			}
			return sb.ToString();
		}

		/// <summary>
		/// Loopback address if it makes sense, <b>null</b> otherwise.
		/// </summary>
		public virtual NetworkAddress Loopback
		{
			get { return null; }
		}
	}

	/// <summary>
	/// A class that implements some abstract methods of <see cref="NetworkAddress"/> in a way
	/// that will be useful for some simple address implementations. Uses an integer as address value.
	/// </summary>
	public abstract class SimpleNetworkAddress : NetworkAddress
	{
		/// <summary>
		/// Address value.
		/// </summary>
		protected UInt32 number;

		/// <summary>
		/// Returns <b>true</b> if <paramref name="obj"/> equals to this instance.
		/// </summary>
		/// <param name="obj">Object to compare.</param>
		/// <returns><b>True</b> if <paramref name="obj"/> is equal.</returns>
		public override bool Equals(object obj)
		{
			SimpleNetworkAddress sna = obj as SimpleNetworkAddress;
			if (sna == null)
				return false;

			return this.number == sna.number;
		}

		/// <summary>
		/// Returns hash code for this object.
		/// </summary>
		/// <returns>Hash code.</returns>
		public override int GetHashCode()
		{
			return unchecked((int)number);
		}

		/// <summary>
		/// Returns always 4 since <see cref="UInt32"/> is used as address value.
		/// </summary>
		public override byte Length
		{
			get { return 4; }
		}

		/// <summary>
		/// Returns address bytes. Byte order is Big Endians.
		/// </summary>
		/// <returns>Address bytes.</returns>
		public override byte[] GetAddressBytes()
		{
			ByteArrayList bal = new ByteArrayList(4);
			bal.AddUInt32(number);
			return bal.Bytes;
		}

		/// <summary>
		/// Do not call this method. Returns always <b>null</b>.
		/// </summary>
		/// <param name="bytes"></param>
		/// <returns></returns>
		public override NetworkAddress CreateNewAddress(byte[] bytes)
		{
			Debug.Fail("Should not be called.");
			return null;
		}
	}

	/// <summary>
	/// Address used by <see cref="SimpleAdapter"/> as its hardware address.
	/// </summary>
	public class SimpleLinkLayerAddress : SimpleNetworkAddress
	{
		/// <summary>
		/// Each time new instance is created, its address value is set to the next number. This is
		/// global counter for all class instances.
		/// </summary>
		private static UInt32 counter = 1;

		/// <summary>
		/// Lock used for incrementing <see cref="counter"/>.
		/// </summary>
		private static object _lock = new object();

		/// <summary>
		/// Creates new <see cref="SimpleLinkLayerAddress"/> object.
		/// </summary>
		public SimpleLinkLayerAddress()
		{
			lock(_lock)
			{
				this.number = counter++;
			}
		}

		/// <summary>
		/// Returns always <see cref="Packet.PacketGroup.LinkLayer"/>.
		/// </summary>
		public override Packet.PacketGroup PacketGroup
		{
			get { return Packet.PacketGroup.LinkLayer; }
		}

		/// <summary>
		/// Returns always zero.
		/// </summary>
		public override int PacketType
		{
			get { return 0; }
		}

	}
}
