/*

 Copyright (c) 2006 Pavel Novak.

 The use and distribution terms for this software are contained in the file named License.txt, 
 which can be found in the root of the NetSim distribution. By using this software 
 in any fashion, you are agreeing to be bound by the terms of this license.
 
 You must not remove this notice from this software.

*/

using System;
using System.Collections;
using System.Diagnostics;
using System.Threading;
using ThreadState = System.Threading.ThreadState;

namespace NetSim.Core
{
	/// <summary>
	/// Instances of <see cref="Node"/> class represent virtual network nodes. They are hosts for
	/// <see cref="IAdapter"/>, <see cref="IModule"/>, and <see cref="INodeApplication"/> objects.
	/// </summary>
	public class Node : StructureObject
	{
		/// <summary>
		/// List of adapters via which the node can communicate.
		/// </summary>
		// generics: <string, IAdapter>
		private Hashtable adapters;

		/// <summary>
		/// List of modules that provides functionality of the node. The order of modules is important - they
		/// are inserted to the list as they are added to the node.
		/// </summary>
		// generics: <IModule>
		private ArrayList modules;

		/// <summary>
		/// Keys: modules names, values: modules. The same modules as are inserted in <see cref="modules"/> list.
		/// If the module has no name, it is not inserted into this list.
		/// </summary>
		// generics: <string, IModule>
		private Hashtable modulesByName;

		/// <summary>
		/// List of <see cref="INodeApplication"/> objects running on this node.
		/// </summary>
		// generics: <INodeApplication>
		private ArrayList applications;

		/// <summary>
		/// Whether the node is enabled or disabled.
		/// </summary>
		public override bool Enabled
		{
			get { return enabled; }
			set
			{
				if (enabled == value)
					return; // no change
				
				foreach (DictionaryEntry de in adapters)
				{
					((IAdapter)de.Value).Enabled = value;
				}

				enabled = value;
			}
		}
		private bool enabled = true;

		/// <summary>
		/// Returns always <b>true</b> since <see cref="Enabled"/> property can be set. For more information
		/// see <see cref="ISuspendable"/>.
		/// </summary>
		public override bool CanDisable
		{
			get { return true; }
		}

		/// <summary>
		/// Node requires default constructor to allow creating from the configuration file.
		/// </summary>
		public Node()
		{
			adapters = new Hashtable();
			modules = new ArrayList();
			modulesByName = new Hashtable();
			applications = new ArrayList();

			// register delegates
			Simulator.GetInstance().StartEvent += new Simulator.StartEventDelegate(this.SimulatorStartEventHandler);
			Simulator.GetInstance().StopEvent += new Simulator.StopEventDelegate(this.SimulatorStopEventHandler);
		}

		/// <summary>
		/// Default gateways for this node. Keys are protocol numbers (<see cref="Packet.NetworkLayerPacketType"/>).
		/// </summary>
		private Hashtable defaultGateways = new Hashtable();
		public void SetDefaultGateway(Packet.NetworkLayerPacketType netType, NetworkAddress gateway)
		{
			this.defaultGateways[netType] = gateway;
		}

		/// <summary>
		/// Registers new adapter with the node. The node will use this adapter to send packets.
		/// </summary>
		/// <param name="a">The adapter.</param>
		public void AddAdapter(IAdapter a)
		{
			if (a.Name == null)
				throw new InitializationException(String.Format("Cannot add adapter without name in node {0}", this.Name));
			if (adapters.Contains(a.Name))
				throw new InitializationException(String.Format("Duplicate adapter name {0} in node {1}", a.Name, this.Name));

			// remember our adapter
			adapters.Add(a.Name, a);

			// inform the adapter that it belongs to this node
			a.SetNode(this);
		}

		/// <summary>
		/// Returns an adapter of specified name.
		/// </summary>
		/// <param name="adapterName">Adapter name as specified in the configuration file.</param>
		/// <returns>An adapter or <b>null</b> if such adapter doesn't exist.</returns>
		public IAdapter GetAdapter(string adapterName)
		{
			return (IAdapter) adapters[adapterName];
		}

		/// <summary>
		/// Returns a module of specified name.
		/// </summary>
		/// <param name="moduleName">Module name as specified in the configuration file.</param>
		/// <returns>A module or <b>null</b> if such module doesn't exist.</returns>
		public IModule GetModule(string moduleName)
		{
			return (IModule) modulesByName[moduleName];
		}

		/// <summary>
		/// Returns an application of specified name.
		/// </summary>
		/// <param name="name">A name of the application.</param>
		/// <returns>Application or <b>null</b> if not found.</returns>
		public INodeApplication GetApplication(string name)
		{
			foreach (INodeApplication app in applications)
			{
				if (app.Name == name)
					return app;
			}
			return null;
		}

		/// <summary>
		/// Returns a specified item (adapter, module, or application).
		/// </summary>
		/// <param name="kind">Which kind of the item should be returned.</param>
		/// <param name="name">The name of the item.</param>
		/// <returns>An item or <b>null</b> if no match is found.</returns>
		public IStructureObject GetItem(NodeItemKind kind, string name)
		{
			if (name == null)
				throw new ArgumentNullException("name");
			if (name == String.Empty)
				throw new ArgumentException("Name should not be an empty string.", "name");

			switch (kind)
			{
				case NodeItemKind.Adapter:
					return GetAdapter(name);
				case NodeItemKind.Module:
					return GetModule(name);
				case NodeItemKind.Application:
					return GetApplication(name);
			}

			return null;
		}

		/// <summary>
		/// Associates the new module with this node. The module is added to the end of the chain.
		/// </summary>
		/// <param name="m"></param>
		public void AddModule(IModule m)
		{
			// remember the module
			int id = modules.Add(m);

			if (m.Name != null)
			{
				try { modulesByName.Add(m.Name, m); }
				catch (ArgumentException e)
				{
					throw new InitializationException(String.Format("Two modules have the same name '{0}'", m.Name), e);
				}
			}

			// Set the ID that will be passed to us every time new packet is being sent by the module.
			// We use this ID to determine at which postion the current module is and which other modules
			// we should call before sending the packet to the adapter.
			m.ID = id;

			// register our method with the event that is rised when new packet to send is created.
			m.SendNewPacketEvent += new Module.NewPacketDelegate(NewPacketToSend);	
			// and the method for new packets to receive
			m.ReceiveNewPacketEvent += new Module.NewPacketDelegate(NewPacketToReceive);
		}

		/// <summary>
		/// Associates an application with the node.
		/// </summary>
		/// <param name="app">An application to add to this node.</param>
		public void AddNodeApplication(INodeApplication app)
		{
			// remember the application
			this.applications.Add(app);

			// create the application main thread
			app.MainThread = new Thread(new ThreadStart(app.Start));
			app.MainThread.Name = "NodeApplicationThread";
		}

		/// <summary>
		/// Sends the packet to the adapter (if specified) or all adapters.
		/// </summary>
		/// <param name="pInfo"></param>
		protected virtual void SendPacket(PacketInfo pInfo)
		{
			// if we have adapter specified, use that adapter
			if (pInfo.UseAdapter != null)
			{
				pInfo.UseAdapter.SendPacket(pInfo);
			}
			else // send to all adapters, every adapter will check if it should send the packet or not
			{
				foreach (DictionaryEntry de in adapters)
				{
					((IAdapter)de.Value).SendPacket(pInfo);
				}
			}
		}

		/// <summary>
		/// Sends a packet to the specified adapter.
		/// </summary>
		/// <param name="pInfo">Packet to send.</param>
		/// <param name="adapter">An adapter where the packet should be sent.</param>
		private void SendPacketToAdapter(PacketInfo pInfo, IAdapter adapter)
		{
			adapter.SendPacket(pInfo);
		}

		/// <summary>
		/// Receives the <see cref="Packet"/> from the adapter and passes it to the modules.
		/// </summary>
		/// <param name="pInfo">The packet to receive.</param>
		public virtual void ReceivePacket(PacketInfo pInfo)
		{
			// receive packet starting at the first module
			ReceivePacketInternal(0, pInfo);
		}

		/// <summary>
		/// Takes the new packet from the module, passes it down to other modules and finally to the adapters.
		/// </summary>
		/// <param name="moduleID">Module ID that is sending the packet.</param>
		/// <param name="pInfo">Packet to send.</param>
		public virtual void NewPacketToSend(object moduleID, PacketInfo pInfo)
		{
			Debug.Assert(moduleID is int && pInfo != null);
			
			// pass the pInfo to all previous modules
			for (int i = (int)moduleID - 1; i >= 0; i--)
			{
				((IModule)modules[i]).SendPacket(ref pInfo);
				if (pInfo == null)
					break;
			}

			// and finally to the adapters
			if (pInfo != null)
				this.SendPacket(pInfo);
		}

		/// <summary>
		/// Takes the new packet from the module and passes it up to other modules above.
		/// </summary>
		/// <param name="moduleID">ID of the module that is passing the packet to receive.</param>
		/// <param name="pInfo">Packet to receive.</param>
		public virtual void NewPacketToReceive(object moduleID, PacketInfo pInfo)
		{
			Debug.Assert(moduleID is int && pInfo != null);

			ReceivePacketInternal((int)moduleID + 1, pInfo);
		}

		/// <summary>
		/// Passes the packet to modules associated with this node starting at position <paramref name="startAt"/>.
		/// </summary>
		/// <param name="startAt">The number of module to start.</param>
		/// <param name="pInfo">Packet to receive.</param>
		private void ReceivePacketInternal(int startAt, PacketInfo pInfo)
		{
			for (int i = startAt; i < modules.Count; i++)
			{
				((IModule)modules[i]).ReceivePacket(ref pInfo);

				if (pInfo == null)
					break; // the packet was processed by the module
			}
		}

		/// <summary>
		/// Sets ThisMachine, NextHopDestination and UseAdapter fields according to RemoteAddress.
		/// ThisMachine will be preserved if it is set already.
		/// </summary>
		/// <param name="pi"></param>
		public void InitPacketInfo(PacketInfo pi)
		{
			Debug.Assert(pi.RemoteAddress != null);

			if (pi.UseAdapter != null || pi.NextHopDestination != null)
				return; // nothing to do, values set already

			// try to find the network where the destination is..
			foreach (DictionaryEntry de in adapters)
			{
				IAdapter a = (IAdapter)de.Value;
				NetworkAddress this_address;
				if (a.CanReach(pi.RemoteAddress, out this_address)) // destination is directly reacheable through this adapter
				{
					pi.UseAdapter = a;
					pi.NextHopDestination = pi.RemoteAddress;
					if (pi.ThisMachine == null) pi.ThisMachine = this_address;
					break;
				}
			}

			if (pi.UseAdapter == null) // ultimate destination is not direcly reacheable, try default gateway
			{
				NetworkAddress def_address = (NetworkAddress) this.defaultGateways[(Packet.NetworkLayerPacketType)pi.RemoteAddress.PacketType];
				if (def_address != null)
				{
					// try to find the network where the default destination is..
					NetworkAddress this_address;
					foreach (DictionaryEntry de in adapters)
					{
						IAdapter a = (IAdapter)de.Value;
						if (a.CanReach(def_address, out this_address))
						{
							pi.UseAdapter = a;
							if (pi.ThisMachine == null) pi.ThisMachine = this_address;
							break;
						}
					}

					// set only if the adapter was found - i.e. set all values or nothing
					if (pi.UseAdapter != null)
						pi.NextHopDestination = def_address;
				}
			}
			return;
		}

		/// <summary>
		/// Returns <b>true</b> if the <paramref name="address"/> is a local address.
		/// </summary>
		/// <param name="address">An address to check.</param>
		/// <returns><b>True</b> if <paramref name="address"/> is a local address.</returns>
		public bool IsLocalAddress(NetworkAddress address)
		{
			// loopback
			NetworkAddress loopback = address.Loopback;
			if (loopback != null) // loopback makes sense for this network address, check it
			{
				if (loopback.Equals(address))
					return true;
			}

			foreach (DictionaryEntry de in adapters)
			{
				IAdapter a = (IAdapter)de.Value;
				if (a.IsLocalAddress(address))
					return true;
			}

			return false;
		}

		/// <summary>
		/// Returns <b>true</b> if the packet with <paramref name="address"/> should be received by this node.
		/// Note that intermediate nodes (routers) need to receive not only packets that will return true here.
		/// </summary>
		/// <param name="address">Destination address of the packet that is being received.</param>
		/// <param name="localAddress">Local address of the interface that will receive this packet.</param>
		/// <returns><b>True</b> if this node is destination of this packet.</returns>
		public bool ShouldReceive(NetworkAddress address, out NetworkAddress localAddress)
		{
			localAddress = null;
			foreach (DictionaryEntry de in adapters)
			{
				IAdapter a = (IAdapter)de.Value;
				if (a.ShouldReceive(address, out localAddress))
					return true;
			}

			return false;
		}

		/// <summary>
		/// Called when the simulator is started.
		/// </summary>
		public void SimulatorStartEventHandler()
		{
			// start the application threads
			foreach (INodeApplication app in this.applications)
			{
					try { app.MainThread.Start(); }
					catch (ThreadStateException) { }
			}
		}

		/// <summary>
		/// Called when the simulator is stopped.
		/// </summary>
		public void SimulatorStopEventHandler()
		{
			// suspend all the application threads
			foreach (INodeApplication app in this.applications)
			{
				for (int i = 0; i <= app.Threads.Length; i++)
				{
					Thread t = i == app.Threads.Length ? app.MainThread : app.Threads[i];

					switch (t.ThreadState)
					{
						case ThreadState.Aborted:
						case ThreadState.AbortRequested:
							continue;

						default:
							t.Abort();
							break;
					}
				}
			}
		}

		/// <summary>
		/// Called when the simulator execution is resumed after being suspended.
		/// </summary>
		public void SimulatorResumeEventHandler()
		{
			// start the application threads
			foreach (INodeApplication app in this.applications)
			{
				for (int i = 0; i <= app.Threads.Length; i++)
				{
					Thread t = i == app.Threads.Length ? app.MainThread : app.Threads[i];
					
					try { t.Resume(); }
					catch (ThreadStateException) { }
				}
			}
		}

		/// <summary>
		/// Called when the simulator execution is being suspended.
		/// </summary>
		public void SimulatorSuspendEventHandler()
		{
			// suspend all the application threads
			foreach (INodeApplication app in this.applications)
			{
				for (int i = 0; i <= app.Threads.Length; i++)
				{
					Thread t = i == app.Threads.Length ? app.MainThread : app.Threads[i];
					
					switch (t.ThreadState)
					{
						case ThreadState.Stopped:
						case ThreadState.StopRequested:
						case ThreadState.Aborted:
						case ThreadState.AbortRequested:
							continue;

						default:
							try { t.Suspend(); }
							catch (ThreadStateException) { }
							break;
					}
				}
			}
		}
	}
}
