/*

 Copyright (c) 2006 Pavel Novak.

 The use and distribution terms for this software are contained in the file named License.txt, 
 which can be found in the root of the NetSim distribution. By using this software 
 in any fashion, you are agreeing to be bound by the terms of this license.
 
 You must not remove this notice from this software.

*/

using System;
using System.Diagnostics;

namespace NetSim.Core
{
	#region PacketInfo
	/// <summary>
	/// The info about packet that is being sent or received.
	/// </summary>
	public class PacketInfo : ICloneable
	{
		/// <summary>
		/// Current postion in the stack.
		/// </summary>
		public Packet Packet;

		/// <summary>
		/// While receiving, contains the packet of the link layer, while sending, can be set to force
		/// link layer to send this packet instead of <see cref="PacketInfo.Packet"/> encapsulated by the
		/// link layer header and footer.
		/// </summary>
		public Packet RawPacket;

		/// <summary>
		/// <p>While receiving, contains the address of adapter that has received the packet (its settings
		/// match the packet destination address) or null if any adapter could receive the packet (for
		/// example if the target of the packet was broadcast address).</p>
		/// <p>While sending, it contains the address of this machine that will be used as source address
		/// of the packet being sent.</p>
		/// </summary>
		public NetworkAddress ThisMachine;

		/// <summary>
		/// While sending, upper layers before the network layer set this value to the destination to where
		/// the packet should be sent, while receiving, network layer set this value to the network address
		/// of the sender.
		/// </summary>
		public NetworkAddress RemoteAddress;

		/// <summary>
		/// <p>Used only while sending the packet. If the destination is not located in the network the adapter
		/// is directly attached to, the packet should be sent to the router which will redirect it to
		/// the next network in the route. This field contains protocol address of that router.</p>
		/// <p>If the value is <b>null</b>, the packet will be broadcasted (if the underlying technology supports
		/// broadcast) or send via simulated broadcast (if possible) to all nodes connected to the "physical" network.</p>
		/// </summary>
		/// <remarks>
		/// For hosts, this value is usually set to default gateway from the configuration or obtained from
		/// DHCP server, routers set this value according to routing tables.
		/// </remarks>
		public NetworkAddress NextHopDestination;

		/// <summary>
		/// Used only while sending and only if the packet is being sent through a socket. Otherwise contains <b>null</b>.
		/// </summary>
		public NetSimSocket Socket;

		/// <summary>
		/// Adapter that should be used for sending the packet or through which the packet has been
		/// received. If no such information is available, value is null.
		/// </summary>
		public IAdapter UseAdapter;

		/// <summary>
		/// The adapter that should be excluded while sending the packet.
		/// </summary>
		public IAdapter ExcludeAdapter;

		/// <summary>
		/// Creates new <see cref="PacketInfo"/> object and sets <see cref="Packet"/> member to
		/// <paramref name="packet"/>.
		/// </summary>
		/// <param name="packet">Packet that should be assigned to <see cref="Packet"/> member.</param>
		public PacketInfo(Packet packet) : this(packet, null)
		{
		}

		/// <summary>
		/// Creates new <see cref="PacketInfo"/> object and sets <see cref="Packet"/> and
		/// <see cref="UseAdapter"/> members.
		/// </summary>
		/// <param name="packet">Packet that should be assigned to <see cref="Packet"/> member.</param>
		/// <param name="useAdapter">Adapter that should be assigned to <see cref="UseAdapter"/> member.</param>
		public PacketInfo(Packet packet, IAdapter useAdapter)
		{
			this.Packet = packet;
			this.UseAdapter = useAdapter;
		}

		#region ICloneable Members
		/// <summary>
		/// Creates a shallow copy of this <see cref="PacketInfo"/> instance.
		/// </summary>
		/// <returns>Shallow copy of this instance.</returns>
		public object Clone()
		{
			PacketInfo result = new PacketInfo(this.Packet);
			result.RawPacket = this.RawPacket;
			result.ThisMachine = this.ThisMachine;
			result.RemoteAddress = this.RemoteAddress;
			result.NextHopDestination = this.NextHopDestination;
			result.UseAdapter = this.UseAdapter;
			result.ExcludeAdapter = this.ExcludeAdapter;
			result.Socket = this.Socket;

			return result;
		}
		#endregion

	}
	#endregion

	#region Packet (abstract class)
	/// <summary>
	/// Common ancestor of all packet classes.
	/// </summary>
	public abstract class Packet
	{
		#region Enums
		/// <summary>
		/// Type of the packet that should be created, or that the packet factory can create.
		/// </summary>
		/// <remarks>
		/// This enum is used in the special way in <see cref="Simulator"/> for storing references to individual
		/// packet factories, so the underlying type should be always <see cref="int"/> and the values should
		/// be zero-based with no spaces ending by dummy entry <see cref="PacketGroup._max"/>.
		/// </remarks>
		public enum PacketGroup
		{
			/// <summary>
			/// Packet at the link layer - layer 2 in ISO/OSI, such as Ethernet frame. The identification used for this types
			/// of packets is based on WinPcap source code (see libpcap/pcap-bpf.h) because these
			/// packets are usually created in behalf of ExternalConnection module located in the NetSimLibrary
			/// assembly.
			/// </summary>
			LinkLayer,

			/// <summary>
			/// Packet at the network layer - layer 3 in ISO/OSI, such as IP packet. The identification used
			/// for this type of packets is based on numbers assigned by IANA for the field "protocol" in
			/// Ethernet frame (see http://www.iana.org/assignments/ethernet-numbers). This solution was chosen
			/// because Ethernet is the most wide-spread technology and these numbers are used not only by
			/// Ethernet (for example ARP).
			/// </summary>
			NetworkLayer,

			/// <summary>
			/// Packet at transport layer - layer 4 in ISO/OSI, such as TCP or UDP. The identification is based on
			/// "protocol" field in the IP packet, because IP is the leading network technology and in most cases
			/// the packets at transport layer are encapsulated into the IP packets.
			/// </summary>
			TransportLayer,

			/// <summary>
			/// This packet type is added for other user-defined packets that do not match any other group or there
			/// is no space for identification.
			/// </summary>
			Other,

			/// <summary>
			/// Dummy entry that should be always the last one.
			/// </summary>
			_max
		}

		/// <summary>
		/// Possibly supported underlying data link types. For easy extension of this enum the values are
		/// always used as <see cref="int"/> type, not this enum type.
		/// </summary>
		/// <remarks>
		/// Copied from WinPcap 3.1 source code, file libpcap/pcap-bpf.h
		/// </remarks>
		public enum LinkLayerPacketType
		{
			DLT_NULL = 0,	  // BSD loopback encapsulation
			DLT_EN10MB = 1, // Ethernet (10Mb)
			DLT_EN3MB = 2,  // Experimental Ethernet (3Mb)
			DLT_AX25 = 3,   // Amateur Radio AX.25
			DLT_PRONET = 4, // Proteon ProNET Token Ring
			DLT_CHAOS = 5,  // Chaos
			DLT_IEEE802 = 6,// IEEE 802 Networks
			DLT_ARCNET = 7, // ARCNET, with BSD-style header
			DLT_SLIP = 8,   // Serial Line IP
			DLT_PPP = 9,    // Point-to-point Protocol
			DLT_FDDI = 10,  // FDDI
		}

		/// <summary>
		/// Packet at the network layer - layer 3 in ISO/OSI, such as IP packet. The identification used
		/// for this type of packets is based on numbers assigned by IANA for the field "protocol" in
		/// Ethernet frame (see http://www.iana.org/assignments/ethernet-numbers).
		/// </summary>
		public enum NetworkLayerPacketType
		{
			Xerox_NS_IDP                = 0x0600,
			IP                          = 0x0800,
			X75                         = 0x0801,
			NBS                         = 0x0802,
			ECMA                        = 0x0803,
			Chaosnet                    = 0x0804,
			X25                         = 0x0805,
			ARP                         = 0x0806,
			XNS_Compatibility           = 0x0807,
			Berkeley_Trailer            = 0x1000,
			BBN_Simnet                  = 0x5208,
			DEC_MOP_DumpLoad            = 0x6001,
			DEC_MOP_RemoteConsole       = 0x6002,
			DECNET_Phase4               = 0x6003,
			DEC_LAT                     = 0x6004,
			DEC_1                       = 0x6005,
			DEC_2                       = 0x6006,
			HP_Probe                    = 0x8005,
			Excelan                     = 0x8010,
			ReverseARP                  = 0x8035,
			DEC_LANBridge               = 0x8038,
			AppleTalk                   = 0x809B,
			AppleTalk_ARP               = 0x80F3,
			IPX_SPX                     = 0x8137
		}

		/// <summary>
		/// <p>Types of protocols at the transport (4) layer. Values assigned by IANA for 
		/// IP "protocol" field are used.</p>
		/// <p>See http://www.iana.org/assignments/protocol-numbers</p>
		/// </summary>
		public enum TransportLayerPacketType
		{
			/// <summary>Internet Control Message Protocol</summary>
			ICMP = 0x01,
			/// <summary>Internet Gateway Management Protocol</summary>
			IGMP = 0x02,
			/// <summary>Transmission Control Protocol</summary>
			TCP  = 0x06,
			/// <summary>External Gateway Protocol</summary>
			EGP  = 0x08,
			/// <summary>Internal Gateway Protocol</summary>
			IGP  = 0x09,
			/// <summary>User Datagram Protocol</summary>
			UDP  = 0x11
		}
		#endregion

		/// <summary>
		/// Gets or sets a packet object that represents a packet of higher layer. (Packet that
		/// is carried as data by current packet.)
		/// </summary>
		public Packet HigherLayerPacket
		{
			get { return higherLayerPacket; }
			set { higherLayerPacket = value; }
		}
		private Packet higherLayerPacket;

		/// <summary>
		/// Computes the space required by this packet + all upper layers.
		/// </summary>
		/// <returns>Number of bytes this packet requires.</returns>
		public int GetBytesCount()
		{
			// number of own header and footer bytes
			int count = this.BytesCount();
			// plus number of bytes of all layer above
			if (HigherLayerPacket != null)
				count += HigherLayerPacket.GetBytesCount();

			return count;
		}

		/// <summary>
		/// Returns the number of bytes this layer requires for header and footer (this layer only!).
		/// </summary>
		/// <returns>Number of bytes for this layer in the packet.</returns>
		protected abstract int BytesCount();

		/// <summary>
		/// Appends bytes of this packet and all upper layer packets to the buffer <paramref name="buf"/>.
		/// </summary>
		/// <param name="buf">The buffer where the bytes will be appended.</param>
		public void AppendBytes(ByteArrayList buf)
		{
			AppendBytesHeader(buf);

			if (higherLayerPacket != null)
				higherLayerPacket.AppendBytes(buf);

			AppendBytesFooter(buf);
		}

		/// <summary>
		/// Appends bytes of a header of this packet.
		/// </summary>
		/// <param name="buf">The buffer where the bytes will be appended.</param>
		protected abstract void AppendBytesHeader(ByteArrayList buf);

		/// <summary>
		/// Appends bytes of a footer of this packet.
		/// </summary>
		/// <param name="buf">The buffer where the bytes will be appended.</param>
		protected abstract void AppendBytesFooter(ByteArrayList buf);

		/// <summary>
		/// Pakcket group to which this packet belongs.
		/// </summary>
		public abstract Packet.PacketGroup Group { get; }

		/// <summary>
		/// A type of this packet.
		/// </summary>
		public abstract int Type { get; }

	}
	#endregion

	#region UnknownPacket
	/// <summary>
	/// A packet that is created when no other packet is suitable (mostly when an unknown packet type
	/// arrives from the real network).
	/// </summary>
	public class UnknownPacket : Packet
	{
		/// <summary>
		/// Gets the data.
		/// </summary>
		public byte[] Data
		{
			get { return data; }
		}
		private byte[] data;

		/// <summary>
		/// Creates new unknown packet. Data are copied from <paramref name="array"/> beginning at index
		/// <paramref name="offset"/> to the end.
		/// </summary>
		/// <param name="array">An array of data to copy.</param>
		/// <param name="offset">Index to start copying at.</param>
		/// <returns>Newly created <see cref="UnknownPacket"/>.</returns>
		public static UnknownPacket CreatePacket(byte[] array, int offset)
		{
			return CreatePacket(array, offset, array.Length - offset);
		}

		/// <summary>
		/// Creates new unknown packet. <paramref name="len"/> bytes are copied from
		/// <paramref name="array"/> beginning at index <paramref name="offset"/>.
		/// </summary>
		/// <param name="array">An array of data to copy.</param>
		/// <param name="offset">Index to start copying at.</param>
		/// <param name="len">The length of data to copy.</param>
		/// <returns>Newly created <see cref="UnknownPacket"/>.</returns>
		public static UnknownPacket CreatePacket(byte[] array, int offset, int len)
		{
			return new UnknownPacket(array, offset, len);
		}

		/// <summary>
		/// Creates new object.
		/// </summary>
		/// <param name="array">An array of data to copy.</param>
		/// <param name="offset">Index to start copying at.</param>
		/// <param name="len">The length of data to copy.</param>
		protected UnknownPacket(byte[] array, int offset, int len)
		{
			this.data = new byte[len];
			Buffer.BlockCopy(array, offset, this.data, 0, len);
		}

		/// <summary>
		/// Appends all data to the buffer.
		/// </summary>
		/// <param name="buf">Buffer where the data are appended.</param>
		protected override void AppendBytesHeader(ByteArrayList buf)
		{
			buf.AddBytes(data, 0, data.Length);
		}

		/// <summary>
		/// Since all data have been appended by <see cref="AppendBytesHeader"/>, this method
		/// does nothing.
		/// </summary>
		/// <param name="buf">Buffer where the data are appended. (Left unmodified.)</param>
		protected override void AppendBytesFooter(ByteArrayList buf)
		{
			// nothing
		}

		/// <summary>
		/// Returns the number of bytes this packet will take in the real network packet.
		/// </summary>
		/// <returns>Number of bytes this packet would occupy.</returns>
		protected override int BytesCount()
		{
			return Data.Length;
		}

		/// <summary>
		/// Returns <see cref="Packet.PacketGroup.Other"/>.
		/// </summary>
		public override Packet.PacketGroup Group
		{
			get { return Packet.PacketGroup.Other; }
		}

		/// <summary>
		/// Returns <see cref="Int32.MaxValue"/>.
		/// </summary>
		public override int Type
		{
			get { return Int32.MaxValue; }
		}

	}
	#endregion

	#region DataPacket
	/// <summary>
	/// Packet containing only arbitrary data, usually some application data.
	/// </summary>
	/// <remarks>
	/// Since the behavior of carrying known data is the same as carrying unknown data,
	/// this class inherits most functionality from <see cref="UnknownPacket"/>.
	/// </remarks>
	public class DataPacket : UnknownPacket
	{
		/// <summary>
		/// Creates new data packet. Data are copied from <paramref name="array"/> beginning at index
		/// <paramref name="offset"/> to the end.
		/// </summary>
		/// <param name="array">An array of data to copy.</param>
		/// <param name="offset">Index to start copying at.</param>
		/// <returns>Newly created <see cref="DataPacket"/>.</returns>
		public static new DataPacket CreatePacket(byte[] array, int offset)
		{
			return CreatePacket(array, offset, array.Length - offset);
		}

		/// <summary>
		/// Creates new data packet. <paramref name="len"/> bytes are copied from <paramref name="array"/> beginning at index
		/// <paramref name="offset"/>.
		/// </summary>
		/// <param name="array">An array of data to copy.</param>
		/// <param name="offset">Index to start copying at.</param>
		/// <param name="len">The length of data to copy.</param>
		/// <returns>Newly created <see cref="DataPacket"/>.</returns>
		public static new DataPacket CreatePacket(byte[] array, int offset, int len)
		{
			return new DataPacket(array, offset, len);
		}

		/// <summary>
		/// Creates new object.
		/// </summary>
		/// <param name="array">An array of data to copy.</param>
		/// <param name="offset">Index to start copying at.</param>
		/// <param name="len">The length of data to copy.</param>
		protected DataPacket(byte[] array, int offset, int len) : base(array, offset, len)
		{
		}

		/// <summary>
		/// Returns always zero.
		/// </summary>
		public override int Type
		{
			get { return 0; }
		}

	}
	#endregion

}
