/*

 Copyright (c) 2006 Pavel Novak.

 The use and distribution terms for this software are contained in the file named License.txt, 
 which can be found in the root of the NetSim distribution. By using this software 
 in any fashion, you are agreeing to be bound by the terms of this license.
 
 You must not remove this notice from this software.

*/

using System;
using System.Net;
using System.Xml;

namespace NetSim.Core
{
	/// <summary>
	/// Base class for classes that hold protocol settings, like IP address/network mask etc.
	/// </summary>
	public abstract class ProtocolSettings
	{
		/// <summary>
		/// Returns <b>true</b> if current protocol settings allows data to be sent to the
		/// <paramref name="address"/>. It is usually if the destination is directly reacheable
		/// (is in the same network).
		/// </summary>
		/// <param name="address">Destination address.</param>
		/// <returns><b>True</b> if the packet can be sent to destination.</returns>
		public abstract bool CanSendPacketTo(NetworkAddress address);

		/// <summary>
		/// Returns <b>true</b> if current protocol settings are set that <paramref name="address"/>
		/// specify the destination equal to this node. In an out parameter returns local address, which
		/// may differ from <paramref name="address"/>; it can be for example broadcast address.
		/// </summary>
		/// <param name="address">Destination address in the packet.</param>
		/// <param name="localAddress">Local address specified in the protocol settings.</param>
		/// <returns><b>True</b> if the destination match local protocol settings.</returns>
		public abstract bool ShouldReceive(NetworkAddress address, out NetworkAddress localAddress);

		/// <summary>
		/// Gets the local address.
		/// </summary>
		public abstract NetworkAddress Address { get; }

		/// <summary>
		/// This method may be overriden in descendant classes to easily read the configuration
		/// from the adapter configuration file.
		/// </summary>
		/// <param name="settingsNode">XML node containing the settings.</param>
		public virtual void Configure(XmlNode settingsNode)
		{
		}
	}

}
