/*

 Copyright (c) 2006 Pavel Novak.

 The use and distribution terms for this software are contained in the file named License.txt, 
 which can be found in the root of the NetSim distribution. By using this software 
 in any fashion, you are agreeing to be bound by the terms of this license.
 
 You must not remove this notice from this software.

*/

using System;
using System.Xml;

namespace NetSim.Core
{
	/// <summary>
	/// Types of objects in the simulator.
	/// </summary>
	[Serializable]
	public enum BasicObjectKind
	{
		Node = 0,
		Link = 1,
		_max = 2,

		None = Int32.MaxValue
	}

	/// <summary>
	/// Types of objects that can be present at the node.
	/// </summary>
	[Serializable]
	public enum NodeItemKind
	{
		Adapter = 0,
		Module = 1,
		Application = 2,
		_max = 3,
		None = Int32.MaxValue
	}

	/// <summary>
	/// Interface that should be implemented by all classes that represents any part of the network structure.
	/// </summary>
	public interface IStructureObject : ISuspendable
	{
		string Name { get; }
	}

	/// <summary>
	/// Common ancestor of all objects used for in the network structure.
	/// </summary>
	/// <remarks>
	/// Descendants include <see cref="Link"/>, <see cref="Node"/>, <see cref="Adapter"/>,
	/// <see cref="Module"/>, <see cref="NodeApplication"/>.
	/// </remarks>
	public abstract class StructureObject : IStructureObject
	{
		#region Members
		/// <summary>
		/// Name of the simulator object.
		/// </summary>
		public string Name
		{
			get { return name; }
			set { name = value; }
		}
		private string name;

		/// <summary>
		/// Configuration file for this object. Used while reinitializing the object. Is <b>null</b>
		/// if no configuration is needed (<see cref="Configure"/> method has not been called).
		/// </summary>
		protected string configurationFile;
		#endregion

		#region Construction
		/// <summary>
		/// Structure objects are created according to the configuration file while the simulator
		/// is starting. The simulator requires a default constructor for each object.
		/// </summary>
		/// <remarks>This does not force the descendant objects to have a default constructor,
		/// however, it a right place where to write such comment.</remarks>
		public StructureObject()
		{
		}
		#endregion

		#region Configuration
		/// <summary>
		/// Configures the object instance. This default implementation expects that configuration file is
		/// an XML file, parses it and calls <see cref="XmlConfigure"/> method. Override this method if
		/// you want to have different behavior. Don't forget to assign <see cref="configurationFile"/>
		/// member.
		/// </summary>
		/// <param name="filePath">Configuration file.</param>
		public virtual void Configure(string filePath)
		{
			// if we have XmlDocument, read it and call XmlConfigure() method
			XmlValidatingReader reader;
			System.IO.Stream stream = null;

			// create validating XML reader
			try
			{
				// path relatively to main configuration file
				string config_file = System.IO.Path.GetDirectoryName(Simulator.GetInstance().MainConfigFile);
				config_file = System.IO.Path.Combine(config_file, filePath);

				stream = new System.IO.FileStream(config_file, System.IO.FileMode.Open);
				XmlTextReader text_reader = new XmlTextReader(stream);
				reader = new XmlValidatingReader(text_reader);

				// parse document
				XmlDocument document = new XmlDocument();
				document.Load(reader);
				
				// save config file name here, when it is clear that it is correct
				configurationFile = filePath;

				// process by configure method for XML documents
				this.XmlConfigure(document);
			}
			catch (System.IO.FileNotFoundException e)
			{
				throw new InitializationException(String.Format("Configuration file {0} was not found", filePath), e);
			}
			catch (System.Xml.Schema.XmlSchemaException e)
			{
				throw new InitializationException(String.Format("Configuration file {0} does not conform to the XML schema", filePath), e);
			}
			catch (System.Xml.XmlException e)
			{
				throw new InitializationException(String.Format("Error reading configuration file {0}. If it is not XML file, you have to override Configure() method", filePath), e);
			}
			catch (Exception e)
			{
				throw new InitializationException(String.Format("Error configuring module according to configuration file {0}", filePath), e);
			}
			finally
			{
				if (stream != null)
					stream.Close();
			}
			
		}

		/// <summary>
		/// Called by the default implementation of <see cref="Configure"/>.
		/// </summary>
		/// <param name="document">XML document read from the configuration file.</param>
		public virtual void XmlConfigure(XmlDocument document)
		{
			// default action - do nothing...
		}
		#endregion

		#region ISuspendable
		/// <summary>
		/// Returns always <b>false</b>. Override to change the behavior.
		/// </summary>
		public virtual bool CanDisable
		{
			get { return false; }
		}

		/// <summary>
		/// Returns always <b>false</b>. Override to change the behavior.
		/// </summary>
		public virtual bool CanReconfigure
		{
			get { return false; }
		}

		/// <summary>
		/// Returns always <b>true</b>, setting throws an exception. Override to change the behavior.
		/// </summary>
		public virtual bool Enabled
		{
			get { return true; }
			set { throw new InvalidOperationException("Enabling/disabling is not supported"); }
		}

		/// <summary>
		/// Always throws <see cref="InvalidOperationException"/>. Override to change the behavior.
		/// </summary>
		public virtual void Reconfigure()
		{
			throw new InvalidOperationException("Reconfiguring is not supported");
		}

		#endregion
	}

}
