/*

 Copyright (c) 2006 Pavel Novak.

 The use and distribution terms for this software are contained in the file named License.txt, 
 which can be found in the root of the NetSim distribution. By using this software 
 in any fashion, you are agreeing to be bound by the terms of this license.
 
 You must not remove this notice from this software.

*/

using System;
using System.Collections;
using System.Xml;
using System.Reflection;
using BasicObjectKind = NetSim.Core.BasicObjectKind;
using NodeItemKind = NetSim.Core.NodeItemKind;

namespace NetSim.Gui
{
	/// <summary>
	/// Data for the GUI application that are not project-specific.
	/// </summary>
	public class GeneralData : XmlData
	{
		/// <summary>
		/// The root directory of GUI application - ie. parent directory of "Templates", "Projects", "Icons" and other
		/// directories.
		/// </summary>
		public string RootDirectory
		{
			get { return rootDirectory; }
		}
		private string rootDirectory;

		/// <summary>
		/// The name (if it is in %PATH% or in current directory) or full path of editor used
		/// for eidtation of configuration files.
		/// </summary>
		public string ExternalEditor
		{
			get { return externalEditor; }
		}
		private string externalEditor;

		/// <summary>
		/// Full path to the simulator console executable file.
		/// </summary>
		public string SimulatorPath
		{
			get { return simulatorPath; }
		}
		private string simulatorPath;

/*
		/// <summary>
		/// List of XML "Library" nodes from the configuration file.
		/// </summary>
		// generics: <XmlNode>
		private ArrayList libraries = new ArrayList();
*/

		public IList PredefinedNodes
		{
			get { return predefinedNodes; }
		}
		// generics: <PredefinedNode>
		private ArrayList predefinedNodes = new ArrayList();

		public IList PredefinedLinks
		{
			get { return predefinedLinks; }
		}
		// generics: <PredefinedLink>
		private ArrayList predefinedLinks = new ArrayList();

		public IList PredefinedAdapters
		{
			get { return predefinedAdapters; }
		}
		// generics: <PredefinedAdapter>
		private ArrayList predefinedAdapters = new ArrayList();

		public IList PredefinedModules
		{
			get { return predefinedModules; }
		}
		// generics: <PredefinedModule>
		private ArrayList predefinedModules = new ArrayList();

		public IList PredefinedApplications
		{
			get { return predefinedApplications; }
		}
		// generics: <PredefinedApplication>
		private ArrayList predefinedApplications = new ArrayList();

		public GeneralData()
		{
			rootDirectory = System.IO.Directory.GetCurrentDirectory();
		}

		public void Initialize(string configFile)
		{
			Load(configFile);
		}

		public override void LoadDocument(System.Xml.XmlDocument document)
		{
			XmlNode node = document.DocumentElement.FirstChild;
			while (node != null)
			{
				switch (node.Name)
				{
					case "PredefinedObjects":
						LoadPredefinedObjects(node);
						break;
					case "RootDirectory":
						try
						{
							string dir = System.IO.Path.GetFullPath(node.InnerText);
							if (!System.IO.Directory.Exists(dir))
								throw new DataProcessingException(String.Format("Directory {0} doesn't exist.", dir));

							rootDirectory = dir;
						}
						catch (Exception e)
						{
							NetSim.Gui.Console.WriteLine("Cannot set root directory: {0}", e.Message);
						}
						break;
					case "ExternalEditor":
						string editor = node.InnerText;
						if (System.IO.File.Exists(System.IO.Path.Combine(rootDirectory, editor)))
						{
							editor = System.IO.Path.Combine(rootDirectory, editor);
						}
						else if (!FileSystem.IsFileInPath(editor))
						{
							NetSim.Gui.Console.WriteLine("Cannot find external editor '{0}'. Defaults to notepad.exe.", editor);
							editor = "notepad.exe";
						}
						this.externalEditor = editor;
						break;
					case "Simulator":
						string simulator = node.InnerText;
						if (System.IO.File.Exists(System.IO.Path.Combine(rootDirectory, simulator)))
						{
							simulator = System.IO.Path.Combine(rootDirectory, simulator);
						}
						else if (!FileSystem.IsFileInPath(simulator))
						{
							throw new DataProcessingException(String.Format("Cannot find the simulator '{0}'.", simulator));
						}
						this.simulatorPath = simulator;
						break;
					case "#comment":
						break;
					default:
						NetSim.Gui.Console.WriteLine("Unknown node '{0}' in main configuration file.", node.Name);
						break;
				}

				node = node.NextSibling;
			}
		}

	/*
		private void LoadLibraries(XmlNode libNode)
		{
			XmlNode node = libNode.FirstChild;
			while (node != null)
			{
				if (node.Name == "Library")
				{
					LoadLibrary(node);
				}
				else
				{
					NetSim.Gui.Console.WriteLine("Unknown node '{0}' in main configuration file.", node.Name);
				}
				node = node.NextSibling;
			}
		}

		private void LoadLibrary(XmlNode node)
		{
			XmlAttribute pathAttribute = node.Attributes["path"];
			if (pathAttribute == null)
			{
				NetSim.Gui.Console.WriteLine("Missing required attribute 'path' in element 'Library' in main configuration file: library ignored.");
				return;
			}

			try
			{
				this.libraries.Add(new Library(pathAttribute.InnerText));
			}
			catch (Exception e)
			{
				NetSim.Gui.Console.WriteLine("Error while loading library '{0}': {1}.", pathAttribute.InnerText, e.Message);
			}
		}
*/
		private void LoadPredefinedObjects(XmlNode objectsNode)
		{
			XmlNode node = objectsNode.FirstChild;
			while (node != null)
			{
				switch (node.Name)
				{
					case "Nodes":
						LoadPredefinedNodes(node);
						break;
					case "Links":
						LoadPredefinedLinks(node);
						break;
					case "Adapters":
						LoadPredefinedAdapters(node);
						break;
					case "Modules":
						LoadPredefinedModules(node);
						break;
					case "Applications":
						LoadPredefinedApplications(node);
						break;

					default:
						NetSim.Gui.Console.WriteLine("Unexpected node '{0}' in '{1}'.", node.Name, objectsNode.Name);
						break;
				}

				node = node.NextSibling;
			}
		}

		private void LoadPredefinedNodes(XmlNode nodesNode)
		{
			XmlNode node = nodesNode.FirstChild;
			while (node != null)
			{
				if (node.Name == "Node")
				{
					try
					{
						this.predefinedNodes.Add(new PredefinedNode(node));
					}
					catch (Exception e)
					{
						XmlAttribute attr = node.Attributes["caption"];
						string name = attr != null ? attr.InnerText : "<missing caption>";
						NetSim.Gui.Console.WriteLine("Error loading predefined node '{0}': {1}.", name, e.Message);
					}
				}
				else
				{
					NetSim.Gui.Console.WriteLine("Unexpected node '{0}' in '{1}'.", node.Name, nodesNode.Name);
				}

				node = node.NextSibling;
			}
		}

		private void LoadPredefinedLinks(XmlNode linksNode)
		{
			XmlNode node = linksNode.FirstChild;
			while (node != null)
			{
				if (node.Name == "Link")
				{
					try
					{
						this.predefinedLinks.Add(new PredefinedLink(node));
					}
					catch (Exception e)
					{
						XmlAttribute attr = node.Attributes["caption"];
						string name = attr != null ? attr.InnerText : "<missing caption>";
						NetSim.Gui.Console.WriteLine("Error loading predefined link '{0}': {1}.", name, e.Message);
					}
				}
				else
				{
					NetSim.Gui.Console.WriteLine("Unexpected node '{0}' in '{1}'.", node.Name, linksNode.Name);
				}

				node = node.NextSibling;
			}
		}

		private void LoadPredefinedAdapters(XmlNode adaptersNode)
		{
			XmlNode node = adaptersNode.FirstChild;
			while (node != null)
			{
				if (node.Name == "Adapter")
				{
					try
					{
						this.predefinedAdapters.Add(new PredefinedAdapter(node));
					}
					catch (Exception e)
					{
						XmlAttribute attr = node.Attributes["caption"];
						string name = attr != null ? attr.InnerText : "<missing caption>";
						NetSim.Gui.Console.WriteLine("Error loading predefined adapter '{0}': {1}.", name, e.Message);
					}
				}
				else
				{
					NetSim.Gui.Console.WriteLine("Unexpected node '{0}' in '{1}'.", node.Name, adaptersNode.Name);
				}

				node = node.NextSibling;
			}
		}

		private void LoadPredefinedModules(XmlNode modulesNode)
		{
			XmlNode node = modulesNode.FirstChild;
			while (node != null)
			{
				if (node.Name == "Module")
				{
					try
					{
						this.predefinedModules.Add(new PredefinedModule(node));
					}
					catch (Exception e)
					{
						XmlAttribute attr = node.Attributes["caption"];
						string name = attr != null ? attr.InnerText : "<missing caption>";
						NetSim.Gui.Console.WriteLine("Error loading predefined module '{0}': {1}.", name, e.Message);
					}
				}
				else
				{
					NetSim.Gui.Console.WriteLine("Unexpected node '{0}' in '{1}'.", node.Name, modulesNode.Name);
				}

				node = node.NextSibling;
			}
		}

		private void LoadPredefinedApplications(XmlNode appNode)
		{
			XmlNode node = appNode.FirstChild;
			while (node != null)
			{
				if (node.Name == "Application")
				{
					try
					{
						this.predefinedApplications.Add(new PredefinedApplication(node));
					}
					catch (Exception e)
					{
						XmlAttribute attr = node.Attributes["caption"];
						string name = attr != null ? attr.InnerText : "<missing caption>";
						NetSim.Gui.Console.WriteLine("Error loading predefined application '{0}': {1}.", name, e.Message);
					}
				}
				else
				{
					NetSim.Gui.Console.WriteLine("Unexpected node '{0}' in '{1}'.", node.Name, appNode.Name);
				}

				node = node.NextSibling;
			}
		}

		public override void SaveDocument(out System.Xml.XmlDocument document)
		{
			throw new NotImplementedException("Saving changes from the Gui is not implemented (in fact it is not needed, because not project-specific values cannot be changed).");
		}

	}

//	/// <summary>
//	/// Represents one library, its nodes, modules and adapters.
//	/// </summary>
//	public class Library
//	{
//		public class Entry
//		{
//			public string Caption
//			{
//				get { return guiVisibleAttr.Caption; }
//			}
//
//			public string Description
//			{
//				get { return guiVisibleAttr.Description; }
//			}
//
//			private NetSim.Core.GuiVisibleAttribute guiVisibleAttr;
//			private Type type;
//
//			public Entry(NetSim.Core.GuiVisibleAttribute guiVisAttr, Type t)
//			{
//				this.guiVisibleAttr = guiVisAttr;
//				this.type = t;
//			}
//		}
//
//		protected string path;
//
//		// generics: <Entry>
//		protected ArrayList nodes = new ArrayList();
//		// generics: <Entry>
//		protected ArrayList modules = new ArrayList();
//		// generics: <Entry>
//		protected ArrayList adapters = new ArrayList();
//		// generics: <Entry>
//		protected ArrayList applications = new ArrayList();
//
//		public Library(string path)
//		{
//			this.path = path;
//			Assembly libraryAssembly = Assembly.LoadFrom(path);
//
//			Type[] types = libraryAssembly.GetTypes();
//			foreach (Type t in types)
//				TestAndAddType(t);
//		}
//
//		/// <summary>
//		/// If the <paramref name="type"/> should be added to the list of nodes, modules or adapters,
//		/// adds it there.
//		/// </summary>
//		/// <param name="type"></param>
//		private void TestAndAddType(Type type)
//		{
//			object[] attributes = type.GetCustomAttributes(typeof(NetSim.Core.GuiVisibleAttribute), false);
//			if (attributes.Length == 0)
//				return;
//
//			NetSim.Core.GuiVisibleAttribute attr = (NetSim.Core.GuiVisibleAttribute)attributes[0];
//
//			if (!type.IsPublic) // public modifier required
//			{
//				NetSim.Gui.Console.WriteLine("'{0}' attribute specified at non-public type '{1}' - type ignored.",
//					typeof(NetSim.Core.GuiVisibleAttribute).Name, type.FullName);
//				return;
//			}
//
//			Type module_interface = type.GetInterface("NetSim.Core.IModule");
//			Type adapter_interface = type.GetInterface("NetSim.Core.IAdapter");
//			bool node_subclass = type.Equals(typeof(NetSim.Core.Node)) || type.IsSubclassOf(typeof(NetSim.Core.Node));
//			Type app_interface = type.GetInterface("NetSim.Core.INodeApplication");
//
//			if (module_interface == null && adapter_interface == null && node_subclass == false && app_interface == null)
//			{
//				NetSim.Gui.Console.WriteLine("'{0}' attribute specified at type '{1}' that doesn't implement IModule, IAdapter, INodeApplication, nor is subclass of Node - type ignored.",
//					typeof(NetSim.Core.GuiVisibleAttribute).Name, type.FullName);
//				return;
//			}
//
//			// one class can implement more things - in that case it is added to more lists
//			if (module_interface != null)
//			{
//				this.modules.Add(new Entry(attr, type));
//				NetSim.Gui.Console.WriteLine("{0}: Module '{1}' loaded.", System.IO.Path.GetFileName(this.path), type.FullName);
//			}
//
//			if (adapter_interface != null)
//			{
//				this.adapters.Add(new Entry(attr, type));
//				NetSim.Gui.Console.WriteLine("{0}: Adapter '{1}' loaded.", System.IO.Path.GetFileName(this.path), type.FullName);
//			}
//
//			if (node_subclass)
//			{
//				this.nodes.Add(new Entry(attr, type));
//				NetSim.Gui.Console.WriteLine("{0}: Node '{1}' loaded.", System.IO.Path.GetFileName(this.path), type.FullName);
//			}
//
//			if (app_interface != null)
//			{
//				this.applications.Add(new Entry(attr, type));
//				NetSim.Gui.Console.WriteLine("{0}: Application '{1}' loaded.", System.IO.Path.GetFileName(this.path), type.FullName);
//			}
//		}
//
//	}

	/// <summary>
	/// Ancestor of all objects that are added to the simulator configuration through GUI.
	/// </summary>
	public abstract class PredefinedObject
	{
		public static ObjInfo GetObjInfo(BasicObjectKind kind)
		{
			switch (kind)
			{
				case BasicObjectKind.Node:
					return PredefinedNode.Info;
				case BasicObjectKind.Link:
					return PredefinedLink.Info;
			}

			System.Diagnostics.Debug.Fail("Invalid kind value");
			return null;
		}

		public static ObjInfo GetObjInfo(NodeItemKind kind)
		{
			switch (kind)
			{
				case NodeItemKind.Adapter:
					return PredefinedAdapter.Info;
				case NodeItemKind.Module:
					return PredefinedModule.Info;
				case NodeItemKind.Application:
					return PredefinedApplication.Info;
			}

			System.Diagnostics.Debug.Fail("Invalid kind value");
			return null;
		}

		public class ObjInfo
		{
			public virtual string Path { get { return "Templates"; } }
			public virtual string ConfigFileName { get { return "Config.xml"; } }
			public virtual string IconsPath { get { return "Icons"; } }
			public virtual System.Drawing.Color DefaultColor { get { return System.Drawing.Color.Transparent; } }

			// these should be accessed only through child objects
			public virtual string TargetProjectDir { get { throw new InvalidOperationException(); } }
			public virtual string[] ParentXmlNodes { get { throw new InvalidOperationException(); } }
			public virtual string XmlNodeName { get { throw new InvalidOperationException(); } }
		}
		public static ObjInfo Info = new ObjInfo();

		/// <summary>
		/// </summary>
		public virtual BasicObjectKind BasicObjKind { get { return BasicObjectKind.None; } }
		public virtual NodeItemKind NodeItemKind { get { return NodeItemKind.None; } }

		/// <summary>
		/// Relative path (according to application root directory) to the directory where the predefined
		/// object's own directory exist.
		/// </summary>
		public virtual string Path { get { return Info.Path; } }

		/// <summary>
		/// The name of file in the object's directory that contains configuration that should be added to
		/// main config file. This file is ignored while copying a template to the project's directory.
		/// </summary>
		public static readonly string ConfigFileName = Info.ConfigFileName;

		/// <summary>
		/// Relative path (according to application root directory) to the icon files.
		/// </summary>
		public static readonly string IconsPath = Info.IconsPath;

		/// <summary>
		/// Relative path (according to project's main configuration file) to the directory where should be
		/// the new object placed.
		/// </summary>
		public abstract string TargetProjectDir { get; }

		/// <summary>
		/// The name of xml nodes in main configuration file under which this object should be placed.
		/// </summary>
		public abstract string[] ParentXmlNodes { get; }

		/// <summary>
		/// The name of XML node this item should have in the main configuration file.
		/// </summary>
		public abstract string XmlNodeName { get; }

		// names of attributes in config files
		protected const string captionAttrName = "caption";
		protected const string descAttrName = "description";
		protected const string iconAttrName = "icon";
		protected const string dirAttrName = "directory";
		protected const string colorAttrName = "color";

		/// <summary>
		/// The user-friendly name that should be displayed in GUI.
		/// </summary>
		public string Caption
		{
			get { return caption; }
		}
		protected string caption;

		/// <summary>
		/// Description of the item usually displayed after the item is selected and ready to be copied to the project.
		/// </summary>
		public string Description
		{
			get { return description; }
		}
		protected string description;

		/// <summary>
		/// File name of the icon in "Icons" directory.
		/// </summary>
		public string Icon
		{
			get { return icon; }
		}
		protected string icon;

		/// <summary>
		/// Index of the icon in <see cref="MainForm.iconsImageList"/>. If no icon is specified, index 0 is returned, which
		/// results in displaying of a default icon.
		/// </summary>
		public int IconIndex
		{
			get { return iconIndex; }
		}
		private int iconIndex = 0;
		
		/// <summary>
		/// The name of directory in templates and equals to the directory that will be created for this object
		/// in <see cref="TargetProjectDir"/>. If one exists, a number will be added to the name.
		/// </summary>
		public string Directory
		{
			get { return directory; }
		}
		protected string directory;

		/// <summary>
		/// The color the new item should have.
		/// </summary>
		public System.Drawing.Color Color
		{
			get { return color; }
		}
		protected System.Drawing.Color color;

		/// <summary>
		/// Relative path (according to application root directory) to the configuration file in the template with
		/// content that should be added to main config file.
		/// </summary>
		protected string configFile;

		/// <summary>
		/// Creates new predefined object, called by the constructors of descendants. Throws
		/// <see cref="DataProcessingException"/> if something is wrong with the template.
		/// </summary>
		/// <param name="node"><see cref="XmlNode"/> of the predefined object in configuration file
		/// that describes all predefined objects.</param>
		protected PredefinedObject(XmlNode node)
		{
			if (node.Attributes[captionAttrName] == null)
				throw new DataProcessingException(String.Format("Missing required XML attribute '{0}'.", captionAttrName));
			if (node.Attributes[descAttrName] == null)
				throw new DataProcessingException(String.Format("Missing required XML attribute '{0}'.", descAttrName));
			if (node.Attributes[iconAttrName] == null)
				throw new DataProcessingException(String.Format("Missing required XML attribute '{0}'.", iconAttrName));
			if (node.Attributes[dirAttrName] == null)
				throw new DataProcessingException(String.Format("Missing required XML attribute '{0}'.", dirAttrName));

			this.caption = node.Attributes[captionAttrName].InnerText;
			this.description = node.Attributes[descAttrName].InnerText;
			this.icon = node.Attributes[iconAttrName].InnerText;
			this.directory = node.Attributes[dirAttrName].InnerText;
			
			// color
			this.color = System.Drawing.Color.Empty;
			System.Drawing.Color c;
			if (node.Attributes[colorAttrName] != null && DisplayElement.ParseColor(node.Attributes[colorAttrName].InnerText, out c))
				this.color = c;

			//
			// icon
			//
			if (Icon != String.Empty)
			{
				string icon_path = System.IO.Path.Combine(IconsPath, Icon);
				if (!System.IO.File.Exists(icon_path))
					throw new DataProcessingException(String.Format("Icon file '{0}' not found.", Icon));

				MainForm.Instance.GetImage(icon_path, out iconIndex);
			}

			//
			// config file
			//
			string dir_path = System.IO.Path.Combine(this.Path, this.Directory);
			if (!System.IO.Directory.Exists(dir_path))
				throw new DataProcessingException(String.Format("Directory '{0}' doesn't exist.", dir_path));

			string config_path = System.IO.Path.Combine(dir_path, PredefinedObject.ConfigFileName);
			if (!System.IO.File.Exists(config_path))
				throw new DataProcessingException(String.Format("File '{0}' is missing.", config_path));

			configFile = config_path;
		}

		public XmlNode GetConfigNode()
		{
			XmlTextReader reader = new XmlTextReader(configFile);
			XmlDocument doc = new XmlDocument();
			doc.Load(reader);
			reader.Close();

			return doc.DocumentElement;
		}
	}

	public class PredefinedNode : PredefinedObject
	{
		public new class ObjInfo : PredefinedObject.ObjInfo
		{
			public override string Path { get { return System.IO.Path.Combine(base.Path, "Nodes"); } }
			public override string TargetProjectDir { get { return "Nodes"; } }
			public override string[] ParentXmlNodes { get { return new string[] {"Network", "Nodes"}; } }
			public override string XmlNodeName { get { return "Node"; } }
			public override System.Drawing.Color DefaultColor { get { return System.Drawing.Color.Transparent; } }
		}
		public static new ObjInfo Info = new ObjInfo();

		public override string Path { get { return Info.Path; } }
		public override string TargetProjectDir { get { return Info.TargetProjectDir; } }
		public override string[] ParentXmlNodes { get { return Info.ParentXmlNodes; } }
		public override string XmlNodeName { get { return Info.XmlNodeName; } }
		public override BasicObjectKind BasicObjKind { get { return BasicObjectKind.Node; } }

		public PredefinedNode(XmlNode xmlNode) : base(xmlNode)
		{
		}
	}

	public class PredefinedLink : PredefinedObject
	{
		public new class ObjInfo : PredefinedObject.ObjInfo
		{
			public override string Path { get { return System.IO.Path.Combine(base.Path, "Links"); } }
			public override string TargetProjectDir { get { return "Links"; } }
			public override string[] ParentXmlNodes { get { return new string[] {"Network", "Links"}; } }
			public override string XmlNodeName { get { return "Link"; } }
			public override System.Drawing.Color DefaultColor { get { return System.Drawing.Color.DarkBlue; } }
		}
		public static new ObjInfo Info = new ObjInfo();

		public override string Path { get { return Info.Path; } }
		public override string TargetProjectDir { get { return Info.TargetProjectDir; } }
		public override string[] ParentXmlNodes { get { return Info.ParentXmlNodes; } }
		public override string XmlNodeName { get { return Info.XmlNodeName; } }
		public override BasicObjectKind BasicObjKind { get { return BasicObjectKind.Link; } }

		public PredefinedLink(XmlNode xmlNode) : base(xmlNode)
		{
		}
	}

	public class PredefinedAdapter : PredefinedObject
	{
		public new class ObjInfo : PredefinedObject.ObjInfo
		{
			public override string Path { get { return System.IO.Path.Combine(base.Path, "Adapters"); } }
			public override string TargetProjectDir { get { return "Nodes/{0}/Adapters"; } }
			public override string[] ParentXmlNodes { get { return new string[] {"Network", "Nodes", "Node"}; } }
			public override string XmlNodeName { get { return "Adapter"; } }
			public override System.Drawing.Color DefaultColor { get { return System.Drawing.Color.Transparent; } }
		}
		public static new ObjInfo Info = new ObjInfo();

		public override NodeItemKind NodeItemKind { get { return NodeItemKind.Adapter; } }
		
		public override string Path { get { return Info.Path; } }
		public override string TargetProjectDir { get { return Info.TargetProjectDir; } }
		public override string[] ParentXmlNodes { get { return Info.ParentXmlNodes; } }
		public override string XmlNodeName { get { return Info.XmlNodeName; } }

		public PredefinedAdapter(XmlNode xmlNode) : base(xmlNode)
		{
		}
	}

	public class PredefinedModule : PredefinedObject
	{
		public new class ObjInfo : PredefinedObject.ObjInfo
		{
			public override string Path { get { return System.IO.Path.Combine(base.Path, "Modules"); } }
			public override string TargetProjectDir { get { return "Nodes/{0}/Modules"; } }
			public override string[] ParentXmlNodes { get { return new string[] {"Network", "Nodes", "Node"}; } }
			public override string XmlNodeName { get { return "Module"; } }
			public override System.Drawing.Color DefaultColor { get { return System.Drawing.Color.Transparent; } }
		}
		public static new ObjInfo Info = new ObjInfo();

		public override NodeItemKind NodeItemKind { get { return NodeItemKind.Module; } }

		public override string Path { get { return Info.Path; } }
		public override string TargetProjectDir { get { return Info.TargetProjectDir; } }
		public override string[] ParentXmlNodes { get { return Info.ParentXmlNodes; } }
		public override string XmlNodeName { get { return Info.XmlNodeName; } }

		public PredefinedModule(XmlNode xmlNode) : base(xmlNode)
		{
		}

	}

	public class PredefinedApplication : PredefinedObject
	{
		public new class ObjInfo : PredefinedObject.ObjInfo
		{
			public override string Path { get { return System.IO.Path.Combine(base.Path, "Applications"); } }
			public override string TargetProjectDir { get { return "Nodes/{0}/Applications"; } }
			public override string[] ParentXmlNodes { get { return new string[] {"Network", "Nodes", "Node"}; } }
			public override string XmlNodeName { get { return "Application"; } }
			public override System.Drawing.Color DefaultColor { get { return System.Drawing.Color.Transparent; } }
		}
		public static new ObjInfo Info = new ObjInfo();

		public override NodeItemKind NodeItemKind { get { return NodeItemKind.Application; } }

		public override string Path { get { return Info.Path; } }
		public override string TargetProjectDir { get { return Info.TargetProjectDir; } }
		public override string[] ParentXmlNodes { get { return Info.ParentXmlNodes; } }
		public override string XmlNodeName { get { return Info.XmlNodeName; } }

		public PredefinedApplication(XmlNode xmlNode) : base(xmlNode)
		{
		}

	}

}
