/*

 Copyright (c) 2006 Pavel Novak.

 The use and distribution terms for this software are contained in the file named License.txt, 
 which can be found in the root of the NetSim distribution. By using this software 
 in any fashion, you are agreeing to be bound by the terms of this license.
 
 You must not remove this notice from this software.

*/

using System;
using System.Xml;
using System.IO;
using System.Windows.Forms;
using System.Collections;
using BasicObjectKind = NetSim.Core.BasicObjectKind;
using NodeItemKind = NetSim.Core.NodeItemKind;

namespace NetSim.Gui
{
	/// <summary>
	/// Interface used for items that have name and can be renamed.
	/// </summary>
	public interface INamedItem
	{
		/// <summary>
		/// Gets the item name.
		/// </summary>
		string Name { get; }

		/// <summary>
		/// Changes the name. Throws <see cref="DataProcessingException"/> if the name cannot be changed.
		/// </summary>
		/// <param name="name">New name.</param>
		void Rename(string name);
	}

	public class DisplayElement : Control, INamedItem
	{
		public const int DefaultIconSize = 40;
		public const float DefaultFontSize = 10;

		public DisplayElementConfig Config
		{
			get { return xmlConfig; }
		}

		private DisplayElementConfig xmlConfig;
		private XmlNode XmlGuiNode;

		public new string Name
		{
			get { return xmlConfig.Name; }
			set
			{
				xmlConfig.Name = value;
				this.Label.Text = value;
			}
		}

		internal PictureBox Picture;
		private Label Label;

		public System.Drawing.Color Color
		{
			get { return color; }
			set
			{
				this.color = value;
				this.Picture.BackColor = value;
			}
		}
		private System.Drawing.Color color;

		// moving elements support
		public bool Dragging;
		public int LastX;
		public int LastY;

		/// <summary>
		/// Whether the element in the network is enabled or disabled.
		/// </summary>
		public new bool Enabled
		{
			get { return enabled; }
			set
			{
				if (enabled == value)
					return;

				enabled = value;

        this.Label.BackColor = enabled ? System.Drawing.Color.FromKnownColor(System.Drawing.KnownColor.Transparent)
					: System.Drawing.Color.FromKnownColor(System.Drawing.KnownColor.Red);
				this.Label.Invalidate();
			}
		}
		private bool enabled = true;

		private ProjectData owner;

		/// <summary>
		/// Last clicked element. Used for recogntion which element should be modified by the context menu.
		/// </summary>
		public static DisplayElement LastSelected;

		public System.Drawing.Point Center
		{
			get
			{
				int c_x = this.Left + Picture.Left + Picture.Width/2;
				int c_y = this.Top + Picture.Top + Picture.Height/2;
				return new System.Drawing.Point(c_x, c_y);
			}
		}

		protected DisplayElement(BasicObjectKind deKind, XmlNode configNode, XmlNode guiNode, PictureBox picture, Label label, ProjectData owner)
		{
			this.xmlConfig = DisplayElementConfig.Create(deKind, configNode);

			this.XmlGuiNode = guiNode;
			this.Picture = picture;
			this.Label = label;
			this.owner = owner;

			// color
			System.Drawing.Color c;
			XmlAttribute color_attr = guiNode.Attributes["color"];
			if (color_attr != null && ParseColor(color_attr.InnerText, out c))
			{
				Color = c;
				// for links draw color to the background
				if (deKind == BasicObjectKind.Link) this.BackColor = c;
			}
			else
			{
				Color = PredefinedObject.GetObjInfo(deKind).DefaultColor;
			}

			this.Name = xmlConfig.Name;
		}

		/// <summary>
		/// Writes position of elements back to the config file
		/// </summary>
		public void Save()
		{
			XmlData.GetXmlAttribute(XmlGuiNode, "x", true).InnerText = (Left * 100 / owner.Zoom).ToString();
			XmlData.GetXmlAttribute(XmlGuiNode, "y", true).InnerText = (Top * 100 / owner.Zoom).ToString();
			XmlData.GetXmlAttribute(XmlGuiNode, "color", true).InnerText = Color.ToArgb().ToString("x");
		}

		public static DisplayElement Create(BasicObjectKind deKind, XmlNode configNode, XmlNode guiNode, ProjectData owner)
		{
			//
			// create picture
			//
			PictureBox pb = new PictureBox();
			pb.Left = 0;
			pb.Top = 0;
			int image_index;
			pb.Image = MainForm.Instance.GetImage(Path.Combine(PredefinedObject.IconsPath, guiNode.Attributes["icon"].InnerText), out image_index);
			pb.Width = (int)(DefaultIconSize * owner.Zoom / 100f);
			pb.Height = (int)(DefaultIconSize * owner.Zoom / 100f);
			pb.SizeMode = PictureBoxSizeMode.StretchImage;
			pb.BorderStyle = BorderStyle.None;
			pb.ContextMenu = MainForm.Instance.nodeContextMenu;
			pb.MouseDown += new MouseEventHandler(MainForm.pictureBox_MouseDown);
			pb.MouseMove += new MouseEventHandler(MainForm.pictureBox_MouseMove);
			pb.MouseUp += new MouseEventHandler(MainForm.pictureBox_MouseUp);
			pb.Resize += new EventHandler(pictureBox_Resize);

			//
			// Create label
			//
			Label l = new Label();
			l.Font = new System.Drawing.Font(l.Font.FontFamily, DefaultFontSize * owner.Zoom / 100);
			l.AutoSize = true;
			l.TextAlign = System.Drawing.ContentAlignment.TopCenter;
			l.Text = configNode.Attributes["name"].InnerText;

			//
			// Create display element
			//
			DisplayElement result = new DisplayElement(deKind, configNode, guiNode, pb, l, owner);
			pb.Tag = result;
			l.Tag = result;
			result.Parent = MainForm.Instance.mainPanel;
			result.Controls.Add(pb);
			result.Width = pb.Width;
			result.Height = pb.Height;
			pb.Parent = result;
			l.Parent = result.Parent;
			result.Left = Int32.Parse(guiNode.Attributes["x"].InnerText);
			result.Top = Int32.Parse(guiNode.Attributes["y"].InnerText);
			l.Left = result.Left + pb.Width/2 - l.Width/2;
			l.Top = result.Top + pb.Height + owner.Zoom / 10;

			return result;
		}

		public void Delete()
		{
			// delete from config xml
			this.xmlConfig.Delete();
			// delete from gui xml
			this.XmlGuiNode.ParentNode.RemoveChild(this.XmlGuiNode);

			// delete from screen
			MainForm.Instance.mainPanel.Controls.Remove(this.Label);
			MainForm.Instance.mainPanel.Controls.Remove(this);

			// if this is a link, delete references from adapters
			if (this.Config.Kind == BasicObjectKind.Link)
				MainForm.Instance.projectData.ReconnectNodes(this.Name, String.Empty);
		}

		public static bool ParseColor(string rgb, out System.Drawing.Color color)
		{
			try
			{
				uint color_int = uint.Parse(rgb, System.Globalization.NumberStyles.HexNumber);
				color = System.Drawing.Color.FromArgb((int)(color_int | 0xff000000)); // always add full alpha
				return true;
			}
			catch
			{
				color = System.Drawing.Color.Empty;
				return false;
			}
		}

		/// <summary>
		/// Moves the element <paramref name="moveX"/> pixels to the right and <paramref name="moveY"/> down.
		/// Values can be negative.
		/// </summary>
		/// <param name="moveX"></param>
		/// <param name="moveY"></param>
		public void MoveXY(int moveX, int moveY)
		{
			this.Left += moveX;
			this.Top += moveY;
			Label.Left += moveX;
			Label.Top += moveY;
		}

		private static void pictureBox_Resize(object sender, EventArgs e)
		{
			// with resizing we also want to change font size, which is not adjusting automaticaly
			DisplayElement de = (DisplayElement)(((PictureBox)sender).Tag);

			float em_size = DefaultFontSize * de.Picture.Width / DefaultIconSize;
			if (em_size < 1) em_size = 1;
			de.Label.Font = new System.Drawing.Font(de.Label.Font.FontFamily, em_size);
		}

		/// <summary>
		/// Renames the node or link. Throws <see cref="DataProcessingException"/> if the name is not valid.
		/// </summary>
		/// <param name="newName"></param>
		public void Rename(string newName)
		{
			string old_name = this.Name;

			if (newName == String.Empty)
				throw new DataProcessingException("The name cannot be empty.");
			
			int invalid_index = newName.IndexOfAny(System.IO.Path.InvalidPathChars);
			if (invalid_index >= 0)
				throw new DataProcessingException(String.Format("Name cannot contain '{0}' character.", newName[invalid_index]));

			int sep_index = newName.IndexOf(System.IO.Path.DirectorySeparatorChar);
			if (sep_index < 0)
				sep_index = newName.IndexOf(System.IO.Path.AltDirectorySeparatorChar);
			if (sep_index >= 0)
				throw new DataProcessingException(String.Format("Name cannot contain directory separator: '{0}'.", newName[sep_index]));

			// basic checks made, however, the name might still be invalid, or match some other item.
			
			// check duplicates
			foreach (DisplayElement de in MainForm.Instance.projectData.GetObjectList(this.Config.Kind))
			{
				if (!de.Equals(this) && de.Name.Equals(newName))
					throw new DataProcessingException(String.Format("The item with name '{0}' already exist.", newName));
			}

			// try to rename directory if one exists
			string base_dir = System.IO.Path.GetDirectoryName(MainForm.Instance.projectData.FileName);
			string relative_dir = Config.ParentDirectory;
			string old_dir_rel = System.IO.Path.Combine(relative_dir, this.Name);
			if (System.IO.Directory.Exists(Path.Combine(base_dir, old_dir_rel)))
			{
				string new_dir_rel = System.IO.Path.Combine(relative_dir, newName);
				try
				{
					System.IO.Directory.Move(System.IO.Path.Combine(base_dir, old_dir_rel),
						System.IO.Path.Combine(base_dir, new_dir_rel)); 
				}
				catch (Exception e)
				{
					throw new DataProcessingException(String.Format("Cannot rename directory {0} to {1}: {2}",
						old_dir_rel, new_dir_rel, e.Message), e);
				}

				// renamed, change the path in all config files...
				this.xmlConfig.ChangePathInConfig(old_dir_rel, new_dir_rel);
				this.XmlGuiNode.Attributes["name"].InnerText = newName;

				// if this is the link, we have to change all adapters connected to that link to the new one
				if (Config.Kind == BasicObjectKind.Link)
					MainForm.Instance.projectData.ReconnectNodes(old_name, newName);
			}

			// finally change the name in config file
			this.Name = newName;
		}

		/// <summary>
		/// Changes path only in "config" attributes in the configuration file. <seealso cref="Rename"/>.
		/// </summary>
		/// <param name="oldName"></param>
		/// <param name="newName"></param>
		public void RenameConfig(string oldName, string newName)
		{
			this.xmlConfig.ChangePathInConfig(System.IO.Path.Combine(Config.ParentDirectory, oldName),
				System.IO.Path.Combine(Config.ParentDirectory, newName));
		}

		public void Reconnect(string linkFrom, string linkTo)
		{
			// should be called only for nodes
			System.Diagnostics.Debug.Assert(Config.Kind == BasicObjectKind.Node);
			
			foreach (AdapterConfig ac in ((NodeConfig)this.xmlConfig).GetItems(NodeItemKind.Adapter))
			{
				if (ac.Link == linkFrom)
					ac.Link = linkTo;
			}
		}
	}

	public abstract class DisplayElementConfig
	{
		protected XmlNode xmlConfig;

		public string Name
		{
			get { return xmlConfig.Attributes["name"].InnerText; }
			set { xmlConfig.Attributes["name"].InnerText = value; }
		}

		/// <summary>
		/// The directory relative to project main configuration file where the directory for this element resides.
		/// </summary>
		public string ParentDirectory { get { return this.parentDirectory; } }
		private string parentDirectory;

		public BasicObjectKind Kind
		{
			get { return kind; }
		}
		private BasicObjectKind kind;

		public static DisplayElementConfig Create(BasicObjectKind kind, XmlNode config)
		{
			DisplayElementConfig result = null;
			switch (kind)
			{
				case BasicObjectKind.Node:
					result = new NodeConfig(config);
					result.parentDirectory = "Nodes";
					break;
				case BasicObjectKind.Link:
					result = new LinkConfig(config);
					result.parentDirectory = "Links";
					break;
				default:
					System.Diagnostics.Debug.Fail("Bad BasicObjectKind while creating DisplayElementConfig");
					break;
			}

			result.kind = kind;
			return result;
		}

		protected DisplayElementConfig(XmlNode config)
		{
			this.xmlConfig = config;

			// test mandatory fields
			XmlAttribute name_attr = xmlConfig.Attributes["name"];
			if (name_attr == null)
			{
				XmlDocumentFragment frag = config.OwnerDocument.CreateDocumentFragment();
				frag.InnerXml = "<X name=\"unnamed\"/>";
				xmlConfig.Attributes.Prepend(frag.FirstChild.Attributes["name"]);
			}
		}

		public void Delete()
		{
			this.xmlConfig.ParentNode.RemoveChild(this.xmlConfig);
		}

		public override string ToString()
		{
			return Name;
		}

		public bool HasConfigFile()
		{
			XmlAttribute config_attr = xmlConfig.Attributes["config"];
			return (config_attr != null);
		}

		public void EditConfigFile()
		{
			XmlAttribute config_attr = xmlConfig.Attributes["config"];
			if (config_attr == null)
				throw new DataProcessingException("No configuration file is associated with this item.");

			string config_path = Path.Combine(Path.GetDirectoryName(MainForm.Instance.projectData.FileName), config_attr.InnerText);
			if (!File.Exists(config_path))
				throw new DataProcessingException(String.Format("File '{0}' doesn't exist.", config_path));

			System.Diagnostics.Process.Start(MainForm.Instance.generalData.ExternalEditor, config_path);
		}

		/// <summary>
		/// Replaces all occurences of <paramref name="oldDir"/> in "config" attributes at the beginning with
		/// <paramref name="newDir"/> path fragment.
		/// </summary>
		/// <param name="oldDir"></param>
		/// <param name="newDir"></param>
		public void ChangePathInConfig(string oldDir, string newDir)
		{
			// make all directory separators forward slashes for simpler comparison
			oldDir = oldDir.Replace('\\', '/');
			newDir = newDir.Replace('\\', '/');
			ChangePathInConfigInternal(oldDir, newDir, this.xmlConfig);
		}

		private void ChangePathInConfigInternal(string oldDir, string newDir, XmlNode xmlNode)
		{
			XmlAttribute config_attr = xmlNode.Attributes["config"];
			if (config_attr != null)
			{
				// test if change is needed
				string c_str = config_attr.InnerText.Replace('\\', '/');;
				if (c_str.ToLower().StartsWith(oldDir.ToLower() + "/"))
				{
					c_str = c_str.Remove(0, oldDir.Length + 1);
					config_attr.InnerText = System.IO.Path.Combine(newDir, c_str).Replace('\\', '/');
				}
			}

			// change in all children
			XmlNode child = xmlNode.FirstChild;
			while (child != null)
			{
				ChangePathInConfigInternal(oldDir, newDir, child);
				child = child.NextSibling;
			}
		}

	}

	public class NodeConfig : DisplayElementConfig
	{
		private ArrayList[] nodeItems = new ArrayList[(int)NodeItemKind._max];

		public NodeConfig(XmlNode config) : base(config)
		{
			for (int i = 0; i < (int)NodeItemKind._max; i++)
				this.nodeItems[i] = new ArrayList();

			XmlNode child = config.FirstChild;
			while (child != null)
			{
				if (child.Name == PredefinedModule.Info.XmlNodeName)
					CreateNodeItem(NodeItemKind.Module, child);
				if (child.Name == PredefinedAdapter.Info.XmlNodeName)
					CreateNodeItem(NodeItemKind.Adapter, child);
				if (child.Name == PredefinedApplication.Info.XmlNodeName)
					CreateNodeItem(NodeItemKind.Application, child);

				child = child.NextSibling;
			}
		}

		private NodeItemConfig CreateNodeItem(NodeItemKind kind, XmlNode xmlNode)
		{
			NodeItemConfig result = null;
			switch (kind)
			{
				case NodeItemKind.Module:
					result = new ModuleConfig(xmlNode, this);
					break;
				case NodeItemKind.Adapter:
					result = new AdapterConfig(xmlNode, this);
					break;
				case NodeItemKind.Application:
					result = new ApplicationConfig(xmlNode, this);
					break;
			}

			System.Diagnostics.Debug.Assert(result != null);
			nodeItems[(int)kind].Add(result);
			return result;
		}

		public void SwapItems(NodeItemKind kind, int index1, int index2)
		{
			System.Diagnostics.Debug.Assert(index1 >= 0 && index2 >= 0 && index1 != index2);

			// make first index smaller
			if (index1 > index2)
			{
				int tmp = index1;
				index1 = index2;
				index2 = tmp;
			}

			//
			// swap in our array list
			//
			ModuleConfig m1 = (ModuleConfig)nodeItems[(int)kind][index1];
			ModuleConfig m2 = (ModuleConfig)nodeItems[(int)kind][index2];
			nodeItems[(int)kind].RemoveAt(index2); // first remove second one - index of the first one doesn't change
			nodeItems[(int)kind].RemoveAt(index1);

			// first insert to the smaller index
			nodeItems[(int)kind].Insert(index1, m2);
			nodeItems[(int)kind].Insert(index2, m1);

			//
			// swap in XML file
			//
			int i = 0;
			XmlNode n = xmlConfig.FirstChild;
			XmlNode prev_first = null;
			XmlNode first = null;
			XmlNode second = null;
			XmlNode after_second = null;
			while (n != null)
			{
				bool is_module = n.Name == PredefinedModule.Info.XmlNodeName;

				if (is_module)
				{
					if (i == index1)
					{
						first = n;
						prev_first = n.PreviousSibling;
					}
					else if (i == index2)
					{
						second = n;
						after_second = n.NextSibling;
						break;
					}
					i++;
				}

				n = n.NextSibling;
			}

			xmlConfig.RemoveChild(first);
			xmlConfig.RemoveChild(second);
			xmlConfig.InsertAfter(second, prev_first);
			xmlConfig.InsertBefore(first, after_second);
		}

		private void RemoveFromXmlAt(int index, string nodeName)
		{
			int i = 0;
			XmlNode n = xmlConfig.FirstChild;
			while (n != null)
			{
				if (n.Name == nodeName)
				{
					if (i == index)
					{
						xmlConfig.RemoveChild(n);
						break;
					}
					i++;
				}
				n = n.NextSibling;
			}
		}

		public void RemoveNodeItem(NodeItemConfig item)
		{
			int index = nodeItems[(int)item.Kind].IndexOf(item);
			if (index < 0)
				return;
			
			// remove from our list
			nodeItems[(int)item.Kind].RemoveAt(index);

			// remove from XML file
			RemoveFromXmlAt(index, item.Info.XmlNodeName);
		}

		public void AddNodeItem(PredefinedObject obj, string name)
		{
			System.Diagnostics.Debug.Assert(obj.NodeItemKind != NodeItemKind.None);

			// append to the last item of the same type
			XmlNode child = this.xmlConfig.LastChild;
			while (child != null)
			{
				for (int i = (int)obj.NodeItemKind; i >= 0; i--)
				{
					// items are odreded in the configuration file according to the value of their kind.
					// if we have traversed to the XML node that is for the kind being inserted or for
					// any previous kind, the new item should be inserted behind it
					if (PredefinedObject.GetObjInfo((NodeItemKind)i).XmlNodeName == child.Name)
						goto found_label;
				}
				child = child.PreviousSibling;
			}
			found_label:

			XmlDocumentFragment frag = this.xmlConfig.OwnerDocument.CreateDocumentFragment();
			frag.InnerXml = obj.GetConfigNode().OuterXml;
			
			// child can be null here, in that case new item should be the first one
			XmlNode result;
			if (child != null)
				result = this.xmlConfig.InsertAfter(frag.FirstChild, child);
			else
				result = this.xmlConfig.PrependChild(frag.FirstChild);

			XmlAttribute name_attr = XmlData.GetXmlAttribute(result, "name", true);
			name_attr.InnerText = name;
			
			// add directories to config attribute
			XmlAttribute config_attr = result.Attributes["config"];
			if (config_attr != null)
			{
				string dir = String.Format(PredefinedObject.GetObjInfo(obj.NodeItemKind).TargetProjectDir, this.Name);
				dir = System.IO.Path.Combine(dir, name);
				config_attr.InnerText = System.IO.Path.Combine(dir, config_attr.InnerText).Replace('\\', '/');
			}

			NodeItemConfig config = CreateNodeItem(obj.NodeItemKind, result);
		}

		/// <summary>
		/// Returns the list of objects of the specified kind.
		/// </summary>
		/// <param name="kind"></param>
		/// <returns>List of <see cref="NodeItemConfig"/> objects.</returns>
		public IList GetItems(NodeItemKind kind)
		{
			return ArrayList.ReadOnly(nodeItems[(int)kind]);
		}

		/// <summary>
		/// Returns the names of objects of the specified kind.
		/// </summary>
		/// <param name="kind"></param>
		/// <returns></returns>
		public IList GetItemNames(NodeItemKind kind)
		{
			string[] result = new string[nodeItems[(int)kind].Count];
			for (int i = 0; i < nodeItems[(int)kind].Count; i++)
			{
				result[i] = ((NodeItemConfig)nodeItems[(int)kind][i]).Name;
			}
			return result;
		}
	}

	public class LinkConfig : DisplayElementConfig
	{
		public LinkConfig(XmlNode config) : base(config)
		{
		}
	}

	public abstract class NodeItemConfig : INamedItem
	{
		protected XmlNode xmlConfig;
		protected NodeConfig ownerNode;

		public abstract NodeItemKind Kind { get; }
		public abstract PredefinedObject.ObjInfo Info { get; }

		public string Name
		{
			get { return xmlConfig.Attributes["name"].InnerText; }
			set { xmlConfig.Attributes["name"].InnerText = value; }
		}

		public void Rename(string newName)
		{
			// do not rename directly here, we should check if the directory can be copied..
			MainForm.Instance.projectData.RenameNodeItem(ownerNode, this, newName);			
		}

		public string Class
		{
			get { return xmlConfig.Attributes["class"].InnerText; }
		}

		public override string ToString()
		{
			return String.Format("{0} ({1})", Name, Class);
		}

		protected NodeItemConfig(XmlNode config, NodeConfig owner)
		{
			this.xmlConfig = config;
			this.ownerNode = owner;

			// test mandatory fields
			XmlAttribute name_attr = xmlConfig.Attributes["name"];
			if (name_attr == null)
			{
				XmlDocumentFragment frag = config.OwnerDocument.CreateDocumentFragment();
				frag.InnerXml = "<X name=\"unnamed\"/>";
				name_attr = xmlConfig.Attributes.Prepend(frag.FirstChild.Attributes["name"]);
			}
			// check name uniquness
			IList forbidden_names = owner.GetItemNames(this.Kind);
			int i = 1;
			while (forbidden_names.Contains(name_attr.InnerText))
			{
				name_attr.InnerText = name_attr.InnerText + i.ToString();
				i++;
			}

			XmlAttribute class_attr = xmlConfig.Attributes["class"];
			if (class_attr == null)
			{
				XmlDocumentFragment frag = config.OwnerDocument.CreateDocumentFragment();
				frag.InnerXml = "<X name=\"error: no class\"/>";
				xmlConfig.Attributes.Prepend(frag.FirstChild.Attributes["class"]);
			}
		}

		public bool HasConfigFile()
		{
			XmlAttribute config_attr = xmlConfig.Attributes["config"];
			return (config_attr != null);
		}

		public void EditConfigFile()
		{
			XmlAttribute config_attr = xmlConfig.Attributes["config"];
			if (config_attr == null)
				throw new DataProcessingException("No configuration file is associated with this item.");

			string config_path = Path.Combine(Path.GetDirectoryName(MainForm.Instance.projectData.FileName), config_attr.InnerText);
			if (!File.Exists(config_path))
				throw new DataProcessingException(String.Format("File '{0}' doesn't exist.", config_path));

			System.Diagnostics.Process.Start(MainForm.Instance.generalData.ExternalEditor, config_path);
		}
	}

	public class ModuleConfig : NodeItemConfig
	{
		public override NodeItemKind Kind { get { return NodeItemKind.Module; } }
		public override PredefinedObject.ObjInfo Info { get { return PredefinedModule.Info; } }

		public ModuleConfig(XmlNode config, NodeConfig owner) : base(config, owner)
		{
		}

	}

	public class AdapterConfig : NodeItemConfig
	{
		public override NodeItemKind Kind { get { return NodeItemKind.Adapter; } }
		public override PredefinedObject.ObjInfo Info { get { return PredefinedAdapter.Info; } }

		public AdapterConfig(XmlNode config, NodeConfig owner) : base(config, owner)
		{
		}

		public string Link
		{
			get
			{
				XmlAttribute link_attr = xmlConfig.Attributes["link"];
				return link_attr != null ? link_attr.InnerText : String.Empty;
			}

			set
			{
				XmlAttribute link_attr = xmlConfig.Attributes["link"];
				if (link_attr == null)
				{
					XmlDocumentFragment frag = xmlConfig.OwnerDocument.CreateDocumentFragment();
					frag.InnerXml = String.Format("<X link=\"{0}\"/>", value);
					xmlConfig.Attributes.Append(frag.FirstChild.Attributes["link"]);
				}
				else
				{
					link_attr.InnerText = value;
				}
			}
		}

		public void SetLink(LinkConfig link)
		{
			this.Link = link.Name;
		}

		public DisplayElement GetLinkElement()
		{
			string link_name = Link;

			foreach (DisplayElement de in MainForm.Instance.projectData.GetObjectList(BasicObjectKind.Link))
			{
				LinkConfig lc = de.Config as LinkConfig;
				if (lc.Name == link_name)
					return de;
			}

			return null;
		}

	}

	public class ApplicationConfig : NodeItemConfig
	{
		public override NodeItemKind Kind { get { return NodeItemKind.Application; } }
		public override PredefinedObject.ObjInfo Info { get { return PredefinedApplication.Info; } }

		public ApplicationConfig(XmlNode config, NodeConfig owner) : base(config, owner)
		{
		}
	}


	/// <summary>
	/// Project-specific data.
	/// </summary>
	public class ProjectData : XmlData
	{
		protected const string GuiConfigFilename = "GuiConfig.xml";

		// generics: <ArrayList<DisplayElement> >
		private ArrayList[] basicObjects = new ArrayList[(int)BasicObjectKind._max];

		/// <summary>
		/// Returns list of <see cref="DisplayElement"/> objects of specified kind.
		/// </summary>
		/// <param name="kind">Kind of objects to return.</param>
		/// <returns>List of <see cref="DisplayElement"/> objects.</returns>
		public IList GetObjectList(BasicObjectKind kind)
		{
			System.Diagnostics.Debug.Assert((int)kind < (int)BasicObjectKind._max);

			return ArrayList.ReadOnly(this.basicObjects[(int)kind]);
		}

		internal delegate void RepaintDelegate(DisplayElement de);
		internal event RepaintDelegate RepaintEvent;

		/// <summary>
		/// Zoom in percents.
		/// </summary>
		public int Zoom
		{
			get { return zoom; }
			set
			{
				//				// we need to resize all items
				//				foreach (DisplayElement de in nodes)
				//					de.Resize(zoom, value);
				//				foreach (DisplayElement de in links)
				//					de.Resize(zoom, value);

				zoom = value;
			}
		}
		private int zoom = 100;

		/// <summary>
		/// 
		/// </summary>
		/// <param name="moveLeftTop">If <b>false</b>, elements will be moved only to the right and bottom.</param>
		internal void CheckPositionAllElements(bool moveLeftTop)
		{
			int x_margin = MainForm.Instance.mainPanel.AutoScrollMargin.Width;
			int y_margin = MainForm.Instance.mainPanel.AutoScrollMargin.Height;
			int min_x = Int32.MaxValue;
			int min_y = Int32.MaxValue;
			int max_x = Int32.MinValue;
			int max_y = Int32.MinValue;

			for (int i = 0; i < (int)BasicObjectKind._max; i++)
			{
				foreach (DisplayElement de in basicObjects[i])
				{
					min_x = Math.Min(min_x, de.Left);
					min_y = Math.Min(min_y, de.Top);
					max_x = Math.Max(max_x, de.Left + de.Width);
					max_y = Math.Max(max_y, de.Top + de.Height);
				}
			}

			if (min_x == Int32.MaxValue) // no objects to move
				return;

			if (!moveLeftTop && x_margin <= min_x && y_margin <= min_y) // moving right bottom not needed
				return;

			// if the diagram is small according to the screen size, it is not necessary to move it
			System.Drawing.Size panel_size = MainForm.Instance.mainPanel.Size;
			int correction_x = Math.Max(0, (panel_size.Width - 2*x_margin - (max_x - min_x))/2);
			int correction_y = Math.Max(0, (panel_size.Height - 2*y_margin - (max_y - min_y))/2);

			int move_x = x_margin - min_x + MainForm.Instance.mainPanel.AutoScrollPosition.X + correction_x;
			int move_y = y_margin - min_y + MainForm.Instance.mainPanel.AutoScrollPosition.Y + correction_y;
			if (move_x != 0 || move_y != 0) // movement needed
			{
				for (int i = 0; i < (int)BasicObjectKind._max; i++)
				{
					foreach (DisplayElement de in basicObjects[i])
						de.MoveXY(move_x, move_y);
				}
			}
		}

		public bool Modified
		{
			get { return modified; }
			set
			{
				modified = value;
				string caption = MainForm.Instance.Text.TrimEnd('*');
				if (modified)
					caption = caption + "*";
				if (caption != MainForm.Instance.Text)
					MainForm.Instance.Text = caption;
			}
		}
		private bool modified;

		private XmlDocument document;
		private GuiProjectData guiData;

		public ProjectData()
		{
			for (int i = 0; i < (int)BasicObjectKind._max; i++)
				basicObjects[i] = new ArrayList();
		}

		public override void LoadDocument(XmlDocument document)
		{
			this.document = document;
			
			this.guiData = new GuiProjectData();
			string gui_path = Path.GetDirectoryName(this.fileName);
			gui_path = Path.Combine(gui_path, GuiConfigFilename);

			// if the gui config file doesn't exist, create one..
			if (!File.Exists(gui_path))
			{
				string src = System.IO.Path.Combine(MainForm.Instance.generalData.RootDirectory, MainForm.NewProjectFile);
				src = System.IO.Path.GetDirectoryName(src);
				src = System.IO.Path.Combine(src, GuiConfigFilename);
				File.Copy(src, gui_path);
			}

			this.guiData.Load(gui_path);

			// init our lists of elements
			for (int i = 0; i < (int)BasicObjectKind._max; i++)
				InitBasicObjects((BasicObjectKind)i);
		}

		private void InitBasicObjects(BasicObjectKind kind)
		{
			PredefinedObject.ObjInfo o_info = PredefinedObject.GetObjInfo(kind);

			XmlNode parent = document.DocumentElement;
			foreach (string s in o_info.ParentXmlNodes)
			{
				parent = GetXmlNode(parent, s, false);
				if (parent == null)
					return;
			}

			foreach (XmlNode node in parent.ChildNodes)
			{
				if (node.Name != o_info.XmlNodeName) continue;

				XmlNode gui_node = guiData.GetBasicItem(kind, node.Attributes["name"].InnerText);
				if (gui_node == null)
					gui_node = this.guiData.AddBasicItem(kind, node.Attributes["name"].InnerText, null,
						System.Drawing.Point.Empty, System.Drawing.Color.Empty);

				DisplayElement de = DisplayElement.Create(kind, node, gui_node, this);
				this.basicObjects[(int)kind].Add(de);
				RepaintEvent(de);
			}
		}

		public override void SaveDocument(out XmlDocument document)
		{
			// save positions
			for (int i = 0; i < (int)BasicObjectKind._max; i++)
			{
				foreach (DisplayElement de in basicObjects[i])
					de.Save();
			}

			this.guiData.Save();
			document = this.document;
			Modified = false;
		}

		/// <summary>
		/// Copies subdirectory <paramref name="name"/> from directory <paramref name="srcDir"/> to
		/// directory <paramref name="dstDir"/>. If the directory in <paramref name="dstDir"/> exists,
		/// a number is added to the end. Changed name is returned in an out parameter.
		/// </summary>
		/// <param name="srcDir">Source directory.</param>
		/// <param name="dstDir">Destination directory.</param>
		/// <param name="name">Name of the subdirectory to copy, can be modified in the destination directory,
		/// modified name is returned.</param>
		/// <param name="forbiddenNames">List of strings containing the names that cannot be assigned
		/// to the destination directory.</param>
		private void CopyPredefinedObject(string srcDir, string dstDir, ref string name, IList forbiddenNames)
		{
			int num = 0;
			while (
				forbiddenNames.Contains(name + (num == 0 ? "" : num.ToString()))
				|| Directory.Exists(Path.Combine(dstDir, name + (num == 0 ? "" : num.ToString())))
				|| File.Exists(Path.Combine(dstDir, name + (num == 0 ? "" : num.ToString())))
				)
				num++;

			// copy directory
			string src_dir = Path.Combine(srcDir, name);
			name = name + (num == 0 ? "" : num.ToString());
			string dst_dir = Path.Combine(dstDir, name);
			FileSystem.CopyDirectory(src_dir, dst_dir, true);

			// and delete the config file of the predefined object that was copied too
			File.Delete(Path.Combine(dst_dir, PredefinedObject.ConfigFileName));
		}

		public void AddBasicItem(PredefinedObject obj, System.Drawing.Point point)
		{
			// copy template
			string proj_dir = Path.GetDirectoryName(this.fileName);
			string items_dir = Path.Combine(proj_dir, obj.TargetProjectDir);
			if (!Directory.Exists(items_dir))
				Directory.CreateDirectory(items_dir);

			string old_name = obj.Directory;
			string result_dir = obj.Directory;
			CopyPredefinedObject(obj.Path, items_dir, ref result_dir, new string[] {}); // there might be a number added to the result_dir

			// add to XML document
			XmlNode parent = document.DocumentElement;
			foreach (string p_node_name in obj.ParentXmlNodes)
				parent = GetXmlNode(parent, p_node_name, true);
			XmlDocumentFragment frag = parent.OwnerDocument.CreateDocumentFragment();
			frag.InnerXml = obj.GetConfigNode().OuterXml;
			XmlNode xml_node = parent.AppendChild(frag.FirstChild);
			XmlAttribute name_attr = xml_node.Attributes["name"];
			name_attr.InnerText = result_dir;

			// add to Gui
			XmlNode gui_node = this.guiData.AddBasicItem(obj.BasicObjKind, result_dir, obj.Icon, point, obj.Color);
			DisplayElement de = DisplayElement.Create(obj.BasicObjKind, xml_node, gui_node, this);
			RepaintEvent(de);

			de.RenameConfig(old_name, result_dir);

			System.Diagnostics.Debug.Assert(obj.BasicObjKind != BasicObjectKind.None);
			this.basicObjects[(int)obj.BasicObjKind].Add(de);

			modified = true;
		}

		/// <summary>
		/// Adds a new node item (adapter, module, application) to specified node.
		/// </summary>
		/// <param name="nodeConf">The target node configuration.</param>
		/// <param name="obj">New item to add.</param>
		public void AddNodeItem(NodeConfig nodeConf, PredefinedObject obj)
		{
			// copy template
			string proj_dir = Path.GetDirectoryName(this.fileName);
			string items_dir = Path.Combine(proj_dir, String.Format(obj.TargetProjectDir, nodeConf.Name));
			if (!Directory.Exists(items_dir))
				Directory.CreateDirectory(items_dir);

			string old_name = obj.Directory;
			string result_dir = obj.Directory;
			CopyPredefinedObject(obj.Path, items_dir, ref result_dir, nodeConf.GetItemNames(obj.NodeItemKind)); // there might be a number added to the result_dir

			// add to XML document
			nodeConf.AddNodeItem(obj, result_dir);

			modified = true;
		}

		public void RenameNodeItem(NodeConfig nodeConf, NodeItemConfig itemConf, string newName)
		{
			if (newName == itemConf.Name) // same name, nothing to do
				return;

			if (newName == String.Empty) // name cannot be empty string
				throw new DataProcessingException("The name cannot be empty.");

			int invalid_index = newName.IndexOfAny(System.IO.Path.InvalidPathChars);
			if (invalid_index >= 0)
				throw new DataProcessingException(String.Format("Name cannot contain '{0}' character.", newName[invalid_index]));

			int sep_index = newName.IndexOf(System.IO.Path.DirectorySeparatorChar);
			if (sep_index < 0)
				sep_index = newName.IndexOf(System.IO.Path.AltDirectorySeparatorChar);
			if (sep_index >= 0)
				throw new DataProcessingException(String.Format("Name cannot contain directory separator: '{0}'.", newName[sep_index]));

			if (nodeConf.GetItemNames(itemConf.Kind).Contains(newName)) // duplicate name
				throw new DataProcessingException("Item with that name already exists");

			// basic checks made, however, the name might still be invalid, or match some other item.

			// copy the directory
			string proj_dir = Path.GetDirectoryName(this.fileName);
			string items_rel_dir = String.Format(PredefinedObject.GetObjInfo(itemConf.Kind).TargetProjectDir, nodeConf.Name);
			string old_dir_rel = Path.Combine(items_rel_dir, itemConf.Name);
			string new_dir_rel = Path.Combine(items_rel_dir, newName);
			if (System.IO.Directory.Exists(Path.Combine(proj_dir, old_dir_rel)))
			{
				try
				{
					System.IO.Directory.Move(Path.Combine(proj_dir, old_dir_rel), Path.Combine(proj_dir, new_dir_rel));
				}
				catch (Exception e)
				{
					throw new DataProcessingException(String.Format("Cannot rename directory {0} to {1}: {2}",
						old_dir_rel, new_dir_rel, e.Message), e);
				}

				// if the directory has been moved, change the references to the files in that
				// directory in main configuration file
				nodeConf.ChangePathInConfig(old_dir_rel, new_dir_rel);
			}

			// rename in XML
			itemConf.Name = newName;
		}

		public void RemoveNodeItem(NodeConfig nodeConf, NodeItemConfig itemConf)
		{
			// remove from XML
			nodeConf.RemoveNodeItem(itemConf);

			// delete item directory if exist
			string proj_dir = Path.GetDirectoryName(this.fileName);
			string items_dir = Path.Combine(proj_dir, String.Format(itemConf.Info.TargetProjectDir, nodeConf.Name));
			string item_dir = Path.Combine(items_dir, itemConf.Name);
			if (System.IO.Directory.Exists(item_dir))
			{
				try { System.IO.Directory.Delete(item_dir, true); }
				catch (Exception e)
				{
					Gui.Console.WriteLine("Cannot delete directory {0}: {1}", item_dir, e.Message);
				}
			}
		}

		public void DeleteBasicItem(DisplayElement de)
		{
			// delete directory
			string proj_dir = Path.GetDirectoryName(this.fileName);
			string parent_dir = Path.Combine(proj_dir, de.Config.ParentDirectory);
			string dir = Path.Combine(parent_dir, de.Name);
			if (Directory.Exists(dir))
				Directory.Delete(dir, true);

			// delete others
			de.Delete();
			this.basicObjects[(int)de.Config.Kind].Remove(de);
		}

		public void RepaintAll()
		{
			if (RepaintEvent != null)
			{
				// save positions
				for (int i = 0; i < (int)BasicObjectKind._max; i++)
				{
					foreach (DisplayElement de in this.basicObjects[i])
						RepaintEvent(de);
				}
			}
		}

		public void ReconnectNodes(string linkFrom, string linkTo)
		{
			foreach (DisplayElement de in this.basicObjects[(int)BasicObjectKind.Node])
			{
				de.Reconnect(linkFrom, linkTo);
			}
		}

		/// <summary>
		/// Called when the simulator is stopped.
		/// </summary>
		public void SimulatorStopped()
		{
			// enable all elements
			for (int i = 0; i < (int)BasicObjectKind._max; i++)
			{
				foreach (DisplayElement de in this.basicObjects[i])
					de.Enabled = true;
			}
		}
	}
}
