/*

 Copyright (c) 2006 Pavel Novak.

 The use and distribution terms for this software are contained in the file named License.txt, 
 which can be found in the root of the NetSim distribution. By using this software 
 in any fashion, you are agreeing to be bound by the terms of this license.
 
 You must not remove this notice from this software.

*/

using System;
using System.Xml;
using System.Text;

namespace NetSim.Gui
{
	/// <summary>
	/// Base class for data files written in XML.
	/// </summary>
	public abstract class XmlData
	{
		public const string NetSimNamespace = "http://pavelnovak.eu/NetSimConfig";

		public string FileName
		{
			get { return fileName; }
		}
		protected string fileName;

		public XmlData()
		{
		}

		public virtual void Load(string filePath)
		{
			this.fileName = System.IO.Path.GetFullPath(filePath);

			// if we have XmlDocument, read it and call XmlConfigure() method
			XmlValidatingReader reader;
			System.IO.Stream stream = null;

			try
			{
				// create validating XML reader
				stream = new System.IO.FileStream(fileName, System.IO.FileMode.Open);
				XmlTextReader text_reader = new XmlTextReader(stream);
				reader = new XmlValidatingReader(text_reader);

				// parse document
				XmlDocument document = new XmlDocument();
				document.Load(reader);

				// process by configure method for XML documents
				this.LoadDocument(document);
			}
			catch (System.IO.FileNotFoundException e)
			{
				throw new DataProcessingException(String.Format("Configuration file {0} was not found", filePath), e);
			}
			catch (System.Xml.Schema.XmlSchemaException e)
			{
				throw new DataProcessingException(String.Format("Configuration file {0} does not conform to the XML schema", filePath), e);
			}
			catch (System.Xml.XmlException e)
			{
				throw new DataProcessingException(String.Format("Error reading configuration file {0}", filePath), e);
			}
			finally
			{
				if (stream != null)
					stream.Close();
			}
		}

		public void Save()
		{
			Save(this.fileName);
		}

		public virtual void Save(string file)
		{
			XmlDocument document;
			SaveDocument(out document);

			XmlTextWriter writer = new XmlTextWriter(this.fileName, Encoding.UTF8);
			writer.Formatting = Formatting.Indented;

			document.Save(writer);
			
			writer.Close();
		}

		public abstract void LoadDocument(XmlDocument document);
		public abstract void SaveDocument(out XmlDocument document);

		internal static XmlNode GetXmlNode(XmlNode parent, string nodeName, bool create)
		{
			XmlNodeList list = parent.ChildNodes;
			foreach (XmlNode n in list)
				if (n.Name == nodeName)
					return n;

			if (!create)
				return null;

			XmlDocumentFragment frag = parent.OwnerDocument.CreateDocumentFragment();
			frag.InnerXml = String.Format("<{0} xmlns=\"{1}\" />", nodeName, NetSimNamespace);
			return parent.AppendChild(frag.FirstChild);
		}

		internal static XmlAttribute GetXmlAttribute(XmlNode node, string attributeName, bool create)
		{
			XmlAttribute result = node.Attributes[attributeName];
			if (result != null)
				return result;

			if (!create)
				return null;

			XmlDocumentFragment frag = node.OwnerDocument.CreateDocumentFragment();
			frag.InnerXml = String.Format("<X {0}=\"\" />", attributeName);
			return node.Attributes.Append(frag.FirstChild.Attributes[0]);
		}
	}
}
