/*

 Copyright (c) 2006 Pavel Novak.

 The use and distribution terms for this software are contained in the file named License.txt, 
 which can be found in the root of the NetSim distribution. By using this software 
 in any fashion, you are agreeing to be bound by the terms of this license.
 
 You must not remove this notice from this software.

*/

using System;
using System.IO;

namespace NetSim.Gui
{
	/// <summary>
	/// Helper methods related to filesystem.
	/// </summary>
	public sealed class FileSystem
	{
		private FileSystem()
		{
		}

		/// <summary>
		/// Copies a directory <paramref name="src"/> to directory <paramref name="dst"/>, the target directory
		/// should not exist. If <paramref name="recursive"/> is set to <b>true</b>, all subdirectories are also
		/// copied. If it is <b>false</b>, only files within <paramref name="src"/> are copied, directories are
		/// ignored.
		/// </summary>
		/// <param name="src">Source directory to copy.</param>
		/// <param name="dst">Destination directory. Shouldn't exist.</param>
		/// <param name="recursive"><b>True</b> if copy also subdirectories.</param>
		public static void CopyDirectory(string src, string dst, bool recursive)
		{
			System.IO.Directory.CreateDirectory(dst);

			string[] files = Directory.GetFiles(src);

			foreach (string f in files)
				File.Copy(f, Path.Combine(dst, Path.GetFileName(f)));

			if (recursive) // copy subdirectories too
			{
				string[] sub_dirs = Directory.GetDirectories(src);
				foreach (string sub_dir in sub_dirs)
					CopyDirectory(sub_dir, Path.Combine(dst, Path.GetFileName(sub_dir)), recursive);
			}
		}

		/// <summary>
		/// Returns <b>true</b> if the file exists in the current path.
		/// </summary>
		/// <param name="file"></param>
		/// <returns><b>True</b> if the file is somwhere in the %PATH%.</returns>
		public static bool IsFileInPath(string file)
		{
			string path_env = Environment.GetEnvironmentVariable("PATH");
			string[] paths = path_env.Split(new char[] {';', ','});
			
			foreach (string p in paths)
			{
				if (File.Exists(Path.Combine(p, file)))
					return true;
			}

			return false;
		}
	}
}
