/*

 Copyright (c) 2006 Pavel Novak.

 The use and distribution terms for this software are contained in the file named License.txt, 
 which can be found in the root of the NetSim distribution. By using this software 
 in any fashion, you are agreeing to be bound by the terms of this license.
 
 You must not remove this notice from this software.

*/

using System;
using System.Diagnostics;
using System.Threading;
using System.Runtime.Remoting.Channels;

namespace NetSim.Gui
{
	/// <summary>
	/// Communicates with the simulator and controls the simulation.
	/// </summary>
	public class SimulatorControl
	{
		public class Callback : MarshalByRefObject, NetSim.Core.RemoteControl.ICallback
		{
			[System.Runtime.Remoting.Messaging.OneWay]
			public void NewEvent(NetSim.Core.RemoteControl.SimulatorEvent ev)
			{
				switch (ev)
				{
					case NetSim.Core.RemoteControl.SimulatorEvent.NewLogEntry:
						SimulatorControl.Instance.WaitForNewLogEntry.Set();
						break;
				}
			}

			[System.Runtime.Remoting.Messaging.OneWay]
			public void EnabledDisabled(NetSim.Core.BasicObjectKind objKind, string name, bool enabled)
			{
				foreach (DisplayElement de in MainForm.Instance.projectData.GetObjectList(objKind))
				{
					if (de.Name == name && de.Enabled != enabled)
						de.Enabled = enabled;
				}
			}

			public override object InitializeLifetimeService()
			{
				return null; // never expires
			}

		}

		private Callback callbackObject = new Callback();

		/// <summary>
		/// The reference to the simulator. Is <b>null</b> if the simulator is not running.
		/// </summary>
		private Process simulatorProcess;

		/// <summary>
		/// Refernece to remote object in the simulator process used for controlling the simulation.
		/// </summary>
		private NetSim.Core.RemoteControl remoteControl;

		/// <summary>
		/// Channel for communication with the simulator.
		/// </summary>
		private IChannel remoteControlChannel;

		/// <summary>
		/// We don't want the method here called remotely by the simulator to call the simulator remotely
		/// again to get log entry. Therefore we have employed the thread waiting for the event to arrive
		/// that is just woken up and it asks for the new entry and pastes it to the list in GUI.
		/// </summary>
		internal AutoResetEvent WaitForNewLogEntry = new AutoResetEvent(false);

		/// <summary><b>True</b> if the simulator is shutting down.</summary>
		private bool shuttingDown;

		/// <summary>How many times we will try to obtain remote reference to control the simulator.</summary>
		private const int tryToConnectCount = 5;
		/// <summary>How many miliseconds to wait between tries to obtain remote reference.</summary>
		private const int tryToConnectDelay = 5000;

		public static SimulatorControl Instance
		{
			get
			{
				if (instance == null)
					instance = new SimulatorControl();
				return instance;
			}
		}
		private static SimulatorControl instance;

		private SimulatorControl()
		{
		}

		public bool Running
		{
			get { return simulatorProcess != null; }
		}

		/// <summary>
		/// Sends a string to the simulator.
		/// </summary>
		/// <param name="str"></param>
		public void Write(string str)
		{
			if (simulatorProcess != null)
				simulatorProcess.StandardInput.Write(str);
		}

		/// <summary>
		/// Starts simulation.
		/// </summary>
		/// <param name="simulatorPath">Path to simulator console executable file.</param>
		/// <param name="configFile">Virtual network configuration file.</param>
		public void Start(string simulatorPath, string configFile)
		{
			// prepare for start
			ProcessStartInfo start_info = new ProcessStartInfo(simulatorPath,
				String.Format("-pt:adv -rs:shm -rc:shm \"{0}\"", configFile));
			start_info.WorkingDirectory = System.IO.Path.GetDirectoryName(configFile);
			start_info.UseShellExecute = false;
			start_info.RedirectStandardOutput = true;
			start_info.RedirectStandardInput = true;
			start_info.RedirectStandardError = true;
			start_info.ErrorDialog = false;
			start_info.CreateNoWindow = true;
			simulatorProcess = new Process();
			simulatorProcess.StartInfo = start_info;
			
			shuttingDown = false;

			// start the simulator
			simulatorProcess.Start();
			MainForm.Instance.SetFocusSimulatorControls(true, true);

			// and create the thread that will monitor when the simulation exits. This is necessary
			// because the user can for example kill the thread and we should be informed about it
			Thread wait_t = new Thread(new ThreadStart(WaitForShutdown));
			wait_t.Name = "WaitingForSimShutdown";
			wait_t.Start();

			// thread reading standard output
			Thread read_std = new Thread(new ThreadStart(ReadSimulatorStdOutput));
			read_std.Name = "ReadingSimStd";
			read_std.Start();
			// thread reading error output
			Thread read_err = new Thread(new ThreadStart(ReadSimulatorErrOutput));
			read_err.Name = "ReadingSimErr";
			read_err.Start();

			// start thread that will wait for the simulator to start up and get it's remote control
			// object reference
			ThreadPool.QueueUserWorkItem(new WaitCallback(StartRemoteControl));
		}
		

		private void StartRemoteControl(object state)
		{
			//
			// Try to register the channel and obtain remote reference
			//
			for (int i = 0; i < tryToConnectCount; i++)
			{
				Thread.Sleep(tryToConnectDelay / (i==0?2:1)); // shorter delay for the first try

				Exception exc = null;
				try
				{
					if (remoteControlChannel == null) // create and register the channel
					{
						System.Collections.Specialized.ListDictionary ch_properties = new System.Collections.Specialized.ListDictionary();
						ch_properties.Add("name", "RemoteControlInNetSimGui");
						BinaryServerFormatterSinkProvider server_prov = new BinaryServerFormatterSinkProvider();
						server_prov.TypeFilterLevel = System.Runtime.Serialization.Formatters.TypeFilterLevel.Full; // because of security restrictions
						remoteControlChannel = new PHP.Core.ShmChannel(ch_properties, new BinaryClientFormatterSinkProvider(), server_prov);
						ChannelServices.RegisterChannel(remoteControlChannel);
					}

					if (remoteControl == null)
					{
						remoteControl = (NetSim.Core.RemoteControl)Activator.GetObject(typeof(NetSim.Core.RemoteControl),
							"shm://NetSimServer/" + NetSim.Core.Simulator.RemoteControlObjectName);

						// force remoting to establish the connection + register our event handler
						remoteControl.RegisterCallback(callbackObject);

						// start thread getting info from the simulator
						Thread read_err = new Thread(new ThreadStart(CheckSimulatorLog));
						read_err.Name = "CheckingSimLog";
						read_err.Start();

					}
				}
				catch (Exception e)
				{
					Gui.Console.WriteLine("Try no. {0} to connect to the simulator failed: {1} ({2} tries remain)",
						i+1, e.Message, tryToConnectCount - i - 1);
					exc = e;
					remoteControl = null;
				}

				if (shuttingDown)
					return;

				if (remoteControlChannel != null && remoteControl != null)
					break;
			}

			//
			// Were we succesful?
			//
			if (remoteControlChannel != null && remoteControl != null)
				Gui.Console.WriteLine("Connected to the simulator.");
			else
				Gui.Console.WriteLine("Cannot connect to the simulator. However, the simulator might still be running and be controlled via commands in the 'Command Window'.");
		}

		const int readBufferLen = 4069; // 4096 == size of underlying buffer
		public void ReadSimulatorStdOutput()
		{
			char[] buffer = new char[readBufferLen]; 
			int len = 1;
			while (simulatorProcess != null && len > 0)
			{
				try
				{
					len = simulatorProcess.StandardOutput.Read(buffer, 0, readBufferLen);
					MainForm.Instance.WriteToCommandTextBox(new String(buffer, 0, len));
				}
				catch (Exception e)
				{
					Gui.Console.WriteLine("Reading std simulator output error: " + e.Message);
				}
			}
		}

		public void ReadSimulatorErrOutput()
		{
			char[] buffer = new char[readBufferLen]; 
			int len = 1;
			while (simulatorProcess != null && len > 0)
			{
				try
				{
					len = simulatorProcess.StandardError.Read(buffer, 0, readBufferLen);
					MainForm.Instance.WriteToCommandTextBox(new String(buffer, 0, len));
				}
				catch (Exception e)
				{
					Gui.Console.WriteLine("Reading error simulator output error: " + e.Message);
				}
			}
		}

		public void Stop()
		{
			shuttingDown = true;
			
			if (simulatorProcess == null)
				return;

			SendToSimulator("exit");
			if (simulatorProcess != null && !simulatorProcess.WaitForExit(10000))
			{
				Gui.Console.WriteLine();
				Gui.Console.Write("Unable to shut down, sending kill signal... ");
				simulatorProcess.Kill();
			}

			//
			// Stop remote control if it has been started
			//
			if (remoteControl != null)
			{
				remoteControl = null;
			}

			if (remoteControlChannel != null)
			{
				ChannelServices.UnregisterChannel(remoteControlChannel);
				remoteControlChannel = null;
			}
		}

		/// <summary>
		/// The entry point for the thread that monitors when the simulator ends its execution.
		/// </summary>
		private void WaitForShutdown()
		{
			// wait for simulator exit
			simulatorProcess.WaitForExit();

			// set controls status appropriately
			MainForm.Instance.SetFocusSimulatorControls(false, true);

			// make all elements enabled
			if (MainForm.Instance.projectData != null) // project can be unloaded before simulator is stopped
				MainForm.Instance.projectData.SimulatorStopped();

			Gui.Console.WriteLine("Simulator stopped.");

			simulatorProcess = null;
		}

		private void Pause(bool pause)
		{
			if (pause)
			{
				SendToSimulator("pause");
			}
			else
			{
				SendToSimulator("resume");
			}
		}

		public void EnableDisable(NetSim.Core.BasicObjectKind objKind, string name, bool enable)
		{
			switch (objKind)
			{
				case NetSim.Core.BasicObjectKind.Node:
					SendToSimulator(String.Format("{0} node \"{1}\"", enable ? "enable" : "disable", name));
					break;

				case NetSim.Core.BasicObjectKind.Link:
					SendToSimulator(String.Format("{0} link \"{1}\"", enable ? "enable" : "disable", name));
					break;
			}
		}

		/// <summary>
		/// Sends the command <paramref name="str"/> to the simulator. Newline is implicitly added to the string.
		/// </summary>
		/// <param name="str">Command to the simulator.</param>
		private void SendToSimulator(string str)
		{
			string command = str + Environment.NewLine;
			try
			{
				simulatorProcess.StandardInput.Write(command);
				MainForm.Instance.WriteToCommandTextBox(command);
			}
			catch (ArgumentOutOfRangeException)
			{
				// occurs while writing to the command text box if the gui is shutting down, that's OK..
			}
			catch (NullReferenceException)
			{
				// occurs when simulatorProcess is not running, e.g. the user clicked the stop button
				// twice.. that's OK.
			}
		}

		private void CheckSimulatorLog()
		{
			while (remoteControl != null && !shuttingDown)
			{
				// wait for new entry to be written to the log
				WaitForNewLogEntry.WaitOne();

				try
				{
					NetSim.Core.Event ev;
					while (remoteControl != null && !shuttingDown && remoteControl.GetLogEntry(out ev, true))
					{
						MainForm.Instance.NewLogEntry(ev);
					}
				}
				catch (Exception e)
				{
					if (!this.shuttingDown)
						Gui.Console.WriteLine("Error reading simulator log file: " + e.Message);
				}
			}
		}
	}
}
