/*

 Copyright (c) 2006 Pavel Novak.

 The use and distribution terms for this software are contained in the file named License.txt, 
 which can be found in the root of the NetSim distribution. By using this software 
 in any fashion, you are agreeing to be bound by the terms of this license.
 
 You must not remove this notice from this software.

*/

// This constant is needed by SDDL:
// See http://msdn.microsoft.com/library/default.asp?url=/library/en-us/secbp/security/creating_a_dacl.asp
#define _WIN32_WINNT 0x0500

#include "netsimspecific.h"
#include "netsimspi.h"

#include <process.h>
#include <sddl.h>

//
// Globals for NetSim support
//

SECURITY_ATTRIBUTES gFileMappingSecurityAttributes;

// NetSim provider structures. These providers are not actually registered with WinSock,
// but are passed as a parameter to functions here if they should communicate with
// virtual network instead of real base providers installed on the computer.
// We create such dummy PROVIDER structure for each provider we are actually layered on.
LPPROVIDER gNetSimBaseInfo = NULL;

// The information whether the LSP Wrapper DLL has been succesfuly loaded and therefore
// we have correct function addresses to redirect etc..
static BOOL gIsWrapperLoaded = FALSE;

#define MAX_REDIRECTIONS 32
#define SHARED_NAME TEXT("NetSimSharedRedirections")
#define SHARED_MUTEX_NAME TEXT("NetSimSharedMutexRedirections")

// variables initialized while creating mapped file
static HANDLE gMappedFile = NULL;
static HANDLE gSharedMutex = NULL; // Critical section for accessing processes shared data
static REDIRECTION * gRedirections = NULL; // Max 32 redirections allowed from one computer

static HMODULE gWrapperDll = NULL;

pfnNestingCount gNestingCount = NULL;
pfnGetThreadStartAddress gGetThreadStartAddress = NULL;


// Returns true if the current call is from the application that should communicate through
// the virtual network. This function checkes current PID and returns true if it is in the
// list of applications maintained by the NetSim.
BOOL NetSimRedirectCurrentCall(LPREDIRECTION lpRedirection)
{
	// Do not redirect if we were unable to load the wrapper DLL...
	if (!gIsWrapperLoaded)
		return FALSE;
	
	if (!gSharedMutex || !gRedirections)
	{
		// try to init again
		InitSharedData();

		// if not loaded now, no redirection is possible
		if (!gSharedMutex || !gRedirections)
			return FALSE;
	}

	// shared memory used, no testing needed
	//if (gNestingCounter > 0) // redirect only if this is not called to handle redirection..
	//	return FALSE;

	int pid = getpid();
	bool redirect = false;
	bool wait = false;

	while (true)
	{
		wait = false;
		WaitForSingleObject(gSharedMutex, INFINITE);

		if (gRedirections[0].Pid > 0)
		{
			wait = true;
		}
		else
		{
			for (int i = 1; i <= MAX_REDIRECTIONS; i++)
				if (gRedirections[i].Pid == pid)
				{
					redirect = true;
					// if redirection info requested, copy it there
					if (lpRedirection != NULL)
						memcpy(lpRedirection, &gRedirections[i], sizeof(REDIRECTION));

					break;
				}
		}
		ReleaseMutex(gSharedMutex);

		if (wait)
			Sleep(1);
		else
			break;
	}

	return redirect;
}

int GetWrapperDllPath(LPTSTR path)
{
  HKEY netsimKey;
	HRESULT hr;

  DWORD cValues = 0;              // number of values for key 
  DWORD cchMaxValue;          // longest value name 
  DWORD cbMaxValueData;       // longest value data 
  DWORD cType;

  DWORD j; 
  DWORD retValue; 

  TCHAR achValue[MAX_PATH]; 
  DWORD cchValue = MAX_PATH; 
  TCHAR achData[MAX_PATH];
  DWORD cchData = MAX_PATH;

  // default value is only phpc.exe - will be enough if phpc is in PATH
  hr = RegOpenKeyEx(HKEY_LOCAL_MACHINE, "SOFTWARE\\Pavel Novak\\Network Simulator", 0, KEY_READ, &netsimKey);
	if (hr != ERROR_SUCCESS)
		return -1;

  // Get the class name and the value count. 
  RegQueryInfoKey(netsimKey,        // key handle 
    NULL,                // buffer for class name 
    0,           // size of class string 
    NULL,                    // reserved 
    NULL,               // number of subkeys 
    NULL,            // longest subkey size 
    NULL,            // longest class string 
    &cValues,                // number of values for this key 
    &cchMaxValue,            // longest value name 
    &cbMaxValueData,         // longest value data 
    NULL,   // security descriptor 
    NULL);       // last write time 

  // Enumerate the key values. 
  if (cValues) 
  {
    for (j = 0, retValue = ERROR_SUCCESS; j < cValues; j++) 
    { 
      cchValue = MAX_PATH; 
      achValue[0] = '\0';
      cchData = MAX_PATH;
      achData[0] = '\0';

      retValue = RegEnumValue(netsimKey, j, achValue, 
        &cchValue, 
        NULL, 
        &cType,    // &dwType, 
        (LPBYTE)achData,    // &bData, 
        &cchData);   // &bcData 

      if (retValue == (DWORD) ERROR_SUCCESS ) 
      { 
        if (!strcmp(achValue, "LSPWrapper") && cType == REG_SZ)
        {
          strcpy(path, achData);
          break;
        }
      } 
    }
  }

  RegCloseKey(netsimKey);
  return 0;
}

int NetSimStartup()
{
	INT iErrno = NO_ERROR;
	pfnInitProcTable pfnInitWrapperProcTable = NULL;

	//
	// Load the wrapper library containing the managed code..
	//
	TCHAR libPath[MAX_PATH];
	if (GetWrapperDllPath(libPath) != 0)
		return -1;

	HMODULE gWrapperDll = LoadLibrary(libPath);

	// initialize CLR
	// http://support.microsoft.com/?id=814472
	pfnEnsureInit pfn_dll = NULL;
	if (gWrapperDll)
	{
		pfn_dll = (pfnEnsureInit)::GetProcAddress(gWrapperDll, "DllEnsureInit");
		if (pfn_dll) pfn_dll();

		pfnInitWrapperProcTable = (pfnInitProcTable)::GetProcAddress(gWrapperDll, "InitProcTable");
		gNestingCount = (pfnNestingCount)::GetProcAddress(gWrapperDll, "NestingCount");
		gGetThreadStartAddress = (pfnGetThreadStartAddress)::GetProcAddress(gWrapperDll, "GetThreadStartAddress");
	}
	if (!pfn_dll || !pfnInitWrapperProcTable || !gNestingCount || !gGetThreadStartAddress)
	{
		return -1;
	}

	pfnInitWrapperProcTable(&gWrapperProcTable, &gMainUpCallTable);

	//
	// initialize our providers
	//

	// copy real base provider's structures
	gNetSimBaseInfo = (LPPROVIDER) LspAlloc(gLayerCount * sizeof(PROVIDER), &iErrno);
	if ( NULL == gNetSimBaseInfo )
    {
        dbgprint("NetSimStartup: LspAlloc failed: %d", iErrno );
        goto cleanup;
    }
	memcpy(gNetSimBaseInfo, gBaseInfo, gLayerCount * sizeof(PROVIDER));
	
	// change some values
	for (INT i = 0; i < gLayerCount; i++)
	{
		// no extra module loaded
		gNetSimBaseInfo[i].ProviderPathW[0] = '\0';
		gNetSimBaseInfo[i].ProviderPathLen = 0;
		gNetSimBaseInfo[i].LibraryPathW[0] = '\0';
		gNetSimBaseInfo[i].Module = NULL;
		// we don't have own startup function
		gNetSimBaseInfo[i].fnWSPStartup = NULL;

		InitProcTable(&gNetSimBaseInfo[i].NextProcTable);

		InitializeCriticalSection(&gNetSimBaseInfo[i].ProviderCritSec);
		InitializeListHead(&gNetSimBaseInfo[i].SocketList);
	}

	// Allow for redirect to us only if everything is OK.
	gIsWrapperLoaded = TRUE;

cleanup:
	
	return iErrno;
}

int NetSimCleanup()
{
	// there is nothing to cleanup if nothing is loaded..
	if (!gIsWrapperLoaded)
		return NO_ERROR;
	
	gIsWrapperLoaded = FALSE;

	for (INT i = 0; i < gLayerCount; i++)
	{
		DeleteCriticalSection(&gNetSimBaseInfo[i].ProviderCritSec);
	}

	LspFree(gNetSimBaseInfo);
	gNetSimBaseInfo = NULL;

	// terminate CLR
	// http://support.microsoft.com/?id=814472
	if (gWrapperDll)
	{
		pfnForceTerm pfnDll = (pfnForceTerm) ::GetProcAddress(gWrapperDll, "DllForceTerm");
		if (pfnDll) pfnDll();
	}

	return NO_ERROR;
}

int AddRedirection(unsigned int pid, char * computer, unsigned short port, char * nodeName)
{
	if (!gSharedMutex || !gRedirections)
	{
		// try to init again
		InitSharedData();

		// if not loaded now, no redirection is possible
		if (!gSharedMutex || !gRedirections)
			return -2;
	}

	WaitForSingleObject(gSharedMutex, INFINITE);

	bool added = false;
	for (int i = 1; i <= MAX_REDIRECTIONS; i++)
	{
		if (gRedirections[i].Pid == pid) // already in the list
		{
			added = true;
			break;
		}

		if (gRedirections[i].Pid == 0) // first empty
		{
			gRedirections[i].Pid = pid;
			strncpy(gRedirections[i].Computer, computer, sizeof(gRedirections[i].Computer));
			gRedirections[i].Computer[sizeof(gRedirections[i].Computer)-1] = '\0';
			gRedirections[i].Port = port;
			strncpy(gRedirections[i].NodeName, nodeName, sizeof(gRedirections[i].NodeName));
			gRedirections[i].NodeName[sizeof(gRedirections[i].NodeName)-1] = '\0';
			added = true;
			break;
		}
	}

	ReleaseMutex(gSharedMutex);

	return added ? 0 : -1;
}

int RemoveRedirection(unsigned int pid)
{
	if (!gSharedMutex || !gRedirections)
	{
		// try to init again
		InitSharedData();

		// if not loaded now, no redirection is possible
		if (!gSharedMutex || !gRedirections)
			return FALSE;
	}

	WaitForSingleObject(gSharedMutex, INFINITE);
	
	bool deleted = false;
	for (int i = 1; i <= MAX_REDIRECTIONS; i++)
	{
		if (gRedirections[i].Pid == 0) // first empty - all others are empty, do not continue
			break;

		if (deleted) // we have deleted, so move current to the previous slot
		{
			gRedirections[i-1].Pid = gRedirections[i].Pid;
			strcpy(gRedirections[i-1].Computer, gRedirections[i].Computer);
			gRedirections[i-1].Port = gRedirections[i].Port;
			gRedirections[i].Pid = 0;
			gRedirections[i].Computer[0] = '\0';
			gRedirections[i].Port = 0;
		}
		else if (gRedirections[i].Pid == pid) // no deletion yet, test if it is the current
		{
			gRedirections[i].Pid = 0;
			gRedirections[i].Computer[0] = '\0';
			gRedirections[i].Port = 0;
			deleted = true;
		}
	}

	ReleaseMutex(gSharedMutex);
	
	return deleted ? 0 : -1;
}

int BlockRedirections()
{
	WaitForSingleObject(gSharedMutex, INFINITE);

	gRedirections[0].Pid++;

	ReleaseMutex(gSharedMutex);
	return 0;
}

int ResumeRedirections()
{
	WaitForSingleObject(gSharedMutex, INFINITE);

	if (gRedirections[0].Pid > 0)
		gRedirections[0].Pid--;

	ReleaseMutex(gSharedMutex);
	return 0;
}

//
// Returns 0 on success, nonzero if failed
//
int GetRedirection(int index, int * pid, char * computer, int computerSize, unsigned short * port, char * nodeName, int nodeNameSize)
{
	if (index <= 0 || index > MAX_REDIRECTIONS)
		return -1;

	int result = 0;

	WaitForSingleObject(gSharedMutex, INFINITE);

	if (gRedirections[index].Pid == 0)
	{
		result = 1;
	}
	else
	{
		*pid = gRedirections[index].Pid;
		strncpy(computer, gRedirections[index].Computer, computerSize-1);
		computer[computerSize-1] = '\0';
		*port = gRedirections[index].Port;
		strncpy(nodeName, gRedirections[index].NodeName, nodeNameSize-1);
		nodeName[nodeNameSize-1] = '\0';
	}

	ReleaseMutex(gSharedMutex);

	return result;
}

// CreateMyDACL.
//    Create a security descriptor that contains the DACL 
//    you want.
//    This function uses SDDL to make Deny and Allow ACEs.
//
// Parameter:
//    SECURITY_ATTRIBUTES * pSA
//    Pointer to a SECURITY_ATTRIBUTES structure. It is your
//    responsibility to properly initialize the 
//    structure and to free the structure's 
//    lpSecurityDescriptor member when you have
//    finished using it. To free the structure's 
//    lpSecurityDescriptor member, call the 
//    LocalFree function.
// 
// Return value:
//    FALSE if the address to the structure is NULL. 
//    Otherwise, this function returns the value from the
//    ConvertStringSecurityDescriptorToSecurityDescriptor 
//    function.
BOOL CreateMyDACL(SECURITY_ATTRIBUTES * pSA)
{
     // Define the SDDL for the DACL. This example sets 
     // the following access:
     //     Built-in guests are denied all access.
     //     Anonymous logon is denied all access.
     //     Authenticated users are allowed 
     //     read/write/execute access.
     //     Administrators are allowed full control.
     // Modify these values as needed to generate the proper
     // DACL for your application. 
     TCHAR * szSD = TEXT("D:")       // Discretionary ACL
        //TEXT("(D;OICI;GA;;;BG)")     // Deny access to 
        //                             // built-in guests
        //TEXT("(D;OICI;GA;;;AN)")     // Deny access to 
        //                             // anonymous logon
        //TEXT("(A;OICI;GRGWGX;;;AU)") // Allow 
        //                             // read/write/execute 
        //                             // to authenticated 
        //                             // users
        TEXT("(A;OICI;GA;;;WD)")    // Allow full control to everyone

        TEXT("(A;OICI;GA;;;BA)");    // Allow full control 
                                     // to administrators

    if (NULL == pSA)
        return FALSE;

     return ConvertStringSecurityDescriptorToSecurityDescriptor(
                szSD,
                SDDL_REVISION_1,
                &(pSA->lpSecurityDescriptor),
                NULL);
}

void InitSharedData()
{
	// the first item is used for blocking redirections counter
	UINT shared_size = (MAX_REDIRECTIONS+1) * sizeof(REDIRECTION);

	if (gMappedFile == NULL)
	{
		// prepare security descriptors for the mapping
		gFileMappingSecurityAttributes.nLength = sizeof(SECURITY_ATTRIBUTES);
		gFileMappingSecurityAttributes.bInheritHandle = FALSE;
		// Set DACL to lpSecurityDescriptor member of security attributes.
		if (!CreateMyDACL(&gFileMappingSecurityAttributes))
		{
			// Error encountered; generate message and exit.
			::MessageBox(NULL, "Failed CreateMyDACL\n", "Error", 0);
			return;
		}

		// if this is the first process, it will create the mapping, if not,
		// it will return the handle to existing
		gMappedFile = CreateFileMapping(
						INVALID_HANDLE_VALUE,    // use paging file
						&gFileMappingSecurityAttributes,
						PAGE_READWRITE,          // read/write access
						0,                       
						shared_size,
						SHARED_NAME);                 // name of mapping object
		 
		if (gMappedFile == NULL || gMappedFile == INVALID_HANDLE_VALUE) 
		{ 
			DWORD code = ::GetLastError();
			char str[50];
			sprintf(str, "%d", code);
			::MessageBox(NULL, "Cannot create file mapping", str, 0);

			gMappedFile = NULL;
			return;
		}
	}

	if (gMappedFile && gRedirections == NULL)
	{
		gRedirections = (REDIRECTION *) MapViewOfFile(gMappedFile,   // handle to map object
					FILE_MAP_ALL_ACCESS, // read/write permission
					0, 0, shared_size);           
		 
		if (gRedirections == NULL) 
		{
			CloseHandle(gMappedFile);
			gMappedFile = NULL;
			return;
		}
	}

	if (gSharedMutex == NULL)
		gSharedMutex = CreateMutex(&gFileMappingSecurityAttributes, FALSE, SHARED_MUTEX_NAME);
}

void DestroySharedData()
{
	if (gSharedMutex)
	{
		CloseHandle(gSharedMutex);
		gSharedMutex = NULL;
	}
	if (gRedirections)
	{
		UnmapViewOfFile(gRedirections);
		gRedirections = NULL;
	}
	if (gMappedFile)
	{
		CloseHandle(gMappedFile);
		gMappedFile = NULL;
	}
}

