
#include <process.h>
#include "netsimspecific.h"
#include "netsimspi.h"

//WORKING:
BOOL mutexInitCalled = FALSE;

//
// Globals for NetSim support
//

// NetSim provider structures. These providers are not actually registered with WinSock,
// but are passed as a parameter to functions here if they should communicate with
// virtual network instead of real base providers installed on the computer.
// We create such dummy PROVIDER structure for each provider we are actually layered on.
LPPROVIDER gNetSimBaseInfo = NULL;

// The information whether the LSP Wrapper DLL has been succesfuly loaded and therefore
// we have correct function addresses to redirect etc..
static BOOL gIsWrapperLoaded = FALSE;

#define MAX_REDIRECTIONS 32
#define SHARED_NAME "NetSimSharedRedirections"
#define SHARED_MUTEX_NAME "NetSimSharedMutexRedirections"

// variables initialized while creating mapped file
static HANDLE gMappedFile = NULL;
static HANDLE gSharedMutex = NULL; // Critical section for accessing processes shared data
static REDIRECTION * gRedirections = NULL; // Max 32 redirections allowed from one computer

static HMODULE gWrapperDll = NULL;

int gRedirectedNestingCounter = 0;

// Returns true if the current call is from the application that should communicate through
// the virtual network. This function checkes current PID and returns true if it is in the
// list of applications maintained by the NetSim.
BOOL NetSimRedirectCurrentCall()
{
	// Do not redirect if we were unable to load the wrapper DLL...
	if (!gIsWrapperLoaded)
		return FALSE;
	
	if (!gSharedMutex || !gRedirections)
	{
		// try to init again
		InitSharedData();

		// if not loaded now, no redirection is possible
		if (!gSharedMutex || !gRedirections)
			return FALSE;
	}

	if (gRedirectedNestingCounter > 0) // redirect only if this is not called to handle redirection..
		return FALSE;

	int pid = getpid();
	bool redirect = false;

	WaitForSingleObject(gSharedMutex, INFINITE);
	for (int i = 0; i < MAX_REDIRECTIONS; i++)
	{
		if (gRedirections[i].Pid == pid)
		{
			redirect = true;
			break;
		}
	}
	ReleaseMutex(gSharedMutex);

	return redirect;
}


int NetSimStartup()
{
	INT iErrno = NO_ERROR;
	pfnInitProcTable pfnInitWrapperProcTable = NULL;

	//
	// Load the wrapper library containing the managed code..
	//
	//TODO: Remove path..
	HMODULE gWrapperDll = LoadLibrary("E:\\Pavel\\diplomka\\NetSim\\LSPWrapper\\Debug\\LSPWrapper.dll");

	// initialize CLR
	// http://support.microsoft.com/?id=814472
	pfnEnsureInit pfn_dll = NULL;
	if (gWrapperDll)
	{
		pfn_dll = (pfnEnsureInit)::GetProcAddress(gWrapperDll, "DllEnsureInit");
		if (pfn_dll) pfn_dll();

		pfnInitWrapperProcTable = (pfnInitProcTable)::GetProcAddress(gWrapperDll, "InitProcTable");
	}
	if (!pfn_dll || !pfnInitWrapperProcTable)
	{
		return -1;
	}

	pfnInitWrapperProcTable(&gWrapperProcTable);

	//
	// initialize our providers
	//

	// copy real base provider's structures
	gNetSimBaseInfo = (LPPROVIDER) LspAlloc(gLayerCount * sizeof(PROVIDER), &iErrno);
	if ( NULL == gNetSimBaseInfo )
    {
        dbgprint("NetSimStartup: LspAlloc failed: %d", iErrno );
        goto cleanup;
    }
	memcpy(gNetSimBaseInfo, gBaseInfo, gLayerCount * sizeof(PROVIDER));
	
	// change some values
	for (INT i = 0; i < gLayerCount; i++)
	{
		// no extra module loaded
		gNetSimBaseInfo[i].ProviderPathW[0] = '\0';
		gNetSimBaseInfo[i].ProviderPathLen = 0;
		gNetSimBaseInfo[i].LibraryPathW[0] = '\0';
		gNetSimBaseInfo[i].Module = NULL;
		// we don't have own startup function
		gNetSimBaseInfo[i].fnWSPStartup = NULL;

		// initialize proc table to our functions...
		InitProcTable(&gNetSimBaseInfo[i].NextProcTable);

		InitializeCriticalSection(&gNetSimBaseInfo[i].ProviderCritSec);
		InitializeListHead(&gNetSimBaseInfo[i].SocketList);
	}

	// Allow for redirect to us only if everything is OK.
	gIsWrapperLoaded = TRUE;

cleanup:
	
	return iErrno;
}

int NetSimCleanup()
{
	// there is nothing to cleanup if nothing is loaded..
	if (!gIsWrapperLoaded)
		return NO_ERROR;
	
	gIsWrapperLoaded = FALSE;

	for (INT i = 0; i < gLayerCount; i++)
	{
		DeleteCriticalSection(&gNetSimBaseInfo[i].ProviderCritSec);
	}

	LspFree(gNetSimBaseInfo);
	gNetSimBaseInfo = NULL;

	// terminate CLR
	// http://support.microsoft.com/?id=814472
	if (gWrapperDll)
	{
		pfnForceTerm pfnDll = (pfnForceTerm) ::GetProcAddress(gWrapperDll, "DllForceTerm");
		if (pfnDll) pfnDll();
	}

	return NO_ERROR;
}

int AddRedirection(unsigned int pid, char * computer, unsigned short port)
{
	if (!gSharedMutex || !gRedirections)
	{
		// try to init again
		InitSharedData();

		// if not loaded now, no redirection is possible
		if (!gSharedMutex || !gRedirections)
			return FALSE;
	}

	if (!gSharedMutex)
	{
		if (mutexInitCalled)
			return -20;
		else
			return -2;
	}
	if (!gRedirections)
		return -3;

    WaitForSingleObject(gSharedMutex, INFINITE);

	bool added = false;
	for (int i = 0; i < MAX_REDIRECTIONS; i++)
	{
		if (gRedirections[i].Pid == pid) // already in the list
		{
			added = true;
			break;
		}

		if (gRedirections[i].Pid == 0) // first empty
		{
			gRedirections[i].Pid = pid;
			strcpy(gRedirections[i].Computer, computer);
			gRedirections[i].Port = port;
			added = true;
			break;
		}
	}

	ReleaseMutex(gSharedMutex);

	return added ? 0 : -1;
}

int RemoveRedirection(unsigned int pid)
{
	if (!gSharedMutex || !gRedirections)
	{
		// try to init again
		InitSharedData();

		// if not loaded now, no redirection is possible
		if (!gSharedMutex || !gRedirections)
			return FALSE;
	}

	WaitForSingleObject(gSharedMutex, INFINITE);
	
	bool deleted = false;
	for (int i = 0; i < MAX_REDIRECTIONS; i++)
	{
		if (gRedirections[i].Pid == 0) // first empty - all others are empty, do not continue
			break;

		if (deleted) // we have deleted, so move current to the previous slot
		{
			gRedirections[i-1].Pid = gRedirections[i].Pid;
			strcpy(gRedirections[i-1].Computer, gRedirections[i].Computer);
			gRedirections[i-1].Port = gRedirections[i].Port;
			gRedirections[i].Pid = 0;
			gRedirections[i].Computer[0] = '\0';
			gRedirections[i].Port = 0;
		}
		else if (gRedirections[i].Pid == pid) // no deletion yet, test if it is the current
		{
			gRedirections[i].Pid = 0;
			gRedirections[i].Computer[0] = '\0';
			gRedirections[i].Port = 0;
			deleted = true;
		}
	}

    ReleaseMutex(gSharedMutex);
	
	return deleted ? 0 : -1;
}


void InitSharedData()
{
	UINT shared_size = MAX_REDIRECTIONS * sizeof(REDIRECTION);

	mutexInitCalled = TRUE;

	if (gMappedFile == NULL)
	{
		// if this is the first process, it will create the mapping, if not,
		// it will return the handle to existing
		gMappedFile = CreateFileMapping(
						INVALID_HANDLE_VALUE,    // use paging file
						NULL,                    // default security 
						PAGE_READWRITE,          // read/write access
						0,                       
						shared_size,
						SHARED_NAME);                 // name of mapping object
		 
		if (gMappedFile == NULL || gMappedFile == INVALID_HANDLE_VALUE) 
		{ 
			gMappedFile = NULL;
			return;
		}
	}

	if (gMappedFile && gRedirections == NULL)
	{
		gRedirections = (REDIRECTION *) MapViewOfFile(gMappedFile,   // handle to map object
					FILE_MAP_ALL_ACCESS, // read/write permission
					0, 0, shared_size);           
		 
		if (gRedirections == NULL) 
		{
			CloseHandle(gMappedFile);
			gMappedFile = NULL;
			return;
		}
	}

	if (gSharedMutex == NULL)
		gSharedMutex = CreateMutex(NULL, FALSE, SHARED_MUTEX_NAME);
}

void DestroySharedData()
{
	if (gSharedMutex)
	{
		CloseHandle(gSharedMutex);
		gSharedMutex = NULL;
	}
	if (gRedirections)
	{
		UnmapViewOfFile(gRedirections);
		gRedirections = NULL;
	}
	if (gMappedFile)
	{
		CloseHandle(gMappedFile);
		gMappedFile = NULL;
	}
}

