/*

 Copyright (c) 2006 Pavel Novak.

 The use and distribution terms for this software are contained in the file named License.txt, 
 which can be found in the root of the NetSim distribution. By using this software 
 in any fashion, you are agreeing to be bound by the terms of this license.
 
 You must not remove this notice from this software.

*/

using System;
using NetSim.Core;
using System.Diagnostics;

namespace NetSim.Library
{
	/// <summary>
	/// Represents an IP address - either IPv4 or IPv6.
	/// </summary>
	public abstract class IPAddress : NetworkAddress, IComparable
	{
		public static IPAddress Parse(string address)
		{
			System.Net.IPAddress sys_addr = System.Net.IPAddress.Parse(address);
			if (sys_addr.AddressFamily == System.Net.Sockets.AddressFamily.InterNetwork)
			{
				return new IPv4Address(sys_addr);
			}
			else if (sys_addr.AddressFamily == System.Net.Sockets.AddressFamily.InterNetworkV6)
			{
				return new IPv6Address(sys_addr);
			}
			else
			{
				throw new ArgumentException(String.Format("Wrong address format ({0}) - not either IPv4 or IPv6", address));
			}
		}

		public override Packet.PacketGroup PacketGroup
		{
			get { return Packet.PacketGroup.NetworkLayer; }
		}

		public override int PacketType
		{
			get {	return (int)Packet.NetworkLayerPacketType.IP; }
		}

		public abstract byte IPVersion { get; }
		public abstract int CompareTo(object obj);

		public static bool IsInNetwork(IPAddress addr, IPAddress networkAddr, IPAddress mask)
		{
			Debug.Assert(networkAddr.Length == mask.Length);

			if (addr.IPVersion != networkAddr.IPVersion)
				return false; // different version

			byte[] addr_bytes = addr.GetAddressBytes();
			byte[] net_bytes = networkAddr.GetAddressBytes();
			byte[] mask_bytes = mask.GetAddressBytes();

			for (int i = 0; i < addr_bytes.Length; i++)
			{
				if (addr_bytes[i] == net_bytes[i] || mask_bytes[i] == 0)
					continue; // this byte is OK

				// bit by bit check is needed
				byte bit = 1 << 7;
				for (int j = 0; j < 8; j++)
				{
					if ((mask_bytes[i] & bit) != 0 // bit here is important
						&& (addr_bytes[i] & bit) != (net_bytes[i] & bit)) // and bits in network and address are different
						return false;

					bit >>= 1;
				}
			}

			return true;
		}

		public abstract System.Net.IPAddress ToSystemIPAddress();
		internal static IPAddress CreateAddress(System.Net.IPAddress systemIP)
		{
			if (systemIP.AddressFamily == System.Net.Sockets.AddressFamily.InterNetwork)
				return new IPv4Address(systemIP);
			else if (systemIP.AddressFamily == System.Net.Sockets.AddressFamily.InterNetworkV6)
				return new IPv6Address(systemIP);
			else
				throw new ArgumentException("Only InterNetwork or InternNetworkV6 supported", "systemIP");
		}

		public override NetworkAddress Loopback
		{
			get
			{
				return CreateAddress(System.Net.IPAddress.Loopback);
			}
		}
	}

	/// <summary>
	/// IPv4 network address.
	/// </summary>
	public class IPv4Address : IPAddress
	{
		public static readonly IPv4Address Broadcast = IPv4Address.Parse("255.255.255.255");

		private System.Net.IPAddress ipAddress;

		internal IPv4Address(System.Net.IPAddress sysAddress)
		{
			this.ipAddress = sysAddress;
		}

		public override System.Net.IPAddress ToSystemIPAddress()
		{
			return this.ipAddress;
		}

		/// <summary>
		/// Creates new IPv4Address according to UInt32 value. Byte order is network order.
		/// </summary>
		/// <param name="address">Address in network order endiannes.</param>
		internal IPv4Address(UInt32 address)
		{
			//Future version: better (faster) implementation
			byte[] bytes = new byte[4];
			bytes[3] = unchecked((byte)address); address >>= 8;
			bytes[2] = unchecked((byte)address); address >>= 8;
			bytes[1] = unchecked((byte)address); address >>= 8;
			bytes[0] = unchecked((byte)address); address >>= 8;
			string addr = bytes[0].ToString() + "." + bytes[1].ToString() + "." + bytes[2].ToString() + "." + bytes[3].ToString();
			this.ipAddress = System.Net.IPAddress.Parse(addr); //new System.Net.IPAddress(ipv6_bytes);
		}

		protected IPv4Address(string address)
		{
			this.ipAddress = System.Net.IPAddress.Parse(address);
		}

		protected IPv4Address(byte[] bytes) : this(bytes, 0)
		{
		}

		protected IPv4Address(byte[] bytes, int offset)
		{
			//Future version: better (faster) implementation
			string addr = bytes[offset].ToString() + "." + bytes[offset+1].ToString() + "." + bytes[offset+2].ToString() + "." + bytes[offset+3].ToString();
			this.ipAddress = System.Net.IPAddress.Parse(addr); //new System.Net.IPAddress(ipv6_bytes);
		}

		public static IPv4Address CreateAddress(byte[] bytes)
		{
			return new IPv4Address(bytes);
		}

		public static IPv4Address CreateAddress(byte[] bytes, int offset)
		{
			return new IPv4Address(bytes, offset);
		}

		public static new IPv4Address Parse(string address)
		{
			return new IPv4Address(address);
		}

		public override byte[] GetAddressBytes()
		{
			return ipAddress.GetAddressBytes();
		}

		public override NetworkAddress CreateNewAddress(byte[] bytes)
		{
			if (this.Length != bytes.Length)
				throw new ArgumentException(String.Format("Byte array of {0} bytes expected", this.Length));

			return IPv4Address.CreateAddress(bytes);
		}

		public override bool Equals(object obj)
		{
			IPv4Address second = obj as IPv4Address;
			if (second == null)
				return false;

			return this.ipAddress.Equals(second.ipAddress);
		}

		public override int GetHashCode()
		{
			return this.ipAddress.GetHashCode();
		}

		public override byte Length
		{
			get { return 4; }
		}

		public override byte IPVersion
		{
			get {	return 4; }
		}

		#region IComparable Members

		public override int CompareTo(object obj)
		{
			IPv4Address second = obj as IPv4Address;
			if (second == null)
				throw new ArgumentException(String.Format("Cannot compare IPAddress to {0}", obj.ToString()));

			byte[] this_bytes = this.ipAddress.GetAddressBytes();
			byte[] second_bytes = second.ipAddress.GetAddressBytes();
			int result = 0;
			for (int i = 0; i < this.Length; i++)
			{
				result = this_bytes[i] - second_bytes[i];
				if (result != 0)
					break;
			}

			return result;
		}

		#endregion
	}


	/// <summary>
	/// IPv6 network address.
	/// </summary>
	public class IPv6Address : IPAddress
	{
		private System.Net.IPAddress ipAddress;

		public override System.Net.IPAddress ToSystemIPAddress()
		{
			return this.ipAddress;
		}

		internal IPv6Address(System.Net.IPAddress sysAddress)
		{
			this.ipAddress = sysAddress;
		}

		protected IPv6Address(string address)
		{
			this.ipAddress = System.Net.IPAddress.Parse(address);
		}

		protected IPv6Address(byte[] bytes)
		{
			this.ipAddress = new System.Net.IPAddress(bytes);
		}

		public static new IPv6Address Parse(string address)
		{
			return new IPv6Address(address);
		}

		public static IPv6Address CreateAddress(byte[] bytes)
		{
			return new IPv6Address(bytes);
		}

		public override byte[] GetAddressBytes()
		{
			return ipAddress.GetAddressBytes();
		}

		public override NetworkAddress CreateNewAddress(byte[] bytes)
		{
			if (this.Length != bytes.Length)
				throw new ArgumentException(String.Format("Byte array of {0} bytes expected", this.Length));

			return IPv6Address.CreateAddress(bytes);
		}

		public override bool Equals(object obj)
		{
			IPv6Address second = obj as IPv6Address;
			if (second == null)
				return false;

			return this.ipAddress.Equals(second.ipAddress);
		}

		public override int GetHashCode()
		{
			return this.ipAddress.GetHashCode();
		}

		public override byte Length
		{
			get { return 16; }
		}

		public override byte IPVersion
		{
			get { return 6; }
		}

		#region IComparable Members

		public override int CompareTo(object obj)
		{
			IPv6Address second = obj as IPv6Address;
			if (second == null)
				throw new ArgumentException(String.Format("Cannot compare IPAddress to {0}", obj.ToString()));

			byte[] this_bytes = this.ipAddress.GetAddressBytes();
			byte[] second_bytes = second.ipAddress.GetAddressBytes();
			int result = 0;
			for (int i = 0; i < this.Length; i++)
			{
				result = this_bytes[i] - second_bytes[i];
				if (result != 0)
					break;
			}

			return result;
		}

		#endregion
	}

//	public class LibAddressFactory : AddressFactory
//	{
//	}
}
