/*

 Copyright (c) 2006 Pavel Novak.

 The use and distribution terms for this software are contained in the file named License.txt, 
 which can be found in the root of the NetSim distribution. By using this software 
 in any fashion, you are agreeing to be bound by the terms of this license.
 
 You must not remove this notice from this software.

*/

using System;
using NetSim.Core;
using System.Threading;
using System.Net;
using System.Net.Sockets;
using IIPRoutingModule = NetSim.Library.Modules.IIPRoutingModule;
using RoutingTable = NetSim.Library.Modules.RoutingTable;
using System.Collections;
using System.Xml;

namespace NetSim.Library.Applications
{
	/// <summary>
	/// Summary description for RIP2.
	/// </summary>
	public class Rip : NodeApplication
	{
		protected class TableEntryData
		{
			public UInt32 Distance;
			public UInt16 RouteTag;
			public long Expires;

			public TableEntryData(UInt32 distance, UInt16 routeTag, long expires)
			{
				this.Distance = distance;
				this.RouteTag = routeTag;
				this.Expires = expires;
			}
		}

		/// <summary>
		/// The port used for sending and receiving UDP rip messages. Defaults to 520.
		/// </summary>
		private UInt16 RipPort = 520;

		/// <summary>
		/// The delay in miliseconds between two anouncements of routing information. Defaults to 30000 ms.
		/// </summary>
		private int RoutingInfoSendInterval = 30000;

		/// <summary>
		/// The time (in miliseconds) for which the route entry is valid without receiving an advertisement.
		/// After this delay, the route becomes invalid and should be removed from the table.
		/// Defaults to 180 seconds (ie. 180000 ms).
		/// </summary>
		private int RouteEntryRemovalTimeout = 180 * 1000;

		/// <summary>
		/// The maximum number of entries in one update message. Defaults to 25.
		/// </summary>
		private int RouteEntriesPerUpdateMessage = 25;

		/// <summary>
		/// The number of entries that will fit in the buffer for incoming update messages. Defaults to 50.
		/// </summary>
		private int ReceiveBufferLength = 50 * 20 + 4;

		/// <summary>
		/// Wheather the RIP application is in passive mode - ie. doesn't send updates. Defaults to false;
		/// </summary>
		private bool PassiveMode = false;

		private Thread updateThread;
		private AutoResetEvent updateThreadWait;
		
		/// <summary>
		/// Socket for incoming data.
		/// </summary>
		private NetSimSocket socket;

		private string routingModuleName;
		private IIPRoutingModule routingModule;

		/// <summary>
		/// <see cref="ArrayList"/> of <see cref="NetSim.Library.IPAddress"/> instances of participating routers
		/// attached to directly accessible networks.
		/// </summary>
		// generics: <IPAddress>
		private ArrayList neighborRouters = new ArrayList();

		public Rip()
		{
		}

		protected override int Run()
		{
			this.routingModule = Node.GetModule(routingModuleName) as IIPRoutingModule;
			if (routingModule == null)
			{
				Simulator.GetInstance().Events.Error(String.Format("Rip application at node '{0}': No routing module with name '{1}'",
					Node.Name, routingModuleName));
				return -1;
			}

			// bind to UDP socket intended for RIP usage
			socket = new NetSimSocket(AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.Udp, this.Node);
			socket.Bind(new System.Net.IPEndPoint(System.Net.IPAddress.Any, RipPort), true);
			
			// run the thread that will announce routing information
			updateThreadWait = new AutoResetEvent(false);
			updateThread = new Thread(new ThreadStart(this.UpdateThread));
			updateThread.Name = "RipUpdateThread";
			ThreadList.Add(updateThread);
			updateThread.Start();

			//
			// the loop receiving routing information
			//
			byte[] buffer = new byte[ReceiveBufferLength * 20 + 4]; // ReceiveBufferLength * ONE-ENTRY-SIZE + HEADER-SIZE
			EndPoint remote_ep = new IPEndPoint(System.Net.IPAddress.Any, 0);

			while (true)
			{
				int count = buffer.Length;
				try
				{
					count = socket.ReceiveFrom(buffer, 0, buffer.Length, SocketFlags.None, ref remote_ep);
				}
				catch (SocketException e)
				{
					Simulator.GetInstance().Events.Notice("RIP: Exception caught while receiving message - buffer not large enough?", e);
				}

				ProcessIncomingMessage(buffer, 0, count, IPAddress.CreateAddress(((IPEndPoint)remote_ep).Address));
			}			

		}

		private void ProcessIncomingMessage(byte[] message, int offset, int length, IPAddress nextHop)
		{
			byte command = message[offset++];
			byte version = message[offset++];
			length -= 2;
			offset += 2;

			if (version != 2) // only v2 supported
				return;

			switch (command)
			{
				case 1: // request for routing information update
					updateThreadWait.Set(); // wake up the update thread that will send the information
					break;
					
				case 2: // routing update message
					while (length >= 20)
					{
						ProcessOneEntry(message, offset, nextHop);
						offset += 20;
						length -= 20;
					}

					break;
			}
		}

		private void ProcessOneEntry(byte[] message, int offset, IPAddress nextHop)
		{
			//
			// Decompose message
			//
			UInt16 family = ByteArrayList.ToUInt16(message, offset);
			offset += 2;
			UInt16 route_tag = ByteArrayList.ToUInt16(message, offset);
			offset += 2;
			IPv4Address ip_addr = IPv4Address.CreateAddress(message, offset);
			offset += 4;
			IPv4Address mask = IPv4Address.CreateAddress(message, offset);
			offset += 4;
			IPv4Address entry_next_hop = IPv4Address.CreateAddress(message, offset);
			offset += 4;
			UInt32 distance = ByteArrayList.ToUInt32(message, offset);
			offset += 4;

			if (family != 2) // not an IP
				return;
			
			//if (Node.IsLocalAddress(entry_next_hop)) // do not accept entries that have our address as a gateway
			//	return;

			RoutingTable rt = this.routingModule.RoutingTable;
			lock (rt.SyncRoot)
			{
				RoutingTable.Entry entry = rt.Get(ip_addr, mask);
				if (entry == null) // not in table
				{
					entry = new RoutingTable.Entry(ip_addr, mask, nextHop, null);
					entry.PrivateData = new TableEntryData(distance + 1, route_tag,
						Simulator.Timer.GetTimeAfterDelay(RouteEntryRemovalTimeout * 10000L));
					rt.Add(entry);
				}
				else // an entry exist
				{
					TableEntryData ted = entry.PrivateData as TableEntryData;
					if (ted != null) // our RIP entry - otherwise do not modify
					{
						bool replace = false;
						if (ted.Distance > distance + 1) // received route is shorter
							replace = true;
						if (ted.Expires < Simulator.Timer.Now) // entry expired
							replace = true;

						if (replace)
						{
							entry.Gateway = nextHop;
							ted.Distance = distance + 1;
							ted.RouteTag = route_tag;
							ted.Expires = Simulator.Timer.GetTimeAfterDelay(RouteEntryRemovalTimeout * 10000L);
						}
						else if (entry.Gateway.Equals(nextHop) && ted.Distance == distance + 1) // update expiration time
						{
							ted.Expires = Simulator.Timer.GetTimeAfterDelay(RouteEntryRemovalTimeout * 10000L);
						}
					}
				}
			}
		}

		private void UpdateThread()
		{
			//generics: <byte[]>
			ArrayList messages = new ArrayList();
			ByteArrayList bytes = new ByteArrayList(RouteEntriesPerUpdateMessage * 20 + 4);
			//generics: <RoutingTable.Entry>
			ArrayList entries_to_remove = new ArrayList();

			while (true)
			{
				RoutingTable table = this.routingModule.RoutingTable;
				lock (table.SyncRoot)
				{
					foreach (RoutingTable.Entry entry in table)
					{
						TableEntryData private_data = entry.PrivateData as TableEntryData;
						if (entry.PrivateData != null && private_data == null)
							continue; // not RIP nor static entry
						
						if (entry.Network.IPVersion != 4 || entry.Mask.IPVersion != 4 || (entry.Gateway != null && entry.Gateway.IPVersion != 4))
							continue; // we support only IPv4

						if (private_data != null && private_data.Expires < Simulator.Timer.Now) // entry expired, remove it
						{
							entries_to_remove.Add(entry);
							continue;
						}

						if (PassiveMode) // just remove entries in passive mode
							continue;

						if (bytes.Count == 0) // new empty message, add header
						{
							bytes.AddByte(2); // response
							bytes.AddByte(2); // version 2
							bytes.AddUInt16(0); // must be zero
						}
						
						// IP family
						bytes.AddUInt16(2);
						
						// Route tag
						if (private_data != null)
							bytes.AddUInt16(private_data.RouteTag);
						else
							bytes.AddUInt16(0);

						// Network address
						bytes.AddBytes(entry.Network.GetAddressBytes());

						// Network mask
						bytes.AddBytes(entry.Mask.GetAddressBytes());

						// Next hop
						if (entry.Gateway != null)
							bytes.AddBytes(entry.Gateway.GetAddressBytes());
						else
							bytes.AddUInt32(0); // Device is specified instead of gateway, it is directly accessible

						// Distance
						if (private_data != null)
							bytes.AddUInt32(private_data.Distance);
						else
							bytes.AddUInt32(1); // assume all static routes to have distance 1


						if (bytes.Count == bytes.Capacity) // the message is full of entries, remember it and create new one
						{
							byte[] m = new byte[bytes.Count];
							bytes.CopyTo(m, 0);
							messages.Add(m);
							bytes.Clear();
						}
					}

					foreach (RoutingTable.Entry e in entries_to_remove)
						table.Remove(e);

				} // lock

				// send the messages
				foreach (IPAddress addr in neighborRouters)
				{
					IPEndPoint remote_ep = new IPEndPoint(addr.ToSystemIPAddress(), RipPort);
					
					foreach (byte[] message in messages)
						socket.SendTo(message, remote_ep);

					if (bytes.Count > 0)
						socket.SendTo(bytes.Bytes, 0, bytes.Count, SocketFlags.None, remote_ep);
				}

				messages.Clear();
				bytes.Clear();
				entries_to_remove.Clear();
				
				// wait an pre-set delay
				int timeout = RoutingInfoSendInterval;
				updateThreadWait.WaitOne(timeout, false);
			}
		}

		public override void XmlConfigure(System.Xml.XmlDocument document)
		{
			XmlElement root_element = document.DocumentElement;
			XmlNode node = root_element.FirstChild;

			while (node != null)
			{
				switch (node.Name)
				{
					case "Port":
						SetPort(node.InnerText);
						break;

					case "RoutingInfoSendInterval":
						SetRoutingInfoSendInterval(node.InnerText);
						break;

					case "RouteEntryRemovalTimeout":
						SetRouteEntryRemovalTimeout(node.InnerText);
						break;

					case "RouteEntriesPerUpdateMessage":
						SetRouteEntriesPerUpdateMessage(node.InnerText);
						break;

					case "ReceiveBufferLength":
						SetReceiveBufferLength(node.InnerText);
						break;

					case "RoutingModuleName":
						SetRoutingModuleName(node.InnerText);
						break;

					case "NeighborRouters":
						SetNeighborRouters(node);
						break;

					case "Passive":
						SetPassive(node.InnerText);
						break;
				}

				node = node.NextSibling;
			}

			// check mandatory fields
			if (this.routingModuleName == null)
				throw new InitializationException(String.Format(
					"No 'RoutingModuleName' specified in RIP configuration file at node '{0}'", Node.Name));
		}

		private void SetPort(string value)
		{
			try
			{
				this.RipPort = UInt16.Parse(value);
			}
			catch (Exception e)
			{
				throw new InitializationException(String.Format(
					"Error parsing 'Port' value in RIP configuration file at node '{0}'", Node.Name), e);
			}
		}

		private void SetRoutingInfoSendInterval(string value)
		{
			try
			{
				this.RoutingInfoSendInterval = int.Parse(value);
			}
			catch (Exception e)
			{
				throw new InitializationException(String.Format(
					"Error parsing 'RoutingInfoSendInterval' value in RIP configuration file at node '{0}'", Node.Name), e);
			}
		}

		private void SetRouteEntryRemovalTimeout(string value)
		{
			try
			{
				this.RouteEntryRemovalTimeout = int.Parse(value);
			}
			catch (Exception e)
			{
				throw new InitializationException(String.Format(
					"Error parsing 'RouteEntryRemovalTimeout' value in RIP configuration file at node '{0}'", Node.Name), e);
			}
		}
	
		private void SetRouteEntriesPerUpdateMessage(string value)
		{
			try
			{
				this.RouteEntriesPerUpdateMessage = int.Parse(value);
			}
			catch (Exception e)
			{
				throw new InitializationException(String.Format(
					"Error parsing 'RouteEntriesPerUpdateMessage' value in RIP configuration file at node '{0}'", Node.Name), e);
			}
		}
	
		private void SetReceiveBufferLength(string value)
		{
			try
			{
				this.ReceiveBufferLength = int.Parse(value);
			}
			catch (Exception e)
			{
				throw new InitializationException(String.Format(
					"Error parsing 'ReceiveBufferLength' value in RIP configuration file at node '{0}'", Node.Name), e);
			}
		}
	
		private void SetRoutingModuleName(string value)
		{
			this.routingModuleName = value;
		}
	
		private void SetNeighborRouters(XmlNode node)
		{
			XmlNode router_node = node.FirstChild;

			try
			{
				while (router_node != null)
				{
					if (router_node.Name == "Router")
					{
						IPAddress addr = IPAddress.Parse(router_node.InnerText);
						this.neighborRouters.Add(addr);
					}

					router_node = router_node.NextSibling;
				}
			}
			catch (Exception e)
			{
				throw new InitializationException(String.Format(
					"Error parsing 'Router' value in RIP configuration file at node '{0}'", Node.Name), e);
			}
		}

		private void SetPassive(string value)
		{
			try
			{
				this.PassiveMode = bool.Parse(value);
			}
			catch (Exception e)
			{
				throw new InitializationException(String.Format(
					"Error parsing 'Passive' value in RIP configuration file at node '{0}'", Node.Name), e);
			}
		}
	}

}
