/*

 Copyright (c) 2006 Pavel Novak.

 The use and distribution terms for this software are contained in the file named License.txt, 
 which can be found in the root of the NetSim distribution. By using this software 
 in any fashion, you are agreeing to be bound by the terms of this license.
 
 You must not remove this notice from this software.

*/

using System;
using NetSim.Core;
using System.Net;
using System.Net.Sockets;
using System.Threading;
using System.Text;
using System.Collections;
using System.Xml;

namespace NetSim.Library.Applications
{
	/// <summary>
	/// Summary description for UdpTestApp.
	/// </summary>
	public class SimpleWebServer : NodeApplication
	{
		//generics: <string, string>
		private Hashtable pages = new Hashtable();
		
		//generics: <int, string>
		private Hashtable errorPages = new Hashtable();

		public SimpleWebServer()
		{
		}

		protected override int Run()
		{
			NetSimSocket socket = new NetSimSocket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp, this.Node);
			socket.Bind(new System.Net.IPEndPoint(System.Net.IPAddress.Any, 80));
			socket.Listen(20);

			while (true)
			{
				NetSimSocket s = socket.Accept();
				ThreadPool.QueueUserWorkItem(new WaitCallback(this.ProcessRequest), s);
			}

		}

		private void ProcessRequest(object state)
		{
			NetSimSocket socket = (NetSimSocket)state;

			byte[] buffer = new byte[10000];
			StringBuilder request_str = new StringBuilder();

			// get request
			int index = 0;
			int count = 0;
			do
			{
				count = socket.Receive(buffer, index, buffer.Length - index, SocketFlags.None);
				for (int i = index; i < index + count; i++)
					request_str.Append(Convert.ToChar(buffer[i]));

				string r = request_str.ToString();
				if (r.IndexOf("\r\n\r\n") >= 0) // whole request header received
				{
					break;
				}

				index += count;
			}
			while (count > 0);

			bool ok = true;
			string response = null;
			string request = request_str.ToString();
			int first_line_end = request.IndexOf("\r\n");
			if (first_line_end < 0)
			{
				ok = false;
			}
			else
			{
				string first_line = request.ToString().Substring(0, first_line_end);
				int http = first_line.IndexOf(" HTTP/1.1");
				if (http < 4 || first_line.Substring(0, 4) != "GET ")
					ok = false;

				if (ok)
					response = (string)pages[first_line.Substring(4, http - 4)];
			}

			if (response == null)
				response = (string)errorPages[404];

			if (response != null)
			{
				// send response
				char[] chars = response.ToCharArray();
				byte[] bytes = new byte[chars.Length];
				for (int i = 0; i < chars.Length; i++)
					bytes[i] = Convert.ToByte(chars[i]);

				socket.Send(bytes);
			}
			socket.Shutdown(SocketShutdown.Both);

			Thread.Sleep(3);
			socket.Close();
		}

		public override void XmlConfigure(System.Xml.XmlDocument document)
		{
			XmlElement root_el = document.DocumentElement;
			XmlNode node = root_el.FirstChild;

			while (node != null)
			{
				switch (node.Name)
				{
					case "WebPage":
						AddWebPage(node);
						break;
					case "ErrorPage":
						AddErrorPage(node);
						break;
				}

				node = node.NextSibling;
			}
		}

		private void AddWebPage(XmlNode node)
		{
			XmlAttribute pathAttribute = node.Attributes["path"];
			if (pathAttribute == null)
				throw new InitializationException(String.Format("'path' attribute is mandatory at 'WebPage' xml node in SimpleWebServer configuration at node '{0}'", Node.Name));

			try
			{
				pages.Add(pathAttribute.InnerText, node.InnerText);
			}
			catch (ArgumentException)
			{
				throw new InitializationException(String.Format("Page with path '{0}' already exist in SimpleWebServer configuration at node '{1}'", pathAttribute.InnerText, Node.Name));
			}
		}

		private void AddErrorPage(XmlNode node)
		{
			XmlAttribute codeAttribute = node.Attributes["code"];
			if (codeAttribute == null)
				throw new InitializationException(String.Format("'code' attribute is mandatory at 'ErrorPage' xml node in SimpleWebServer configuration at node '{0}'", Node.Name));

			int code;
			try
			{
				code = Int32.Parse(codeAttribute.InnerText);
			}
			catch (Exception)
			{
				throw new InitializationException(String.Format("Wrong format of 'code' attribute at 'ErrorPage' xml node in SimpleWebServer configuration at node '{0}'", Node.Name));
			}

			try
			{
				errorPages.Add(code, node.InnerText);
			}
			catch (ArgumentException)
			{
				throw new InitializationException(String.Format("Error page with code '{0}' already exist in SimpleWebServer configuration at node '{1}'", code, Node.Name));
			}
		}
	}
}
