/*

 Copyright (c) 2006 Pavel Novak.

 The use and distribution terms for this software are contained in the file named License.txt, 
 which can be found in the root of the NetSim distribution. By using this software 
 in any fashion, you are agreeing to be bound by the terms of this license.
 
 You must not remove this notice from this software.

*/

using System;
using System.Collections;
using System.Xml;
using NetSim.Core;
using Tamir.IPLib;

namespace NetSim.Library.Modules
{
	/// <summary>
	/// Summary description for ExternalConnection.
	/// </summary>
	public class ExternalConnection : Module
	{
		private PcapDevice device;
		private PacketIdentifier sentPackets;

		public ExternalConnection()
		{
			this.sentPackets = new PacketIdentifier(0);

			// register our methods with simulator events
			Simulator.GetInstance().StartEvent += new Simulator.StartEventDelegate(this.SimulatorStartEventHandler);
			Simulator.GetInstance().StopEvent += new Simulator.StopEventDelegate(this.SimulatorStopEventHandler);
		}

		~ExternalConnection()
		{
			// unregister methods
			Simulator.GetInstance().StartEvent -= new Simulator.StartEventDelegate(this.SimulatorStartEventHandler);
			Simulator.GetInstance().StopEvent -= new Simulator.StopEventDelegate(this.SimulatorStopEventHandler);
		}

		public override void XmlConfigure(System.Xml.XmlDocument document)
		{
			XmlNodeList devices_list = document.DocumentElement.GetElementsByTagName("Device");
			if (devices_list.Count == 0 || devices_list[0].InnerText == String.Empty)
				throw new InitializationException("Device not specified");

			XmlNode device_xml = devices_list[0];

			// Retrieve the device list
			PcapDeviceList devices = SharpPcap.GetAllDevices();

			// Collect available devices to string for error message
			System.Text.StringBuilder devices_names = new System.Text.StringBuilder();
			// try to find our device
			PcapDevice found_device = null;
			foreach (PcapDevice device in devices)
			{
				// for searching by device description if available
				NetworkDevice ndev = device as NetworkDevice;

				if (devices_names.Length > 0)
					devices_names.Append("; ");
				devices_names.Append(device.PcapName);
				devices_names.Append(", ");
				devices_names.Append(device.PcapDescription);
				devices_names.Append(", ");
				if (device is NetworkDevice) devices_names.Append(((NetworkDevice)device).Description);

				if ((device.PcapDescription != null && device.PcapDescription.IndexOf(device_xml.InnerText) >= 0)
					|| (device.PcapName != null && device.PcapName.IndexOf(device_xml.InnerText) >= 0)
					|| (ndev != null && ndev.Description.IndexOf(device_xml.InnerText) >= 0))
				{
					if (found_device == null)
						found_device = device;
					else
						throw new InitializationException(String.Format("Two network device names match the string '{0}': '{1}' and '{2}'",
							device_xml.InnerText,
							String.Concat(found_device.PcapName, ", ", found_device.PcapDescription, found_device is NetworkDevice ? ", " + ((NetworkDevice)found_device).Description : ""),
							String.Concat(device.PcapName, ", ", device.PcapDescription, device is NetworkDevice ? ", " + ((NetworkDevice)device).Description : "")
							));
				}
			}

			if (found_device == null)
			{
				if (devices_names.Length == 0)
					throw new InitializationException("No network device for capturing packets found. Do you have sufficient rights to capture packets?");
				else
					throw new InitializationException(String.Format("No device matching string '{0}' found. Available devices are: '{1}'",
						device_xml.InnerText, devices_names.ToString()));
			}

			this.device = found_device;

			//Register our handler function to the 'packet arrival' event
			this.device.PcapOnPacketArrival += 
				new SharpPcap.PacketArrivalEvent(this.PcapOnPacketArrival);

			// open device
			this.device.PcapOpen(true, 1); // smallest timeout possible to process the packet immediately
		}

		private void SimulatorStartEventHandler()
		{
			// start capture
			if (device != null)
				device.PcapStartCapture();
		}

		private void SimulatorStopEventHandler()
		{
			// stop capture
			if (device != null)
				device.PcapStopCapture();
		}

		public override void ReceivePacket(ref PacketInfo pInfo)
		{
			// send the packet to the real network..
			ByteArrayList data = new ByteArrayList(pInfo.RawPacket.GetBytesCount());
			pInfo.RawPacket.AppendBytes(data);

			// remember the packet and do not send it back to virtual network if it is captured.
			sentPackets.Insert(data.Bytes);

			device.PcapSendPacket(data.Bytes);
		}

		protected void PcapOnPacketArrival(object sender, byte[] packetBytes)
		{
			// do not send packet to the virtual network that has just been sent to the real network
			if (sentPackets.CheckAndRemove(packetBytes))
				return; // our packet was captured

			Packet p = Simulator.GetInstance().CreatePacket(Packet.PacketGroup.LinkLayer, this.device.PcapDataLink, packetBytes);
			if (p != null) // packet has been recognised and created
			{
				PacketInfo pi = new PacketInfo(p);
				pi.RawPacket = p;

				this.SendNewPacket(pi);
			}
		}
	}

	/// <summary>
	/// Identifies the packets and can compare the other packet if it has been added here or not.
	/// </summary>
	public class PacketIdentifier
	{
		/// <summary>
		/// An entry for each packet.
		/// </summary>
		private class Entry
		{
			public byte[] Bytes = new byte[20];
			public long Expires;
			UInt16 Checksum;

			public Entry(byte[] bytes, long expires, UInt16 checksum)
			{
				int count = Math.Min(bytes.Length, this.Bytes.Length);
				Buffer.BlockCopy(bytes, 0, this.Bytes, 0, count);

				this.Expires = expires;
				this.Checksum = checksum;
			}

			public bool Compare(byte[] data)
			{
				int count = Math.Min(data.Length, this.Bytes.Length);
				for (int i = 0; i < count; i++)
					if (data[i] != this.Bytes[i])
						return false;

				return true;
			}

			public override bool Equals(object obj)
			{
				Entry e = obj as Entry;
				if (e == null)
					return false;

				if (this.Checksum != e.Checksum)
					return false;

				for (int i = 0; i < Bytes.Length; i++)
					if (e.Bytes[i] != this.Bytes[i])
						return false;

				return true;
			}

			public override int GetHashCode()
			{
				UInt16 check2 = ByteArrayList.OneComplementChecksum2(this.Bytes, 0, this.Bytes.Length, 0);
				return (int) ( (((UInt32)check2) << 16) | this.Checksum);
			}

		}

		/// <summary>
		/// Remembered entries.
		/// </summary>
		//generics: <UInt16, Entry>
		private Hashtable entries;
		
		/// <summary>
		/// Timeout in ticks after that an entry becomes invalid.
		/// </summary>
		private long timeout;

		public PacketIdentifier(long timeout)
		{
			this.timeout = timeout;
			this.entries = Hashtable.Synchronized(new Hashtable());
		}

		public void Insert(byte[] data)
		{
			UInt16 check = ByteArrayList.OneComplementChecksum2(data, 0, data.Length, 0);
			this.entries[check] = new Entry(data, 0, check);
		}

		/// <summary>
		/// Returns <b>true</b> if the packet is present and is not expired. The packet is removed.
		/// </summary>
		/// <param name="data"></param>
		/// <returns></returns>
		public bool CheckAndRemove(byte[] data)
		{
			bool found = false;

			UInt16 check = ByteArrayList.OneComplementChecksum2(data, 0, data.Length, 0);
			Entry e = (Entry)entries[check];

			if (e != null)
			{
				if (e.Compare(data))
					found = true;

				// no need to sync here, no exception is thrown if the key doesn't exist
				entries.Remove(check);
			}

			return found;
		}
	}

}
