/*

 Copyright (c) 2006 Pavel Novak.

 The use and distribution terms for this software are contained in the file named License.txt, 
 which can be found in the root of the NetSim distribution. By using this software 
 in any fashion, you are agreeing to be bound by the terms of this license.
 
 You must not remove this notice from this software.

*/

using System;
using NetSim.Core;
using System.Collections;
using System.Diagnostics;
using System.Xml;

namespace NetSim.Library.Modules
{
	/// <summary>
	/// This interface should implement all modules capable of IP routing. They can be than used
	/// by other modules or applications that need read or can modify routing iformation.
	/// </summary>
	public interface IIPRoutingModule
	{
		RoutingTable RoutingTable { get; }
	}

	/// <summary>
	/// This module handles the basics of IP functionality (currently only IPv4). It means packet fragmentation
	/// and reassembly, extracting higher layer packets.
	/// </summary>
	public class IPModule : Module, IIPRoutingModule
	{
		private UInt16 nextPacketIdentification = 1;
		private object _nextPacketIdentificationLock = new object();
		protected UInt16 NextPacketIdentification
		{
			get
			{
				UInt16 result;
				lock (_nextPacketIdentificationLock)
				{
					result = nextPacketIdentification;
					unchecked { nextPacketIdentification++; } // allow overflow back to zero..
				}

				return result;
			}
		}

		private byte newPacketTimeToLive = 128;
		protected byte NewPacketTimeToLive
		{
			get { return this.newPacketTimeToLive; }
			set { this.newPacketTimeToLive = value; }
		}

		/// <summary>
		/// Hashtable containing <see cref="IPv4Fragments"/> objects, keys are "identification" numbers specified
		/// in the IP packets.
		/// </summary>
		private Hashtable fragments = new Hashtable();

		public RoutingTable RoutingTable { get { return this.routingTable; } }
		protected RoutingTable routingTable;

		/// <summary>
		/// The name of the module that is used for sending ICMP messages generated by the router. Can be <b>null</b>.
		/// </summary>
		private string icmpModuleName;

		public IPModule()
		{
			routingTable = new RoutingTable();
		}

		public override void ReceivePacket(ref PacketInfo pInfo)
		{
			IPPacket ip_p = pInfo.Packet as IPPacket;
			if (ip_p == null) // this is not an IP packet...
				return;
			
			NetworkAddress local_address;
			// if the destination is not this node, routing is needed
			if (!Node.ShouldReceive(ip_p.DestinationIP, out local_address))
			{
				// the packet should be sent further
				PacketInfo new_pi = new PacketInfo(ip_p);
				new_pi.RemoteAddress = ip_p.DestinationIP;
				ReRoutePacket(new_pi);
				pInfo = null;
				return;
			}

			if (ip_p.HigherLayerPacket is UnknownPacket)
			{
				// The packet was not parsed yet - fragmentation possible..
				switch (ip_p.Version)
				{
					case 4:
						ReceiveIPv4Fragment(ref pInfo);
						break;
					default:
						// not supported version - do nothing with the packet, maybe the next module will handle it..
						return;
				}
			}

			// if the packet was fragment, it is possible that pInfo is null..
			if (pInfo != null)
			{
				// the higher layer packet has been parsed, pass it further..
				pInfo.Packet = pInfo.Packet.HigherLayerPacket;
				pInfo.RemoteAddress = ip_p.SourceIP;
				pInfo.ThisMachine = local_address;
			}
		}

		protected void ReceiveIPv4Fragment(ref PacketInfo pInfo)
		{
			this.CombineIPv4Fragments(ref pInfo);
			if (pInfo != null) // packet combined
			{
			}
		}

		protected void CombineIPv4Fragments(ref PacketInfo pInfo)
		{
			Debug.Assert(pInfo.Packet is IPv4Packet && pInfo.Packet.HigherLayerPacket is UnknownPacket);

			IPv4Packet ipv4_p = pInfo.Packet as IPv4Packet;
			UnknownPacket unknown_p = ipv4_p.HigherLayerPacket as UnknownPacket;

			// is the packet fragment?
			if (ipv4_p.FragmentOffset > 0 || ((ipv4_p.FlagsField & IPv4Packet.Flags.MoreFragments) != 0))
			{
				IPv4Fragments frag = (IPv4Fragments) this.fragments[ipv4_p.Identification];
				if (frag == null)
				{
					frag = new IPv4Fragments();
					this.fragments.Add(ipv4_p.Identification, frag);
				}
				
				if (frag.Add(ipv4_p))
				{
					// all fragments have arrived
					IPv4Packet combined = frag.GetCombinedPacket();
					pInfo.Packet = combined;
					// delete fragments
					this.fragments.Remove(ipv4_p.Identification);
				}
				else
				{
					// wait for more fragments..
					pInfo = null;
				}
			}


		}

		public override void SendPacket(ref PacketInfo pInfo)
		{
			// check loopback
			if (pInfo.RemoteAddress.Loopback != null && pInfo.RemoteAddress.Loopback.Equals(pInfo.RemoteAddress))
			{
				Node.NewPacketToReceive(this.ID, pInfo);
				pInfo = null;
				return;
			}

			// if the packet is already network layer or lower, do not encapsulate in IP
			if (pInfo.Packet.Group == Packet.PacketGroup.NetworkLayer || pInfo.Packet.Group == Packet.PacketGroup.LinkLayer
				|| !(pInfo.RemoteAddress is IPAddress))
			{
				return;
			}

			ReRoutePacket(pInfo);
			
			// new packets are sent from the previous function, do not allow to send further twice
			// and possibly not fragmented etc..
			pInfo = null;
		}

		private void SendFragmentsIPv4(byte headerLen, byte serviceType, UInt16 identification, IPv4Packet.Flags flags,
			UInt16 fragmentOffset, byte ttl, byte protocol, IPv4Address sourceIP, IPv4Address destinationIP,
			Packet higherLayerPacket, IAdapter adapter, NetworkAddress nextHopAddress, NetworkAddress thisMachineAddress)
		{
			int upper_layer_length = higherLayerPacket.GetBytesCount();

			// clear "More Fragments" bit
			flags &= ~IPv4Packet.Flags.MoreFragments;

			int max_data_len = adapter.MTU - headerLen * 4;
			if (upper_layer_length <= max_data_len) // will fit in the single packet
			{
				IPv4Packet ipv4_p = new IPv4Packet(headerLen, serviceType, (UInt16)(headerLen * 4 + upper_layer_length),
					identification, flags, fragmentOffset, ttl, protocol, sourceIP, destinationIP);
				ipv4_p.HigherLayerPacket = higherLayerPacket;

				SendOneFragmentIPv4(ipv4_p, adapter, nextHopAddress, thisMachineAddress);
				return;
			}

			// fragmentation needed

			// data in the fragment should be 8 bytes aligned
			max_data_len =(max_data_len / 8) * 8;

			// make an array of bytes from the packet
			ByteArrayList whole_packet = new ByteArrayList(upper_layer_length);
			higherLayerPacket.AppendBytes(whole_packet);

			int offset = 0;
			while (offset < upper_layer_length)
			{
				// create upper layer packet
				int current_len = Math.Min(upper_layer_length - offset, max_data_len);
				UnknownPacket unk_p = UnknownPacket.CreatePacket(whole_packet.Bytes, offset, current_len);

				// create IP packet
				bool last_fragment = offset + current_len == upper_layer_length;
				UInt16 total_len = (UInt16)(headerLen * 4 + current_len);
				IPv4Packet.Flags flag = last_fragment ? flags : flags | IPv4Packet.Flags.MoreFragments;
				UInt16 frag_offset = (UInt16) (fragmentOffset + offset / 8);

				IPv4Packet ipv4_p = new IPv4Packet(headerLen, serviceType, total_len, identification, flag, frag_offset,
					ttl, protocol, sourceIP, destinationIP);
				ipv4_p.HigherLayerPacket = unk_p;

				// send
				SendOneFragmentIPv4(ipv4_p, adapter, nextHopAddress, thisMachineAddress);

				offset += current_len;
			}

		}

		private void SendOneFragmentIPv4(IPv4Packet packet, IAdapter adapter, NetworkAddress nextHopAddress,
			NetworkAddress thisMachineAddress)
		{
			PacketInfo pi = new PacketInfo(packet);
			pi.UseAdapter = adapter;
			pi.RemoteAddress = packet.DestinationIP;
			pi.NextHopDestination = nextHopAddress;
			pi.ThisMachine = thisMachineAddress;

			this.SendNewPacket(pi);
		}

		public override void XmlConfigure(System.Xml.XmlDocument document)
		{
			// find 'Routes' node
			XmlNode routesNode = document.DocumentElement.FirstChild;
			while (routesNode != null && routesNode.Name != "Routes")
				routesNode = routesNode.NextSibling;
			if (routesNode == null)
				throw new InitializationException("No 'Routes' node found");

			// go through all 'Route' nodes
			XmlNode routeNode = routesNode.FirstChild;
			while (routeNode != null)
			{
				if (routeNode.Name == "Route")
					this.XmlConfigureRoute(routeNode);
				routeNode = routeNode.NextSibling;
			}

			// find 'IcmpModule' node
			XmlNode icmpNode = document.DocumentElement.FirstChild;
			while (icmpNode != null && icmpNode.Name != "IcmpModule")
				icmpNode = icmpNode.NextSibling;
			if (icmpNode == null || icmpNode.InnerText == String.Empty)
				Simulator.GetInstance().Events.Warning(String.Format("No 'IcmpModule' for the router at node {0} specified - the router cannot send ICMP messages.", this.Node.Name));
			else
				this.icmpModuleName = icmpNode.InnerText;
		}

		private void XmlConfigureRoute(XmlNode routeNode)
		{
			RoutingTable.Entry new_entry = new RoutingTable.Entry();

			XmlNode child = routeNode.FirstChild;
			while (child != null)
			{
				switch (child.Name)
				{
					case "Network":
						new_entry.Network = IPAddress.Parse(child.InnerText);
						break;
					case "Mask":
						new_entry.Mask = IPAddress.Parse(child.InnerText);
						break;
					case "Gateway":
						new_entry.Gateway = IPAddress.Parse(child.InnerText);
						break;
					case "Device":
						new_entry.Device = child.InnerText;
						break;
				}

				child = child.NextSibling;
			}
			
			// this is static routing table, no distances measured..
			//new_entry.Distance = 1;

			if (new_entry.Network == null || new_entry.Mask == null
				|| (new_entry.Gateway == null && new_entry.Device == null))
				throw new InitializationException("Not all mandatory values specified for the route");

			this.routingTable.Add(new_entry);
		}

		/// <summary>
		/// Sends the packet that is either incoming but not for this node, or outgoing. In pInfo should
		/// be only fields "Packet" and "RemoteAddress" set, packet can be either IP packet or something
		/// else. In that case it is upper layer packet and it will be wrapped into IP packet and possibly
		/// fragmented.
		/// </summary>
		/// <remarks>Currently only IPv4 is supported.</remarks>
		/// <param name="pInfo"></param>
		private void ReRoutePacket(PacketInfo pInfo)
		{
			Debug.Assert(pInfo.RemoteAddress is IPAddress);

			switch (((IPAddress)pInfo.RemoteAddress).IPVersion)
			{
				case 4:
					ReRoutePacketIPv4(pInfo);
					break;

				default:
					// not supported version, cannot route
					return;
			}
		}

		/// <summary>
		/// Takes the packet - either incoming with destination not directly reacheable, or outgoing,
		/// also with destination not reacheable by networks directly attached to this node.
		/// Searches routing tables and if found, the packet is sent to the apropriate gateway, or
		/// ICMP message is generated and sent to the source.
		/// </summary>
		/// <param name="pInfo"><see cref="PacketInfo"/> containing the packet to redirect.</param>
		private void ReRoutePacketIPv4(PacketInfo pInfo)
		{
			IPv4Packet ipv4_p = pInfo.Packet as IPv4Packet; // can be null - in that case it is outgoing packet and should be wrapped

			// try to find out where should be the packet sent
			InitRoutingPacketInfo(pInfo);

			if (pInfo.UseAdapter == null) // no info where to send
			{
				if (ipv4_p != null) // incoming packet - sender should be notified
					HandleUnreacheableDestination(ipv4_p);

				return;
			}

			if (ipv4_p != null)
			{
				if (ipv4_p.TimeToLive == 1) // time to live of incoming packet exceeded
				{
					HandleTtlExceeded(ipv4_p);
					return;
				}
				ipv4_p.TimeToLive--;
			}

			if (pInfo.Socket != null) // if this is the first I/O for the socket bound to ANY, IPAddress should be set
			{
				IPAddress loc_ip = (IPAddress)pInfo.ThisMachine;
				pInfo.Socket.SetLocalEndPoint(loc_ip.ToSystemIPAddress(), 0);
			}

			if (ipv4_p != null && pInfo.UseAdapter.MTU >= ipv4_p.TotalLength)
			{
				// can be sent immediately, IP packet is ready and doesn't exceed adatpers MTU
				this.SendNewPacket(pInfo);
				return;
			}

			// send - will be wrapped into IP packet and fragmented if needed
			SendWrappedAndFragmentedIPv4(pInfo, ipv4_p);
		}

		/// <summary>
		/// 
		/// </summary>
		/// <param name="pInfo"></param>
		/// <param name="ipv4Packet">The packet containted in <paramref name="pInfo" /> if it is already an
		/// <see cref="IPv4Packet"/>, or <b>null</b> if the pInfo.Packet should be wrapped into IP packet.</param>
		private void SendWrappedAndFragmentedIPv4(PacketInfo pInfo, IPv4Packet ipv4Packet)
		{
			if (ipv4Packet != null)
			{
				SendFragmentsIPv4(ipv4Packet.HeaderLen, ipv4Packet.ServiceType, ipv4Packet.Identification, ipv4Packet.FlagsField,
					ipv4Packet.FragmentOffset, ipv4Packet.TimeToLive, ipv4Packet.Protocol,
					ipv4Packet.SourceIPv4, ipv4Packet.DestinationIPv4, ipv4Packet.HigherLayerPacket, pInfo.UseAdapter,
					pInfo.NextHopDestination, pInfo.ThisMachine);
			}
			else
			{
				byte protocol;
				try {	protocol = (byte)pInfo.Packet.Type; }
				catch (OverflowException)
				{
					Simulator.GetInstance().Events.Warning(String.Format("Packet type {0} is too large to fit into IPv4 header.",
						pInfo.Packet.Type));
					return;
				}

				SendFragmentsIPv4(5, 0, this.NextPacketIdentification, IPv4Packet.Flags.None, 0, this.NewPacketTimeToLive,
					(byte)pInfo.Packet.Type, (IPv4Address)pInfo.ThisMachine, (IPv4Address)pInfo.RemoteAddress, pInfo.Packet,
					pInfo.UseAdapter, pInfo.NextHopDestination, pInfo.ThisMachine);
			}
		}


		private void InitRoutingPacketInfo(PacketInfo pInfo)
		{
			// try networks directly connected
			Node.InitPacketInfo(pInfo);
			if (pInfo.NextHopDestination != null) // the destination was found
				return;

			// and finally routes from the routing table
			RoutingTable.Entry e = routingTable.Get((IPAddress)pInfo.RemoteAddress);
			if (e != null)
			{
				if (e.Gateway != null)
				{
					// remember original remote address
					NetworkAddress ultimate_destination = pInfo.RemoteAddress;
					// set remote address to gateway
					pInfo.RemoteAddress = e.Gateway;
					// should init pi.UseAdapter according to gateway
					Node.InitPacketInfo(pInfo);
					// set back the original destination
					pInfo.RemoteAddress = ultimate_destination;
				}
				else
				{
					pInfo.UseAdapter = Node.GetAdapter(e.Device); // should get the adapter if the name is correct
				}

				if (pInfo.UseAdapter != null) // initialized
					return;

				// if we are here the entry is wrong
				Simulator.GetInstance().Events.Warning(String.Format("Wrong entry in the StaticRouter routing table: '{0}'", e));
			}
		}

		private void HandleUnreacheableDestination(IPv4Packet packet)
		{
			if (icmpModuleName == null) // no ICMP module - cannot send notification
				return;

			IcmpModule err_handler = Node.GetModule(icmpModuleName) as IcmpModule;
			if (err_handler == null)
				return;

			err_handler.UnreacheableDestination(packet, IcmpUnreacheablePacket.Code.NetUnreacheable);
		}

		private void HandleTtlExceeded(IPv4Packet packet)
		{
			if (icmpModuleName == null) // no ICMP module - cannot send notification
				return;

			IcmpModule err_handler = Node.GetModule(icmpModuleName) as IcmpModule;
			if (err_handler == null)
				return;

			err_handler.TimeExceeded(packet, IcmpTimeExceededPacket.Code.TtlCountExceeded);
		}

	}

	internal class IPv4Fragments
	{
		/// <summary>
		/// One packet entry int the list.
		/// </summary>
		private class Fragment
		{
			/// <summary>
			/// Offset of data carried in this fragment measured in bytes (!).
			/// </summary>
			public UInt16 Offset;

			/// <summary>
			/// Length of data carried by this fragment.
			/// </summary>
			public UInt16 Length;

			/// <summary>
			/// <b>True</b> if more fragments are needed, <b>false</b> otherwise.
			/// </summary>
			public bool MoreFragments;

			public IPv4Packet FragmentPacket;

			/// <summary>
			/// Next fragment in the chain.
			/// </summary>
			public Fragment Next;

			public Fragment(IPv4Packet packet)
			{
				// this won't overflow because fragment offset field in IPv4 header is 13 bits long - exactly
				// 3 bits remain to UInt16
				this.Offset = (UInt16)(packet.FragmentOffset * 8);

				this.Length = (UInt16)(packet.TotalLength - 4 * packet.HeaderLen);
				this.MoreFragments = (packet.FlagsField & IPv4Packet.Flags.MoreFragments) != 0;

				this.FragmentPacket = packet;
			}
		}

		/// <summary>
		/// First <see cref="Fragment"/> in the chain. The chain is ordered by the <see cref="Fragment.Offset"/>.
		/// </summary>
		private Fragment firstFragment;

		/// <summary>
		/// <b>True</b> if all fragments have been added and <see cref="GetCombinedPacket"/> method will
		/// return the original packet.
		/// </summary>
		public bool HasAllFragments
		{
			get { return hasAllFragments; }
		}
		private bool hasAllFragments = false;

		/// <summary>
		/// Use only if <see cref="hasAllFragments"/> is set to <b>true</b>. Contains the total length of all data
		/// from the fragments.
		/// </summary>
		private int originalDataLength;

		public IPv4Fragments()
		{
		}

		/// <summary>
		/// Inserts new fragment to the list. Returns <b>true</b> if this was the last fragment
		/// and the original packet can be combined, otherwise <b>false</b>.
		/// </summary>
		/// <param name="fragmentPacket">Packet fragment to insert.</param>
		/// <returns><b>True</b> if the original packet can be combined.</returns>
		public bool Add(IPv4Packet fragmentPacket)
		{
			// assert this packet is fragment
			Debug.Assert(fragmentPacket.FragmentOffset > 0 || ((fragmentPacket.FlagsField & IPv4Packet.Flags.MoreFragments) != 0));

			if (this.hasAllFragments) // we have already all fragments, do nothing
				return true;

			// create new fragment, after the insertion to the chain this value is set to null
			Fragment new_fragment = new Fragment(fragmentPacket);

			Fragment prev = null;
			Fragment f = this.firstFragment;
			bool complete = true; // assume there are all packets
			while (f != null)
			{
				if (new_fragment != null)
				{
					if (f.Offset == new_fragment.Offset) // the same packet has arrived twice, no change
						return this.hasAllFragments;

					if (f.Offset > new_fragment.Offset) // fragment should be inserted here
					{
						new_fragment.Next = f;
						if (prev == null) this.firstFragment = new_fragment; else prev.Next = new_fragment;
						f = new_fragment; // new fragment is now the next in the chain
						new_fragment = null; // prevent from insertion twice
					}
				}

				// check the completeness of fragments in the chain
				if (prev == null) // this should be the first fragment
				{
					if (f.Offset != 0)
						complete = false;
				}
				else
				{
					if (prev.Offset + prev.Length < f.Offset)
						complete = false;
				}

				if (!complete && new_fragment == null) // some fragment is missing and new fragment was inserted - no sense to continue
					return false;

				prev = f;
				f = f.Next;
			}


			if (new_fragment != null) // fragment not inserted yet, insert at the end of the chain
			{
				if (prev == null) this.firstFragment = new_fragment; else prev.Next = new_fragment;
				if (prev == null) // this should be the first fragment
				{
					if (new_fragment.Offset != 0)
						complete = false;
				}
				else
				{
					if (prev.Offset + prev.Length < new_fragment.Offset)
						complete = false;
				}
				prev = new_fragment;
			}

			if (prev.MoreFragments) // more fragments needed
				complete = false;

			if (complete)
			{
				this.originalDataLength = prev.Offset + prev.Length; // prev is the last fragment in the chain
				this.hasAllFragments = true;
			}

			return this.hasAllFragments;
		}

		/// <summary>
		/// Assemblies the original packet from all fragments.
		/// </summary>
		/// <returns>The original packet or <b>null</b> if not all fragments are available.</returns>
		public IPv4Packet GetCombinedPacket()
		{
			if (!hasAllFragments)
				return null;

			ByteArrayList data = new ByteArrayList(originalDataLength);
			Fragment f = firstFragment;
			while (f != null)
			{
				f.FragmentPacket.HigherLayerPacket.AppendBytes(data);
				f = f.Next;
			}

			Debug.Assert(originalDataLength == data.Count);

			IPv4Packet first = firstFragment.FragmentPacket;
			IPv4Packet combined_p = new IPv4Packet(5, first.ServiceType, (UInt16)(5*4 + originalDataLength), first.Identification,
				IPv4Packet.Flags.None, 0, first.TimeToLive, first.Protocol, first.SourceIPv4, first.DestinationIPv4);
			combined_p.HigherLayerPacket = Simulator.GetInstance().CreatePacket(Packet.PacketGroup.TransportLayer, first.Protocol, data.Bytes);

			return combined_p;
		}
	}

	public class RoutingTable : IEnumerable
	{
		/// <summary>
		/// Routing table entry.
		/// </summary>
		public class Entry
		{
			/// <summary>
			/// Specifies network address of the target network. (Together with <see cref="Mask"/>.)
			/// </summary>
			public IPAddress Network;

			/// <summary>
			/// Mask for the network address.
			/// </summary>
			public IPAddress Mask;

			/// <summary>
			/// IP address of the next router in the route (which attaches to the network that is directly
			/// accessible by this node). Can be <b>null</b> if <see cref="Device"/> is specified.
			/// </summary>
			public IPAddress Gateway;

			/// <summary>
			/// The name of the device (<see cref="Adapter"/>) the packet should be sent over. Can be <b>null</b>
			/// if a <see cref="Gateway"/> is specified.
			/// </summary>
			public string Device;

			/// <summary>
			/// The field used by applications exchanging routing information to store their specific data.
			/// </summary>
			public object PrivateData;

			public Entry()
			{}

			public Entry(IPAddress network, IPAddress mask, IPAddress gateway, string device)
			{
				this.Network = network;
				this.Mask = mask;
				this.Gateway = gateway;
				this.Device = device;
			}
		}

		/// <summary>
		/// List of <see cref="Entry"/> objects sorted by the first occurency of non-zero bit in the mask.
		/// </summary>
		private SortedList2 routes;

		public object SyncRoot { get { return routes.SyncRoot; } }
		public RoutingTable()
		{
			routes = SortedList2.Synchronized(new SortedList2());
		}

		/// <summary>
		/// Inserts new entry to the table.
		/// </summary>
		/// <param name="e">New entry to add.</param>
		public void Add(Entry e)
		{
			routes.Add(e.Mask, e);
		}

		public void Remove(Entry e)
		{
			lock (SyncRoot)
			{
				int index = routes.IndexOfValue(e);
				if (index >= 0)
					routes.RemoveAt(index);
			}
		}

		/// <summary>
		/// Returns routing table <see cref="Entry"/> that describes the destination for the specified
		/// <paramref name="addr"/>. If there are more entries in the table that match the address,
		/// one of them is returned - there is no guarantee which one.
		/// </summary>
		/// <param name="addr">An addres to search for.</param>
		/// <returns>Found entry or <b>null</b> if no match is found.</returns>
		public Entry Get(IPAddress addr)
		{
			lock (SyncRoot)
			{
				// thanks to sorted list the values need to be checked only sequentially, but from the end of the list,
				// there are "host routes", "default routes" are at the beginning (see IPAddress.CompareTo() method).
				for (int i = routes.Count-1; i >= 0; i--)
				{
					Entry e = (Entry)routes.GetByIndex(i);
					if (IPAddress.IsInNetwork(addr, e.Network, e.Mask))
						return e;
				}
				return null;
			}
		}

		/// <summary>
		/// Returns routing table <see cref="Entry"/> that exactly matches criteria specified.
		/// </summary>
		/// <param name="network"></param>
		/// <param name="mask"></param>
		/// <returns></returns>
		public Entry Get(IPAddress network, IPAddress mask)
		{
			lock (SyncRoot)
			{
				for (int i = routes.Count-1; i >= 0; i--)
				{
					Entry e = (Entry)routes.GetByIndex(i);
					if (e.Mask.Equals(mask) && IPAddress.IsInNetwork(network, e.Network, e.Mask))
						return e;
				}
				return null;
			}
		}

		#region IEnumerable Members

		public class Enumerator : IEnumerator
		{
			private IEnumerator innerEnumerator;
			public Enumerator(IEnumerator innerEnumerator)
			{
				this.innerEnumerator = innerEnumerator;
			}

			public void Reset()
			{
				innerEnumerator.Reset();
			}

			public object Current
			{
				get
				{
					return innerEnumerator.Current;
				}
			}

			public bool MoveNext()
			{
				return innerEnumerator.MoveNext();
			}
		}

		public IEnumerator GetEnumerator()
		{
			return new Enumerator(((IEnumerable)routes).GetEnumerator());
		}

		#endregion
	}

}
