/*

 Copyright (c) 2006 Pavel Novak.

 The use and distribution terms for this software are contained in the file named License.txt, 
 which can be found in the root of the NetSim distribution. By using this software 
 in any fashion, you are agreeing to be bound by the terms of this license.
 
 You must not remove this notice from this software.

*/

using System;
using System.Net;
using NetSim.Core;
using System.Collections;
using SocketFlags = System.Net.Sockets.SocketFlags;
using ProtocolType = System.Net.Sockets.ProtocolType;
using SocketType = System.Net.Sockets.SocketType;
using AddressFamily = System.Net.Sockets.AddressFamily;
using System.Diagnostics;
using System.Threading;

namespace NetSim.Library.Modules
{
	/// <summary>
	/// Summary description for SocketModule.
	/// </summary>
	public class SocketModule : Module, ISocketModule
	{
		/// <summary>The minimum port number that can be assigned if no port number is specified.</summary>
		/// <remarks>See http://www.iana.org/assignments/port-numbers</remarks>
		protected const UInt16 minDynamicPortNumber = 49152;
		/// <summary>The maximum port number that can be assigned if no port number is specified.</summary>
		/// <remarks>See http://www.iana.org/assignments/port-numbers</remarks>
		protected const UInt16 maxDynamicPortNumber = 65535;

		/// <summary>
		/// The minimum port number that can be assigned if no port number is specified and
		/// <see cref="NetSimSocket.Bind"/> is called. See <see cref="System.Net.Sockets.Socket.Bind"/> documentation.
		/// </summary>
		protected const UInt16 minBindPortNumber = 1024;
		/// <summary>
		/// The maximum port number that can be assigned if no port number is specified and
		/// <see cref="NetSimSocket.Bind"/> is called. See <see cref="System.Net.Sockets.Socket.Bind"/> documentation.
		/// </summary>
		protected const UInt16 maxBindPortNumber = 5000;

		/// <summary>
		/// Returns a number that can be used as a new port number.
		/// </summary>
		/// <param name="protocol">Protocol type for which the port number is requested. (Only TCP and UDP are supported.)</param>
		/// <param name="bind">Whether this number is requested as the result of <see cref="NetSimSocket.Bind"/> invocation.</param>
		/// <returns>New port number.</returns>
		protected UInt16 GetUnassignedPortNumber(System.Net.Sockets.ProtocolType protocol, bool bind)
		{
			SocketCollection table;
			if (protocol == System.Net.Sockets.ProtocolType.Udp)
				table = this.udpSockets;
			else if (protocol == System.Net.Sockets.ProtocolType.Tcp)
				table = this.tcpSockets;
			else
				throw new ArgumentException("Only TCP and UDP protocols are supported", "protocol");

			UInt16 min = bind ? minBindPortNumber : minDynamicPortNumber;
			UInt16 max = bind ? maxBindPortNumber : maxDynamicPortNumber;

			while (true)
			{
				UInt16 result = (UInt16)Simulator.Random.Next(min, max);
				if (!table.ContainsLocalPort(result, null))
				{
					return result;
				}
			}
		}

		/// <summary>
		/// The name of this module specified in the configuration file should be always this name.
		/// It is used for retrieving reference to the module while creating a socket (for example from
		/// the remote computer).
		/// </summary>
		/// <remarks>Should be equal to <see cref="NetSimSocket.StdSocketModuleName"/>.</remarks>
		public static readonly string StdModuleName = "SocketModule";

		/// <summary>
		/// Mapping of TCP endpoints to sockets.
		/// </summary>
		private SocketCollection tcpSockets;

		/// <summary>
		/// Mapping of UDP endpoints to sockets.
		/// </summary>
		private SocketCollection udpSockets;

		public SocketModule()
		{
			tcpSockets = new SocketCollection();
			udpSockets = new SocketCollection();
		}

		/// <summary>
		/// Returns the socket associated with specified connection.
		/// </summary>
		/// <param name="protocol"></param>
		/// <param name="remoteAddress"></param>
		/// <param name="remotePort"></param>
		/// <param name="localAddress"></param>
		/// <param name="localPort"></param>
		/// <returns></returns>
		public NetSimSocket FindSocket(ProtocolType protocol, IPAddress remoteAddress, UInt16 remotePort,
			IPAddress localAddress, UInt16 localPort)
		{
			NetSimSocket sock = null;
			switch (protocol)
			{
				case ProtocolType.Tcp:
					sock = tcpSockets.Get(remoteAddress, remotePort, localAddress, localPort);
					break;
				case ProtocolType.Udp: // UDP can receive packet from arbitrary remote host
					sock = udpSockets.Get(localAddress, localPort);
					break;
			}

			return sock;
		}

		/// <summary>
		/// Processes <see cref="SocketDataPacket"/> sent by <see cref="TcpModule"/> or <see cref="UdpModule"/>.
		/// </summary>
		/// <param name="pInfo"></param>
		public override void ReceivePacket(ref PacketInfo pInfo)
		{
			Debug.Assert(pInfo != null);

			SocketDataPacket sdp = pInfo.Packet as SocketDataPacket;
			if (sdp == null) // packet is not for us
				return;

			IPAddress remote_ip = pInfo.RemoteAddress as IPAddress;
			IPAddress local_ip = pInfo.ThisMachine as IPAddress;
			if (remote_ip == null || local_ip == null)
				return;

			NetSimSocket socket = pInfo.Socket;
			if (socket == null)
				socket = FindSocket(sdp.Protocol, remote_ip, sdp.RemotePort, local_ip, sdp.LocalPort);

			if (socket == null) // no socket for this packet
			{
				pInfo = null;
				return;
			}

			// set local endpoint if needed
			socket.SetLocalEndPoint(local_ip.ToSystemIPAddress(), sdp.LocalPort);


			// pass data to the socket, if any
			if (sdp.Data.Length > 0)
			{
				SocketFlags flags = SocketFlags.None;
				if ((sdp.Options & SocketDataPacket.TransferOptions.OutOfBandData) != 0) flags |= SocketFlags.OutOfBand;
				IPEndPoint remote_ep = new IPEndPoint(remote_ip.ToSystemIPAddress(), sdp.RemotePort);
				socket.NewData(sdp.Data, flags, remote_ep);
			}

			// handle special cases like connection request, close etc. Will return true if the packet will be processed
			ReceivePacketNoData(sdp, remote_ip, local_ip, socket);

			pInfo = null;
		}

		/// <summary>
		/// 
		/// </summary>
		/// <param name="sdp"></param>
		/// <param name="remoteIP"></param>
		/// <param name="localIP"></param>
		/// <param name="socket"></param>
		/// <returns><b>True</b> if any action has been taken.</returns>
		private bool ReceivePacketNoData(SocketDataPacket sdp, IPAddress remoteIP, IPAddress localIP, NetSimSocket socket)
		{
			// check parameters
			Debug.Assert(sdp != null && socket != null);

			if ((sdp.Options & SocketDataPacket.TransferOptions.Connect) != 0)
			{
				Debug.Assert(socket.SocketType == SocketType.Stream);

				// sdp should have HigherLayerPacket set to TCP packet with connection request
				TcpPacket tcp_p = (TcpPacket) sdp.HigherLayerPacket;

				lock (socket.CInfo)
				{
					if (socket.CInfo.CurrentState == ConnectionInfo.State.Listen)
					{
						IPEndPoint remote_ep = new IPEndPoint(remoteIP.ToSystemIPAddress(), sdp.RemotePort);
						IPEndPoint local_ep = new IPEndPoint(localIP.ToSystemIPAddress(), sdp.LocalPort);
						socket.NewConnectionRequest(remote_ep, local_ep, tcp_p.SeqNumber, tcp_p.Window);
					}
				}
				return true;
			}
			
			if ((sdp.Options & SocketDataPacket.TransferOptions.Disconnect) != 0)
			{
				socket.Reset();
				return true;
			}
			
			if ((sdp.Options & SocketDataPacket.TransferOptions.EndOfData) != 0)
			{
				ShutdownReceive(socket);
				return true;
			}

			return false;
		}

		/// <summary>
		/// Called by the "send" operations to ensure that local endpoint is set properly. If the local endpoint of
		/// the <paramref name="socket"/> is null or port doesn't contain proper value, allocates the port number
		/// and assigns it to the port. If the address contains <see cref="System.Net.IPAddress.Any"/>, it will be assigned
		/// later at the network layer.
		/// </summary>
		/// <param name="socket"></param>
		private void CheckLocalEndPoint(NetSimSocket socket)
		{
			if (socket.LocalEndPoint == null) // no endpoint at all, create one
			{
				socket.SetLocalEndPoint(System.Net.IPAddress.Any, GetUnassignedPortNumber(socket.ProtocolType, false));
				return;
			}

			// some endpoint is allocated
			IPEndPoint ip_ep = (IPEndPoint)socket.LocalEndPoint;
			if (ip_ep.Port == 0)
				ip_ep.Port = GetUnassignedPortNumber(socket.ProtocolType, false);
		}

		/// <summary>
		/// Checks if the <paramref name="socket"/> binding is correct.
		/// </summary>
		/// <remarks>
		/// Requires socket local endpoint to be set. Checks if it contains local address and unused port
		/// number. Note that an endpoint can contain <see cref="System.Net.IPAddress.Any"/> or port 0.
		/// In case of any it is left to be assigned later, in case of port equal to zero unused port number
		/// is assigned.
		/// </remarks>
		/// <param name="socket">The socket to be bound to the endpoint.</param>
		/// <returns><b>True</b> if the binding is correct.</returns>
		public bool CheckBind(NetSimSocket socket)
		{
			//
			// Check which table to use
			//
			SocketCollection table;
			if (socket.ProtocolType == ProtocolType.Tcp)
				table = tcpSockets;
			else if (socket.ProtocolType == ProtocolType.Udp)
				table = udpSockets;
			else
				return false; // not supported protocol


			IPEndPoint localEndPoint = (IPEndPoint)socket.LocalEndPoint;

			//
			// check if the address specified is local address
			//
			if (!localEndPoint.Address.Equals(System.Net.IPAddress.Any))
			{
				if (!Node.IsLocalAddress(IPAddress.CreateAddress(localEndPoint.Address)))
					return false;
			}

			//
			// check or get the port number
			//
			lock (table)
			{
				if (table.ContainsLocalPort((UInt16)localEndPoint.Port, socket))
					return false; // socket with that number already opened
				// (note that zero cannot be in the table so it won't fail if assigning the port is requested)

				if (localEndPoint.Port == 0)
					localEndPoint.Port = GetUnassignedPortNumber(socket.ProtocolType, true);

				socket.SetLocalEndPoint(localEndPoint.Address, (UInt16)localEndPoint.Port);
			}

			return true;
		}

		/// <summary>
		/// Creates new TCP connection to the <paramref name="remoteEP"/>.
		/// </summary>
		/// <param name="socket">Socket to connect.</param>
		/// <param name="remoteEP">Remote end point to connect to.</param>
		public void Connect(NetSimSocket socket, IPEndPoint remoteEP)
		{
			SocketDataPacket sdp = SocketDataPacket.CreatePacket(new byte[0], 0, 0, ProtocolType.Tcp);
			sdp.RemotePort = (UInt16)remoteEP.Port;
			sdp.Options |= SocketDataPacket.TransferOptions.Connect;

			CheckLocalEndPoint(socket);
			sdp.LocalPort = (UInt16) ((IPEndPoint)socket.LocalEndPoint).Port;

			// prepare packet info
			PacketInfo p_info = new PacketInfo(sdp);
			p_info.Socket = socket;
			p_info.RemoteAddress = IPAddress.CreateAddress(remoteEP.Address);


			// send
			this.Node.NewPacketToSend(this.ID, p_info);
		}

		/// <summary>
		/// Closes the socket and releases the port number, which can be used by another socket.
		/// </summary>
		/// <param name="socket">Socket to close.</param>
		public void Close(NetSimSocket socket)
		{
			ThreadPool.QueueUserWorkItem(new WaitCallback(this.SocketTimeoutRemove), socket);
		}

		/// <summary>
		/// Processed by thread-pool working thread. Removes the socket entry after the specified
		/// timeout.
		/// </summary>
		/// <param name="obj"></param>
		private void SocketTimeoutRemove(object obj)
		{
			Thread.Sleep(30*1000); //Future version: move to config file

			NetSimSocket socket = (NetSimSocket)obj;

			SocketCollection table;
			if (socket.ProtocolType == ProtocolType.Tcp)
				table = tcpSockets;
			else if (socket.ProtocolType == ProtocolType.Udp)
				table = udpSockets;
			else
			{
				Debug.Fail("NetSimSocket has protocol type other than Tcp or Udp");
				return; // should be unreacheable
			}

			table.Remove(socket);
		}

		/// <summary>
		/// Registers new socket with the module.
		/// </summary>
		/// <param name="socket">New socket to be registered.</param>
		public void RegisterNewSocket(NetSimSocket socket)
		{
			switch (socket.ProtocolType)
			{
				case ProtocolType.Tcp:
					tcpSockets.Insert(socket);
					break;
				case ProtocolType.Udp:
					udpSockets.Insert(socket);
					break;
				default:
					throw new InvalidOperationException("SocketModule: Unsupported protocol type");
			}
		}

		/// <summary>
		/// Sends notification to the remote endpoint that we have accepted the connection (SYN + ACK).
		/// </summary>
		/// <param name="socket">The newly created socket for this connection.</param>
		public void Accept(NetSimSocket socket)
		{
			IPEndPoint remote_ep = (IPEndPoint)socket.RemoteEndPoint;
			IPEndPoint local_ep = (IPEndPoint)socket.LocalEndPoint;

			SocketDataPacket sdp = SocketDataPacket.CreatePacket(new byte[0], 0, socket.ProtocolType);
			sdp.RemotePort = (UInt16)remote_ep.Port;
			sdp.LocalPort = (UInt16)local_ep.Port;

			sdp.Options = SocketDataPacket.TransferOptions.ConnectionAccepted;

			PacketInfo p_info = new PacketInfo(sdp);
			p_info.Socket = socket;
			p_info.RemoteAddress = IPAddress.CreateAddress(remote_ep.Address);
			p_info.ThisMachine = IPAddress.CreateAddress(local_ep.Address);
			Node.NewPacketToSend(this.ID, p_info);
		}

		/// <summary>
		/// Notifies the remote endpoint that we have sent all data on the specified socket and
		/// changes the state of the socket appropriately.
		/// </summary>
		/// <param name="socket">Socket to use.</param>
		public void ShutdownSend(NetSimSocket socket)
		{
			socket.CInfo.SendEnd = true;

			IPEndPoint remote_ep = (IPEndPoint)socket.RemoteEndPoint;
			IPEndPoint local_ep = (IPEndPoint)socket.LocalEndPoint;

			if (remote_ep == null) // not connected, nothing to shutdown
				return;

			SocketDataPacket sdp = SocketDataPacket.CreatePacket(new byte[0], 0, socket.ProtocolType);
			sdp.RemotePort = (UInt16)remote_ep.Port;
			sdp.LocalPort = (UInt16)local_ep.Port;

			sdp.Options = SocketDataPacket.TransferOptions.EndOfData;

			PacketInfo p_info = new PacketInfo(sdp);
			p_info.Socket = socket;
			p_info.RemoteAddress = IPAddress.CreateAddress(remote_ep.Address);
			p_info.ThisMachine = IPAddress.CreateAddress(local_ep.Address);
			Node.NewPacketToSend(this.ID, p_info);
		}

		/// <summary>
		/// We have received information that a remote endpoint finished sending; this mehtod changes
		/// a state of the socket appropriately.
		/// </summary>
		/// <param name="socket">Socket to use.</param>
		public void ShutdownReceive(NetSimSocket socket)
		{
			socket.CInfo.ReceiveEnd = true;
		}
	
		/// <summary>
		/// Sends data in <paramref name="buffer"/> to the <paramref name="socket"/>.
		/// </summary>
		/// <param name="socket">Socket to use.</param>
		/// <param name="buffer">An array containing data to send.</param>
		/// <param name="offset">Index to <paramref name="buffer"/> where the data start.</param>
		/// <param name="size">Length of data to send.</param>
		/// <param name="socketFlags">Flags that should be used to modify send operation behavior.</param>
		/// <param name="remoteEP">Remote destination where the data should be sent.</param>
		/// <returns>Number of bytes sent.</returns>
		public int SendTo(NetSimSocket socket, byte[] buffer, int offset, int size, SocketFlags socketFlags, IPEndPoint remoteEP)
		{
			SocketDataPacket sdp = SocketDataPacket.CreatePacket(buffer, offset, size, socket.ProtocolType);
			sdp.RemotePort = (UInt16)remoteEP.Port;
			if ((socketFlags & SocketFlags.OutOfBand) != 0)
				sdp.Options |= SocketDataPacket.TransferOptions.OutOfBandData;

			CheckLocalEndPoint(socket);
			sdp.LocalPort = (UInt16) ((IPEndPoint)socket.LocalEndPoint).Port;
			
			// prepare packet info
			PacketInfo p_info = new PacketInfo(sdp);
			p_info.Socket = socket;
			p_info.RemoteAddress = IPAddress.CreateAddress(remoteEP.Address);
			// set this machine only if the socket is already bound to an address
			IPEndPoint local_ep = (IPEndPoint)socket.LocalEndPoint;
			if (local_ep.Address != null && !local_ep.Address.Equals(System.Net.IPAddress.Any))
				p_info.ThisMachine = IPAddress.CreateAddress(local_ep.Address);

			// send
			this.Node.NewPacketToSend(this.ID, p_info);
			
			return size;
		}
	}

	public class SocketCollection
	{
		private Hashtable sockets = new Hashtable();
		private Hashtable reservedPorts = new Hashtable();

		/// <summary>
		/// Returns a socket from the collection that match specified parameters. If there is no local or
		/// remote endpoint set for the socket, that socket match any parameter.
		/// </summary>
		/// <remarks>
		/// <p>This method is thread-safe.</p>
		/// <p>Currently implemented by traversing the collection item by item.</p>
		/// </remarks>
		/// <param name="remoteAddress"></param>
		/// <param name="remotePort"></param>
		/// <param name="localAddress"></param>
		/// <param name="localPort"></param>
		/// <returns></returns>
		[System.Runtime.CompilerServices.MethodImpl(System.Runtime.CompilerServices.MethodImplOptions.Synchronized)]
		public NetSimSocket Get(IPAddress remoteAddress, UInt16 remotePort, IPAddress localAddress, UInt16 localPort)
		{
			NetSimSocket result = null;

			foreach (DictionaryEntry de in sockets)
			{
				NetSimSocket s = (NetSimSocket)de.Value;
				IPEndPoint local_ep = (IPEndPoint) s.LocalEndPoint;
				IPEndPoint remote_ep = (IPEndPoint) s.RemoteEndPoint;

				bool local_exact_match = local_ep != null && local_ep.Address.Equals(localAddress.ToSystemIPAddress());
				bool local_match = local_exact_match || s.BoundToAny || local_ep == null || local_ep.Address.Equals(System.Net.IPAddress.Any);
				bool local_port_match = local_ep == null || local_ep.Port == localPort;

				bool remote_exact_match = remote_ep != null && remote_ep.Address.Equals(remoteAddress.ToSystemIPAddress());
				bool remote_match = remote_exact_match || remote_ep == null || remote_ep.Address.Equals(System.Net.IPAddress.Any);
				bool remote_port_match = remote_ep == null || remote_ep.Port == remotePort;

				if (local_exact_match && local_port_match && remote_exact_match && remote_port_match) // exact match, no need to search further
					return s;

				// if the match is not exact, save the result, but continue to search for better match
				if (local_match && (localPort == 0 || local_port_match)
					&& remote_match && (remotePort == 0 || remote_port_match))
					result = s;
			}

			return result;
		}

		[System.Runtime.CompilerServices.MethodImpl(System.Runtime.CompilerServices.MethodImplOptions.Synchronized)]
		public NetSimSocket Get(IPAddress localAddress, UInt16 localPort)
		{
			NetSimSocket result = null;

			foreach (DictionaryEntry de in sockets)
			{
				NetSimSocket s = (NetSimSocket)de.Value;
				IPEndPoint local_ep = (IPEndPoint) s.LocalEndPoint;

				if (local_ep.Port == localPort) // port is equal
				{
					if (local_ep.Address.Equals(localAddress.ToSystemIPAddress())
						|| local_ep.Address.Equals(System.Net.IPAddress.Any)) // local EP bound to Any or exact match
						return s;

					if (s.BoundToAny) // if the socket was bound to any address, remember it to return that if no better match is found
						result = s;
				}
			}

			return result;
		}

		[System.Runtime.CompilerServices.MethodImpl(System.Runtime.CompilerServices.MethodImplOptions.Synchronized)]
		public void Insert(NetSimSocket socket)
		{
			sockets[socket] = socket;
		
			IPEndPoint ip_ep = socket.LocalEndPoint as IPEndPoint;
			if (ip_ep != null && ip_ep.Port != 0)
			{
				UInt16 port = (UInt16) ((IPEndPoint)socket.LocalEndPoint).Port;
				if (reservedPorts.Contains(port))
					reservedPorts.Remove(port);
			}
		}

		[System.Runtime.CompilerServices.MethodImpl(System.Runtime.CompilerServices.MethodImplOptions.Synchronized)]
		public void Remove(NetSimSocket socket)
		{
			sockets.Remove(socket);
		}

		[System.Runtime.CompilerServices.MethodImpl(System.Runtime.CompilerServices.MethodImplOptions.Synchronized)]
		public bool ContainsLocalPort(UInt16 port, NetSimSocket excludeSocket)
		{
			if (reservedPorts.Contains(port))
				return true;

			foreach (DictionaryEntry de in sockets)
			{
				NetSimSocket s = (NetSimSocket)de.Value;
				if (excludeSocket != null && s.Equals(excludeSocket))
					continue;

				IPEndPoint loc_ep = (IPEndPoint)s.LocalEndPoint;
				if (loc_ep != null && loc_ep.Port == port)
					return true;
			}

			return false;
		}

		[System.Runtime.CompilerServices.MethodImpl(System.Runtime.CompilerServices.MethodImplOptions.Synchronized)]
		public void ReserveLocalPortNumber(UInt16 port)
		{
			this.reservedPorts[port] = true;
		}
	}
}
