/*

 Copyright (c) 2006 Pavel Novak.

 The use and distribution terms for this software are contained in the file named License.txt, 
 which can be found in the root of the NetSim distribution. By using this software 
 in any fashion, you are agreeing to be bound by the terms of this license.
 
 You must not remove this notice from this software.

*/

using System;
using NetSim.Core;
using ProtocolType = System.Net.Sockets.ProtocolType;
using System.Diagnostics;
using System.Collections;

namespace NetSim.Library.Modules
{
	/// <summary>
	/// TCP protocol implementation.
	/// </summary>
	public class TcpModule : Module
	{
		/// <summary>
		/// The delay before the non-acknowledged packet is retransmitted (in ticks). Currently 2 seconds.
		/// </summary>
		// Future work: move to config file or implement adaptive delay
		private const long RetransmissionDelay = 2 * 10000000;

		/// <summary>
		/// The number of retransmissions if no acknowledgement is received (excluding the first send).
		/// Currently equals to two.
		/// </summary>
		// Future work: move to config file
		private const int RetransmissionCount = 2;

		/// <summary>
		/// Local window size advertised. Currently no limit (<see cref="UInt16.MaxValue"/> is used).
		/// </summary>
		protected UInt16 WindowSize = UInt16.MaxValue;
		
		protected SocketModule SockModule
		{
			get
			{
				// lazy initialization - cannot be initialized in the constructor, because we have no info
				// about in which order the modules are added to the node
				if (sockModule != null)
					return sockModule;

				sockModule = (SocketModule)Node.GetModule(SocketModule.StdModuleName);
				if (sockModule == null)
					throw new InitializationException(String.Format("TcpModule requires SocketModule which is not present at node '{0}'", Node.Name));

				return sockModule;
			}
		}
		private SocketModule sockModule;

		public TcpModule()
		{
		}

		public override void ReceivePacket(ref PacketInfo pInfo)
		{
			TcpPacket tcp_p = pInfo.Packet as TcpPacket;
			if (tcp_p == null)
				return;

			IPAddress remoteIP = pInfo.RemoteAddress as IPAddress;
			IPAddress localIP = pInfo.ThisMachine as IPAddress;

			if (remoteIP == null || localIP == null) // addresses were not set by IPModule, cannot proceed
			{
				Simulator.GetInstance().Events.Notice(String.Format("TcpModule at node '{0}' received packet with no IP addresses set: Probably malformed packet - Missing IPModule at some neighbour node?", this.Node.Name));
				return;
			}

			// find the socket for the packet
			NetSimSocket socket = SockModule.FindSocket(System.Net.Sockets.ProtocolType.Tcp,
				remoteIP, tcp_p.SourcePort,
				localIP, tcp_p.DestinationPort);

			if (socket == null) // no socket associtated with the port
			{
				HandleClosedSocket(ref pInfo, tcp_p);
				return;
			}

			ReceivePacketInternal(ref pInfo, tcp_p, socket);
		}

		private void ReceivePacketInternal(ref PacketInfo pInfo, TcpPacket packet, NetSimSocket socket)
		{
			Debug.Assert(socket != null && pInfo != null && packet != null);

			// 0: do nothing after the state exploration
			// 1: reset
			// 2: syn + ack
			// 3: ack - means send ack immediately
			// 4: fin
			// note that ack is always sent if anything (but reset) is sent to the sender.
			int action_down = 0;

			// -1: discard packet
			// 0: do nothing special, if the packet contains data, pass it to the SocketModule
			// 1: send new connection request
			// 2: incoming connection closed
			// 3: both connections closed
			int action_up = 0;

			//---------------------------------------------------------------------------------
			// Determine what to do with the packet
			//

			// the state of the socket after the actions are taken
			ConnectionInfo.State result_state;

			lock (socket.CInfo)
			{
				// if reset bit is present regardless the state close the connection immediately
				if ((packet.Bits & TcpPacket.CodeBits.RST) != 0)
				{
					action_up = 3; // disconnect immediately
					socket.CInfo.CurrentState = ConnectionInfo.State.Closed;
				}
				else // the action taken depends on the current socket state
				{
					switch (socket.CInfo.CurrentState)
					{
						case ConnectionInfo.State.Closed:
							action_down = 1; // reset
							action_up = -1; // discard packet
							break;

						case ConnectionInfo.State.Listen:
							if (packet.SynBitOnly) // new connection requested
							{
								action_up = 1; // put new connection request to the queue through the SocketModule
							}
							else // refuse everything else
							{
								action_down = 1;
								action_up = -1;
							}
							break;

						case ConnectionInfo.State.SynRecvd: // syn has been sent and received, waiting for last ack
							if ((packet.Bits & TcpPacket.CodeBits.SYN) != 0) // syn resend by the remote party, ignore
							{
								action_up = -1;
							}
							else if ((packet.Bits & TcpPacket.CodeBits.FIN) != 0 // fin cannot be set while establishing connection
								|| (packet.Bits & TcpPacket.CodeBits.ACK) == 0) // if ack is not present, something is wrong
							{
								action_down = 1; // reset
								action_up = -1; // discard packet
								socket.CInfo.CurrentState = ConnectionInfo.State.Closed;
							}
							else if ((packet.Bits & TcpPacket.CodeBits.ACK) != 0)
							{
								if ((socket.CInfo.OutWindowStart + 1) != packet.AckNumber) // unexpected ack number
								{
									action_down = 1; // reset
									action_up = -1; // discard packet
									socket.CInfo.CurrentState = ConnectionInfo.State.Closed;
								}
								else // ok, connection is now established, no reply needed
								{
									// move out window - one position consumed by SYN just acked
									unchecked { socket.CInfo.OutWindowStart++; }
									// connection established
									socket.CInfo.CurrentState = ConnectionInfo.State.Established;
								}
							}
							break;

						case ConnectionInfo.State.SynSent: // first syn has been sent, waiting for ack+syn
							if ((packet.Bits & TcpPacket.CodeBits.FIN) != 0) // fin cannot be set while establishing connection
							{
								action_down = 1; // reset
								action_up = -1; // discard packet
								socket.CInfo.CurrentState = ConnectionInfo.State.Closed;
							}
							else if ((packet.Bits & TcpPacket.CodeBits.SYN) != 0) // ok, syn received, however, it can be the first one when attempting to establish connection from the other side
							{
								if ((packet.Bits & TcpPacket.CodeBits.ACK) != 0) // ok, this is the reply for our syn
								{
									UInt32 expected_seq_num = unchecked(socket.CInfo.OutWindowStart+1);
									if (packet.AckNumber != expected_seq_num) // unexpected ack number
									{
										action_down = 1; // reset
										action_up = -1; // discard packet
										socket.CInfo.CurrentState = ConnectionInfo.State.Closed;
									}
									else // ack was ok save the remote sequence number
									{
										socket.CInfo.OutWindowStart = expected_seq_num;
										socket.CInfo.InWindowStart = unchecked(packet.SeqNumber + 1);
										socket.CInfo.OutWindowLen = packet.Window;
										action_down = 3; // send ack
										socket.CInfo.CurrentState = ConnectionInfo.State.Established;
									}
								}
								else // no ack, remote side is trying to establish connection too
								{
									// save the sequence number
									socket.CInfo.InWindowStart = unchecked(packet.SeqNumber + 1);
									socket.CInfo.OutWindowLen = packet.Window;
									action_down = 2; // send syn + ack
									socket.CInfo.CurrentState = ConnectionInfo.State.SynRecvd; // wait for ack
								}
							}
							else // incoming packet has no syn bit set - something is wrong
							{
								action_down = 1; // reset
								action_up = -1; // discard packet
								socket.CInfo.CurrentState = ConnectionInfo.State.Closed;
							}
							break;

						case ConnectionInfo.State.Established:
							if ((packet.Bits & TcpPacket.CodeBits.ACK) != 0)
							{
								// move the window if we have received greater ack number
								if (PacketReorder.SeqNumCompare(packet.AckNumber, socket.CInfo.OutWindowStart) > 0)
									socket.CInfo.OutWindowStart = packet.AckNumber;
							}
							if ((packet.Bits & TcpPacket.CodeBits.FIN) != 0)
							{
								action_up = 2;
								//action_down = 3; // ack
								socket.CInfo.CurrentState = ConnectionInfo.State.CloseWait;
							}
							break;

						case ConnectionInfo.State.CloseWait:
							if ((packet.Bits & TcpPacket.CodeBits.ACK) != 0)
							{
								// move the window if we have received greater ack number
								if (PacketReorder.SeqNumCompare(packet.AckNumber, socket.CInfo.OutWindowStart) > 0)
									socket.CInfo.OutWindowStart = packet.AckNumber;
							}

							if (packet.HigherLayerPacket != null && packet.HigherLayerPacket.GetBytesCount() > 0)
								packet.HigherLayerPacket = null; // refuse incoming data - fin has been received

							break;

						case ConnectionInfo.State.LastAck:
							if ((packet.Bits & TcpPacket.CodeBits.ACK) != 0)
							{
								if (packet.AckNumber == socket.CInfo.OutWindowStart) // our fin acknowledged
								{
									action_up = 3; // both connections closed
									socket.CInfo.CurrentState = ConnectionInfo.State.Closed;
								}
							}

							break;

						case ConnectionInfo.State.FinWait1: // we have just sent fin
							if ((packet.Bits & TcpPacket.CodeBits.ACK) != 0)
							{
								// move the window if we have received greater ack number
								if (PacketReorder.SeqNumCompare(packet.AckNumber, socket.CInfo.OutWindowStart) > 0)
									socket.CInfo.OutWindowStart = packet.AckNumber;
							}

							if ((packet.Bits & TcpPacket.CodeBits.FIN) != 0) // remote side wants to close connection too
							{
								if ((packet.Bits & TcpPacket.CodeBits.ACK) != 0) // also ack to our fin received
								{
									action_down = 3; // send ack
									action_up = 3; // both connections closed
									socket.CInfo.CurrentState = ConnectionInfo.State.TimedWait;
								}
								else // just fin received - ack it
								{
									action_down = 3; // send ack
									action_up = 3; // both connections closed
									socket.CInfo.CurrentState = ConnectionInfo.State.Closing;
								}
							}
							else // no fin
							{
								if ((packet.Bits & TcpPacket.CodeBits.ACK) != 0
									&& packet.AckNumber == socket.CInfo.OutWindowStart) // ack to our fin received, wait for remote side to close connection
								{
									socket.CInfo.CurrentState = ConnectionInfo.State.FinWait2;
								}
							}

							break;

						case ConnectionInfo.State.FinWait2: // waiting for close by the remote side
							if ((packet.Bits & TcpPacket.CodeBits.FIN) != 0)
							{
								action_down = 3; // send ack
								action_up = 3; // both connections closed
								socket.CInfo.CurrentState = ConnectionInfo.State.TimedWait;
							}

							break;

						case ConnectionInfo.State.Closing: // wait for ack to our fin
							if ((packet.Bits & TcpPacket.CodeBits.ACK) != 0
								&& packet.AckNumber == socket.CInfo.OutWindowStart)
							{
								socket.CInfo.CurrentState = ConnectionInfo.State.TimedWait;
							}
							
							break;

						case ConnectionInfo.State.TimedWait:
							// ignore everything here
							action_up = -1; // discard packet
							break;
					}
				} // if reset present

				// save the state before unlock
				result_state = socket.CInfo.CurrentState;

			} // lock

			// is se to true if ack should be sent. We do not send ack immediately because also acking data might be needed
			bool send_ack = false;
			
			//-----------------------------------------------------------------------
			// Do "action_down"
			//
			TcpPacket new_tcp_p;
			PacketInfo p_i;
			switch (action_down)
			{
				case 0: // do nothing after the state exploration
					break;

				case 1: // reset
					new_tcp_p = new TcpPacket(packet.DestinationPort, packet.SourcePort, socket.CInfo.OutNewDataSeqNum(0),
						0, TcpPacket.CodeBits.RST, socket.CInfo.InWindowLen, 0);
					p_i = new PacketInfo(new_tcp_p);
					p_i.RemoteAddress = pInfo.RemoteAddress;
					p_i.ThisMachine = pInfo.ThisMachine;
					p_i.Socket = socket;
					
					// this call will go through all modules below, through an adapter to the link
					// It can take some time, however, it is not crucial here, if the application will get data
					// or some other notification after this is sent..
					Node.NewPacketToSend(this.ID, p_i);
					break;

				case 2: // syn + ack
					new_tcp_p = new TcpPacket(packet.DestinationPort, packet.SourcePort, socket.CInfo.OutNewDataSeqNum(1),
						socket.CInfo.InWindowStart, TcpPacket.CodeBits.SYN | TcpPacket.CodeBits.ACK, socket.CInfo.InWindowLen, 0);
					p_i = new PacketInfo(new_tcp_p);
					p_i.RemoteAddress = pInfo.RemoteAddress;
					p_i.ThisMachine = pInfo.ThisMachine;
					p_i.Socket = socket;

					// same note applies as in previous case
					Node.NewPacketToSend(this.ID, p_i);
					break;

				case 3: // ack - means send ack immediately
					send_ack = true;
					break;

				case 4: // fin
					new_tcp_p = new TcpPacket(packet.DestinationPort, packet.SourcePort, socket.CInfo.OutNewDataSeqNum(1),
						0, TcpPacket.CodeBits.FIN, socket.CInfo.InWindowLen, 0);
					p_i = new PacketInfo(new_tcp_p);
					p_i.RemoteAddress = pInfo.RemoteAddress;
					p_i.ThisMachine = pInfo.ThisMachine;
					p_i.Socket = socket;

					// same note applies as in previous case
					Node.NewPacketToSend(this.ID, p_i);
					break;
			}

			//-----------------------------------------------------------------------
			// Do "action_up"
			//
			SocketDataPacket.TransferOptions sdp_options = SocketDataPacket.TransferOptions.None;
			switch (action_up)
			{
				case -1: // discard packet
					pInfo = null;
					break;

				case 0: // do nothing special, if the packet contains data, pass it to the SocketModule
					// will be done next
					break;

				case 1: // send new connection request
					sdp_options |= SocketDataPacket.TransferOptions.Connect;
					break;

				case 2: // incoming connection closed
					// do not add end of data now, some packet could be lost.. We have to wait for all data
					// before the application is notified
					// sdp_options |= SocketDataPacket.TransferOptions.EndOfData;
					break;

				case 3: // both connections closed
					sdp_options |= SocketDataPacket.TransferOptions.Disconnect;
					break;
			}

			if (pInfo == null)
			{
				if (send_ack)
					AckData(socket, packet, pInfo);

				return;
			}

			//-----------------------------------------------------------------------
			// take data if any and pass them to the SocketModule
			//

			// if the connection is not established yet, the data will be queued and not passed to the application.
			// after establishing the connection, they will be sent.
			DataPacket data_p = packet.HigherLayerPacket as DataPacket;
			SocketDataPacket sdp = null;
			
			if (data_p != null)
			{
				byte[] data;
				bool last_data;
				bool moved = false;
				lock (socket.CInfo) // do not allow other threads to modify incoming window
				{
					// should we pass the packet up?
					bool return_packet = result_state == ConnectionInfo.State.Established 
						|| result_state == ConnectionInfo.State.CloseWait;

					socket.CInfo.InOutOfOrderPackets.Add(packet.SeqNumber, data_p.Data, socket.CInfo.InWindowStart,
						!return_packet, action_up == 2 || action_up == 3,	out data, out last_data);

					if (data != null && data.Length > 0)
					{
						socket.CInfo.InWindowMove((UInt32)data.Length);
						moved = true;
					}

					if (last_data)
					{
						sdp_options |= SocketDataPacket.TransferOptions.EndOfData;
						socket.CInfo.InWindowMove(1); // FIN has one number in the stream
						moved = true;
					}

					if (moved)// more data in the incoming stream, ack it
						send_ack = true;

					if (data != null)
						sdp = SocketDataPacket.CreatePacket(data, 0, System.Net.Sockets.ProtocolType.Tcp);
				}
			}

			// send ack here - acks also data
			if (send_ack)
				AckData(socket, packet, pInfo);

			if (sdp != null || sdp_options != SocketDataPacket.TransferOptions.None)
			{
				if (sdp == null) // just options set, create empty packet
					sdp = SocketDataPacket.CreatePacket(new byte[0], 0, ProtocolType.Tcp);

				sdp.Options = sdp_options;

				// set endpoints appropriately
				sdp.LocalPort = packet.DestinationPort;
				sdp.RemotePort = packet.SourcePort;
				sdp.HigherLayerPacket = packet;

				pInfo.Packet = sdp;
				pInfo.Socket = socket;
			}
			else // data have been queued (and no options), do not send anything up right now
			{
				pInfo = null;
			}
			
		}

		private void AckData(NetSimSocket socket, TcpPacket packet, PacketInfo pInfo)
		{
			TcpPacket new_tcp_p = new TcpPacket(packet.DestinationPort, packet.SourcePort, socket.CInfo.OutNewDataSeqNum(0),
				socket.CInfo.InWindowStart, TcpPacket.CodeBits.ACK, socket.CInfo.InWindowLen, 0);
			PacketInfo p_i = new PacketInfo(new_tcp_p);
			p_i.RemoteAddress = pInfo.RemoteAddress;
			p_i.ThisMachine = pInfo.ThisMachine;
			p_i.Socket = socket;

			Node.NewPacketToSend(this.ID, p_i);
		}

		protected void HandleClosedSocket(ref PacketInfo pInfo, TcpPacket packet)
		{
			// send connection reset back to source
			TcpPacket new_tcp_p = new TcpPacket(packet.DestinationPort, packet.SourcePort, 0,
				0, TcpPacket.CodeBits.RST, 0, 0);
			PacketInfo p_i = new PacketInfo(new_tcp_p);
			p_i.RemoteAddress = pInfo.RemoteAddress;
			p_i.ThisMachine = pInfo.ThisMachine;
					
			pInfo = null;

			// this call will go through all modules below, through an adapter to the link
			// It can take some time, however, it is not crucial here, if the application will get data
			// or some other notification after this is sent..
			Node.NewPacketToSend(this.ID, p_i);
		}

		public override void SendPacket(ref PacketInfo pInfo)
		{
			// if the packet is for us, wrap with TCP packet..

			SocketDataPacket sdp = pInfo.Packet as SocketDataPacket;
			if (sdp == null || sdp.Protocol != ProtocolType.Tcp)
				return;

			Debug.Assert(pInfo.Socket != null); // SocketDataPacket requires socket instance set to the PacketInfo

			SendPacketNoData(pInfo, sdp, pInfo.Socket); // handle options

			if (sdp.Data.Length > 0) // send data
			{
				lock (pInfo.Socket.CInfo)
				{
					DataPacket data_p = DataPacket.CreatePacket(sdp.Data, 0);
					if (data_p != null)
					{
						TcpPacket new_tcp_p = new TcpPacket(sdp.LocalPort, sdp.RemotePort, pInfo.Socket.CInfo.OutNewDataSeqNum((UInt32)data_p.Data.Length),
							pInfo.Socket.CInfo.InWindowStart, TcpPacket.CodeBits.ACK, pInfo.Socket.CInfo.InWindowLen, 0);
						new_tcp_p.HigherLayerPacket = data_p;
						pInfo.Packet = new_tcp_p;
						RemeberForRetransmission(pInfo);
					}
					else
					{
						pInfo = null; // do not send data right now
					}
				}
			}
			else // do not continue sending packet
			{
				pInfo = null;
			}
		}

		/// <summary>
		/// Handles special cases such as connection establishing, closing etc.
		/// </summary>
		/// <param name="pInfo"><see cref="PacketInfo"/> coming from the upper layer.</param>
		/// <param name="sdp"><see cref="SocketDataPacket"/> containing information to be sent.</param>
		/// <param name="socket">Socket requesting the action.</param>
		private void SendPacketNoData(PacketInfo pInfo, SocketDataPacket sdp, NetSimSocket socket)
		{
			// Connection accept
			if ((sdp.Options & SocketDataPacket.TransferOptions.ConnectionAccepted) != 0)
			{
				TcpPacket new_tcp_p = new TcpPacket(sdp.LocalPort, sdp.RemotePort, socket.CInfo.OutNewDataSeqNum(1),
					socket.CInfo.InWindowStart, TcpPacket.CodeBits.ACK | TcpPacket.CodeBits.SYN, socket.CInfo.InWindowLen, 0);
				PacketInfo p_i = new PacketInfo(new_tcp_p);
				p_i.RemoteAddress = pInfo.RemoteAddress;
				p_i.ThisMachine = pInfo.ThisMachine;
				p_i.Socket = pInfo.Socket;

				socket.CInfo.CurrentState = ConnectionInfo.State.SynRecvd;
				RemeberForRetransmission(p_i);
				Node.NewPacketToSend(this.ID, p_i);
			}

			// end of data
			else if ((sdp.Options & SocketDataPacket.TransferOptions.EndOfData) != 0)
			{
				TcpPacket new_tcp_p = new TcpPacket(sdp.LocalPort, sdp.RemotePort, socket.CInfo.OutNewDataSeqNum(1),
					socket.CInfo.InWindowStart, TcpPacket.CodeBits.ACK | TcpPacket.CodeBits.FIN, socket.CInfo.InWindowLen, 0);
				PacketInfo p_i = new PacketInfo(new_tcp_p);
				p_i.RemoteAddress = pInfo.RemoteAddress;
				p_i.ThisMachine = pInfo.ThisMachine;
				p_i.Socket = pInfo.Socket;

				socket.CInfo.CurrentState = ConnectionInfo.State.FinWait1;
				RemeberForRetransmission(p_i);
				Node.NewPacketToSend(this.ID, p_i);
			}

			// connect
			else if ((sdp.Options & SocketDataPacket.TransferOptions.Connect) != 0)
			{
				TcpPacket new_tcp_p = new TcpPacket(sdp.LocalPort, sdp.RemotePort, socket.CInfo.OutNewDataSeqNum(1),
					socket.CInfo.InWindowStart, TcpPacket.CodeBits.SYN, socket.CInfo.InWindowLen, 0);
				PacketInfo p_i = new PacketInfo(new_tcp_p);
				p_i.RemoteAddress = pInfo.RemoteAddress;
				p_i.ThisMachine = pInfo.ThisMachine;
				p_i.Socket = pInfo.Socket;

				socket.CInfo.CurrentState = ConnectionInfo.State.SynSent;
				RemeberForRetransmission(p_i);
				Node.NewPacketToSend(this.ID, p_i);
			}
		}

		/// <summary>
		/// Called when the TCP packet is about to send to a lower layer. We have to remember the packet
		/// for the case it needs to be retransmitted.
		/// </summary>
		/// <param name="pInfo"></param>
		private void RemeberForRetransmission(PacketInfo pInfo)
		{
			// we have to create a copy - original packet info might change while transferring through
			// lower layers
			PacketInfo p_info_copy = (PacketInfo)pInfo.Clone();

			TcpPacket tcp_p = (TcpPacket)pInfo.Packet;
			int data_len = 0;
			if (tcp_p.HigherLayerPacket != null)
				data_len = tcp_p.HigherLayerPacket.GetBytesCount();
			if (data_len == 0)
				data_len = 1; // each packet requrires at least one position in the sequence, even if it contains
			                // no data (connect, end of data...)

			p_info_copy.Socket.CInfo.SentPackets.Add(p_info_copy,
				PacketReorder.SeqNumIncrement(tcp_p.SeqNumber, (uint)data_len), RetransmissionCount);

			ManagedThreadPool.QueueUserWorkItem(new System.Threading.WaitCallback(this.CheckPacketsToResend),
				p_info_copy.Socket, RetransmissionDelay);
		}

		private void CheckPacketsToResend(object state)
		{
			// socket to check the packets is passed in the state param
			NetSimSocket sock = (NetSimSocket)state;
			// get the packet to retransmission if any
			bool last_try;
			PacketInfo p_i = sock.CInfo.SentPackets.GetOldPacket(RetransmissionDelay, out last_try);
			if (p_i == null)
				return;
			
			// send again
			this.Node.NewPacketToSend(this.ID, p_i);
			// check again if this is not the last send operation
			if (!last_try)
				ManagedThreadPool.QueueUserWorkItem(new System.Threading.WaitCallback(this.CheckPacketsToResend),
					sock, RetransmissionDelay);
		}
	}
}
