/*

 Copyright (c) 2006 Pavel Novak.

 The use and distribution terms for this software are contained in the file named License.txt, 
 which can be found in the root of the NetSim distribution. By using this software 
 in any fashion, you are agreeing to be bound by the terms of this license.
 
 You must not remove this notice from this software.

*/

using System;
using NetSim.Core;
using System.Diagnostics;
using ProtocolType = System.Net.Sockets.ProtocolType;

namespace NetSim.Library
{
	#region IP

	public abstract class IPPacket : Packet
	{
		public byte Version
		{
			get { return version; }
		}
		protected byte version;

		/// <summary>
		/// Returns the number of bytes occupied by the header - including all options or next headers.
		/// </summary>
		public abstract int HeaderLengthBytes { get; }

		/// <summary>
		/// Gets source IP address.
		/// </summary>
		public IPAddress SourceIP { get { return sourceIP; } }
		protected IPAddress sourceIP;

		/// <summary>
		/// Gets destination IP address.
		/// </summary>
		public IPAddress DestinationIP { get { return destinationIP; } }
		protected IPAddress destinationIP;

		/// <summary>
		/// Creates new IPv4 or IPv6 packet.
		/// </summary>
		/// <param name="data"></param>
		/// <param name="offset"></param>
		/// <param name="len"></param>
		/// <returns></returns>
		public static IPPacket CreatePacket(byte[] data, int offset, int len)
		{
			byte version = (byte) (data[offset] >> 4);
			switch (version)
			{
				case 4:
					return IPv4Packet.CreatePacket(data, offset, len);
				case 6:
					return IPv6Packet.CreatePacket(data, offset, len);
				default:
					return null;
			}
		}

		public override NetSim.Core.Packet.PacketGroup Group
		{
			get { return Packet.PacketGroup.NetworkLayer; }
		}

		public override int Type
		{
			get { return (int)Packet.NetworkLayerPacketType.IP; }
		}

	}

	public class IPv4Packet : IPPacket
	{
		[Flags]
		public enum Flags : byte
		{
			None = 0,
			DoNotFragment = 0x2,
			MoreFragments = 0x1
		}

		public byte HeaderLen
		{
			get { return headerLen; }
		}
		private byte headerLen;

		public override int HeaderLengthBytes
		{
			get { return headerLen * 4; }
		}

		public byte ServiceType
		{
			get { return serviceType; }
		}
		private byte serviceType;

		public UInt16 TotalLength
		{
			get { return totalLength; }
		}
		private UInt16 totalLength;

		public UInt16 Identification
		{
			get { return identification; }
		}
		private UInt16 identification;

		public Flags FlagsField
		{
			get { return this.flags; }
		}
		private Flags flags;

		/// <summary>
		/// Fragment offset specified in units of 8 bytes.
		/// </summary>
		public UInt16 FragmentOffset
		{
			get { return fragmentOffset; }
		}
		private UInt16 fragmentOffset;

		public byte TimeToLive
		{
			get { return ttl; }
			set { ttl = value; }
		}
		private byte ttl;

		public byte Protocol
		{
			get { return protocol; }
		}
		private byte protocol;

		public IPv4Address SourceIPv4
		{
			get { return (IPv4Address)sourceIP; }
		}

		public IPv4Address DestinationIPv4
		{
			get { return (IPv4Address)destinationIP; }
		}

		public static new IPv4Packet CreatePacket(byte[] data, int offset, int len)
		{
			return new IPv4Packet(data, offset, len);
		}

		protected IPv4Packet(byte[] data, int offset, int len)
		{
			int original_offset = offset;

			// version
			version = (byte)(data[offset] >> 4);
			// header length
			headerLen = (byte) (data[offset] & 0x0F);
			offset++;
			// service type
			serviceType = data[offset];
			offset++;
			// total length
			totalLength = ByteArrayList.ToUInt16(data, offset);
			offset += 2;
			// identification
			identification = ByteArrayList.ToUInt16(data, offset);
			offset += 2;
			// flags
			flags = (Flags) (data[offset] >> 5);
			// fragment offset
			fragmentOffset = (UInt16)(ByteArrayList.ToUInt16(data, offset) & (UInt16)0x1FFF);
			offset += 2;
			// time to live
			ttl = data[offset];
			offset++;
			// protocol
			protocol = data[offset];
			offset++;
			// skip header checksum
			offset += 2;
			// source IP
			sourceIP = new IPv4Address(ByteArrayList.ToUInt32(data, offset));
			offset += 4;
			// target IP
			destinationIP = new IPv4Address(ByteArrayList.ToUInt32(data, offset));
			offset += 4;

			// options
			if (headerLen > 5) // there are some options
			{
				// Future version: implement options
				offset += 4 * (headerLen - 5);
			}

			// try to parse the upper layer
			int data_len = totalLength - 4 * headerLen;
			if (data_len <= (len - (offset - original_offset))) // there is enough data for the data according to totalLength
			{
				if (((flags & Flags.MoreFragments) == 0) && fragmentOffset == 0) // packet is not fragmented, continue parsing
				{
					HigherLayerPacket = Simulator.GetInstance().CreatePacket(
						Packet.PacketGroup.TransportLayer, protocol, data, offset, data_len);
				}

				if (HigherLayerPacket == null) // packet is fragmented, we cannot parse upper layers until reassembly (or not supported packet type)
				{
					this.HigherLayerPacket = UnknownPacket.CreatePacket(data, offset, data_len);
				}
			}
		}

		public IPv4Packet(byte headerLen, byte serviceType, UInt16 totalLength,
			UInt16 identification, Flags flags, UInt16 fragmentOffset, byte ttl, byte protocol,
			IPv4Address sourceIP, IPv4Address destIP)
		{
			this.version = 4;
			this.headerLen = headerLen;
			this.serviceType = serviceType;
			this.totalLength = totalLength;
			this.identification = identification;
			this.flags = flags | Flags.DoNotFragment;
			this.fragmentOffset = fragmentOffset;
			this.ttl = ttl;
			this.protocol = protocol;
			this.sourceIP = sourceIP;
			this.destinationIP = destIP;
		}

		protected override void AppendBytesHeader(ByteArrayList buf)
		{
			// remember where the header starts for computing the checksum
			int header_start_offset = buf.Count;

			// version + header length
			byte first_byte = (byte) ((this.version << 4) | this.headerLen);
			buf.AddByte(first_byte);
			// service type
			buf.AddByte(serviceType);
			// total length
			buf.AddUInt16(totalLength);
			// identification
			buf.AddUInt16(identification);
			// flags + fragment offset
			UInt16 flags_offset = this.fragmentOffset;
			flags_offset |= (UInt16) ((UInt16)flags << 13);
			buf.AddUInt16(flags_offset);
			// time to live
			buf.AddByte(ttl);
			// protocol
			buf.AddByte(protocol);
			// header checksum - add zero, to not affect computing which follows
			buf.AddUInt16(0);
			// source + destination IP
			buf.AddBytes(sourceIP.GetAddressBytes());
			buf.AddBytes(destinationIP.GetAddressBytes());
			
			//Future version: options
			if (headerLen > 5)
			{
				byte[] options = new byte[headerLen*4 - 20];
				buf.AddBytes(options);
			}

			UInt16 checksum = buf.OneComplementChecksum2(header_start_offset, headerLen * 4);
			buf.SetUInt16(header_start_offset + 10, checksum);

			//
			// compute pseudoheader checksum and set it to the UDP or TCP if present
			//
			TcpPacket tcp_p = HigherLayerPacket as TcpPacket;
			UdpPacket udp_p = HigherLayerPacket as UdpPacket;
			if (tcp_p != null || udp_p != null)
			{
				ByteArrayList pseudoheader = new ByteArrayList(12);
				pseudoheader.AddBytes(sourceIP.GetAddressBytes());
				pseudoheader.AddBytes(destinationIP.GetAddressBytes());
				pseudoheader.AddByte(0);
				pseudoheader.AddByte(protocol);
				pseudoheader.AddUInt16((UInt16)(totalLength - headerLen*4));

				UInt32 ph_sum = pseudoheader.SumForOneComplementComputation(0, 12);
				if (tcp_p != null)
					tcp_p.SetPseudoHeaderSum(ph_sum);
				if (udp_p != null)
					udp_p.SetPseudoHeaderSum(ph_sum);
			}
		}

		protected override void AppendBytesFooter(ByteArrayList buf)
		{
			// nothing
		}
	
		/// <summary>
		/// Returns number of bytes needed by this layer headers and footers
		/// </summary>
		/// <returns></returns>
		protected override int BytesCount()
		{
			return headerLen * 4;
		}
	}

	/// <summary>
	/// IPv6 packet. Not implemented yet.
	/// </summary>
	public class IPv6Packet : IPPacket
	{
		public override int HeaderLengthBytes
		{
			get
			{
				throw new NotImplementedException();
			}
		}

		public static new IPv6Packet CreatePacket(byte[] data, int offset, int len)
		{
			throw new NotImplementedException();
		}

		protected override void AppendBytesHeader(ByteArrayList buf)
		{
			throw new NotImplementedException();
		}

		protected override void AppendBytesFooter(ByteArrayList buf)
		{
			throw new NotImplementedException();
		}
	
		protected override int BytesCount()
		{
			throw new NotImplementedException();
		}
	}

	#endregion

	#region ICMP
	/// <summary>
	/// Common ancestor for IcmpPacket classes. This class is instatiated if there is no implementation
	/// for required ICMP packet type.
	/// </summary>
	public class IcmpPacket : Packet
	{
		/// <summary>
		/// Possible values of "type" filed in the ICMP message.
		/// </summary>
		public enum MessageType : byte
		{
			EchoReply = 0,
			DestinationUnreacheable = 3,
			SourceQuench = 4,
			Redirect = 5,
			AlternateHostAddress = 6,
			EchoRequest = 8,
			RouterAdvertisement = 9,
			RouterSolicitation = 10,
			TimeToLiveExceeded = 11,
			ParameterProblem = 12,
			TimestampRequest = 13,
			TimestampReply = 14,
			InformationRequest = 15,
			InformationReply = 16,
			AddressMaskRequest = 17,
			AddressMaskReply = 18,
			Traceroute = 30,
			DatagramConversionError = 31,
			MobileHostRedirect = 32,
			IPv6WhereAreYou = 33,
			IPv6IAmHere = 34,
			MobileRegistrationRequest = 35,
			MobileRegistrationReply = 36,
			DomainNameRequest = 37,
			DomainNameReply = 38,
			Skip = 39,
			Photuris = 40
		}

		/// <summary>
		/// ICMP message type.
		/// </summary>
		public MessageType MsgType { get { return this.type; } }
		protected MessageType type;

		/// <summary>
		/// Code field in ICMP message.
		/// </summary>
		protected byte code;

		/// <summary>
		/// Creates new <see cref="IcmpPacket"/> from raw data.
		/// </summary>
		/// <param name="data">Array of packet bytes.</param>
		/// <param name="offset">Index where this packet starts.</param>
		/// <param name="len">Maximum lenght of this packet in bytes.</param>
		/// <returns>Newly created IcmpPacket.</returns>
		public static IcmpPacket CreatePacket(byte[] data, int offset, int len)
		{
			MessageType message_type = (MessageType)data[offset];

			switch (message_type)
			{
				case MessageType.EchoRequest: goto case MessageType.EchoReply;
				case MessageType.EchoReply:
					return IcmpEchoPacket.CreatePacket(data, offset, len);

				case MessageType.DestinationUnreacheable:
					return IcmpUnreacheablePacket.CreatePacket(data, offset, len);

				default:
					// create general ICMP packet..
					return new IcmpPacket(data, offset, len);
			}
		}

		protected IcmpPacket()
		{
		}

		protected IcmpPacket(byte[] data, int offset, int len)
		{
			this.type = (MessageType)data[offset];
			this.code = data[offset+1];
			offset += 4; // skip checksum
			len -= 4;

			this.HigherLayerPacket = DataPacket.CreatePacket(data, offset, len);
		}

		[ThreadStatic]
		private static int headerStartOffset;
		protected override void AppendBytesHeader(ByteArrayList buf)
		{
			headerStartOffset = buf.Count;
			buf.AddByte((byte)this.type);
			buf.AddByte(this.code);
			buf.AddUInt16(0); // Checksum is computed in AppendBytesFooter
		}

		protected override void AppendBytesFooter(ByteArrayList buf)
		{
			UInt16 checksum = buf.OneComplementChecksum2(headerStartOffset, buf.Count - headerStartOffset);
			buf.SetUInt16(headerStartOffset + 2, checksum);
		}

		protected override int BytesCount()
		{
			return 4;
		}

		public override Packet.PacketGroup Group
		{
			get { return Packet.PacketGroup.TransportLayer; }
		}

		public override int Type
		{
			get { return (int)Packet.TransportLayerPacketType.ICMP; }
		}

	}

	public class IcmpEchoPacket : IcmpPacket
	{
		public UInt16 Identifier
		{
			get { return identifier; }
		}
		protected UInt16 identifier;

		public UInt16 SequenceNumber
		{
			get { return sequenceNumber; }
		}
		protected UInt16 sequenceNumber;

		/// <summary>
		/// A method for instantiating <see cref="IcmpEchoPacket"/> class. Use this instead of the constructor.
		/// </summary>
		/// <param name="data"></param>
		/// <param name="offset"></param>
		/// <param name="len"></param>
		/// <returns></returns>
		public static new IcmpEchoPacket CreatePacket(byte[] data, int offset, int len)
		{
			return new IcmpEchoPacket(data, offset, len);
		}

		protected IcmpEchoPacket(byte[] data, int offset, int len)
		{
			int original_offset = offset;

			this.type = (MessageType)data[offset];
			this.code = data[offset+1];
			offset += 4; // skip checksum

			Debug.Assert(type == MessageType.EchoReply || type == MessageType.EchoRequest);

			identifier = ByteArrayList.ToUInt16(data, offset);
			offset += 2;
			sequenceNumber = ByteArrayList.ToUInt16(data, offset);
			offset += 2;

			this.HigherLayerPacket = DataPacket.CreatePacket(data, offset, len - (offset - original_offset));
		}
	
		public IcmpEchoPacket(IcmpPacket.MessageType msgType, byte code, UInt16 identifier, UInt16 sequenceNum,
			byte[] data)
		{
			Debug.Assert(msgType == IcmpPacket.MessageType.EchoRequest || msgType == IcmpPacket.MessageType.EchoReply);

			this.type = msgType;
			this.code = code;
			this.identifier = identifier;
			this.sequenceNumber = sequenceNum;
			this.HigherLayerPacket = DataPacket.CreatePacket(data, 0);
		}

		[ThreadStatic]
		private static int headerStartOffset;
		protected override void AppendBytesHeader(ByteArrayList buf)
		{
			headerStartOffset = buf.Count;
			buf.AddByte((byte)this.type);
			buf.AddByte(this.code);
			buf.AddUInt16(0); // Checksum is computed in AppendBytesFooter
			buf.AddUInt16(this.identifier);
			buf.AddUInt16(this.sequenceNumber);
		}

		protected override void AppendBytesFooter(ByteArrayList buf)
		{
			UInt16 checksum = buf.OneComplementChecksum2(headerStartOffset, buf.Count - headerStartOffset);
			buf.SetUInt16(headerStartOffset + 2, checksum);
		}

		protected override int BytesCount()
		{
			return 8;
		}

		public static IcmpEchoPacket SwitchToEchoReply(IcmpEchoPacket request)
		{
			Debug.Assert(request.MsgType == IcmpPacket.MessageType.EchoRequest);

			byte[] opt_data;
			DataPacket data_p = request.HigherLayerPacket as DataPacket;
			if (data_p != null)
				opt_data = data_p.Data;
			else
				opt_data = new byte[0];

			return new IcmpEchoPacket(IcmpPacket.MessageType.EchoReply, 0, request.identifier, request.sequenceNumber,
				opt_data);
		}
	}

	public class IcmpUnreacheablePacket : IcmpPacket
	{
		public enum Code : byte
		{
			NetUnreacheable = 0,
			HostUnreacheable = 1,
			ProtocolUnreacheable = 2,
			PortUnreacheable = 3,
			FragmentationNeeded = 4,
			SourceRouteFailed = 5,
			DestNetUnknown = 6,
			DestHostUnknown = 7,
			SourceHostIsolated = 8,
			CommNetProhibited = 9,
			CommHostProhibited = 10,
			NetUnreacheableForTos = 11,
			HostUnreacheableForTos = 12,
			CommAdministrativelyProhibited = 13,
			HostPrecedenceViolation = 14,
			PrecedenceCutoff = 15
		}

		/// <summary>
		/// Method for instantiating <see cref="IcmpUnreacheablePacket"/> class. Use this instead of the constructor.
		/// </summary>
		/// <param name="data"></param>
		/// <param name="offset"></param>
		/// <param name="len"></param>
		/// <returns></returns>
		public static new IcmpUnreacheablePacket CreatePacket(byte[] data, int offset, int len)
		{
			return new IcmpUnreacheablePacket(data, offset, len);
		}

		public static IcmpUnreacheablePacket CreatePacket(IPPacket packet, Code code)
		{
			return new IcmpUnreacheablePacket(packet, code);	
		}

		protected IcmpUnreacheablePacket(IPPacket packet, Code code)
		{
			this.type = IcmpPacket.MessageType.DestinationUnreacheable;
			this.code = (byte)code;
			
			// get first 64 bits of packet
			ByteArrayList buf = new ByteArrayList(packet.GetBytesCount());
			packet.AppendBytes(buf);
			int len = Math.Min(packet.HeaderLengthBytes + 8, buf.Count); // max 8 bytes
			DataPacket d_p = DataPacket.CreatePacket(buf.Bytes, 0, len);
			this.HigherLayerPacket = d_p;
		}

		protected IcmpUnreacheablePacket(byte[] data, int offset, int len)
		{
			int original_offset = offset;

			this.type = (IcmpPacket.MessageType)data[offset];
			this.code = data[offset+1];
			offset += 4; // skip checksum

			Debug.Assert(type == IcmpPacket.MessageType.DestinationUnreacheable);

			offset += 4; // skip unused

			// copy as data packet the internet header + 64 bits of datagram..
			this.HigherLayerPacket = DataPacket.CreatePacket(data, offset, len - (offset - original_offset));
		}
	
		[ThreadStatic]
		private static int headerStartOffset;
		protected override void AppendBytesHeader(ByteArrayList buf)
		{
			headerStartOffset = buf.Count;
			buf.AddByte((byte)this.type);
			buf.AddByte(this.code);
			buf.AddUInt16(0); // Checksum is computed in AppendBytesFooter
			buf.AddUInt32(0); // unused
		}

		protected override void AppendBytesFooter(ByteArrayList buf)
		{
			UInt16 checksum = buf.OneComplementChecksum2(headerStartOffset, buf.Count - headerStartOffset);
			buf.SetUInt16(headerStartOffset + 2, checksum);
		}

		protected override int BytesCount()
		{
			return 8;
		}
	}


	public class IcmpTimeExceededPacket : IcmpPacket
	{
		public enum Code : byte
		{
			TtlCountExceeded = 0,
			FragmentReassemblyTimeout = 1
		}

		/// <summary>
		/// Method for instantiating <see cref="IcmpTimeExceededPacket"/> class. Use this instead of the constructor.
		/// </summary>
		/// <param name="data"></param>
		/// <param name="offset"></param>
		/// <param name="len"></param>
		/// <returns></returns>
		public static new IcmpTimeExceededPacket CreatePacket(byte[] data, int offset, int len)
		{
			return new IcmpTimeExceededPacket(data, offset, len);
		}

		public static IcmpTimeExceededPacket CreatePacket(IPPacket packet, Code code)
		{
			return new IcmpTimeExceededPacket(packet, code);	
		}

		protected IcmpTimeExceededPacket(IPPacket packet, Code code)
		{
			this.type = IcmpPacket.MessageType.TimeToLiveExceeded;
			this.code = (byte)code;
			
			// get IP header + first 64 bits of packet
			ByteArrayList buf = new ByteArrayList(packet.GetBytesCount());
			packet.AppendBytes(buf);
			int len = Math.Min(packet.HeaderLengthBytes + 8, buf.Count); // max 8 bytes
			DataPacket d_p = DataPacket.CreatePacket(buf.Bytes, 0, len);
			this.HigherLayerPacket = d_p;
		}

		protected IcmpTimeExceededPacket(byte[] data, int offset, int len)
		{
			int original_offset = offset;

			this.type = (IcmpPacket.MessageType)data[offset];
			this.code = data[offset+1];
			offset += 4; // skip checksum

			Debug.Assert(type == IcmpPacket.MessageType.TimeToLiveExceeded);

			offset += 4; // skip unused

			// copy as data packet the internet header + 64 bits of datagram..
			this.HigherLayerPacket = DataPacket.CreatePacket(data, offset, len - (offset - original_offset));
		}
	
		[ThreadStatic]
		private static int headerStartOffset;
		protected override void AppendBytesHeader(ByteArrayList buf)
		{
			headerStartOffset = buf.Count;
			buf.AddByte((byte)this.type);
			buf.AddByte(this.code);
			buf.AddUInt16(0); // Checksum is computed in AppendBytesFooter
			buf.AddUInt32(0); // unused
		}

		protected override void AppendBytesFooter(ByteArrayList buf)
		{
			UInt16 checksum = buf.OneComplementChecksum2(headerStartOffset, buf.Count - headerStartOffset);
			buf.SetUInt16(headerStartOffset + 2, checksum);
		}

		protected override int BytesCount()
		{
			return 8;
		}
	}

	#endregion

	#region TCP

	public class TcpPacket : Packet
	{
		[Flags]
			public enum CodeBits : byte
		{
			FIN = 1,
			SYN = 2,
			RST = 4,
			PSH = 8,
			ACK = 16,
			URG = 32
		}

		public UInt16 SourcePort { get { return this.sourcePort; } }
		private UInt16 sourcePort;

		public UInt16 DestinationPort { get { return this.destinationPort; } }
		private UInt16 destinationPort;

		public UInt32 SeqNumber { get { return this.seqNumber; } }
		private UInt32 seqNumber;

		public UInt32 AckNumber { get { return this.ackNumber; } }
		private UInt32 ackNumber;

		private byte headerLen;
		
		public CodeBits Bits { get { return this.codeBits; } }
		private CodeBits codeBits;

		/// <summary>Returns <b>true</b> if only bit "SYN" is set (PSH and URG are ignored).</summary>
		public bool SynBitOnly
		{
			get
			{
				return (codeBits & CodeBits.SYN) != 0 && (codeBits & CodeBits.ACK) == 0
					&& (codeBits & CodeBits.FIN) == 0 && (codeBits & CodeBits.RST) == 0;
			}
		}

		/// <summary>Returns <b>true</b> if only bit "ACK" is set (PSH and URG are ignored).</summary>
		public bool AckBitOnly
		{
			get
			{
				return (codeBits & CodeBits.SYN) == 0 && (codeBits & CodeBits.ACK) != 0
					&& (codeBits & CodeBits.FIN) == 0 && (codeBits & CodeBits.RST) == 0;
			}
		}

		public UInt16 Window { get { return this.window; } }
		private UInt16 window;

		public void SetPseudoHeaderSum(UInt32 sum)
		{
			this.pseudoHeaderSum = sum;
		}
		private UInt32 pseudoHeaderSum;

		private UInt16 urgentPointer;
		
		//Future version: options

		/// <summary>
		/// Sets the IP packet in which this packet resides.
		/// </summary>
		public IPPacket IPPacket { set { this.ipPacket = value; } }
		/// <summary>
		/// <see cref="IPPacket"/> in which this packet resides.
		/// </summary>
		protected IPPacket ipPacket;

		public static TcpPacket CreatePacket(byte[] data, int offset, int len)
		{
			return new TcpPacket(data, offset, len);
		}

		protected TcpPacket(byte[] data, int offset, int len)
		{
			int original_offset = offset;
			int original_len = len;

			// ports
			sourcePort = ByteArrayList.ToUInt16(data, offset);
			offset += 2;
			destinationPort = ByteArrayList.ToUInt16(data, offset);
			offset += 2;
			// sequence number
			seqNumber = ByteArrayList.ToUInt32(data, offset);
			offset += 4;
			// acknowledgement number
			ackNumber = ByteArrayList.ToUInt32(data, offset);
			offset += 4;
			// hlen + code bits
			headerLen = (byte) (data[offset] >> 4);
			codeBits = (CodeBits) (data[offset+1] & 0x3F);
			offset += 2;
			// window
			window = ByteArrayList.ToUInt16(data, offset);
			offset += 2;
			// checksum
			//checksum = ByteArrayList.ToUInt16(data, offset);
			offset += 2;
			// urgent pointer
			urgentPointer = ByteArrayList.ToUInt16(data, offset);
			offset += 2;
			
			// Future version: options
			offset += (headerLen - 5) * 4; // skip options

			HigherLayerPacket = SocketDataPacket.CreatePacket(data, offset, len - 4*headerLen, ProtocolType.Tcp);
		}

		/// <summary>
		/// Creates new <see cref="TcpPacket"/> without options.
		/// </summary>
		/// <param name="sourcePort"></param>
		/// <param name="destPort"></param>
		/// <param name="seqNumber"></param>
		/// <param name="ackNumber"></param>
		/// <param name="codeBits"></param>
		/// <param name="window"></param>
		/// <param name="urgentPointer"></param>
		public TcpPacket(UInt16 sourcePort, UInt16 destPort, UInt32 seqNumber, UInt32 ackNumber, CodeBits codeBits,
			UInt16 window, UInt16 urgentPointer)
		{
			bool is_syn = (codeBits & TcpPacket.CodeBits.SYN) != 0;

			this.sourcePort = sourcePort;
			this.destinationPort = destPort;
			this.seqNumber = seqNumber;
			this.ackNumber = ackNumber;
			
			this.headerLen = 5;
			if (is_syn) headerLen += 2; // for MSS and SACK

			this.codeBits = codeBits;
			this.window = window;
			this.urgentPointer = urgentPointer;
		}

		[ThreadStatic]
		private static int headerStartOffset;
		protected override void AppendBytesHeader(ByteArrayList buf)
		{
			bool is_syn = (codeBits & TcpPacket.CodeBits.SYN) != 0;

			headerStartOffset = buf.Count;
			// ports
			buf.AddUInt16(sourcePort);
			buf.AddUInt16(destinationPort);
			// sequence + ack number
			buf.AddUInt32(seqNumber);
			buf.AddUInt32(ackNumber);
			
			buf.AddByte( (byte) (((byte)headerLen) << 4) );
			buf.AddByte( ((byte)codeBits) );
			// window + checksum + urgent pointer
			buf.AddUInt16(window);
			buf.AddUInt16(0); // checksum will be added in the AppendBytesFooter()
			buf.AddUInt16(urgentPointer);

			// Future version: options
			if (is_syn)
			{
				if (headerLen >= 6)
				{
					buf.AddByte(2); // MSS
					buf.AddByte(4); // MSS length
					buf.AddUInt16(1460); // MSS value
				}
				if (headerLen >= 7)
				{
					buf.AddByte(1); // NOP
					buf.AddByte(1); // NOP
					buf.AddByte(4); // SACK
					buf.AddByte(2); // SACK Permitted
				}
			}
		}

		protected override void AppendBytesFooter(ByteArrayList buf)
		{
			UInt16 checksum = buf.OneComplementChecksum2(headerStartOffset, buf.Count - headerStartOffset, pseudoHeaderSum);
			buf.SetUInt16(headerStartOffset + 16, checksum);
		}
	
		protected override int BytesCount()
		{
			return headerLen * 4;
		}

		public override NetSim.Core.Packet.PacketGroup Group
		{
			get { return Packet.PacketGroup.TransportLayer; }
		}

		public override int Type
		{
			get { return (int)Packet.TransportLayerPacketType.TCP; }
		}


	}

	#endregion

	#region UDP

	public class UdpPacket : Packet
	{
		public UInt16 SourcePort { get { return this.sourcePort; } }
		private UInt16 sourcePort;
	
		public UInt16 DestinationPort { get { return this.destinationPort; } }
		private UInt16 destinationPort;

		public void SetPseudoHeaderSum(UInt32 sum)
		{
			this.pseudoHeaderSum = sum;
		}
		private UInt32 pseudoHeaderSum;

		public static UdpPacket CreatePacket(byte[] data, int offset, int len)
		{
			return new UdpPacket(data, offset, len);
		}

		protected UdpPacket(byte[] data, int offset, int len)
		{
			int original_offset = offset;

			// ports
			sourcePort = ByteArrayList.ToUInt16(data, offset);
			offset += 2;
			destinationPort = ByteArrayList.ToUInt16(data, offset);
			offset += 2;
			// message length
			UInt16 message_lengt = ByteArrayList.ToUInt16(data, offset);
			offset += 2;
			// checksum
			//UInt16 checksum = ByteArrayList.ToUInt16(data, offset); // Future version: check checksum
			offset += 2;

			this.HigherLayerPacket = SocketDataPacket.CreatePacket(data, offset, message_lengt - (offset - original_offset),
				ProtocolType.Udp);
		}

		public UdpPacket(UInt16 sourcePort, UInt16 destinationPort)
		{
			this.sourcePort = sourcePort;
			this.destinationPort = destinationPort;
		}

		[ThreadStatic]
		private static int headerStartOffset;
		protected override void AppendBytesHeader(ByteArrayList buf)
		{
			headerStartOffset = buf.Count;

			buf.AddUInt16(sourcePort);
			buf.AddUInt16(destinationPort);
			buf.AddUInt16((UInt16)(HigherLayerPacket.GetBytesCount() + 8)); //Future version: can be set more effectively in AppendBytesFooter?
			buf.AddUInt16(0); // checksum computation in AppendBytesFooter() requires zero here
		}

		protected override void AppendBytesFooter(ByteArrayList buf)
		{
			UInt16 checksum = buf.OneComplementChecksum2(headerStartOffset, buf.Count - headerStartOffset, pseudoHeaderSum);
			buf.SetUInt16(headerStartOffset + 6, checksum);
		}

		protected override int BytesCount()
		{
			return 8;
		}

		public override NetSim.Core.Packet.PacketGroup Group
		{
			get { return Packet.PacketGroup.TransportLayer; }
		}

		public override int Type
		{
			get { return (int)Packet.TransportLayerPacketType.UDP; }
		}
	}

	#endregion

	#region ARP

	/// <summary>
	/// ARP packet implementation. Includes all ARP request/response, RARP request/response.
	/// </summary>
	public class ArpPacket : Packet
	{
		/// <summary>
		/// Hardware type used in ARP packets.
		/// </summary>
		public enum HardwareType : ushort
		{
			Ethernet = 1,

			Unknown = 0xFFFF
		}

		/// <summary>
		/// Operation = ARP packet type.
		/// </summary>
		public enum Operation : ushort
		{
			ArpRequest = 1,
			ArpResponse = 2,
			RArpRequest = 3,
			RArpResponse = 4
		}

		public HardwareType HardwareTypeField
		{
			get { return this.hardwareType; }
		}
		private HardwareType hardwareType;

		public Packet.NetworkLayerPacketType ProtocolTypeField
		{
			get { return this.protocolType; }
		}
		private Packet.NetworkLayerPacketType protocolType;

		public byte HardwareAddressLen
		{
			get { return this.hardwareAddressLen; }
		}
		private byte hardwareAddressLen;

		public byte ProtocolAddressLen
		{
			get { return this.protocolAddressLen; }
		}
		private byte protocolAddressLen;

		public Operation Op
		{
			get { return this.operation; }
		}
		private Operation operation;

		public byte[] SenderHardwareAddress
		{
			get { return this.senderHardwareAddress; }
		}
		private byte[] senderHardwareAddress;

		public byte[] SenderProtocolAddress
		{
			get { return this.senderProtocolAddress; }
		}
		private byte[] senderProtocolAddress;

		public byte[] TargetHardwareAddress
		{
			get { return this.targetHardwareAddress; }
		}
		private byte[] targetHardwareAddress;

		public byte[] TargetProtocolAddress
		{
			get { return this.targetProtocolAddress; }
		}
		private byte[] targetProtocolAddress;

		public static ArpPacket CreatePacket(byte[] data, int offset, int len)
		{
			return new ArpPacket(data, offset, len);
		}

		/// <summary>
		/// Changes the packet <paramref name="p"/> to be a response. New packet is returned, original one
		/// is unchanged. <paramref name="p"/> has to be ARP Request packet.
		/// </summary>
		/// <param name="p">Original packet.</param>
		/// <param name="hwAddress">Our hardware address to be placed into the response (sender HW address).</param>
		/// <returns>Newly created ARP response packet.</returns>
		public static ArpPacket SwitchToArpResponse(ArpPacket p, NetworkAddress hwAddress)
		{
			if (p.Op != Operation.ArpRequest)
				throw new ArgumentException("ArpPacket with operation set to request expected.");

			ArpPacket reply = CreatePacket(p.HardwareTypeField, p.ProtocolTypeField, p.HardwareAddressLen, p.ProtocolAddressLen,
				Operation.ArpResponse, hwAddress.GetAddressBytes(), p.TargetProtocolAddress, p.SenderHardwareAddress, p.SenderProtocolAddress);
			return reply;
		}

		/// <summary>
		/// Creates new <see cref="ArpPacket"/>. Values passed as in parameters are not copied, so they shouldn't be
		/// modified after this method call.
		/// </summary>
		/// <param name="hwType">Hardware type.</param>
		/// <param name="protoType">Protocol type.</param>
		/// <param name="hwAddrLen">Hardware address length in bytes.</param>
		/// <param name="protoAddrLen">Protocol address length in bytes.</param>
		/// <param name="operation">Operation (request, response, rarp request/response)</param>
		/// <param name="senderHwAddr">Sender hardware address.</param>
		/// <param name="senderProtoAddr">Sender protocol address.</param>
		/// <param name="targetHwAddr">Target (receiver) hardware address.</param>
		/// <param name="targetProtoAddr">Target protocol address.</param>
		/// <returns></returns>
		public static ArpPacket CreatePacket(HardwareType hwType, Packet.NetworkLayerPacketType protoType, byte hwAddrLen, byte protoAddrLen,
			Operation operation, byte[] senderHwAddr, byte[] senderProtoAddr, byte[] targetHwAddr, byte[] targetProtoAddr)
		{
			return new ArpPacket(hwType, protoType, hwAddrLen, protoAddrLen, operation, senderHwAddr, senderProtoAddr,
				targetHwAddr, targetProtoAddr);
		}

		protected ArpPacket(HardwareType hwType, Packet.NetworkLayerPacketType protoType, byte hwAddrLen, byte protoAddrLen,
			Operation operation, byte[] senderHwAddr, byte[] senderProtoAddr, byte[] targetHwAddr, byte[] targetProtoAddr)
		{
			this.hardwareType = hwType;
			this.protocolType = protoType;
			this.hardwareAddressLen = hwAddrLen;
			this.protocolAddressLen = protoAddrLen;
			this.operation = operation;
			
			//do not copy data
			this.senderHardwareAddress = senderHwAddr != null ? senderHwAddr : new byte[hwAddrLen];
			this.senderProtocolAddress = senderProtoAddr != null ? senderProtoAddr : new byte[protoAddrLen];
			this.targetHardwareAddress = targetHwAddr != null ? targetHwAddr : new byte[hwAddrLen];
			this.targetProtocolAddress = targetProtoAddr != null ? targetProtoAddr : new byte[protoAddrLen];
		}

		/// <summary>
		/// Creates new ARP packet form the array of bytes.
		/// </summary>
		/// <remarks>
		/// Future version: check data length (not checked right now..)
		/// </remarks>
		/// <param name="data">Packet data.</param>
		/// <param name="offset">Offset in <paramref name="data"/> where the ARP packet starts.</param>
		/// <param name="len">Length of <paramref name="data"/>.</param>
		protected ArpPacket(byte[] data, int offset, int len)
		{
			// hardware + protocol
			hardwareType = (HardwareType) ByteArrayList.ToUInt16(data, offset);
			offset += 2;
			protocolType = (Packet.NetworkLayerPacketType) ByteArrayList.ToUInt16(data, offset);
			offset += 2;
			// addresses length
			hardwareAddressLen = data[offset];
			offset++;
			protocolAddressLen = data[offset];
			offset++;
			// operation
			operation = (Operation) ByteArrayList.ToUInt16(data, offset);
			offset += 2;

			// sender hardware address
			senderHardwareAddress = new byte[hardwareAddressLen];
			Buffer.BlockCopy(data, offset, senderHardwareAddress, 0, hardwareAddressLen);
			offset += hardwareAddressLen;
			// sender protocol address
			senderProtocolAddress = new byte[protocolAddressLen];
			Buffer.BlockCopy(data, offset, senderProtocolAddress, 0, protocolAddressLen);
			offset += protocolAddressLen;

			// target hardware address
			targetHardwareAddress = new byte[hardwareAddressLen];
			Buffer.BlockCopy(data, offset, targetHardwareAddress, 0, hardwareAddressLen);
			offset += hardwareAddressLen;
			// target protocol address
			targetProtocolAddress = new byte[protocolAddressLen];
			Buffer.BlockCopy(data, offset, targetProtocolAddress, 0, protocolAddressLen);
			offset += protocolAddressLen;
		}

		protected override void AppendBytesHeader(ByteArrayList buf)
		{
			// hw + protocol types
			buf.AddUInt16((UInt16)hardwareType);
			buf.AddUInt16((UInt16)protocolType);
			// lengths + operation
			buf.AddByte(hardwareAddressLen);
			buf.AddByte(protocolAddressLen);
			buf.AddUInt16((UInt16)operation);
			// sender
			buf.AddBytes(senderHardwareAddress);
			buf.AddBytes(senderProtocolAddress);
			// target
			buf.AddBytes(targetHardwareAddress);
			buf.AddBytes(targetProtocolAddress);
		}

		protected override void AppendBytesFooter(ByteArrayList buf)
		{
			// nothing
		}

		protected override int BytesCount()
		{
			return
				8 // fixed size
				+ 2 * (hardwareAddressLen + protocolAddressLen) // variable size
				;
		}

		public override NetSim.Core.Packet.PacketGroup Group
		{
			get { return Packet.PacketGroup.NetworkLayer; }
		}

		public override int Type
		{
			get { return (int)Packet.NetworkLayerPacketType.ARP; }
		}

	}

	#endregion

	#region SocketDataPacket
	/// <summary>
	/// Packet used for encapsulating data that belongs to a specific socket.
	/// </summary>
	public class SocketDataPacket : DataPacket
	{
		/// <summary>
		/// Specifies protocol that is used for communication through the socket.
		/// </summary>
		public ProtocolType Protocol
		{
			get { return protocol; }
		}
		private ProtocolType protocol;

		[Flags]
			public enum TransferOptions
		{
			None = 0,
			Connect = 1,
			Disconnect = 2,
			EndOfData = 4,
			OutOfBandData = 8,
			ConnectionAccepted = 16
		}

		public TransferOptions Options;

		//public IPAddress LocalAddress;
		public UInt16 LocalPort;

		//public IPAddress RemoteAddress;
		public UInt16 RemotePort;

		public static SocketDataPacket CreatePacket(byte[] array, int offset, ProtocolType protocol)
		{
			return CreatePacket(array, offset, array.Length - offset, protocol);
		}

		public static SocketDataPacket CreatePacket(byte[] array, int offset, int len, ProtocolType protocol)
		{
			return new SocketDataPacket(array, offset, len, protocol);
		}

		protected SocketDataPacket(byte[] array, int offset, int len, ProtocolType protocol) : base(array, offset, len)
		{
			this.protocol = protocol;
		}
	}

	#endregion

	#region Factory
	/// <summary>
	/// Packet factory for all packets in NetSim.Library.
	/// </summary>
	public class LibPacketFactory : PacketFactory
	{
		public override PacketFactory.SupportedPacket[] Supports
		{
			get
			{
				return new PacketFactory.SupportedPacket[] {
					new PacketFactory.SupportedPacket(Packet.PacketGroup.NetworkLayer, (int)Packet.NetworkLayerPacketType.IP),
					new PacketFactory.SupportedPacket(Packet.PacketGroup.TransportLayer, (int)Packet.TransportLayerPacketType.TCP),
					new PacketFactory.SupportedPacket(Packet.PacketGroup.TransportLayer, (int)Packet.TransportLayerPacketType.UDP),
					new PacketFactory.SupportedPacket(Packet.PacketGroup.NetworkLayer, (int)Packet.NetworkLayerPacketType.ARP),
					new PacketFactory.SupportedPacket(Packet.PacketGroup.TransportLayer, (int)Packet.TransportLayerPacketType.ICMP)
																									 };
			}
		}

		public override Packet CreatePacket(NetSim.Core.Packet.PacketGroup group, int identification, byte[] data, int offset, int len)
		{
			switch (group)
			{
				case Packet.PacketGroup.NetworkLayer:
				switch (identification)
				{
					case (int)Packet.NetworkLayerPacketType.IP:
						return IPPacket.CreatePacket(data, offset, len);
					case (int)Packet.NetworkLayerPacketType.ARP:
						return ArpPacket.CreatePacket(data, offset, len);
				}
					break;

				case Packet.PacketGroup.TransportLayer:
				switch (identification)
				{
					case (int)Packet.TransportLayerPacketType.ICMP:
						return IcmpPacket.CreatePacket(data, offset, len);
					case (int)Packet.TransportLayerPacketType.TCP:
						return TcpPacket.CreatePacket(data, offset, len);
					case (int)Packet.TransportLayerPacketType.UDP:
						return UdpPacket.CreatePacket(data, offset, len);
				}
					break;
			}

			// cannot create this packet
			throw new PacketCreationException(String.Format("Packet '{0}, {1}' is not supported by LibPacketFactory", group, identification));
		}

	}
	#endregion
}
