/*

 Copyright (c) 2006 Pavel Novak.

 The use and distribution terms for this software are contained in the file named License.txt, 
 which can be found in the root of the NetSim distribution. By using this software 
 in any fashion, you are agreeing to be bound by the terms of this license.
 
 You must not remove this notice from this software.

*/

using System;
using System.Text;
using System.Collections;
using System.Reflection;
using NetSim.Core;

namespace NetSim.Console
{
	/// <summary>
	/// Type of remote control to start.
	/// </summary>
	[Flags]
	public enum RemoteControl
	{
		None = 0,
		SharedMemory = 1
	}

	/// <summary>
	/// How to hanle remote sockets.
	/// </summary>
	[Flags]
	public enum RemoteSockets
	{
		None = 0,
		TcpChannel = 1,
		ShmChannel = 2
	}

	public enum Prompt
	{
		None = 0,
		Simple = 1,
		Advanced = 2
	}

	/// <summary>
	/// 
	/// </summary>
	class NetSimConsole
	{
		//
		// variables assigned according to command line parameters
		//

		/// <summary>
		/// Which support of remote control use.
		/// </summary>
		private static RemoteControl remoteControl = RemoteControl.None;

		/// <summary>
		/// Which support for remote sockets use.
		/// </summary>
		private static RemoteSockets remoteSockets = RemoteSockets.None;

		/// <summary>
		/// Which prompt to use? Advanced is the default.
		/// </summary>
		private static Prompt prompt = Prompt.Advanced;

		/// <summary>
		/// Configuration file describing the virtual network.
		/// </summary>
		private static string configFile;

		/// <summary>
		/// Prints info about the simulator.
		/// </summary>
		private static void Logo()
		{
			Version version = Assembly.GetAssembly(typeof(NetSim.Core.Simulator)).GetName().Version;
			System.Console.WriteLine("Network Simulator {0}.{1} build {2}", version.Major, version.Minor, version.Revision);
		}

		/// <summary>
		/// Prints usage of this tool if wrong parameters were specified.
		/// </summary>
		/// <param name="error">The error string that should be printed out.</param>
		private static void Usage()
		{
			System.Console.WriteLine("Usage:");
#if MONO
			System.Console.WriteLine("netsim [-pt:simple|adv] [-h] <configuration file>");
#else
			System.Console.WriteLine("netsim [-rc:shm] [-rs:shm] [-pt:simple|adv] [-h] <configuration file>");
#endif
//			System.Console.WriteLine("netsim [-rc:shm] [-rs:tcp] [-pt:none|simple|adv] [-h] <configuration file>");
			System.Console.WriteLine();
#if !MONO
			System.Console.WriteLine("    -rc  Start remote control for the simulator");
			System.Console.WriteLine("           shm - use shared memory for remote control");
			System.Console.WriteLine("    -rs  Start remote sockets which allows redirection of any");
			System.Console.WriteLine("         application using WinSock to some virtual node");
			System.Console.WriteLine("           shm - use shared memory for communication");
//			System.Console.WriteLine("           tcp - use TCP Channel for communication");
#endif
			System.Console.WriteLine("    -pt  Prompt for controlling simulation directly from the command line");
			System.Console.WriteLine("         Defaults to simple");
#if !MONO
			System.Console.WriteLine("           none - no control, neither the simulator cannot be stopped.");
			System.Console.WriteLine("                  used together with -rc");
#endif
			System.Console.WriteLine("           simple - only shutting down the simulator is available");
			System.Console.WriteLine("           adv - advanced command prompt allowing disabling and enabling");
			System.Console.WriteLine("                 parts of the simulator while it is running (default)");
			System.Console.WriteLine("    -h  Prints this help and exits");
			System.Console.WriteLine();
		}

		/// <summary>
		/// Parses arguments and sets this class members according to params.
		/// </summary>
		/// <param name="args">Command line arguments.</param>
		/// <returns><b>True</b> if succesfuly parsed, <b>false</b> otherwise.</returns>
		private static bool ParseArgs(string[] args, out string error)
		{
			for (int i = 0; i < args.Length; i++)
			{
				if (args[i][0] == '-') // switch
				{
					int colon_pos = args[i].IndexOf(':', 1);

					//----------------------------------------------------
					// Commands without value
					if (colon_pos < 0)
					{
						string command = args[i].Substring(1);
						switch (command)
						{
							case "h":
								error = null; // no error, just help requested
								return false;
								
							default:
								error = String.Format("Invalid command '{0}'.", command);
								return false;
						}
					}
						// ----------------------------------------------
						// Commands with value
					else
					{

						string command = args[i].Substring(1, colon_pos-1);
						string value = args[i].Substring(colon_pos+1);
						switch (command)
						{
								//
								// remote control
								//
							case "rc":
							switch (value)
							{
								case "shm":
									remoteControl |= RemoteControl.SharedMemory;
									break;
								default:
									error = String.Format("Invalid value '{0}' for -{1}.", value, command);
									return false;
							}
								break;

								//
								// remote sockets
								//
							case "rs":
							switch (value)
							{
								case "tcp":
									remoteSockets |= RemoteSockets.TcpChannel;
									break;
								case "shm":
									remoteSockets |= RemoteSockets.ShmChannel;
									break;
								default:
									error = String.Format("Invalid value '{0}' for -{1}.", value, command);
									return false;
							}
								break;

								//
								// prompt
								//
							case "pt":
							switch (value)
							{
								case "none":
									prompt = Prompt.None;
									break;
								case "simple":
									prompt = Prompt.Simple;
									break;
								case "adv":
									prompt = Prompt.Advanced;
									break;
								default:
									error = String.Format("Invalid value '{0}' for -{1}.", value, command);
									return false;
							}
								break;

							default:
								error = String.Format("Invalid command -{0}.", command);
								return false;
						}
					}
					//---------------------------------------------------
				}
				else // configuration file
				{
					if (configFile != null) // config file specified twice
					{
						error = String.Format("Configuration file specified twice: '{0}'; '{1}'", configFile, args[i]);
						return false;
					}
					configFile = args[i];
				}
			} // for

			//
			// check mandatory parameters
			//
			if (configFile == null)
			{
				error = "No configuration file specified.";
				return false;
			}

			//
			// check dependencies
			//
			if (prompt == Prompt.None && remoteControl == RemoteControl.None)
			{
				error = "Neither controling the simulator via command line nor remote control is specified.";
				return false;
			}

			error = "OK";
			return true; // succesfuly parsed
		}

		private static void ControlRunningSimulator()
		{
			if (prompt == Prompt.Simple)
			{
				System.Console.WriteLine();
				System.Console.WriteLine("Press Enter to exit...");
				System.Console.ReadLine();
			}
			else if (prompt == Prompt.Advanced)
			{
				AdvancedPrompt();
			}
			else // (prompt == Prompt.None)
			{
				// watit for shutting down remotely
				Simulator.GetInstance().WaitForStop.WaitOne();
			}
		}

		private static void AdvancedPrompt()
		{
			System.Console.WriteLine("Type 'help' for the list of options.");
			System.Console.WriteLine();

			bool exit = false;
			while (!exit)
			{
				System.Console.Write("netsim> ");
				string line = System.Console.ReadLine();
				if (line == null) // no more characters available, probably the program is closing
					break;

				line = line.Trim();
				if (line.Length == 0)
					continue;

				// split line to commands
				IList commands = LineToCommands(line);

				bool error = false;
				switch ((string)commands[0])
				{
					case "exit":
						exit = true;
						break;

					case "help":
						System.Console.WriteLine("disable link <name>         list nodes         help");
						System.Console.WriteLine("enable link <name>          list links         exit");
						System.Console.WriteLine("disable node <name>         list events");
						System.Console.WriteLine("enable node <name>          ");
						break;

					case "list":
						commands.RemoveAt(0);
						PromptList(commands, out error);
						break;

					case "disable":
						commands.RemoveAt(0);
						PromptEnableDisable(commands, false, out error);
						break;

					case "enable":
						commands.RemoveAt(0);
						PromptEnableDisable(commands, true, out error);
						break;

					default:
						error = true;
						break;
				}

				if (error)
					System.Console.WriteLine("Wrong command. Type 'help' for the list of options.");
			}
		}

		/// <summary>
		/// Splits a string to list of strings by white characters. Strings in double quotes are left together.
		/// </summary>
		/// <param name="line">String to split.</param>
		/// <returns>List of strings.</returns>
		private static IList LineToCommands(string line)
		{
			ArrayList result = new ArrayList();
			bool in_quotes = false;
			bool escaped = false;
			StringBuilder sb = new StringBuilder();

			foreach (char ch in line)
			{
				if (!escaped && ch == '\\')
				{
					escaped = true;
					continue;
				}

				switch (ch)
				{
					case '\t': goto case ' ';
					case ' ':
						if (in_quotes)
						{
							sb.Append(ch);
						}
						else
						{
							if (sb.Length > 0)
							{
								result.Add(sb.ToString());
								sb.Remove(0, sb.Length);
							}
						}
						break;

					case '"':
						if (escaped)
						{
							sb.Append(ch);
						}
						else
						{
							in_quotes = !in_quotes;
						}
						break;

					default:
						sb.Append(ch);
						break;
				}
			}

			if (sb.Length > 0)
				result.Add(sb.ToString());

			return result;
		}

		/// <summary>
		/// Processes "list" command.
		/// </summary>
		/// <param name="commands">List of subcommands.</param>
		/// <returns><b>True</b> on success.</returns>
		private static void PromptList(IList commands, out bool error)
		{
			error = true;
			if (commands.Count != 1)
				return;

			IList elements;
			switch ((string)commands[0])
			{
				case "nodes":
					elements = Simulator.GetInstance().Nodes;
					break;

				case "links":
					elements = Simulator.GetInstance().Links;
					break;

				case "events":
					PromptListEvents(out error);
					return;

				default:
					return;
			}

			foreach (StructureObject so in elements)
				System.Console.WriteLine("{0} [{1}]{2}", so.Name, so.Enabled ? "enabled" : "disabled",
					so.CanDisable ? "" : "*");

			error = false;
		}

		private static void PromptListEvents(out bool error)
		{
			error = false;
			bool found = false;

			NetSim.Core.Event ev;
			while (Simulator.GetInstance().Events.GetLogEntry(out ev, true))
			{
				found = true;
				System.Console.WriteLine("{0} {1}: {2}", ev.Time, ev.Type, ev.Message);
				if (ev.Exception != null)
				{
					System.Console.WriteLine("   Exception:");
					System.Console.WriteLine("   {0}", ev.Exception.Message);
				}
			}

			if (!found)
				System.Console.WriteLine("No events in the event log.");
		}

		private static void PromptEnableDisable(IList commands, bool enable, out bool error)
		{
			error = true;
			if (commands.Count != 2)
				return;

			StructureObject so;
			BasicObjectKind obj_kind = BasicObjectKind.None;
			switch ((string)commands[0])
			{
				case "node":
					so = Simulator.GetInstance().GetNode((string)commands[1]);
					obj_kind = BasicObjectKind.Node;
					break;

				case "link":
					so = Simulator.GetInstance().GetLink((string)commands[1]);
					obj_kind = BasicObjectKind.Link;
					break;

				default:
					return;
			}

			if (so == null)
				System.Console.WriteLine("{0} \"{1}\" doesn't exist", commands[0], commands[1]);
			else if (!so.CanDisable)
				System.Console.WriteLine("{0} \"{1}\" doesn't support enabling/disabling", commands[0], so.Name);
			else if (so.Enabled == enable)
				System.Console.WriteLine("{0} \"{1}\" is already {2}", commands[0], so.Name, enable ? "enabled" : "disabled");
			else
			{
				if (obj_kind == BasicObjectKind.Node)
					Simulator.GetInstance().EnableDisableNode(so.Name, enable);
				else
					Simulator.GetInstance().EnableDisableLink(so.Name, enable);

				System.Console.WriteLine("{0} \"{1}\" {2}", commands[0], so.Name, enable ? "enabled" : "disabled");
			}

			error = false;
		}

		/// <summary>
		/// The main entry point for the application.
		/// </summary>
		[STAThread]
		static void Main(string[] args)
		{
			Logo();

			string error;
			if (!ParseArgs(args, out error))
			{
				if (error != null) // some error
				{
					System.Console.WriteLine("Wrong command line arguments: " + error);
					System.Console.WriteLine("Use -h for more information.");
				}
				else // no error, just helep requested
				{
					Usage();
				}

#if DEBUG
				System.Console.WriteLine("Press enter to exit..");
				System.Console.ReadLine();
#endif
				return;
			}

			// check if file exists
			if (!System.IO.File.Exists(configFile))
			{
				System.Console.WriteLine("Configuration file ({0}) not found.", configFile);
				return;
			}
			configFile = System.IO.Path.GetFullPath(configFile);

			// file exists, set current working directory to its dir because it can contain relative
			// paths and their should be treated relative to that directory
			Environment.CurrentDirectory = System.IO.Path.GetDirectoryName(configFile);

			int stage = 0;
			try
			{
				System.Console.Write("Configuring simulator... ");
				Simulator.GetInstance().Configure(configFile);
				System.Console.WriteLine("Done.");
				stage = 1;

				if ((remoteControl & RemoteControl.SharedMemory) == RemoteControl.SharedMemory)
					Simulator.GetInstance().UseRemoteControl = true;
				if ((remoteSockets & RemoteSockets.TcpChannel) == RemoteSockets.TcpChannel)
					Simulator.GetInstance().UseRemoteSocketsTcp = true;
				if ((remoteSockets & RemoteSockets.ShmChannel) == RemoteSockets.ShmChannel)
					Simulator.GetInstance().UseRemoteSocketsShm = true;

				System.Console.Write("Starting simulator... ");
				Simulator.GetInstance().Start();
				System.Console.WriteLine("Done.");
				stage = 2;

				ControlRunningSimulator();
				stage = 3;

				System.Console.Write("Shutting down simulator... ");
				Simulator.GetInstance().Stop();
				System.Console.WriteLine("Done.");
				stage = 4;

			}
			catch (Exception e)
			{
				string message = String.Empty;
				switch (stage)
				{
					case 0:
						message = "while configuring simulator"; break;
					case 1:
						message = "while starting the simulator"; break;
					case 2:
						message = "while the simulator is running"; break;
					case 3:
						message = "while stopping the simulator"; break;
				}

				System.Console.WriteLine("Error occured {0}: {1}", message, e.Message);
				if (e.InnerException != null)
					System.Console.WriteLine("Details:");
				while (e.InnerException != null)
				{
					System.Console.WriteLine(e.InnerException.Message);
					e = e.InnerException;
				}
			}
		}
	}
}
